@@ -45,6 +45,9 @@ typedef struct
 	/* Builtin conversion */
 	const ILConversion *builtin;
 
+	/* Coerce "null" to a pointer type */
+	int			pointerNull;
+
 } ConvertRules;
 
 /*
@@ -331,42 +334,54 @@ static int GetUnsafeConvertRules(ILGenInfo *info, ILType *fromType,
 		ILType *toType, int explicit, int kinds, ConvertRules *rules)
 {
 	const ILConversion *conv;
-	/* TODO : Handle the rest of the cases */
-	/*  Numeric/pointer casts  */
-	if (explicit)
+
+	/* Handle pointer/pointer casts */
+	if(ILType_IsPointer(toType) && ILType_IsPointer(fromType))
 	{
-		/*  Numeric -> pointer conversion  */
-		if ( ILIsBuiltinNumeric(toType) && ILType_IsComplex(fromType) )
+		/* Can always cast between pointer types if explicit */
+		if(explicit)
 		{
-			if (ILType_Kind(fromType) == IL_TYPE_COMPLEX_PTR)
-			{
-				conv = ILFindConversion(fromType, toType, explicit,1);
-				if(conv)
-				{
-					rules->builtin = conv;
-					return 1;
-				}
-			}
+			return 1;
 		}
 
-		if ( ILType_IsComplex(toType) && ILIsBuiltinNumeric(fromType) )
+		/* Otherwise can only cast implicitly to "void *" */
+		return (ILType_Ref(toType) == ILType_Void);
+	}
+
+	/* Can cast implicitly from "null" to any pointer type */
+	if(ILType_IsPointer(toType) && fromType == ILType_Null)
+	{
+		rules->pointerNull = 1;
+		return 1;
+	}
+
+	/* Handle explicit numeric/pointer casts */
+	if(explicit)
+	{
+		/* Pointer to numeric conversion */
+		if(ILIsBuiltinNumeric(toType) && ILType_IsPointer(fromType))
 		{
-			if (ILType_Kind(toType) == IL_TYPE_COMPLEX_PTR)
+			conv = ILFindConversion(fromType, toType, explicit, 1);
+			if(conv)
 			{
+				rules->builtin = conv;
 				return 1;
 			}
 		}
 
-		if ( ILType_IsComplex(toType) && ILType_IsComplex(fromType) )
+		/* Numberic to pointer conversion */
+		if(ILType_IsPointer(toType) && ILIsBuiltinNumeric(fromType))
 		{
-			if (ILType_Kind(toType) == IL_TYPE_COMPLEX_PTR
-					&& ILType_Kind(fromType) == IL_TYPE_COMPLEX_PTR)
+			conv = ILFindConversion(fromType, toType, explicit, 1);
+			if(conv)
 			{
+				rules->builtin = conv;
 				return 1;
 			}
 		}
 	}
 
+	/* Could not find an appropriate conversion */
 	return 0;
 }
 
@@ -388,6 +403,7 @@ static int GetConvertRules(ILGenInfo *info, ILType *fromType,
 	rules->method = 0;
 	rules->castType = 0;
 	rules->builtin = 0;
+	rules->pointerNull = 0;
 
 	/* Strip type prefixes before we start */
 	fromType = ILTypeStripPrefixes(fromType);
@@ -405,11 +421,16 @@ static int GetConvertRules(ILGenInfo *info, ILType *fromType,
 		return 0;
 	}
 
-	/* If "fromType" is null, then "toType" must be a reference type */
+	/* If "fromType" is null, then "toType" must be a reference or pointer */
 	if(fromType == ILType_Null)
 	{
 		if((kinds & IL_CONVERT_REFERENCE) != 0)
 		{
+			if(ILType_IsPointer(toType) && info->unsafeLevel > 0)
+			{
+				rules->pointerNull = 1;
+				return 1;
+			}
 			return ILTypeIsReference(toType);
 		}
 		else
@@ -530,6 +551,76 @@ static int GetConvertRules(ILGenInfo *info, ILType *fromType,
 							 IL_META_METHODDEF_SPECIAL_NAME | \
 							 IL_META_METHODDEF_RT_SPECIAL_NAME)
 
+static int ILBetterIndirectConversion(ILGenInfo *info, ILType *from ,
+						ILType *to ,
+						ILType *conversion1,
+						ILType *conversion2)
+{
+	switch(ILBetterConversion(info,from, ILTypeGetParam(conversion1,1), 
+							  ILTypeGetParam(conversion2,1)))
+	{
+		case IL_BETTER_S1:
+		{
+			switch(ILBetterConversionFrom(info, ILTypeGetReturn(conversion1), 
+										ILTypeGetReturn(conversion2),to))
+			{
+				case IL_BETTER_S1:
+				case IL_BETTER_NEITHER:
+				{
+					return IL_BETTER_S1;
+				}
+				break;
+				
+				case IL_BETTER_S2:
+				{
+					return IL_BETTER_NEITHER;
+				}
+				break;
+			}
+		}
+		break;
+		
+		case IL_BETTER_S2:
+		{
+			switch(ILBetterConversionFrom(info, ILTypeGetReturn(conversion1), 
+										ILTypeGetReturn(conversion2),to))
+			{
+				case IL_BETTER_S1:
+				{
+					return IL_BETTER_NEITHER;
+				}
+				break;
+				case IL_BETTER_S2:
+				case IL_BETTER_NEITHER:
+				{
+					return IL_BETTER_S2;
+				}
+				break;
+			}
+		}
+		break;
+		
+		case IL_BETTER_NEITHER:
+		{
+			switch(ILBetterConversionFrom(info, ILTypeGetReturn(conversion1), 
+										ILTypeGetReturn(conversion2),to))
+			{
+				case IL_BETTER_S1:
+				{
+					return IL_BETTER_S1;
+				}
+				case IL_BETTER_S2:
+				{
+					return IL_BETTER_S2;
+				}
+			}
+		}
+		break;
+		
+	}
+	return IL_BETTER_NEITHER;
+}
+
 /*
  * multiple convert rules are checked here . I have implemented only
  * a 2 step conversion fromType->itype1->itype2->toType , more levels
@@ -538,12 +629,11 @@ static int GetConvertRules(ILGenInfo *info, ILType *fromType,
  */
 static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 						   ILType *toType, int explicit,
-						   int kinds, ConvertRules *rules1,
-						   ConvertRules *rules2,
-						   ILType **itype1,ILType **itype2)
+						   int kinds, ILType **itype1,ILType **itype2)
 {
 	ILMethod *method;
 	ILMember *member;
+	ILMember *bestMember = 0;
 	ILType *signature;
 	ILClass *arg1Class = ILTypeToClass(info, fromType);
 	ILClass *arg2Class = ILTypeToClass(info, toType);	
@@ -560,12 +650,12 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 	{
 		arg1Class= ILClassResolve(arg1Class);
 		member=0;
-		while((member = ILClassNextMemberByKind
-					(arg1Class, member, IL_META_MEMBERKIND_METHOD)) != 0)
+		while((member = ILClassNextMemberMatch
+					(arg1Class, member,
+					 IL_META_MEMBERKIND_METHOD, "op_Implicit", 0)) != 0)
 		{
 			/* Filter out members that aren't interesting */
-			if(strcmp(ILMember_Name(member), "op_Implicit") != 0 ||
-			   (ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
+			if((ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
    			(IL_META_METHODDEF_STATIC |IL_META_METHODDEF_SPECIAL_NAME))
 			{
 				continue;
@@ -588,11 +678,11 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 				|| ((explicit && ILCanCastKind(info,returnType,toType,
 				kinds,0) && ILCanCastKind(info,fromType,argType,kinds,0))))
 			{
-				if(itype1)(*itype1)=argType;
-				if(itype2)(*itype2)=returnType;
-				GetConvertRules(info,fromType,argType,0,kinds,rules1);
-				GetConvertRules(info,returnType,toType,0,kinds,rules2);
-				return 1;
+				if(bestMember == NULL || (ILBetterIndirectConversion(info,
+					fromType, toType, ILMethod_Signature(member),									ILMethod_Signature(bestMember)) == IL_BETTER_S1))
+				{
+					bestMember = member;
+				}
 			}
 		}
 		arg1Class = ILClass_Parent(arg1Class);
@@ -602,12 +692,12 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 	{
 		arg2Class= ILClassResolve(arg2Class);
 		member=0;
-		while((member = ILClassNextMemberByKind
-					(arg2Class, member, IL_META_MEMBERKIND_METHOD)) != 0)
+		while((member = ILClassNextMemberMatch
+					(arg2Class, member,
+					 IL_META_MEMBERKIND_METHOD, "op_Implicit", 0)) != 0)
 		{
 			/* Filter out members that aren't interesting */
-			if(strcmp(ILMember_Name(member), "op_Implicit") != 0 ||
-			   (ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
+			if((ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
    			(IL_META_METHODDEF_STATIC |IL_META_METHODDEF_SPECIAL_NAME))
 			{
 				continue;
@@ -631,15 +721,25 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 				|| ((explicit && ILCanCastKind(info,returnType,toType,
 				kinds,0) && ILCanCastKind(info,fromType,argType,kinds,0))))
 			{
-				if(itype1)(*itype1)=argType;
-				if(itype2)(*itype2)=returnType;
-				GetConvertRules(info,fromType,argType,0,kinds,rules1);
-				GetConvertRules(info,returnType,toType,0,kinds,rules2);
-				return 1;
+				if(bestMember == NULL || (ILBetterIndirectConversion(info,
+					fromType, toType, ILMethod_Signature(member),									ILMethod_Signature(bestMember)) == IL_BETTER_S1))
+				{
+					bestMember = member;
+				}
 			}
 		}
 		arg2Class = ILClass_Parent(arg2Class);
 	}
+	
+	if(bestMember)
+	{
+		signature = ILMethod_Signature(bestMember);
+		argType = ILTypeGetParam(signature, 1);
+		returnType = ILTypeGetReturn(signature);
+		if(itype1)(*itype1)=argType;
+		if(itype2)(*itype2)=returnType;
+		return 1;
+	}
 
 	if(!explicit)return 0;
 	
@@ -652,12 +752,12 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 	{
 		arg1Class= ILClassResolve(arg1Class);
 		member=0;
-		while((member = ILClassNextMemberByKind
-					(arg1Class, member, IL_META_MEMBERKIND_METHOD)) != 0)
+		while((member = ILClassNextMemberMatch
+					(arg1Class, member,
+					 IL_META_MEMBERKIND_METHOD, "op_Explicit", 0)) != 0)
 		{
 			/* Filter out members that aren't interesting */
-			if(strcmp(ILMember_Name(member), "op_Explicit") != 0 ||
-			   (ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
+			if((ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
    			(IL_META_METHODDEF_STATIC |IL_META_METHODDEF_SPECIAL_NAME))
 			{
 				continue;
@@ -680,11 +780,11 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 			if(ILCanCastKind(info,returnType,toType,kinds,0) && 
 				ILCanCastKind(info,fromType,argType,kinds,0))
 			{
-				if(itype1)(*itype1)=argType;
-				if(itype2)(*itype2)=returnType;
-				GetConvertRules(info,fromType,argType,1,kinds,rules1);
-				GetConvertRules(info,returnType,toType,1,kinds,rules2);
-				return 1;
+				if(bestMember == NULL || (ILBetterIndirectConversion(info,
+					fromType, toType, ILMethod_Signature(member),									ILMethod_Signature(bestMember)) == IL_BETTER_S1))
+				{
+					bestMember = member;
+				}
 			}
 		}
 		arg1Class = ILClass_Parent(arg1Class);
@@ -694,12 +794,12 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 	{
 		arg2Class= ILClassResolve(arg2Class);
 		member=0;
-		while((member = ILClassNextMemberByKind
-					(arg2Class, member, IL_META_MEMBERKIND_METHOD)) != 0)
+		while((member = ILClassNextMemberMatch
+					(arg2Class, member,
+					 IL_META_MEMBERKIND_METHOD, "op_Explicit", 0)) != 0)
 		{
 			/* Filter out members that aren't interesting */
-			if(strcmp(ILMember_Name(member), "op_Explicit") != 0 ||
-			   (ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
+			if((ILMember_Attrs(member) & METHOD_TYPE_ATTRS) !=
    			(IL_META_METHODDEF_STATIC |IL_META_METHODDEF_SPECIAL_NAME))
 			{
 				continue;
@@ -721,15 +821,26 @@ static int GetIndirectConvertRules(ILGenInfo *info, ILType *fromType,
 			if(ILCanCastKind(info,returnType,toType,kinds,0) && 
 				ILCanCastKind(info,fromType,argType,kinds,0))
 			{
-				if(itype1)(*itype1)=argType;
-				if(itype2)(*itype2)=returnType;
-				GetConvertRules(info,fromType,argType,0,kinds,rules1);
-				GetConvertRules(info,returnType,toType,1,kinds,rules2);
-				return 1;
+				if(bestMember == NULL || (ILBetterIndirectConversion(info,
+					fromType, toType, ILMethod_Signature(member),									ILMethod_Signature(bestMember)) == IL_BETTER_S1))
+				{
+					bestMember = member;
+				}
 			}
 		}
 		arg2Class = ILClass_Parent(arg2Class);
 	}
+
+	if(bestMember)
+	{
+		signature = ILMethod_Signature(bestMember);
+		argType = ILTypeGetParam(signature, 1);
+		returnType = ILTypeGetReturn(signature);
+		if(itype1)(*itype1)=argType;
+		if(itype2)(*itype2)=returnType;
+		return 1;
+	}
+
 	return 0;
 }
 
@@ -788,6 +899,14 @@ static void ApplyRules(ILGenInfo *info, ILNode *node,
 		yysetfilename(*parent, yygetfilename(node));
 		yysetlinenum(*parent, yygetlinenum(node));
 	}
+
+	/* Convert the object reference "null" into a pointer "null" */
+	if(rules->pointerNull)
+	{
+		*parent = ILNode_NullPtr_create();
+		yysetfilename(*parent, yygetfilename(node));
+		yysetlinenum(*parent, yygetlinenum(node));
+	}
 }
 
 /*
@@ -891,13 +1010,12 @@ static ILMachineType CanCoerceConst(ILGenInfo *info, ILNode *node,
 int ILCanCoerce(ILGenInfo *info, ILType *fromType, ILType *toType,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	if(GetConvertRules(info, fromType, toType, 0, IL_CONVERT_ALL, &rules))
 	{
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType, 
-							0, IL_CONVERT_ALL,&rules,&rules2,NULL,NULL))
+												0, IL_CONVERT_ALL,NULL,NULL))
 	{
 		return 1;
 	}
@@ -908,13 +1026,12 @@ int ILCanCoerceKind(ILGenInfo *info, ILType *fromType,
 					ILType *toType, int kinds,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	if(GetConvertRules(info, fromType, toType, 0, kinds, &rules))
 	{
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,0, 
-							kinds,&rules,&rules2,NULL,NULL))
+												kinds,NULL,NULL))
 	{
 		return 1;
 	}
@@ -925,7 +1042,6 @@ int ILCanCoerceNode(ILGenInfo *info, ILNode *node,
 				    ILType *fromType, ILType *toType, int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	if(GetConvertRules(info, fromType, toType, 0, IL_CONVERT_ALL, &rules))
 	{
 		return 1;
@@ -935,7 +1051,7 @@ int ILCanCoerceNode(ILGenInfo *info, ILNode *node,
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,0, 
-							IL_CONVERT_ALL,&rules,&rules2,NULL,NULL))
+												IL_CONVERT_ALL,NULL,NULL))
 	{
 		return 1;
 	}
@@ -950,7 +1066,6 @@ int ILCanCoerceNodeKind(ILGenInfo *info, ILNode *node,
 						int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	if(GetConvertRules(info, fromType, toType, 0, kinds, &rules))
 	{
 		return 1;
@@ -961,7 +1076,7 @@ int ILCanCoerceNodeKind(ILGenInfo *info, ILNode *node,
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,0, 
-							kinds,&rules,&rules2,NULL,NULL))
+												kinds,NULL,NULL))
 	{
 		return 1;
 	}
@@ -975,7 +1090,6 @@ int ILCoerce(ILGenInfo *info, ILNode *node, ILNode **parent,
 			 ILType *fromType, ILType *toType,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	ILType *t1,*t2;
 	ILMachineType constType;
 	if(GetConvertRules(info, fromType, toType, 0, IL_CONVERT_ALL, &rules))
@@ -990,11 +1104,11 @@ int ILCoerce(ILGenInfo *info, ILNode *node, ILNode **parent,
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,0, 
-							IL_CONVERT_ALL,&rules,&rules2,&t1,&t2))
+												IL_CONVERT_ALL,&t1,&t2))
 	{
-		ApplyRules(info, node, parent, &rules, t1);
+		ILCoerce(info,*parent,parent,fromType,t1,0);
 		ILCoerce(info,*parent,parent,t1,t2,0);
-		ApplyRules(info, *parent, parent, &rules2, toType);
+		ILCoerce(info,*parent,parent,t2,toType,0);
 		return 1;
 	}
 	else
@@ -1007,7 +1121,6 @@ int ILCoerceKind(ILGenInfo *info, ILNode *node, ILNode **parent,
 			     ILType *fromType, ILType *toType, int kinds,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	ILMachineType constType;
 	ILType *t1,*t2;
 	if(GetConvertRules(info, fromType, toType, 0, kinds, &rules))
@@ -1023,11 +1136,11 @@ int ILCoerceKind(ILGenInfo *info, ILNode *node, ILNode **parent,
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,0, 
-								kinds,&rules,&rules2,&t1,&t2))
+												kinds,&t1,&t2))
 	{
-		ApplyRules(info, node, parent, &rules, t1);
+		ILCoerceKind(info,*parent,parent,fromType,t1,kinds,0);
 		ILCoerceKind(info,*parent,parent,t1,t2,kinds,0);
-		ApplyRules(info, *parent, parent, &rules2, toType);
+		ILCoerceKind(info,*parent,parent,t2,toType,kinds,0);
 		return 1;
 	}
 	else
@@ -1039,13 +1152,12 @@ int ILCoerceKind(ILGenInfo *info, ILNode *node, ILNode **parent,
 int ILCanCast(ILGenInfo *info, ILType *fromType, ILType *toType,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	if(GetConvertRules(info, fromType, toType, 1, IL_CONVERT_ALL, &rules))
 	{
 		return 1;
 	}
-	else if(indirect && GetIndirectConvertRules(info,fromType,toType,0, 
-							IL_CONVERT_ALL,&rules,&rules2,NULL,NULL))
+	else if(indirect && GetIndirectConvertRules(info,fromType,toType,1, 
+												IL_CONVERT_ALL,NULL,NULL))
 	{
 		return 1;
 	}
@@ -1056,13 +1168,12 @@ int ILCanCastKind(ILGenInfo *info, ILType *fromType,
 				  ILType *toType, int kinds,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	if(GetConvertRules(info, fromType, toType, 1, kinds, &rules))
 	{
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,1, 
-									kinds,&rules,&rules2,NULL,NULL))
+												kinds,NULL,NULL))
 	{
 		return 1;
 	}
@@ -1073,7 +1184,6 @@ int ILCast(ILGenInfo *info, ILNode *node, ILNode **parent,
 		   ILType *fromType, ILType *toType,int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	ILType *t1,*t2;
 	if(GetConvertRules(info, fromType, toType, 1, IL_CONVERT_ALL, &rules))
 	{
@@ -1081,11 +1191,11 @@ int ILCast(ILGenInfo *info, ILNode *node, ILNode **parent,
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,1, 
-							IL_CONVERT_ALL,&rules,&rules2,&t1,&t2))
+												IL_CONVERT_ALL,&t1,&t2))
 	{
-		ApplyRules(info, node, parent, &rules, t1);
+		ILCast(info,*parent,parent,fromType,t1,0);
 		ILCast(info,*parent,parent,t1,t2,0);
-		ApplyRules(info, *parent, parent, &rules2, toType);
+		ILCast(info,*parent,parent,t2,toType,0);
 		return 1;
 	}
 	else
@@ -1099,7 +1209,6 @@ int ILCastKind(ILGenInfo *info, ILNode *node, ILNode **parent,
 			   int indirect)
 {
 	ConvertRules rules;
-	ConvertRules rules2;
 	ILType *t1,*t2;
 	if(GetConvertRules(info, fromType, toType, 1, kinds, &rules))
 	{
@@ -1107,11 +1216,11 @@ int ILCastKind(ILGenInfo *info, ILNode *node, ILNode **parent,
 		return 1;
 	}
 	else if(indirect && GetIndirectConvertRules(info,fromType,toType,1, 
-									kinds,&rules,&rules2,&t1,&t2))
+												kinds,&t1,&t2))
 	{
-		ApplyRules(info, node, parent, &rules, t1);
+		ILCastKind(info,*parent,parent,fromType,t1,kinds,0);
 		ILCastKind(info,*parent,parent,t1,t2,kinds,0);
-		ApplyRules(info, *parent, parent, &rules2, toType);
+		ILCastKind(info,*parent,parent,t2,toType,kinds,0);
 		return 1;
 	}
 	else
@@ -1226,6 +1335,92 @@ int ILBetterConversion(ILGenInfo *info, ILType *sType,
 	}
 }
 
+
+/* TODO : Figure out the actual conversion rules before using
+ * 		  this in the rest of the code . But it's a reasonably
+ * 		  good approximation of what I understand about the spec */
+
+int ILBetterConversionFrom(ILGenInfo *info, ILType *s1Type,
+					   ILType *s2Type, ILType *tType)
+{
+	if(ILTypeIdentical(s1Type, s2Type))
+	{
+		return IL_BETTER_NEITHER;
+	}
+	else if(ILTypeIdentical(s1Type, tType))
+	{
+		return IL_BETTER_S1;
+	}
+	else if(ILTypeIdentical(s2Type, tType))
+	{
+		return IL_BETTER_S2;
+	}
+	else if(ILCanCoerce(info, s1Type, s2Type,0) &&
+	        !ILCanCoerce(info, s2Type, s1Type,0))
+	{
+		return IL_BETTER_S2;
+	}
+	else if(ILCanCoerce(info, s2Type, s1Type,0) &&
+	        !ILCanCoerce(info, s1Type, s2Type,0))
+	{
+		return IL_BETTER_S1;
+	}
+	else if(s1Type == ILType_Int8 &&
+	        (s2Type == ILType_UInt8 ||
+			 s2Type == ILType_UInt16 ||
+			 s2Type == ILType_UInt32 ||
+			 s2Type == ILType_UInt64))
+	{
+		return IL_BETTER_S1;
+	}
+	else if(s2Type == ILType_Int8 &&
+	        (s1Type == ILType_UInt8 ||
+			 s1Type == ILType_UInt16 ||
+			 s1Type == ILType_UInt32 ||
+			 s1Type == ILType_UInt64))
+	{
+		return IL_BETTER_S2;
+	}
+	else if(s1Type == ILType_Int16 &&
+			(s2Type == ILType_UInt16 ||
+			 s2Type == ILType_UInt32 ||
+			 s2Type == ILType_UInt64))
+	{
+		return IL_BETTER_S1;
+	}
+	else if(s2Type == ILType_Int16 &&
+			(s1Type == ILType_UInt16 ||
+			 s1Type == ILType_UInt32 ||
+			 s1Type == ILType_UInt64))
+	{
+		return IL_BETTER_S2;
+	}
+	else if(s1Type == ILType_Int32 &&
+			(s2Type == ILType_UInt32 ||
+			 s2Type == ILType_UInt16))
+	{
+		return IL_BETTER_S1;
+	}
+	else if(s2Type == ILType_Int32 &&
+			(s1Type == ILType_UInt32 ||
+			 s1Type == ILType_UInt16))
+	{
+		return IL_BETTER_S2;
+	}
+	else if(s1Type == ILType_Int64 && s2Type == ILType_UInt64)
+	{
+		return IL_BETTER_S1;
+	}
+	else if(s2Type == ILType_Int64 && s1Type == ILType_UInt64)
+	{
+		return IL_BETTER_S2;
+	}
+	else
+	{
+		return IL_BETTER_NEITHER;
+	}
+}
+
 #ifdef	__cplusplus
 };
 #endif
