@@ -97,17 +97,20 @@ void ILGenInfoInit(ILGenInfo *info, char *progname,
 	info->nextLabel = 1;
 	info->overflowInsns = 1;
 	info->overflowGlobal = 1;
+	info->overflowChanged = 0;
 	info->pedanticArith = 0;
 	info->clsCompliant = 0;
 	info->semAnalysis = 0;
 	info->typeGather = 0;
 	info->inSemType = 0;
+	info->inAttrArg = 0;
 	info->useJavaLib = 0;
 	info->outputIsJava = 0;
 	info->debugFlag = 0;
 	info->hasUnsafe = 0;
 	info->needSwitchPop = 0;
 	info->hasGotoScopes = 1;
+	info->resolvingAlias = 0;
 	info->decimalRoundMode = IL_DECIMAL_ROUND_HALF_EVEN;
 	info->stackHeight = 0;
 	info->maxStackHeight = 0;
@@ -135,6 +138,8 @@ void ILGenInfoInit(ILGenInfo *info, char *progname,
 	info->currentNamespace = 0;
 	info->arrayInit = 0;
 	info->itemHash = 0;
+	info->currentTypeFormals = 0;
+	info->currentMethodFormals = 0;
 	if(useBuiltinLibrary)
 	{
 		ILGenMakeLibrary(info);
