@@ -279,6 +279,7 @@ typedef struct
 	int				checked;
 	int				unsafe;
 	int				noStdLib;
+	int				addConfig;
 	int				optimize;
 	int				warnAsError;
 	int				saneWarnings;
@@ -329,6 +330,7 @@ static int ParseCompileArgs(CSAntTask *task, CSAntCompileArgs *args,
 	args->checked = COMP_FLAG_NOT_SET;
 	args->unsafe = COMP_FLAG_NOT_SET;
 	args->noStdLib = COMP_FLAG_NOT_SET;
+	args->addConfig = COMP_FLAG_NOT_SET;
 	args->optimize = COMP_FLAG_NOT_SET;
 	args->warnAsError = COMP_FLAG_NOT_SET;
 	args->saneWarnings = COMP_FLAG_NOT_SET;
@@ -465,28 +467,68 @@ static int ParseCompileArgs(CSAntTask *task, CSAntCompileArgs *args,
 	}
 
 	/* Collect up additional arguments */
-	node = task->taskChildren;
-	while(node != 0)
+	if(!CSAntRedirectCsc)
 	{
-		if(!strcmp(node->name, "arg"))
+		node = task->taskChildren;
+		while(node != 0)
 		{
-			/* Does this argument only apply to a specific compiler? */
-			value = CSAntTaskParam(node, "compiler");
-			if(!value || !ILStrICmp(value, compilerName))
+			if(!strcmp(node->name, "arg"))
 			{
+				/* Does this argument only apply to a specific compiler? */
+				value = CSAntTaskParam(node, "compiler");
+				if(!value || !ILStrICmp(value, compilerName))
+				{
+					value = CSAntTaskParam(node, "value");
+					if(value)
+					{
+						copyValue = ILDupString(value);
+						if(!copyValue)
+						{
+							CSAntOutOfMemory();
+						}
+						AddArg(&(args->args), &(args->numArgs), copyValue);
+					}
+				}
+			}
+			node = node->next;
+		}
+	}
+	else
+	{
+		/* We are simulating csc, so we need to parse out some
+		   of the csc-style options in the "arg" elements */
+		args->addConfig = COMP_FLAG_TRUE;
+		node = task->taskChildren;
+		while(node != 0)
+		{
+			if(!strcmp(node->name, "arg"))
+			{
+				/* Does this argument only apply to a specific compiler? */
 				value = CSAntTaskParam(node, "value");
 				if(value)
 				{
-					copyValue = ILDupString(value);
-					if(!copyValue)
+					if(!ILStrICmp(value, "/unsafe"))
 					{
-						CSAntOutOfMemory();
+						args->unsafe = COMP_FLAG_TRUE;
+					}
+					else if(!ILStrICmp(value, "/nostdlib"))
+					{
+						args->noStdLib = COMP_FLAG_TRUE;
+					}
+					else if(!ILStrICmp(value, "/noconfig"))
+					{
+						args->addConfig = COMP_FLAG_FALSE;
+					}
+					else if(!strncmp(value, "/r:", 3) ||
+					        !strncmp(value, "/R:", 3))
+					{
+						args->references = CSAntFileSetAdd
+							(args->references, value + 3);
 					}
-					AddArg(&(args->args), &(args->numArgs), copyValue);
 				}
 			}
+			node = node->next;
 		}
-		node = node->next;
 	}
 
 	/* Done */
@@ -581,6 +623,18 @@ static char **BuildCsccCommandLine(CSAntCompileArgs *args)
 		AddValueArg(&argv, &argc, "-D", args->defines[posn]);
 	}
 
+	/* We need to force Latin-1 if we are simulating csc */
+	if(CSAntRedirectCsc)
+	{
+		AddArg(&argv, &argc, "-flatin1-charset");
+	}
+
+	/* Add "-fstdlib-name=corlib" if forced to do so */
+	if(CSAntForceCorLib)
+	{
+		AddArg(&argv, &argc, "-fstdlib-name=corlib");
+	}
+
 	/* Add any extra arguments that were supplied */
 	for(posn = 0; posn < args->numArgs; ++posn)
 	{
@@ -642,6 +696,13 @@ static char **BuildCsccCommandLine(CSAntCompileArgs *args)
 		}
 	}
 
+	/* Add "-lSystem.Xml" and "-lSystem" if necessary to simulate csc */
+	if(args->addConfig == COMP_FLAG_TRUE)
+	{
+		AddArg(&argv, &argc, "-lSystem.Xml");
+		AddArg(&argv, &argc, "-lSystem");
+	}
+
 	/* Terminate the command-line */
 	AddArg(&argv, &argc, (char *)0);
 
@@ -839,18 +900,46 @@ static char **BuildMcsCommandLine(CSAntCompileArgs *args)
 	AddArg(&argv, &argc, "-o");
 	AddArg(&argv, &argc, (char *)(args->output));
 
+	/* Enable debugging if necessary */
+	if(args->debug == COMP_FLAG_TRUE)
+	{
+		AddArg(&argv, &argc, "-g");
+	}
+
 	/* Set the checked compilation state */
 	if(args->checked == COMP_FLAG_TRUE)
 	{
 		AddArg(&argv, &argc, "--checked");
 	}
 
+	/* Set the unsafe compilation state */
+	if(args->unsafe == COMP_FLAG_TRUE)
+	{
+		AddArg(&argv, &argc, "--unsafe");
+	}
+
 	/* Disable the standard library if necessary */
 	if(args->noStdLib == COMP_FLAG_TRUE)
 	{
 		AddArg(&argv, &argc, "--nostdlib");
 	}
 
+	/* Set the optimization level */
+	if(args->optimize == COMP_FLAG_TRUE)
+	{
+		AddArg(&argv, &argc, "-optimize+");
+	}
+	else if(args->optimize == COMP_FLAG_FALSE)
+	{
+		AddArg(&argv, &argc, "-optimize-");
+	}
+
+	/* Convert warnings into errors if requested */
+	if(args->warnAsError == COMP_FLAG_TRUE)
+	{
+		AddArg(&argv, &argc, "-warnaserror+");
+	}
+
 	/* Add any extra arguments that were supplied */
 	for(posn = 0; posn < args->numArgs; ++posn)
 	{
@@ -974,16 +1063,24 @@ int CSAntTask_Cscc(CSAntTask *task)
  */
 int CSAntTask_Csc(CSAntTask *task)
 {
-	CSAntCompileArgs args;
-
-	/* Parse the arguments */
-	if(!ParseCompileArgs(task, &args, 1, "csc"))
+	if(CSAntRedirectCsc)
 	{
-		return 0;
+		/* Redirect NAnt-style <csc> tags to the <compile> functionality */
+		return CSAntTask_Compile(task);
 	}
+	else
+	{
+		CSAntCompileArgs args;
 
-	/* Execute the command */
-	return BuildAndExecute(&args, BuildCscCommandLine);
+		/* Parse the arguments */
+		if(!ParseCompileArgs(task, &args, 1, "csc"))
+		{
+			return 0;
+		}
+
+		/* Execute the command */
+		return BuildAndExecute(&args, BuildCscCommandLine);
+	}
 }
 
 /*
