@@ -34,6 +34,7 @@ static unsigned markerVar = 0;
 static unsigned valueVar = 0;
 static ILMethod *currMethod = 0;
 static ILType *localVarSig = 0;
+static int initNumber = 0;
 
 /*
  * Report an error that resulted from the compiler trying to
@@ -318,6 +319,15 @@ ILMethod *CFunctionCreate(ILGenInfo *info, char *name, ILNode *node,
 					     IL_META_METHODDEF_SPECIAL_NAME,
 					     IL_META_METHODDEF_SPECIAL_NAME);
 	}
+	if(CAttrPresent(decl.attrs, ".init", ".init"))
+	{
+		/* This is a "normal" initializer, generated by the compiler */
+		AddInitOrFini(info, ILToProgramItem(method),
+					  "InitializerAttribute", "InitializerOrderAttribute", 0);
+		ILMemberSetAttrs((ILMember *)method,
+					     IL_META_METHODDEF_SPECIAL_NAME,
+					     IL_META_METHODDEF_SPECIAL_NAME);
+	}
 	if(CAttrPresent(decl.attrs, "destructor", "__destructor__"))
 	{
 		order = CAttrGetInt(decl.attrs, "dorder", "__dorder__", 1);
@@ -369,7 +379,8 @@ void CFunctionDeclareParams(ILGenInfo *info, ILMethod *method)
 	}
 }
 
-void CFunctionOutput(ILGenInfo *info, ILMethod *method, ILNode *body)
+void CFunctionOutput(ILGenInfo *info, ILMethod *method, ILNode *body,
+					 int initGlobalVars)
 {
 	FILE *stream = info->asmOutput;
 	unsigned indexVar = 0;
@@ -423,6 +434,16 @@ void CFunctionOutput(ILGenInfo *info, ILMethod *method, ILNode *body)
 	info->returnType = ILTypeGetReturn(signature);
 	returnMachineType = ILTypeToMachineType(info->returnType);
 
+	/* Force initialization of global variables if we were
+	   called from C# code and ".init" hasn't been run yet */
+	if(initGlobalVars)
+	{
+		fputs("\tldtoken 'init-on-demand'\n", stream);
+		fputs("\tcall\tvoid [.library]System.Runtime.CompilerServices"
+				".RuntimeHelpers::RunClassConstructor"
+				"(valuetype [.library]System.RuntimeTypeHandle)\n", stream);
+	}
+
 	/* Create the "setjmp" header if necessary */
 	if(numSetJmpRefs > 0)
 	{
@@ -581,7 +602,42 @@ void CFunctionOutput(ILGenInfo *info, ILMethod *method, ILNode *body)
 
 void CFunctionFlushInits(ILGenInfo *info, ILNode *list)
 {
-	/* TODO */
+	CDeclSpec spec;
+	CDeclarator decl;
+	char buf[64];
+	char *name;
+	ILMethod *method;
+	ILNode *attributes;
+	ILScope *scope;
+
+	/* Set the return type, which is "void" */
+	CDeclSpecSetType(spec, ILType_Void);
+	spec.specifiers = C_SPEC_STATIC;
+
+	/* Build the initializer name */
+	sprintf(buf, ".init-%d", initNumber);
+	++initNumber;
+	name = ILInternString(buf, -1).string;
+
+	/* Build the attribute list, containing just ".init" */
+	attributes = ILNode_List_create();
+	ILNode_List_Add(attributes, ILNode_CAttribute_create(".init", 0));
+
+	/* Build a prototype for the initializer */
+	CDeclSetName(decl, name, ILQualIdentSimple(name));
+	decl = CDeclCreatePrototype(info, decl, 0, attributes);
+
+	/* Create the method header for the initializer */
+	method = CFunctionCreate
+		(info, name, list, spec, decl, ILNode_List_create());
+
+	/* Wrap the function body in a new scope */
+	scope = ILScopeCreate(info, CGlobalScope);
+	list = ILNode_NewScope_create(list);
+	((ILNode_NewScope *)list)->scope = scope;
+
+	/* Output the initializer */
+	CFunctionOutput(info, method, list, 0);
 }
 
 void CFunctionPredeclare(ILGenInfo *info)
