@@ -346,8 +346,7 @@ static int LoadLibraryFromPath(const char *path, int freePath)
 
 	/* Attempt to load the image */
 	loadError = ILImageLoadFromFile(path, CCCodeGen.context, &image,
-									IL_LOADFLAG_FORCE_32BIT |
-									IL_LOADFLAG_PRE_VALIDATE, 1);
+									IL_LOADFLAG_FORCE_32BIT, 1);
 	if(loadError == 0)
 	{
 		if(!ILAssemblyCreateImport(CCCodeGen.image, image))
@@ -399,6 +398,13 @@ static int LoadLibrary(const char *name, int nostdlib_flag)
 	image = 0;
 	while((image = ILContextNextImage(CCCodeGen.context, image)) != 0)
 	{
+		if(image == CCCodeGen.image)
+		{
+			/* Skip the image that we are building, as it may
+			   have the same name as a library if we had to
+			   build the library in two passes (e.g. System.dll) */
+			continue;
+		}
 		assem = ILAssembly_FromToken(image, (IL_META_TOKEN_ASSEMBLY | 1));
 		if(assem)
 		{
@@ -440,6 +446,13 @@ static int LoadLibrary(const char *name, int nostdlib_flag)
 		return LoadLibraryFromPath(path, 1);
 	}
 
+	/* If this is the C compiler, then ignore the missing library,
+	   since libraries are normally fixed up at link time */
+	if(CCPluginUsesPreproc == CC_PREPROC_C)
+	{
+		return 1;
+	}
+
 	/* Could not locate the library */
 	fprintf(stderr, _("%s: No such library\n"), name);
 	return 0;
@@ -575,7 +588,14 @@ static void CloseCodeGen(void)
 static char *FindCpp(void)
 {
 	char *newPath = 0;
-	if(ILFileExists("/usr/bin/cpp", &newPath))
+	if(ILFileExists("/usr/bin/cpp3", &newPath))
+	{
+		if(newPath)
+			return newPath;
+		else
+			return "/usr/bin/cpp3";
+	}
+	else if(ILFileExists("/usr/bin/cpp", &newPath))
 	{
 		if(newPath)
 			return newPath;
@@ -783,6 +803,16 @@ static void ParseFile(const char *filename, int is_stdin)
 		   touchy when we redefine some of its builtin symbols.  This
 		   wouldn't be necessary if "-undef" actually did what it is
 		   supposed to do, but it only undefines some symbols, not all */
+#ifdef __APPLE__ 
+#ifdef __MACH__
+		/* Apple shipped a special "pre-compiled header" CPP with the
+		   Darwin OS.  That obviously won't do us much good here, so
+		   we need to set a flag to fall back on standard headers.
+
+			TODO:  Can this end be achieved dynamically?  */
+		CCStringListAdd(&argv, &argc, "-no-cpp-precomp");
+#endif
+#endif
 		CCStringListAdd(&argv, &argc, "-w");
 		CCStringListAdd(&argv, &argc, "-nostdinc");
 		CCStringListAdd(&argv, &argc, "-nostdinc++");
@@ -797,13 +827,11 @@ static void ParseFile(const char *filename, int is_stdin)
 		}
 		for(posn = 0; posn < num_include_dirs; ++posn)
 		{
-			CCStringListAdd(&argv, &argc, "-I");
-			CCStringListAdd(&argv, &argc, include_dirs[posn]);
+			CCStringListAddOption(&argv, &argc, "-I", include_dirs[posn]);
 		}
 		for(posn = 0; posn < num_sys_include_dirs; ++posn)
 		{
-			CCStringListAdd(&argv, &argc, "-I");
-			CCStringListAdd(&argv, &argc, sys_include_dirs[posn]);
+			CCStringListAddOption(&argv, &argc, "-I", sys_include_dirs[posn]);
 		}
 		for(posn = 0; posn < num_imacros_files; ++posn)
 		{
@@ -826,24 +854,20 @@ static void ParseFile(const char *filename, int is_stdin)
 									 num_user_defined_symbols,
 									 ILCppUndefines[posn]))
 			{
-				CCStringListAdd(&argv, &argc, "-U");
-				CCStringListAdd(&argv, &argc, ILCppUndefines[posn]);
+				CCStringListAddOption(&argv, &argc, "-U", ILCppUndefines[posn]);
 			}
 		}
 		for(posn = 0; posn < num_pre_defined_symbols; ++posn)
 		{
-			CCStringListAdd(&argv, &argc, "-D");
-			CCStringListAdd(&argv, &argc, pre_defined_symbols[posn]);
+			CCStringListAddOption(&argv, &argc, "-D", pre_defined_symbols[posn]);
 		}
 		for(posn = 0; posn < num_user_defined_symbols; ++posn)
 		{
-			CCStringListAdd(&argv, &argc, "-D");
-			CCStringListAdd(&argv, &argc, user_defined_symbols[posn]);
+			CCStringListAddOption(&argv, &argc, "-D", user_defined_symbols[posn]);
 		}
 		for(posn = 0; posn < num_undefined_symbols; ++posn)
 		{
-			CCStringListAdd(&argv, &argc, "-U");
-			CCStringListAdd(&argv, &argc, undefined_symbols[posn]);
+			CCStringListAddOption(&argv, &argc, "-U", undefined_symbols[posn]);
 		}
 		if(is_stdin)
 		{
