@@ -1,7 +1,7 @@
 /*
  * cc_options.c - Command-line option processing.
  *
- * Copyright (C) 2001  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2001, 2002, 2003  Southern Storm Software, Pty Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -330,6 +330,16 @@ static void gnomeOption(char *arg)
 	gtkOption(arg);
 }
 
+/*
+ * Process a -winforms option.
+ */
+static void winformsOption(char *arg)
+{
+	lOption("System");
+	lOption("System.Drawing");
+	lOption("System.Windows.Forms");
+}
+
 /*
  * Process a -f option.
  */
@@ -383,6 +393,26 @@ static void xOption(char *arg)
 	{
 		prog_language = PROG_LANG_C;
 	}
+	else if(!strcmp(arg, "vb"))
+	{
+		prog_language = PROG_LANG_VB;
+	}
+	else if(!strcmp(arg, "visualbasic"))
+	{
+		prog_language = PROG_LANG_VB;
+	}
+	else if(!strcmp(arg, "java"))
+	{
+		prog_language = PROG_LANG_JAVA;
+	}
+	else if(!strcmp(arg, "bf"))
+	{
+		prog_language = PROG_LANG_BF;
+	}
+	else if(!strcmp(arg, "brainfuck"))
+	{
+		prog_language = PROG_LANG_BF;
+	}
 	else if(!strcmp(arg, "none"))
 	{
 		prog_language = PROG_LANG_DEFAULT;
@@ -477,6 +507,8 @@ static CmdLineOpt const options[] = {
 			N_("-gtk"), N_("Link against the Gtk# libraries")},
 	{"-gnome",		0,	0,						0,	gnomeOption,
 			N_("-gnome"), N_("Link against the Gnome# libraries")},
+	{"-winforms",		0,	0,						0,	winformsOption,
+			N_("-winforms"), N_("Link against the System.Windows.Forms libraries")},
 	{"-f",			2,	0,						0,	fOption, 0, 0},
 	{"-m",			2,	0,						0,	mOption, 0, 0},
 	{"-Wl,",		4,	0,						0,	0,	0, 0},
@@ -532,6 +564,8 @@ static CmdLineOpt const options[] = {
 			"-v", N_("Print names of source files as they are compiled")},
 	{"-vv",			0,	&verbose_mode,			VERBOSE_CMDLINES,	0,
 			"-vv", N_("Print command-lines of tools as they are executed")},
+	{"--version",0,	0,							0,	vOption,
+			"--version", N_("output version information and exit")},
 	{"--help",		0,	0,						0,	helpOption,
 			"--help", N_("Display this information")},
 
@@ -582,6 +616,7 @@ void CCParseCommandLine(int argc, char *argv[], int mode, char *versname)
 	int compat_mode = ((mode & CMDLINE_PARSE_COMPAT) != 0);
 	char *optname;
 	char *value;
+	int outlen;
 	mode &= ~(CMDLINE_PARSE_PLUGIN_NOSTDLIB | CMDLINE_PARSE_COMPAT);
 
 	/* Get the translated out of memory message */
@@ -857,6 +892,24 @@ done_compat:
 		pre_defined_symbols = 0;
 		num_pre_defined_symbols = 0;
 	}
+
+	/* Force "-fstdlib-name=corlib" if compiling "corlib.dll" or if
+	   "-lcorlib" is supplied on the command-line.  Needed for
+	   interoperability with Mono build scripts */
+	outlen = (output_filename ? strlen(output_filename) : 0);
+	if(output_filename != 0 && outlen >= 10 &&
+	   !ILStrICmp(output_filename + outlen - 10, "corlib.dll") &&
+	   (outlen == 10 || output_filename[outlen - 11] == '/' ||
+	    output_filename[outlen - 11] == '\\'))
+	{
+		AddString(&extension_flags, &num_extension_flags,
+				  "stdlib-name=corlib");
+	}
+	else if(CCStringListContains(libraries, num_libraries, "corlib"))
+	{
+		AddString(&extension_flags, &num_extension_flags,
+				  "stdlib-name=corlib");
+	}
 }
 
 /*
@@ -996,6 +1049,14 @@ void CCStringListAdd(char ***list, int *num, char *str)
 	AddString(list, num, str);
 }
 
+void CCStringListAddOption(char ***list, int *num, char *option, char *str)
+{
+	char * concat = malloc(strlen(str) + 3);
+	strcpy(concat, option);
+	strcat(concat, str);
+	AddString(list, num, concat);
+}
+
 #ifdef	__cplusplus
 };
 #endif
