@@ -53,6 +53,9 @@ void CCPreProcInit(CCPreProc *preproc, FILE *stream, const char *filename,
 	preproc->lineNumber = 1;
 	preproc->filename =
 		(filename ? (ILInternString((char *)filename, -1)).string : 0);
+	preproc->defaultLinenumber = 1;
+	preproc->defaultFilename =
+		(filename ? (ILInternString((char *)filename, -1)).string : 0);
 	preproc->currentScope = 0;
 	preproc->reportedUnmatched = 0;
 	preproc->lexerLineNumber = 1;
@@ -218,6 +221,7 @@ static int GetLines(CCPreProc *preproc)
 	int comment, scomment;
 	int dcomment, quote;
 	int lineStart, sawEOF;
+	int literal;
 	int directive;
 
 	/* Bail out early if we have already seen EOF */
@@ -237,6 +241,7 @@ static int GetLines(CCPreProc *preproc)
 	scomment = 0;
 	dcomment = 0;
 	quote = 0;
+	literal = 0;
 	lineStart = bufLen;
 	directive = 0;
 	preproc->numLines = 0;
@@ -468,24 +473,44 @@ static int GetLines(CCPreProc *preproc)
 			{
 				/* End of the string literal */
 				ADD_CH(ch);
-				quote = 0;
-			}
-			else if(ch == '\\')
-			{
-				/* Escape sequence of some type */
-				ADD_CH('\\');
-				ch = getc(stream);
-				if(ch == '"' || ch == '\'' || ch == '\\')
+				if(literal)
 				{
-					ADD_CH(ch);
-				}
-				else if(ch != EOF)
-				{
-					ungetc(ch, stream);
+					ch=getc(stream);	
+					if(ch==quote)
+					{
+						ADD_CH(quote);
+					}
+					else
+					{
+						ungetc(ch,stream);
+						quote=0;
+						literal=0;
+					}
 				}
 				else
 				{
-					sawEOF = 1;
+					quote = 0;
+				}
+			}
+			else if(ch == '\\')
+			{
+				ADD_CH('\\');
+				if(!literal)
+				{
+					/* Escape sequence of some type */
+					ch = getc(stream);
+					if(ch == '"' || ch == '\'' || ch == '\\')
+					{
+						ADD_CH(ch);
+					}
+					else if(ch != EOF)
+					{
+						ungetc(ch, stream);
+					}
+					else
+					{
+						sawEOF = 1;
+					}
 				}
 			}
 			else
@@ -558,6 +583,22 @@ static int GetLines(CCPreProc *preproc)
 			directive = 1;
 			ADD_CH(ch);
 		}
+		else if(ch == '@')
+		{
+			/* This may be the start of a literal string */
+			ADD_CH(ch);
+			ch = getc(stream);
+			if(ch == '"' || ch == '\'')
+			{
+				quote = ch;
+				literal = 1;
+				ADD_CH(ch);
+			}
+			else
+			{
+				ungetc(ch, stream);
+			}
+		}
 		else
 		{
 			/* Normal character: add it to the current line */
@@ -1084,6 +1125,7 @@ static int RefillLineBuffer(CCPreProc *preproc)
 		/* Set the line number information for this line */
 		lines[line].number = (preproc->lineNumber)++;
 		lines[line].filename = preproc->filename;
+		preproc->defaultLinenumber++;
 
 		/* Is this a directive? */
 		if(lines[line].directive &&
@@ -1298,6 +1340,13 @@ static int RefillLineBuffer(CCPreProc *preproc)
 					}
 					CheckAtEnd(preproc, &(lines[line]), &dirname);
 				}
+				else if(!strncmp(dirname,"default",7))
+				{
+					preproc->lineNumber = preproc->defaultLinenumber;
+					preproc->filename = preproc->defaultFilename;
+					dirname+=7;
+					CheckAtEnd(preproc, &(lines[line]), &dirname);
+				}
 				else
 				{
 					Message(&(lines[line]),
