@@ -81,7 +81,23 @@ static void AddMember(CSMemberLookupInfo *results,
 					  ILProgramItem *member, ILClass *owner,
 					  int kind)
 {
-	CSMemberInfo *info = (CSMemberInfo *)ILMalloc(sizeof(CSMemberInfo));
+	CSMemberInfo *info;
+
+	/* Check to make sure that the member isn't already in the list.
+	   This can happen when the same member is located along different
+	   paths in an interface inheritance hierarchy */
+	info = results->members;
+	while(info != 0)
+	{
+		if(info->member == member)
+		{
+			return;
+		}
+		info = info->next;
+	}
+
+	/* Add the new member to the list */
+	info = (CSMemberInfo *)ILMalloc(sizeof(CSMemberInfo));
 	if(!info)
 	{
 		CCOutOfMemory();
@@ -316,10 +332,10 @@ static void FindMembers(ILGenInfo *genInfo, ILClass *info,
 		if(!(genInfo->inSemType))
 		{
 			member = 0;
-			while((member = ILClassNextMember(info, member)) != 0)
+			while((member = ILClassNextMemberMatch
+						(info, member, 0, name, 0)) != 0)
 			{
-				if(!strcmp(ILMember_Name(member), name) &&
-				   ILMemberAccessible(member, accessedFrom))
+				if(ILMemberAccessible(member, accessedFrom))
 				{
 					kind = ILMemberGetKind(member);
 					if(literalType && kind != CS_MEMBERKIND_TYPE)
@@ -660,7 +676,7 @@ static int MemberLookup(ILGenInfo *genInfo, ILClass *info,
  * Process a class and add all indexers to a set of member lookup results.
  */
 static void FindIndexers(ILClass *info, ILClass *accessedFrom,
-					     CSMemberLookupInfo *results)
+					     CSMemberLookupInfo *results, int baseAccess)
 {
 	ILImplements *impl;
 	ILMember *member;
@@ -691,25 +707,38 @@ static void FindIndexers(ILClass *info, ILClass *accessedFrom,
 					/* Static indexers are not legal, so skip them */
 					continue;
 				}
-				if(ILMethod_IsVirtual(method) && !ILMethod_IsNewSlot(method))
+				if((!baseAccess) && 
+					ILMethod_IsVirtual(method) && !ILMethod_IsNewSlot(method))
 				{
 					/* This is a virtual override, so skip it */
 					continue;
 				}
+				if(ILOverrideFromMethod(method) != 0)
+				{
+					/* Ignore private interface overrides for now, so that
+					   we can pick up a non-interface indexer instead if
+					   one exists.  If one doesn't exist, then we will
+					   re-find the interface indexer when we scan the
+					   base interfaces below.  This situation occurs with
+					   code like this:
+
+					        X this[int i] { ... }
+							X I.this[int i] { ... }
+					*/
+					continue;
+				}
 				AddMember(results, (ILProgramItem *)member, info,
 						  IL_META_MEMBERKIND_PROPERTY);
 			}
 		}
 
-		/* If this is an interface, then scan its base interfaces */
-		if(ILClass_IsInterface(info))
+		/* Scan the base interfaces.  Duplicate entries will be cleaned
+		   up by the TrimMemberList call in "IndexerLookup" */
+		impl = 0;
+		while((impl = ILClassNextImplements(info, impl)) != 0)
 		{
-			impl = 0;
-			while((impl = ILClassNextImplements(info, impl)) != 0)
-			{
-				FindIndexers(ILImplementsGetInterface(impl),
-						     accessedFrom, results);
-			}
+			FindIndexers(ILImplementsGetInterface(impl),
+					     accessedFrom, results, baseAccess);
 		}
 
 		/* Move up to the parent */
@@ -721,7 +750,8 @@ static void FindIndexers(ILClass *info, ILClass *accessedFrom,
  * Perform an indexer lookup on a type.
  */
 static int IndexerLookup(ILGenInfo *genInfo, ILClass *info,
-				         ILClass *accessedFrom, CSMemberLookupInfo *results)
+				         ILClass *accessedFrom, CSMemberLookupInfo *results,
+						 int baseAccess)
 {
 	/* Initialize the results */
 	InitMembers(results);
@@ -729,7 +759,7 @@ static int IndexerLookup(ILGenInfo *genInfo, ILClass *info,
 	/* Collect up all indexers */
 	if(info)
 	{
-		FindIndexers(info, accessedFrom, results);
+		FindIndexers(info, accessedFrom, results, baseAccess);
 	}
 
 	/* Trim the list and determine the kind for the result */
@@ -791,13 +821,17 @@ static void AmbiguousError(ILNode *node, const char *name,
 
 			case IL_META_MEMBERKIND_METHOD:
 			{
-				/* TODO */
+				typeName = CSItemToName((ILProgramItem *)(member->member));
+				CCErrorOnLine(yygetfilename(node), yygetlinenum(node),
+				              "  method %s", typeName);
 			}
 			break;
 
 			case IL_META_MEMBERKIND_PROPERTY:
 			{
-				/* TODO */
+				typeName = CSItemToName((ILProgramItem *)(member->member));
+				CCErrorOnLine(yygetfilename(node), yygetlinenum(node),
+				              "  property %s", typeName);
 			}
 			break;
 
@@ -960,239 +994,6 @@ ILClass *CSGetAccessScope(ILGenInfo *genInfo, int defIsModule)
 	}
 }
 
-CSSemValue CSResolveSimpleName(ILGenInfo *genInfo, ILNode *node,
-										const char *name, int literalType)
-{
-	ILClass *startType;
-	ILClass *accessedFrom;
-	CSMemberLookupInfo results;
-	ILNode_Namespace *namespace;
-	ILNode_UsingNamespace *using;
-	int result;
-	ILNode_ClassDefn *nestedParent;
-	ILNode *child;
-
-	/* If we are within type gathering, then search the nesting
-	   parents for a nested type that matches our requirements */
-	if(genInfo->typeGather)
-	{
-		nestedParent = (ILNode_ClassDefn *)(genInfo->currentClass);
-		while(nestedParent != 0)
-		{
-			child = FindNestedClass(0, nestedParent, name);
-			if(child)
-			{
-				CSSemValue value;
-				CSSemSetTypeNode(value, child);
-				return value;
-			}
-			nestedParent = nestedParent->nestedParent;
-		}
-	}
-
-	/* Find the type to start looking at and the scope to use for accesses */
-	startType = CSGetAccessScope(genInfo, 0);
-	accessedFrom = ILClassResolve(CSGetAccessScope(genInfo, 1));
-
-	/* Scan the start type and its nested parents */
-	while(startType != 0)
-	{
-		/* Resolve cross-image references */
-		startType = ILClassResolve(startType);
-
-		/* Look for members */
-		result = MemberLookup(genInfo, startType, name,
-							  accessedFrom, &results, 1, 0, literalType);
-		if(result != CS_SEMKIND_VOID)
-		{
-			return LookupToSem(node, name, &results, result);
-		}
-
-		/* Move up to the nested parent */
-		startType = ILClass_NestedParent(startType);
-	}
-
-	/* Clear the results buffer */
-	InitMembers(&results);
-
-	/* Scan all namespaces that enclose the current context */
-	namespace = (ILNode_Namespace *)(genInfo->currentNamespace);
-	while(namespace != 0 && !(results.num))
-	{
-		/* Look for the type in the current namespace */
-		result = FindTypeInNamespace(genInfo, name, namespace->name,
-									 accessedFrom, &results);
-		if(result != CS_SEMKIND_VOID)
-		{
-			break;
-		}
-	
-		/* Find the types in all using namespaces */
-		using = namespace->using;
-		while(using != 0)
-		{
-			FindTypeInNamespace(genInfo, name, using->name,
-								accessedFrom, &results);
-			using = using->next;
-		}
-
-		/* Move up to the enclosing namespace */
-		namespace = namespace->enclosing;
-	}
-
-	/* We should have 0, 1, or many types at this point */
-	if(results.num > 1)
-	{
-		/* The result is ambiguous */
-		AmbiguousError(node, name, &results);
-	}
-	if(results.num != 0)
-	{
-		/* Return the first type in the results list */
-		if(results.members->kind == CS_MEMBERKIND_TYPE)
-		{
-			return LookupToSem(node, name, &results, CS_SEMKIND_TYPE);
-		}
-		else if(results.members->kind == CS_MEMBERKIND_TYPE_NODE)
-		{
-			return LookupToSem(node, name, &results, CS_SEMKIND_TYPE_NODE);
-		}
-		else
-		{
-			return LookupToSem(node, name, &results, CS_SEMKIND_NAMESPACE);
-		}
-	}
-	FreeMembers(&results);
-
-	/* Could not resolve the name */
-	CCErrorOnLine(yygetfilename(node), yygetlinenum(node),
-				  "`%s' is not declared in the current scope", name);
-
-	if (literalType)
-	{
-		/* Resolve it cleanly if a type was not found */
-		return CSResolveSimpleName(genInfo, node, name, 0);
-	}
-	else
-	{
-		return CSSemValueDefault;
-	}
-}
-
-CSSemValue CSResolveSimpleNameQuiet(ILGenInfo *genInfo, ILNode *node,
-										const char *name, int literalType)
-{
-	ILClass *startType;
-	ILClass *accessedFrom;
-	CSMemberLookupInfo results;
-	ILNode_Namespace *namespace;
-	ILNode_UsingNamespace *using;
-	int result;
-	ILNode_ClassDefn *nestedParent;
-	ILNode *child;
-
-	/* If we are within type gathering, then search the nesting
-	   parents for a nested type that matches our requirements */
-	if(genInfo->typeGather)
-	{
-		nestedParent = (ILNode_ClassDefn *)(genInfo->currentClass);
-		while(nestedParent != 0)
-		{
-			child = FindNestedClass(0, nestedParent, name);
-			if(child)
-			{
-				CSSemValue value;
-				CSSemSetTypeNode(value, child);
-				return value;
-			}
-			nestedParent = nestedParent->nestedParent;
-		}
-	}
-
-	/* Find the type to start looking at and the scope to use for accesses */
-	startType = CSGetAccessScope(genInfo, 0);
-	accessedFrom = ILClassResolve(CSGetAccessScope(genInfo, 1));
-
-	/* Scan the start type and its nested parents */
-	while(startType != 0)
-	{
-		/* Resolve cross-image references */
-		startType = ILClassResolve(startType);
-
-		/* Look for members */
-		result = MemberLookup(genInfo, startType, name,
-							  accessedFrom, &results, 1, 0, literalType);
-		if(result != CS_SEMKIND_VOID)
-		{
-			return LookupToSem(node, name, &results, result);
-		}
-
-		/* Move up to the nested parent */
-		startType = ILClass_NestedParent(startType);
-	}
-
-	/* Clear the results buffer */
-	InitMembers(&results);
-
-	/* Scan all namespaces that enclose the current context */
-	namespace = (ILNode_Namespace *)(genInfo->currentNamespace);
-	while(namespace != 0 && !(results.num))
-	{
-		/* Look for the type in the current namespace */
-		result = FindTypeInNamespace(genInfo, name, namespace->name,
-									 accessedFrom, &results);
-		if(result != CS_SEMKIND_VOID)
-		{
-			break;
-		}
-	
-		/* Find the types in all using namespaces */
-		using = namespace->using;
-		while(using != 0)
-		{
-			FindTypeInNamespace(genInfo, name, using->name,
-								accessedFrom, &results);
-			using = using->next;
-		}
-
-		/* Move up to the enclosing namespace */
-		namespace = namespace->enclosing;
-	}
-
-	/* We should have 0, 1, or many types at this point */
-	if(results.num > 1)
-	{
-		/* The result is ambiguous */
-		AmbiguousError(node, name, &results);
-	}
-	if(results.num != 0)
-	{
-		/* Return the first type in the results list */
-		if(results.members->kind == CS_MEMBERKIND_TYPE)
-		{
-			return LookupToSem(node, name, &results, CS_SEMKIND_TYPE);
-		}
-		else if(results.members->kind == CS_MEMBERKIND_TYPE_NODE)
-		{
-			return LookupToSem(node, name, &results, CS_SEMKIND_TYPE_NODE);
-		}
-		else
-		{
-			return LookupToSem(node, name, &results, CS_SEMKIND_NAMESPACE);
-		}
-	}
-	FreeMembers(&results);
-
-	if (literalType)
-	{
-		/* Resolve it cleanly if a type was not found */
-		return CSResolveSimpleNameQuiet(genInfo, node, name, 0);
-	}
-	else
-	{
-		return CSSemValueDefault;
-	}
-}
 /*
  * Filter a member lookup results list to include only static entries.
  */
@@ -1356,6 +1157,201 @@ static int FilterNonStatic(CSMemberLookupInfo *results, int kind)
 	return kind;
 }
 
+/*
+ * Inner version of CSResolveSimpleName and CSResolveSimpleNameQuiet.
+ */
+static CSSemValue ResolveSimpleName(ILGenInfo *genInfo, ILNode *node,
+							   		const char *name, int literalType,
+									int reportErrors)
+{
+	ILClass *startType;
+	ILClass *accessedFrom;
+	CSMemberLookupInfo results;
+	ILNode_Namespace *namespace;
+	ILNode_UsingNamespace *using;
+	int result;
+	ILNode_ClassDefn *nestedParent;
+	ILNode *child;
+	ILNode_ListIter iter;
+	int formalNum;
+	ILType *formalType;
+	CSSemValue value;
+	ILMethod *caller;
+
+	/* If we are within type gathering, then search the nesting
+	   parents for a nested type that matches our requirements */
+	if(genInfo->typeGather)
+	{
+		nestedParent = (ILNode_ClassDefn *)(genInfo->currentClass);
+		while(nestedParent != 0)
+		{
+			child = FindNestedClass(0, nestedParent, name);
+			if(child)
+			{
+				CSSemSetTypeNode(value, child);
+				return value;
+			}
+			nestedParent = nestedParent->nestedParent;
+		}
+	}
+
+	/* Scan the method formals for a match */
+	ILNode_ListIter_Init(&iter, genInfo->currentMethodFormals);
+	formalNum = 0;
+	while((child = ILNode_ListIter_Next(&iter)) != 0)
+	{
+		if(!strcmp(ILQualIdentName(child, 0), name))
+		{
+			formalType = ILTypeCreateVarNum
+				(genInfo->context, IL_TYPE_COMPLEX_MVAR, formalNum);
+			if(!formalType)
+			{
+				CCOutOfMemory();
+			}
+			CSSemSetType(value, formalType);
+			return value;
+		}
+		++formalNum;
+	}
+
+	/* Find the type to start looking at and the scope to use for accesses */
+	startType = CSGetAccessScope(genInfo, 0);
+	accessedFrom = ILClassResolve(CSGetAccessScope(genInfo, 1));
+
+	/* Scan the start type and its nested parents */
+	/* Note: do not lookup class members while resolving the simple names
+	 * inside an attribute argument */
+	while(startType != 0 && !genInfo->inAttrArg)
+	{
+		/* Resolve cross-image references */
+		startType = ILClassResolve(startType);
+
+		/* Look for members */
+		result = MemberLookup(genInfo, startType, name,
+							  accessedFrom, &results, 1, 0, literalType);
+		if(result != CS_SEMKIND_VOID)
+		{
+			if(genInfo->currentMethod && !reportErrors)
+			{
+				caller = ((ILNode_MethodDeclaration *)(genInfo->currentMethod))
+							->methodInfo;
+				if(caller && ILMethod_IsStatic(caller))
+				{
+					/* We are in a static context, so filter out
+					   non-static members */
+					result = FilterStatic(&results, result);
+				}
+			}
+		}
+		if(result != CS_SEMKIND_VOID)
+		{
+			return LookupToSem(node, name, &results, result);
+		}
+
+		/* Move up to the nested parent */
+		startType = ILClass_NestedParent(startType);
+	}
+
+	/* Clear the results buffer */
+	InitMembers(&results);
+
+	/* Scan the type formals for a match */
+	ILNode_ListIter_Init(&iter, genInfo->currentTypeFormals);
+	formalNum = 0;
+	while((child = ILNode_ListIter_Next(&iter)) != 0)
+	{
+		if(!strcmp(ILQualIdentName(child, 0), name))
+		{
+			formalType = ILTypeCreateVarNum
+				(genInfo->context, IL_TYPE_COMPLEX_VAR, formalNum);
+			if(!formalType)
+			{
+				CCOutOfMemory();
+			}
+			CSSemSetType(value, formalType);
+			return value;
+		}
+		++formalNum;
+	}
+
+	/* Scan all namespaces that enclose the current context */
+	namespace = (ILNode_Namespace *)(genInfo->currentNamespace);
+	while(namespace != 0 && !(results.num))
+	{
+		/* Look for the type in the current namespace */
+		result = FindTypeInNamespace(genInfo, name, namespace->name,
+									 accessedFrom, &results);
+		if(result != CS_SEMKIND_VOID)
+		{
+			break;
+		}
+	
+		/* Find the types in all using namespaces */
+		using = namespace->using;
+		while(using != 0)
+		{
+			FindTypeInNamespace(genInfo, name, using->name,
+								accessedFrom, &results);
+			using = using->next;
+		}
+
+		/* Move up to the enclosing namespace */
+		namespace = namespace->enclosing;
+	}
+
+	/* We should have 0, 1, or many types at this point */
+	if(results.num > 1)
+	{
+		/* The result is ambiguous */
+		AmbiguousError(node, name, &results);
+	}
+	if(results.num != 0)
+	{
+		/* Return the first type in the results list */
+		if(results.members->kind == CS_MEMBERKIND_TYPE)
+		{
+			return LookupToSem(node, name, &results, CS_SEMKIND_TYPE);
+		}
+		else if(results.members->kind == CS_MEMBERKIND_TYPE_NODE)
+		{
+			return LookupToSem(node, name, &results, CS_SEMKIND_TYPE_NODE);
+		}
+		else
+		{
+			return LookupToSem(node, name, &results, CS_SEMKIND_NAMESPACE);
+		}
+	}
+	FreeMembers(&results);
+
+	/* Could not resolve the name */
+	if(reportErrors)
+	{
+		CCErrorOnLine(yygetfilename(node), yygetlinenum(node),
+					  "`%s' is not declared in the current scope", name);
+	}
+	if (literalType)
+	{
+		/* Resolve it cleanly if a type was not found */
+		return ResolveSimpleName(genInfo, node, name, 0, reportErrors);
+	}
+	else
+	{
+		return CSSemValueDefault;
+	}
+}
+
+CSSemValue CSResolveSimpleName(ILGenInfo *genInfo, ILNode *node,
+							   const char *name, int literalType)
+{
+	return ResolveSimpleName(genInfo, node, name, literalType, 1);
+}
+
+CSSemValue CSResolveSimpleNameQuiet(ILGenInfo *genInfo, ILNode *node,
+									const char *name, int literalType)
+{
+	return ResolveSimpleName(genInfo, node, name, literalType, 0);
+}
+
 /*
  * Look for a type or sub-namespace with a given name within the 
  * namespace.
@@ -1365,15 +1361,40 @@ CSSemValue CSResolveNamespaceMemberName(ILGenInfo *genInfo,
 {
 	CSMemberLookupInfo results;
 	int result;
+	ILIntString nspace;
 
 	InitMembers(&results);
+
+	/* Look for a type first */
 	result = FindTypeInNamespace(genInfo, name,
-			CSSemGetNamespace(value), 
-			ILClassResolve(CSGetAccessScope(genInfo, 1)),
-			&results);
-			
-	return result == CS_SEMKIND_VOID ?
-		CSSemValueDefault : LookupToSem(node, name, &results, result);
+								 CSSemGetNamespace(value), 
+								 ILClassResolve(CSGetAccessScope(genInfo, 1)),
+								 &results);
+
+	/* See if the namespace is valid, according to the loaded libraries */
+	if(result == CS_SEMKIND_VOID)
+	{
+		nspace = ILInternAppendedString
+			(ILInternString(CSSemGetNamespace(value), -1),
+			 ILInternString(".", 1));
+		nspace = ILInternAppendedString
+			(nspace, ILInternString((char *)name, -1));
+		if(ILClassNamespaceIsValid(genInfo->context, nspace.string))
+		{
+			CSSemSetNamespace(value, nspace.string);
+			return value;
+		}
+	}
+
+	/* Return the result to the caller */
+	if(result == CS_SEMKIND_VOID)
+	{
+		return CSSemValueDefault;
+	}
+	else
+	{
+		return LookupToSem(node, name, &results, result);
+	}
 }
 
 /*
@@ -1637,7 +1658,7 @@ CSSemValue CSResolveConstructor(ILGenInfo *genInfo, ILNode *node,
 }
 
 CSSemValue CSResolveIndexers(ILGenInfo *genInfo, ILNode *node,
-							 ILClass *classInfo)
+							 ILClass *classInfo, int baseAccess)
 {
 	CSSemValue value;
 	ILClass *accessedFrom;
@@ -1648,7 +1669,8 @@ CSSemValue CSResolveIndexers(ILGenInfo *genInfo, ILNode *node,
 	accessedFrom = ILClassResolve(CSGetAccessScope(genInfo, 1));
 
 	/* Perform a member lookup based on the class */
-	result = IndexerLookup(genInfo, classInfo, accessedFrom, &results);
+	result = IndexerLookup(genInfo, classInfo, accessedFrom, &results, 
+							baseAccess);
 	if(result != CS_SEMKIND_VOID)
 	{
 		CSSemSetKind(value, result, results.members);
