@@ -109,7 +109,7 @@ ILCallFrame *_ILAllocCallFrame(ILExecThread *thread)
 #define	REPORT_METHOD_CALL()	\
 			do { \
 				fprintf(IL_DUMP_CVM_STREAM, "Entering %s::%s (%ld)\n", \
-					    methodToCall->member.owner->name, \
+					    methodToCall->member.owner->className->name, \
 					    methodToCall->member.name, \
 					    (long)(stacktop - thread->stackBase)); \
 				DUMP_STACK(); \
@@ -439,6 +439,7 @@ VMCASE(COP_CALL):
 		callFrame->pc = CVM_ARG_CALL_RETURN(pc);
 		callFrame->frame = frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Pass control to the new method */
 		pc = (unsigned char *)(methodToCall->userData);
@@ -465,6 +466,7 @@ VMCASE(COP_CALL):
 		callFrame->pc = CVM_ARG_CALL_RETURN(thread->pc);
 		callFrame->frame = thread->frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Restore the state information and jump to the new method */
 		RESTORE_STATE_FROM_THREAD();
@@ -472,9 +474,6 @@ VMCASE(COP_CALL):
 		method = methodToCall;
 		CVM_OPTIMIZE_BLOCK();
 	}
-#ifdef IL_PROFILE_CVM_METHODS
-	++(method->count);
-#endif
 }
 VMBREAK(COP_CALL);
 
@@ -524,6 +523,7 @@ VMCASE(COP_CALL_CTOR):
 		callFrame->pc = CVM_ARG_CALL_RETURN(pc);
 		callFrame->frame = frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Pass control to the new method */
 		pc = ((unsigned char *)(methodToCall->userData)) - CVM_CTOR_OFFSET;
@@ -550,6 +550,7 @@ VMCASE(COP_CALL_CTOR):
 		callFrame->pc = CVM_ARG_CALL_RETURN(thread->pc);
 		callFrame->frame = thread->frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Restore the state information and jump to the new method */
 		RESTORE_STATE_FROM_THREAD();
@@ -557,9 +558,6 @@ VMCASE(COP_CALL_CTOR):
 		method = methodToCall;
 		CVM_OPTIMIZE_BLOCK();
 	}
-#ifdef IL_PROFILE_CVM_METHODS
-	++(method->count);
-#endif
 }
 VMBREAK(COP_CALL_CTOR);
 
@@ -734,7 +732,7 @@ VMCASE(COP_CALL_VIRTUAL):
 	if(tempptr)
 	{
 		/* Locate the method to be called */
-		methodToCall = ((ILClassPrivate *)(GetObjectClass(tempptr)->userData))
+		methodToCall = (GetObjectClassPrivate(tempptr))
 							->vtable[CVM_ARG_DWIDE2_SMALL];
 
 		/* Has the method already been converted? */
@@ -748,6 +746,7 @@ VMCASE(COP_CALL_VIRTUAL):
 			callFrame->pc = CVM_ARG_CALLV_RETURN_SMALL(pc);
 			callFrame->frame = frame;
 			callFrame->exceptHeight = thread->exceptHeight;
+			callFrame->permissions = 0;
 
 			/* Pass control to the new method */
 			pc = (unsigned char *)(methodToCall->userData);
@@ -774,6 +773,7 @@ VMCASE(COP_CALL_VIRTUAL):
 			callFrame->pc = CVM_ARG_CALLV_RETURN_SMALL(thread->pc);
 			callFrame->frame = thread->frame;
 			callFrame->exceptHeight = thread->exceptHeight;
+			callFrame->permissions = 0;
 
 			/* Restore the state information and jump to the new method */
 			RESTORE_STATE_FROM_THREAD();
@@ -781,9 +781,6 @@ VMCASE(COP_CALL_VIRTUAL):
 			method = methodToCall;
 			CVM_OPTIMIZE_BLOCK();
 		}
-	#ifdef IL_PROFILE_CVM_METHODS
-		++(method->count);
-	#endif
 	}
 	else
 	{
@@ -831,13 +828,29 @@ VMCASE(COP_CALL_INTERFACE):
 	if(tempptr)
 	{
 		/* Locate the method to be called */
+	#ifdef IL_USE_IMTS
+		methodToCall = GetObjectClassPrivate(tempptr)
+			->imt[CVM_ARG_DWIDE2_SMALL];
+		if(!methodToCall)
+		{
+			methodToCall = CVM_ARG_DWIDE_PTR_SMALL(ILMethod *);
+			methodToCall = _ILLookupInterfaceMethod
+				(GetObjectClassPrivate(tempptr),
+				 methodToCall->member.owner, methodToCall->index);
+			if(!methodToCall)
+			{
+				MISSING_METHOD_EXCEPTION();
+			}
+		}
+	#else
 		methodToCall = _ILLookupInterfaceMethod
-			(GetObjectClass(tempptr), CVM_ARG_DWIDE_PTR_SMALL(ILClass *),
+			(GetObjectClassPrivate(tempptr), CVM_ARG_DWIDE_PTR_SMALL(ILClass *),
 			 CVM_ARG_DWIDE2_SMALL);
 		if(!methodToCall)
 		{
 			MISSING_METHOD_EXCEPTION();
 		}
+	#endif
 
 		/* Has the method already been converted? */
 		if(methodToCall->userData)
@@ -850,6 +863,7 @@ VMCASE(COP_CALL_INTERFACE):
 			callFrame->pc = CVM_ARG_CALLI_RETURN_SMALL(pc);
 			callFrame->frame = frame;
 			callFrame->exceptHeight = thread->exceptHeight;
+			callFrame->permissions = 0;
 
 			/* Pass control to the new method */
 			pc = (unsigned char *)(methodToCall->userData);
@@ -876,6 +890,7 @@ VMCASE(COP_CALL_INTERFACE):
 			callFrame->pc = CVM_ARG_CALLI_RETURN_SMALL(thread->pc);
 			callFrame->frame = thread->frame;
 			callFrame->exceptHeight = thread->exceptHeight;
+			callFrame->permissions = 0;
 
 			/* Restore the state information and jump to the new method */
 			RESTORE_STATE_FROM_THREAD();
@@ -883,9 +898,6 @@ VMCASE(COP_CALL_INTERFACE):
 			method = methodToCall;
 			CVM_OPTIMIZE_BLOCK();
 		}
-	#ifdef IL_PROFILE_CVM_METHODS
-		++(method->count);
-	#endif
 	}
 	else
 	{
@@ -977,7 +989,7 @@ popFrame:
 #ifdef IL_DUMP_CVM
 	/* Dump the name of the method we are returning to */
 	fprintf(IL_DUMP_CVM_STREAM, "Returning to %s::%s (%ld)\n",
-			method->member.owner->name,
+			method->member.owner->className->name,
 		    method->member.name, (long)(stacktop - thread->stackBase));
 	DUMP_STACK();
 #endif
@@ -1255,6 +1267,7 @@ VMCASE(COP_CALLI):
 		callFrame->pc = CVM_ARG_CALLIND_RETURN(pc);
 		callFrame->frame = frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Pass control to the new method */
 		pc = (unsigned char *)(methodToCall->userData);
@@ -1281,6 +1294,7 @@ VMCASE(COP_CALLI):
 		callFrame->pc = CVM_ARG_CALLIND_RETURN(thread->pc);
 		callFrame->frame = thread->frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Restore the state information and jump to the new method */
 		RESTORE_STATE_FROM_THREAD();
@@ -1292,9 +1306,6 @@ VMCASE(COP_CALLI):
 	{
 		NULL_POINTER_EXCEPTION();
 	}
-#ifdef IL_PROFILE_CVM_METHODS
-	++(method->count);
-#endif
 }
 VMBREAK(COP_CALLI);
 
@@ -1315,7 +1326,7 @@ case COP_CALL_VIRTUAL:
 	if(tempptr)
 	{
 		/* Locate the method to be called */
-		methodToCall = ((ILClassPrivate *)(GetObjectClass(tempptr)->userData))
+		methodToCall = (GetObjectClassPrivate(tempptr))
 							->vtable[CVM_ARG_DWIDE2_LARGE];
 
 		/* Copy the state back into the thread object */
@@ -1336,15 +1347,13 @@ case COP_CALL_VIRTUAL:
 		callFrame->pc = CVM_ARG_CALLV_RETURN_LARGE(thread->pc);
 		callFrame->frame = thread->frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Restore the state information and jump to the new method */
 		RESTORE_STATE_FROM_THREAD();
 		pc = (unsigned char *)tempptr;
 		method = methodToCall;
 		CVM_OPTIMIZE_BLOCK();
-	#ifdef IL_PROFILE_CVM_METHODS
-		++(method->count);
-	#endif
 	}
 	else
 	{
@@ -1360,13 +1369,29 @@ case COP_CALL_INTERFACE:
 	if(tempptr)
 	{
 		/* Locate the method to be called */
+	#ifdef IL_USE_IMTS
+		methodToCall = GetObjectClassPrivate(tempptr)
+			->imt[CVM_ARG_DWIDE2_LARGE];
+		if(!methodToCall)
+		{
+			methodToCall = CVM_ARG_DWIDE_PTR_LARGE(ILMethod *);
+			methodToCall = _ILLookupInterfaceMethod
+				(GetObjectClassPrivate(tempptr),
+				 methodToCall->member.owner, methodToCall->index);
+			if(!methodToCall)
+			{
+				MISSING_METHOD_EXCEPTION();
+			}
+		}
+	#else
 		methodToCall = _ILLookupInterfaceMethod
-			(GetObjectClass(tempptr), CVM_ARG_DWIDE_PTR_LARGE(ILClass *),
+			(GetObjectClassPrivate(tempptr), CVM_ARG_DWIDE_PTR_LARGE(ILClass *),
 			 CVM_ARG_DWIDE2_LARGE);
 		if(!methodToCall)
 		{
 			MISSING_METHOD_EXCEPTION();
 		}
+	#endif
 
 		/* Copy the state back into the thread object */
 		COPY_STATE_TO_THREAD();
@@ -1386,15 +1411,13 @@ case COP_CALL_INTERFACE:
 		callFrame->pc = CVM_ARG_CALLI_RETURN_LARGE(thread->pc);
 		callFrame->frame = thread->frame;
 		callFrame->exceptHeight = thread->exceptHeight;
+		callFrame->permissions = 0;
 
 		/* Restore the state information and jump to the new method */
 		RESTORE_STATE_FROM_THREAD();
 		pc = (unsigned char *)tempptr;
 		method = methodToCall;
 		CVM_OPTIMIZE_BLOCK();
-	#ifdef IL_PROFILE_CVM_METHODS
-		++(method->count);
-	#endif
 	}
 	else
 	{
@@ -1443,6 +1466,7 @@ VMCASE(COP_PREFIX_TAIL_CALL):
 	/* Retrieve the target method */
 	methodToCall = CVM_ARG_TAIL_METHOD;
 
+performTailCall:
 	/* Convert the method if necessary */
 	if(methodToCall->userData)
 	{
@@ -1471,12 +1495,121 @@ VMCASE(COP_PREFIX_TAIL_CALL):
 	REPORT_METHOD_CALL();
 	pc = (unsigned char *)tempptr;
 	method = methodToCall;
-#ifdef IL_PROFILE_CVM_METHODS
-	++(method->count);
-#endif
 }
 VMBREAK(COP_PREFIX_TAIL_CALL);
 
+/**
+ * <opcode name="tail_calli" group="Call management instructions">
+ *   <operation>Call a method using indirect tail call semantics</operation>
+ *
+ *   <format>prefix<fsep/>tail_calli</format>
+ *   <dformat>{tail_calli}</dformat>
+ *
+ *   <form name="tail_calli" code="COP_PREFIX_TAIL_CALLI"/>
+ *
+ *   <description>This instruction is identical to <i>calli</i>, except
+ *   that it performs a tail-optimized call.</description>
+ * </opcode>
+ */
+VMCASE(COP_PREFIX_TAIL_CALLI):
+{
+	/* Retrieve the target method */
+	methodToCall = (ILMethod *)(stacktop[-1].ptrValue);
+	--stacktop;
+	if(methodToCall)
+	{
+		goto performTailCall;
+	}
+	else
+	{
+		NULL_POINTER_EXCEPTION();
+	}
+}
+VMBREAK(COP_PREFIX_TAIL_CALLI);
+
+/**
+ * <opcode name="tail_callvirt" group="Call management instructions">
+ *   <operation>Call a virtual method using tail call semantics</operation>
+ *
+ *   <format>prefix<fsep/>tail_callvirt<fsep/>N[1]<fsep/>M[1]</format>
+ *   <dformat>{tail_callvirt}<fsep/>N<fsep/>M</dformat>
+ *
+ *   <form name="tail_callvirt" code="COP_PREFIX_TAIL_CALLVIRT"/>
+ *
+ *   <description>The <i>tail_callvirt</i> instruction is identical
+ *   to <i>call_virtual</i>, except that it uses tail call semantics.
+ *   </description>
+ * </opcode>
+ */
+VMCASE(COP_PREFIX_TAIL_CALLVIRT):
+{
+	/* Call a virtual method */
+	tempptr = stacktop[-((ILInt32)CVMP_ARG_WORD)].ptrValue;
+	if(tempptr)
+	{
+		/* Locate the method to be called */
+		methodToCall = (GetObjectClassPrivate(tempptr))->vtable[CVMP_ARG_WORD2];
+		goto performTailCall;
+	}
+	else
+	{
+		NULL_POINTER_EXCEPTION();
+	}
+}
+VMBREAK(COP_PREFIX_TAIL_CALLVIRT);
+
+/**
+ * <opcode name="tail_callintf" group="Call management instructions">
+ *   <operation>Call an interface method using tail call semantics</operation>
+ *
+ *   <format>prefix<fsep/>tail_callintf<fsep/>N[1]<fsep/>M[1]<fsep/>cptr</format>
+ *   <dformat>{tail_callintf}<fsep/>N<fsep/>M<fsep/>cptr</dformat>
+ *
+ *   <form name="tail_callintf" code="COP_PREFIX_TAIL_CALLINTF"/>
+ *
+ *   <description>The <i>tail_callintf</i> instruction is identical
+ *   to <i>call_interface</i>, except that it uses tail call semantics.
+ *   </description>
+ * </opcode>
+ */
+VMCASE(COP_PREFIX_TAIL_CALLINTF):
+{
+	/* Call an interface method */
+	tempptr = stacktop[-((ILInt32)CVMP_ARG_WORD)].ptrValue;
+	if(tempptr)
+	{
+		/* Locate the method to be called */
+	#ifdef IL_USE_IMTS
+		methodToCall = GetObjectClassPrivate(tempptr)->imt[CVMP_ARG_WORD2];
+		if(!methodToCall)
+		{
+			methodToCall = CVMP_ARG_WORD2_PTR(ILMethod *);
+			methodToCall = _ILLookupInterfaceMethod
+				(GetObjectClassPrivate(tempptr),
+				 methodToCall->member.owner, methodToCall->index);
+			if(!methodToCall)
+			{
+				MISSING_METHOD_EXCEPTION();
+			}
+		}
+	#else
+		methodToCall = _ILLookupInterfaceMethod
+			(GetObjectClassPrivate(tempptr), CVMP_ARG_WORD2_PTR(ILClass *),
+			 CVMP_ARG_WORD2);
+		if(!methodToCall)
+		{
+			MISSING_METHOD_EXCEPTION();
+		}
+	#endif
+		goto performTailCall;
+	}
+	else
+	{
+		NULL_POINTER_EXCEPTION();
+	}
+}
+VMBREAK(COP_PREFIX_TAIL_CALLINTF);
+
 /**
  * <opcode name="ldftn" group="Call management instructions">
  *   <operation>Load the address of a function method onto the stack</operation>
@@ -1530,8 +1663,7 @@ VMCASE(COP_PREFIX_LDVIRTFTN):
 	if(tempptr)
 	{
 		stacktop[-1].ptrValue =
-			((ILClassPrivate *)(GetObjectClass(tempptr)->userData))
-					->vtable[CVMP_ARG_WORD];
+			(GetObjectClassPrivate(tempptr))->vtable[CVMP_ARG_WORD];
 		MODIFY_PC_AND_STACK(CVMP_LEN_WORD, 0);
 	}
 	else
@@ -1571,7 +1703,7 @@ VMCASE(COP_PREFIX_LDINTERFFTN):
 	if(tempptr)
 	{
 		stacktop[-1].ptrValue =
-			_ILLookupInterfaceMethod(GetObjectClass(tempptr),
+			_ILLookupInterfaceMethod(GetObjectClassPrivate(tempptr),
 									 CVMP_ARG_WORD_PTR(ILClass *),
 									 CVMP_ARG_WORD);
 		MODIFY_PC_AND_STACK(CVMP_LEN_WORD_PTR, 0);
@@ -1631,4 +1763,51 @@ VMCASE(COP_PREFIX_PACK_VARARGS):
 }
 VMBREAK(COP_PREFIX_PACK_VARARGS);
 
+/**
+ *<opcode name="profile_count" group="Profiling Instructions">
+ *	<operation>Count the number of times the current method is
+ *             invoked</operation>
+ *
+ * 	<format>profile_count</format>
+ * 	<dformat>{profile_count}</dformat>
+ *
+ * 	<form name="profile_count" code="COP_PREFIX_PROFILE_COUNT"/>
+ *
+ *  <description>This instruction adds 1 to the profiling count for
+ *  the current method.  It is normally inserted at the beginning
+ *  of a method's code by the CVM coder when the engine is running
+ *  in the profiling mode.</description>
+ * </opcode>
+ */
+VMCASE(COP_PREFIX_PROFILE_COUNT):
+{
+	++(method->count);
+	MODIFY_PC_AND_STACK(CVMP_LEN_NONE,0);
+}
+VMBREAK(COP_PREFIX_PROFILE_COUNT);
+
+/**
+ * <opcode name="waddr_native_n" group="Call management instructions">
+ *   <operation>Set position <i>n</i> of the native argument buffer
+ *              to the address of a local variable</operation>
+ *
+ *   <format>prefix<fsep/>waddr_native_n<fsep/>N[4]<fsep/>V[4]</format>
+ *   <dformat>{waddr_native_n}<fsep/>N<fsep/>V</dformat>
+ *
+ *   <form name="waddr_native_n" code="COP_PREFIX_WADDR_NATIVE_N"/>
+ *
+ *   <description>Set position <i>N</i> of the native argument buffer
+ *   to the address of local variable <i>V</i>.  For an "InternalCall"
+ *   method, 0 is the first argument.  For a "PInvoke" method,
+ *   -1 (<i>m1</i>) is the first argument.</description>
+ * </opcode>
+ */
+VMCASE(COP_PREFIX_WADDR_NATIVE_N):
+{
+	/* Set a value within the native argument stack */
+	nativeArgs[CVMP_ARG_WORD + 1] = (void *)(&(frame[CVMP_ARG_WORD2]));
+	MODIFY_PC_AND_STACK(CVMP_LEN_WORD2, 0);
+}
+VMBREAK(COP_PREFIX_WADDR_NATIVE_N);
+
 #endif /* IL_CVM_PREFIX */
