@@ -19,6 +19,7 @@
  */
 
 #ifdef IL_CVMC_CODE
+/*#define DONT_UNROLL_SYSTEM */
 
 /*
  * Allocate enough space for "n" arguments.
@@ -340,12 +341,29 @@ static int CVMEntryGen(CVMEntryContext *ctx, ILCVMCoder *coder,
 	coder->stackCheck = CVM_POSN();
 	CVM_OUT_CKHEIGHT();
 
-	/* Mark this method as perhaps needing to be unrolled later */
-	if(ordinaryMethod && _ILCVMUnrollPossible() && !(coder->debugEnabled))
+	/* Output the CVM code to do the profiling of the methods*/
+	if((coder->flags & IL_CODER_FLAG_METHOD_PROFILE) != 0)
 	{
-		CVMP_OUT_NONE(COP_PREFIX_UNROLL_METHOD);
+		CVMP_OUT_NONE(COP_PREFIX_PROFILE_COUNT);
 	}
 
+#if defined(DONT_UNROLL_SYSTEM)
+{
+	const char *assemName;
+	assemName = ILImageGetAssemblyName(ILProgramItem_Image(method));
+	if(strcmp(assemName,"mscorlib") != 0 && strcmp(assemName,"I18N") != 0)
+	{
+#endif
+		/* Mark this method as perhaps needing to be unrolled later */
+		if(ordinaryMethod && _ILCVMUnrollPossible() && !(coder->debugEnabled))
+		{
+			CVMP_OUT_NONE(COP_PREFIX_UNROLL_METHOD);
+		}
+#if defined(DONT_UNROLL_SYSTEM)
+	}
+}
+#endif
+
 	/* Output CVM code to allocate space for the local variables */
 	numLocals = ctx->numLocalWords + ctx->extraLocals;
 	if(numLocals == 1)
@@ -520,6 +538,82 @@ static void CVMEntryPushLeadIn(CVMEntryContext *ctx, ILCVMCoder *coder,
 	}
 }
 
+#ifdef IL_CONFIG_PINVOKE
+
+/*
+ * Determine if a parameter will need marshalling conversion on
+ * the members of a by-ref structure.
+ */
+static int NeedsStructConversion(ILMethod *method, ILPInvoke *pinv,
+								 unsigned long param, ILType *type)
+{
+	ILClass *classInfo;
+	ILField *field;
+	ILUInt32 marshalType;
+	char *customName;
+	int customNameLen;
+	char *customCookie;
+	int customCookieLen;
+	ILParameter *paramInfo;
+
+	/* Bail out if not a reference to a struct type */
+	if(!type || !ILType_IsComplex(type) ||
+	   ILType_Kind(type) != IL_TYPE_COMPLEX_BYREF)
+	{
+		return 0;
+	}
+	type = ILTypeStripPrefixes(ILType_Ref(type));
+	if(!ILType_IsValueType(type))
+	{
+		return 0;
+	}
+
+	/* Don't bother if this is an "out" parameter */
+	paramInfo = 0;
+	while((paramInfo = ILMethodNextParam(method, paramInfo)) != 0)
+	{
+		if(ILParameter_Num(paramInfo) == param &&
+		   ILParameter_IsOut(paramInfo))
+		{
+			return 0;
+		}
+	}
+
+	/* Determine if any of the fields need conversion */
+	classInfo = ILType_ToValueType(type);
+	field = 0;
+	while((field = (ILField *)ILClassNextMemberByKind
+				(classInfo, (ILMember *)field,
+				 IL_META_MEMBERKIND_FIELD)) != 0)
+	{
+		if(ILField_IsStatic(field))
+		{
+			continue;
+		}
+		type = ILField_Type(field);
+		if(NeedsStructConversion(method, pinv, 0, type))
+		{
+			/* An inner type needs conversion, so convert the outer type */
+			return 1;
+		}
+	    marshalType = ILPInvokeGetMarshalType(pinv, method, 0,
+	   										  &customName, &customNameLen,
+											  &customCookie,
+											  &customCookieLen, type);
+		/* TODO: convert other kinds of fields, not just delegates */
+		if(marshalType == IL_META_MARSHAL_FNPTR)
+		{
+			/* We have found a field that needs to be marshalled */
+			return 1;
+		}
+	}
+
+	/* We don't need conversion if we get here */
+	return 0;
+}
+
+#endif /* IL_CONFIG_PINVOKE */
+
 /*
  * Push native arguments onto the stack.  If "coder" is NULL
  * then determine the number of extra locals that we need only.
@@ -537,6 +631,10 @@ static void CVMEntryPushNativeArgs(CVMEntryContext *ctx, ILCVMCoder *coder,
 	ILPInvoke *pinv;
 #ifdef IL_CONFIG_PINVOKE
 	ILUInt32 marshalType;
+	char *customName;
+	int customNameLen;
+	char *customCookie;
+	int customCookieLen;
 #endif
 
 	/* Get the PInvoke information */
@@ -570,7 +668,11 @@ static void CVMEntryPushNativeArgs(CVMEntryContext *ctx, ILCVMCoder *coder,
 	#ifdef IL_CONFIG_PINVOKE
 		/* Perform PInvoke expansions on the parameter */
 		if(!isInternal &&
-		   (marshalType = ILPInvokeGetMarshalType(pinv, method, param))
+		   (marshalType = ILPInvokeGetMarshalType(pinv, method, param,
+		   										  &customName, &customNameLen,
+												  &customCookie,
+												  &customCookieLen,
+												  paramType))
 				!= IL_META_MARSHAL_DIRECT)
 		{
 			offset = coder->argOffsets[param - thisAdjust];
@@ -597,6 +699,12 @@ static void CVMEntryPushNativeArgs(CVMEntryContext *ctx, ILCVMCoder *coder,
 				}
 				break;
 
+				case IL_META_MARSHAL_UTF16_STRING:
+				{
+					CVMP_OUT_NONE(COP_PREFIX_STR2UTF16);
+				}
+				break;
+
 				case IL_META_MARSHAL_FNPTR:
 				{
 					CVMP_OUT_NONE(COP_PREFIX_DELEGATE2FNPTR);
@@ -620,6 +728,28 @@ static void CVMEntryPushNativeArgs(CVMEntryContext *ctx, ILCVMCoder *coder,
 					CVMP_OUT_NONE(COP_PREFIX_REFARRAY2UTF8);
 				}
 				break;
+
+				case IL_META_MARSHAL_CUSTOM:
+				{
+					CVMP_OUT_WORD2_PTR2(COP_PREFIX_TOCUSTOM,
+									    (ILUInt32)(ILInt32)customNameLen,
+									    (ILUInt32)(ILInt32)customCookieLen,
+									    (void *)customName,
+									    (void *)customCookie);
+				}
+				break;
+
+				case IL_META_MARSHAL_ANSI_ARRAY:
+				{
+					CVMP_OUT_NONE(COP_PREFIX_ARRAY2ANSI);
+				}
+				break;
+
+				case IL_META_MARSHAL_UTF8_ARRAY:
+				{
+					CVMP_OUT_NONE(COP_PREFIX_ARRAY2UTF8);
+				}
+				break;
 			}
 			if(offset < 4)
 			{
@@ -631,6 +761,25 @@ static void CVMEntryPushNativeArgs(CVMEntryContext *ctx, ILCVMCoder *coder,
 			}
 			CVM_ADJUST(-1);
 		}
+
+		/* Check for by-ref structures which include fields that
+		   need to be marshalled explicitly on entry */
+		if(!isInternal &&
+		   NeedsStructConversion(method, pinv, param, paramType))
+		{
+			offset = coder->argOffsets[param - thisAdjust];
+			if(offset < 4)
+			{
+				CVM_OUT_NONE(COP_PLOAD_0 + offset);
+			}
+			else
+			{
+				CVM_OUT_WIDE(COP_PLOAD, offset);
+			}
+			CVM_ADJUST(1);
+			CVMP_OUT_PTR(COP_PREFIX_STRUCT2NATIVE, ILType_Ref(paramType));
+			CVM_ADJUST(-1);
+		}
 	#endif /* IL_CONFIG_PINVOKE */
 
 		/* Generate the code for the parameter */
@@ -759,13 +908,31 @@ static void CVMEntryPushNativeArgs(CVMEntryContext *ctx, ILCVMCoder *coder,
 					CVM_OUT_WIDE(COP_PSTORE, offset);
 				}
 				CVM_ADJUST(-1);
-				CVM_OUT_WIDE(COP_WADDR_NATIVE_0 + ctx->nativeArg, offset);
+				if(ctx->nativeArg < 8)
+				{
+					CVM_OUT_WIDE(COP_WADDR_NATIVE_0 + ctx->nativeArg, offset);
+				}
+				else
+				{
+					CVMP_OUT_WORD2(COP_PREFIX_WADDR_NATIVE_N,
+								   (ILUInt32)(ILInt32)(ctx->nativeArg),
+								   offset);
+				}
 			}
 			else
 			{
 				/* Push a pointer to the actual arg onto the native stack */
-				CVM_OUT_WIDE(COP_WADDR_NATIVE_0 + ctx->nativeArg,
-							 coder->argOffsets[param - thisAdjust]);
+				if(ctx->nativeArg < 8)
+				{
+					CVM_OUT_WIDE(COP_WADDR_NATIVE_0 + ctx->nativeArg,
+								 coder->argOffsets[param - thisAdjust]);
+				}
+				else
+				{
+					CVMP_OUT_WORD2(COP_PREFIX_WADDR_NATIVE_N,
+								   (ILUInt32)(ILInt32)(ctx->nativeArg),
+								   coder->argOffsets[param - thisAdjust]);
+				}
 			}
 			++(ctx->nativeArg);
 		}
@@ -784,6 +951,10 @@ static void CVMEntryCallNative(CVMEntryContext *ctx, ILCVMCoder *coder,
 	int hasReturn;
 	ILUInt32 size;
 	ILUInt32 marshalType;
+	char *customName;
+	int customNameLen;
+	char *customCookie;
+	int customCookieLen;
 
 	/* Push the address of the argument array onto the stack */
 	if(useRawCalls)
@@ -1005,7 +1176,8 @@ static void CVMEntryCallNative(CVMEntryContext *ctx, ILCVMCoder *coder,
 		{
 			/* Marshal the PInvoke return value back into a CLR object */
 			marshalType = ILPInvokeGetMarshalType
-				(ILPInvokeFind(method), method, 0);
+				(ILPInvokeFind(method), method, 0, &customName, &customNameLen,
+				 &customCookie, &customCookieLen, returnType);
 			switch(marshalType)
 			{
 				case IL_META_MARSHAL_ANSI_STRING:
@@ -1019,6 +1191,22 @@ static void CVMEntryCallNative(CVMEntryContext *ctx, ILCVMCoder *coder,
 					CVMP_OUT_NONE(COP_PREFIX_UTF82STR);
 				}
 				break;
+
+				case IL_META_MARSHAL_UTF16_STRING:
+				{
+					CVMP_OUT_NONE(COP_PREFIX_UTF162STR);
+				}
+				break;
+
+				case IL_META_MARSHAL_CUSTOM:
+				{
+					CVMP_OUT_WORD2_PTR2(COP_PREFIX_FROMCUSTOM,
+									    (ILUInt32)(ILInt32)customNameLen,
+									    (ILUInt32)(ILInt32)customCookieLen,
+									    (void *)customName,
+									    (void *)customCookie);
+				}
+				break;
 			}
 		}
 		CVM_OUT_RETURN(1);
@@ -1260,6 +1448,15 @@ static int CVMCoder_Finish(ILCoder *_coder)
 	return result;
 }
 
+/*
+ * Set the flags for profiling debugging etc.
+ */
+static void CVMCoder_SetFlags(ILCoder *_coder,int flags)
+{
+	ILCVMCoder *coder = (ILCVMCoder*)_coder;
+	coder->flags = flags;
+}
+
 /*
  * Mark the end of a method's bytecode, just prior to the exception tables.
  */
@@ -1274,4 +1471,126 @@ static void CVMCoder_MarkEnd(ILCoder *coder)
 	}
 }
 
+/*
+ * Allocate an extra local variable in the current method frame.
+ * Returns the local variable index.
+ */
+static ILUInt32 CVMCoder_AllocExtraLocal(ILCoder *coder, ILType *type)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * Push a thread value onto the stack for an internalcall.
+ */
+static void CVMCoder_PushThread(ILCoder *_coder, int useRawCalls)
+{
+	ILCVMCoder *coder = (ILCVMCoder *)_coder;
+	if(useRawCalls)
+	{
+		CVM_OUT_NONE(COP_PUSH_THREAD_RAW);
+		CVM_ADJUST(1);
+	}
+	else
+	{
+		CVM_OUT_NONE(COP_PUSH_THREAD);
+		++(coder->nativeArgPosn);
+	}
+}
+
+/*
+ * Load the address of an argument onto the native argument stack.
+ */
+static void CVMCoder_LoadNativeArgAddr(ILCoder *_coder, ILUInt32 num)
+{
+	ILCVMCoder *coder = (ILCVMCoder *)_coder;
+	ILUInt32 offset = ((ILCVMCoder *)coder)->argOffsets[num];
+	if(coder->nativeArgPosn < 8)
+	{
+		CVM_OUT_WIDE(COP_WADDR_NATIVE_0 + coder->nativeArgPosn, offset);
+	}
+	else
+	{
+		CVMP_OUT_WORD2(COP_PREFIX_WADDR_NATIVE_N,
+					   (ILUInt32)(ILInt32)(coder->nativeArgPosn),
+					   offset);
+	}
+	++(coder->nativeArgPosn);
+}
+
+/*
+ * Load the address of a local onto the native argument stack.
+ */
+static void CVMCoder_LoadNativeLocalAddr(ILCoder *_coder, ILUInt32 num)
+{
+	ILCVMCoder *coder = (ILCVMCoder *)_coder;
+	ILUInt32 offset = ((ILCVMCoder *)coder)->localOffsets[num];
+	if(coder->nativeArgPosn < 8)
+	{
+		CVM_OUT_WIDE(COP_WADDR_NATIVE_0 + coder->nativeArgPosn, offset);
+	}
+	else
+	{
+		CVMP_OUT_WORD2(COP_PREFIX_WADDR_NATIVE_N,
+					   (ILUInt32)(ILInt32)(coder->nativeArgPosn),
+					   offset);
+	}
+	++(coder->nativeArgPosn);
+}
+
+/*
+ * Start pushing arguments for a "libffi" call onto the stack.
+ */
+static void CVMCoder_StartFfiArgs(ILCoder *_coder)
+{
+	ILCVMCoder *coder = (ILCVMCoder *)_coder;
+	coder->nativeArgPosn = -1;
+	coder->nativeArgHeight = coder->height;
+}
+
+/*
+ * Push the address of the raw argument block onto the stack.
+ */
+static void CVMCoder_PushRawArgPointer(ILCoder *_coder)
+{
+	ILCVMCoder *coder = (ILCVMCoder *)_coder;
+	ILUInt32 height = (ILUInt32)(coder->height - coder->nativeArgHeight);
+	CVM_OUT_WIDE(COP_MADDR, height);
+	CVM_ADJUST(1);
+}
+
+/*
+ * Perform a function call using "libffi".
+ */
+static void CVMCoder_CallFfi(ILCoder *coder, void *fn, void *cif,
+					  		 int useRawCalls, int hasReturn)
+{
+	if(useRawCalls)
+	{
+		if(hasReturn)
+		{
+			CVM_OUT_PTR2(COP_CALL_NATIVE_RAW, fn, cif);
+			CVM_ADJUST(-2);
+		}
+		else
+		{
+			CVM_OUT_PTR2(COP_CALL_NATIVE_VOID_RAW, fn, cif);
+			CVM_ADJUST(-1);
+		}
+	}
+	else
+	{
+		if(hasReturn)
+		{
+			CVM_OUT_PTR2(COP_CALL_NATIVE, fn, cif);
+			CVM_ADJUST(-1);
+		}
+		else
+		{
+			CVM_OUT_PTR2(COP_CALL_NATIVE_VOID, fn, cif);
+		}
+	}
+}
+
 #endif	/* IL_CVMC_CODE */
