@@ -20,7 +20,6 @@
 
 #ifndef	_ENGINE_ENGINE_H
 #define	_ENGINE_ENGINE_H
-
 #include "il_thread.h"
 #include "il_engine.h"
 #include "il_system.h"
@@ -53,6 +52,13 @@ extern	"C" {
 #define	IL_CONFIG_FRAME_STACK_SIZE	512
 #endif
 
+/*
+ * Determine if we should use interface method tables.
+ */
+#ifndef IL_CONFIG_REDUCE_CODE
+	/*#define	IL_USE_IMTS	1*/
+#endif
+
 /*
  * Structure that keeps track of a loaded external module.
  */
@@ -77,6 +83,20 @@ struct _tagILExecDebugWatch
 
 };
 
+/*
+ *	Passed to objects when they are finalized.  Objects being finalized should
+ * verify the information in the context valid before continuuing finalization
+ * because it is possible for objects to be finalized *long* after their owner
+ * process has been destroyed.
+ */
+typedef struct __tagILFinalizationContext ILFinalizationContext;
+struct __tagILFinalizationContext
+{
+	/* The process that created the object and registered the finalizer or 0
+	    if that process has been destroyed. */
+	ILExecProcess *volatile process;
+};
+
 /*
  * Execution control context for a process.
  */
@@ -91,6 +111,9 @@ struct _tagILExecProcess
 	/* The "main" thread for the process */
 	ILExecThread   *mainThread;
 
+	/* The finalizer thread for the process */
+	ILExecThread   *finalizerThread;
+
 	/* Default stack size for new threads */
 	ILUInt32	  	stackSize;
 	ILUInt32	  	frameStackSize;
@@ -112,6 +135,7 @@ struct _tagILExecProcess
 	ILClass        *stringClass;
 	ILClass        *exceptionClass;
 	ILClass        *clrTypeClass;
+	ILClass		 *threadAbortClass;
 
 	/* The object to throw when the system runs out of memory */
 	ILObject	   *outOfMemoryObject;
@@ -125,6 +149,15 @@ struct _tagILExecProcess
 	/* Hash table that contains all intern'ed strings within the system */
 	void		   *internHash;
 
+#ifdef USE_HASHING_MONITORS
+	/* Hash table that contains all monitors */
+	void			*monitorHash;
+	ILMutex		*monitorSystemLock;
+#endif
+
+	/* Finalization context used by this process */
+	ILFinalizationContext *finalizationContext;
+
 	/* Hash table that maps program items to reflection objects */
 	void		   *reflectionHash;
 
@@ -157,6 +190,16 @@ struct _tagILExecProcess
 	unsigned char		randomPool[20];
 	int					randomCount;
 
+	/* Size of the global thread-static allocation */
+	ILUInt32			numThreadStaticSlots;
+
+#ifdef IL_USE_IMTS
+
+	/* Last-allocated base identifier for interface method tables */
+	ILUInt32			imtBase;
+
+#endif
+
 };
 
 /*
@@ -171,6 +214,7 @@ typedef struct _tagILCallFrame
 	unsigned char  *pc;				/* PC to return to in the parent method */
 	CVMWord		   *frame;			/* Base of the local variable frame */
 	CVMWord		   *exceptHeight;	/* Height of the frame for exceptions */
+	void           *permissions;	/* Permissions for this stack frame */
 
 } ILCallFrame;
 #define	IL_INVALID_PC		((unsigned char *)(ILNativeInt)(-1))
@@ -187,8 +231,8 @@ struct _tagILExecThread
 	ILExecThread   *nextThread;
 	ILExecThread   *prevThread;
 
-	/* Operating system thread object */
-	ILThread       *osThread;
+	/* Support thread object */
+	ILThread       *supportThread;
 
 	/* Extent of the execution stack */
 	CVMWord		   *stackBase;
@@ -200,18 +244,32 @@ struct _tagILExecThread
 	CVMWord		   *frame;			/* Base of the local variable frame */
 	CVMWord        *stackTop;		/* Current stack top */
 	ILMethod       *method;			/* Current method being executed */
-
+	
 	/* Last exception that was thrown */
 	ILObject       *thrownException;
 
-	/* Security manager in use by this thread */
-	ILObject	   *securityManager;
+	/* Indicates if an abort is in progress */
+	int				aborting;
+
+	/* System.Threading.Thread object */
+	ILObject *clrThread;
+
+	/* Marks this thread as a finalizer thread */
+	/* Finalizer threads are destroyed last */
+	int isFinalizerThread;
+
+	/* Free monitors list */
+	ILExecMonitor *freeMonitor;
 
 	/* Stack of call frames in use */
 	ILCallFrame	   *frameStack;
 	ILUInt32		numFrames;
 	ILUInt32		maxFrames;
 
+	/* Thread-static values for this thread */
+	void		  **threadStaticSlots;
+	ILUInt32		threadStaticSlotsUsed;
+
 };
 
 /*
@@ -226,27 +284,65 @@ struct _tagILImplPrivate
 };
 #define	ILImplPrivate_Table(priv)	((ILUInt16 *)((priv) + 1))
 
+/*
+ * Define the size of the interface method table.
+ */
+#ifdef	IL_USE_IMTS
+#define	IL_IMT_SIZE		32
+#endif
+
 /*
  * Private information that is associated with a class.
  */
 typedef struct _tagILClassPrivate ILClassPrivate;
 struct _tagILClassPrivate
 {
+	ILClass		   *classInfo;			/* Back-pointer to the class */
 	ILUInt32		size;				/* Full instance size */
 	ILUInt32		nativeSize;			/* Full native instance size */
 	ILUInt32		staticSize;			/* Size of static data */
 	ILUInt32		inLayout : 1;		/* Non-zero if in layout algorithm */
 	ILUInt32		hasFinalizer : 1;	/* Non-zero if non-trivial finalizer */
-	ILUInt32		alignment : 7;		/* Preferred instance alignment */
-	ILUInt32		nativeAlignment : 7;/* Preferred native alignment */
+	ILUInt32		managedInstance : 1;/* Non-zero if managed instance field */
+	ILUInt32		managedStatic : 1;	/* Non-zero if managed static field */
+	ILUInt32		alignment : 6;		/* Preferred instance alignment */
+	ILUInt32		nativeAlignment : 6;/* Preferred native alignment */
 	ILUInt32		vtableSize : 16;	/* Size of the vtable */
 	ILMethod      **vtable;				/* Methods within the vtable */
 	ILObject       *clrType;			/* Associated CLR type object */
 	ILObject       *staticData;			/* Static data area object */
 	ILImplPrivate  *implements;			/* Interface implementation records */
+#ifdef IL_USE_IMTS
+	ILUInt32		imtBase;			/* Base for IMT identifiers */
+	ILMethod	   *imt[IL_IMT_SIZE];	/* Interface method table */
+#endif
 
 };
 
+typedef void * ILLockWord;
+
+struct _tagILExecMonitor
+{
+	ILWaitHandle *supportMonitor;
+	volatile ILInt32 waiters;
+	ILExecMonitor *next;
+};
+
+/*
+*	Header of an object.
+*/
+typedef struct _tagObjectHeader ILObjectHeader;
+
+struct _tagObjectHeader
+{
+	ILClassPrivate *classPrivate;
+#ifdef USE_HASHING_MONITORS
+	/* NOTHING */
+#else
+	ILLockWord lockWord;
+#endif
+};
+
 /*
  * Class information for the CVM coder.
  */
@@ -271,14 +367,14 @@ int _ILCallPackVParams(ILExecThread *thread, ILMethod *method,
  * a direct result buffer.
  */
 void _ILCallUnpackDirectResult(ILExecThread *thread, ILMethod *method,
-					           int isCtor, void *result);
+					           int isCtor, void *result, void *userData);
 
 /*
  * Unpack a method result from the CVM stack and store it into
  * an ILExecValue result buffer.
  */
 void _ILCallUnpackVResult(ILExecThread *thread, ILMethod *method,
-				          int isCtor, void *result);
+				          int isCtor, void *result, void *userData);
 
 /*
  * Prototype for a parameter packing function.
@@ -290,7 +386,7 @@ typedef int (*ILCallPackFunc)(ILExecThread *thread, ILMethod *method,
  * Prototype for a return value unpacking function.
  */
 typedef void (*ILCallUnpackFunc)(ILExecThread *thread, ILMethod *method,
-					             int isCtor, void *result);
+					             int isCtor, void *result, void *userData);
 
 /*
  * Call a method using the supplied packing and unpacking rules.
@@ -312,6 +408,11 @@ int _ILCVMInterpreter(ILExecThread *thread);
  */
 ILExecThread *_ILExecThreadCreate(ILExecProcess *process);
 
+/*
+ *	Destroy an engine level thread.
+ */
+void _ILExecThreadDestroy(ILExecThread *thread);
+
 /*
  * Lay out a class's fields, virtual methods, and interfaces.
  * Returns zero if there is something wrong with the definition.
@@ -417,7 +518,7 @@ int _ILGetInternalDelegate(ILMethod *method, int *isCtor,  ILInternalInfo *info)
 /*
  * Look up an interface method.  Returns NULL if not found.
  */
-ILMethod *_ILLookupInterfaceMethod(ILClass *objectClass,
+ILMethod *_ILLookupInterfaceMethod(ILClassPrivate *objectClassPrivate,
 								   ILClass *interfaceClass,
 								   ILUInt32 index);
 
@@ -447,12 +548,16 @@ ILString *_ILStringInternFromConstant(ILExecThread *thread, void *data,
  */
 ILInt32 _ILStringToBuffer(ILExecThread *thread, ILString *str, ILUInt16 **buf);
 
+#ifndef REDUCED_STDIO
+
 /*
  * Dump a CVM instruction starting at a particular address.
  * Returns the size of the instruction.
  */
 int _ILDumpCVMInsn(FILE *stream, ILMethod *currMethod, unsigned char *pc);
 
+#endif
+
 /*
  * Get a call frame which is a specific number of frames up the stack,
  * not counting the active frame.  Returns NULL if "n" is invalid.
@@ -495,6 +600,12 @@ int _ILCVMUnrollPossible(void);
  */
 int _ILCVMUnrollMethod(ILCoder *coder, unsigned char *pc, ILMethod *method);
 
+/*
+ * Unroll a CVM method to native code - locked version.
+ */
+int _ILUnrollMethod(ILExecThread *thread, ILCoder *coder,
+					unsigned char *pc, ILMethod *method);
+
 /*
  * Determine the size of a type's values in bytes.  This assumes
  * that the caller has obtained the metadata write lock.
@@ -505,7 +616,7 @@ ILUInt32 _ILSizeOfTypeLocked(ILType *type);
  * Get the native closure associated with a delegate.  Returns NULL
  * if the closure could not be created for some reason.
  */
-void *_ILDelegateGetClosure(ILObject *delegate);
+void *_ILDelegateGetClosure(ILExecThread *thread, ILObject *delegate);
 
 /*
  * Get the number of parameter words for a method.
@@ -541,6 +652,57 @@ void _ILBreak(ILExecThread *thread, int type);
 
 #endif /* IL_CONFIG_DEBUG_LINES */
 
+/*
+ * Perform custom marshalling to convert an object reference
+ * into a native pointer.
+ */
+void *_ILObjectToCustom(ILExecThread *thread, ILObject *obj,
+						const char *customName, int customNameLen,
+						const char *customCookie, int customCookieLen);
+
+/*
+ * Perform custom marshalling to convert a native pointer
+ * into an object reference.
+ */
+ILObject *_ILCustomToObject(ILExecThread *thread, void *ptr,
+							const char *customName, int customNameLen,
+							const char *customCookie, int customCookieLen);
+
+/*
+ *	Gets the current managed thread object from an engine thread.
+ */
+ILObject *_ILGetCurrentClrThread(ILExecThread *thread);
+
+/*
+ *	Makes the given support thread execute in the context of the given engine thread.
+ */
+void _ILThreadExecuteOn(ILThread *thread, ILExecThread *execThread);
+
+/*
+ *	Throws a thread abort exception on the given thread.
+ */
+void _ILExecThreadThrowThreadAbortException(ILExecThread *thread, ILObject *stateInfo);
+
+/*
+ *	Aborts the current thread.
+ */
+void _ILAbortThread(ILExecThread *thread);
+
+/*
+ *	Handles thread aborts & interruption.
+ */
+void _ILHandleWaitResult(ILExecThread *thread, int result);
+
+/*
+ *	Creates a monitor used by the execution engine.
+ */
+ILExecMonitor *ILExecMonitorCreate(void);
+
+/*
+ *	Destroys a monitor used by the execution engine.
+ */
+void ILExecMonitorDestroy(ILExecMonitor *monitor);
+
 #ifdef	__cplusplus
 };
 #endif
