@@ -22,7 +22,9 @@
 #include "il_system.h"
 #include "il_image.h"
 #include "il_utils.h"
-#if HAVE_UNISTD_H
+#include "il_thread.h"
+#include "il_coder.h"
+#if defined(HAVE_UNISTD_H) && !defined(_MSC_VER)
 #include <unistd.h>
 #endif
 
@@ -36,6 +38,8 @@ extern	"C" {
 int _ILRegisterWithKernel(const char *progname);
 int _ILUnregisterFromKernel(void);
 
+#ifndef IL_WITHOUT_TOOLS
+
 /*
  * Imports from "cvm_dasm.c".
  */
@@ -47,6 +51,13 @@ int _ILDumpVarProfile(FILE *stream);
  */
 int _ILDumpMethodProfile(FILE *stream, ILExecProcess *process);
 
+/*
+ * Imports from "dumpconfig.c"
+ */
+void _ILDumpConfig(FILE *stream,int level);
+
+#endif
+
 /*
  * Table of command-line options.
  */
@@ -59,6 +70,10 @@ static ILCmdLineOption const options[] = {
 	{"--stack-size", 'S', 1, 
 	        "--stack-size value  or -S value",
 	        "Set the operation stack size to `value' kilobytes."},
+	{"-C", 'C', 1, 0, 0},
+	{"--method-cache-page", 'C', 1, 
+	        "--method-cache-page value  or -C value",
+	        "Set the method cache page size to `value' kilobytes."},
 	{"-L", 'L', 1, 0, 0},
 	{"--library-dir", 'L', 1,
 		"--library-dir dir       or -L dir",
@@ -94,6 +109,8 @@ static ILCmdLineOption const options[] = {
 	{"--var-profile", 'V', 0, 0, 0},
 	{"-P", 'P', 0, 0, 0},
 	{"--dump-params", 'P', 0, 0, 0},
+	{"-D", 'D', 0, 0, 0},
+	{"--dump-config", 'D', 0, 0, 0},
 #endif
 
 	{0, 0, 0, 0, 0}
@@ -102,11 +119,32 @@ static ILCmdLineOption const options[] = {
 static void usage(const char *progname);
 static void version(void);
 
+static int CallStaticConstructor(ILExecThread *thread, ILMethod * method)
+{
+	ILClass *classInfo=ILMethod_Owner(method);
+	ILMethod *cctor = 0;
+	while((cctor = (ILMethod *)ILClassNextMemberByKind
+					(classInfo, (ILMember *)cctor,
+					 IL_META_MEMBERKIND_METHOD)) != 0)
+	{
+		if(ILMethod_IsStaticConstructor(cctor))
+		{
+			if(ILExecThreadCall(thread, cctor, NULL))
+			{
+				/* An exception was thrown while executing the program */
+				return 1;
+			}
+		}
+	}
+	return 0;
+}
+
 int main(int argc, char *argv[])
 {
 	char *progname = argv[0];
 	unsigned long heapSize = IL_CONFIG_GC_HEAP_SIZE;
 	unsigned long stackSize = IL_CONFIG_STACK_SIZE;
+	unsigned long methodCachePageSize = IL_CONFIG_CACHE_PAGE_SIZE;
 	char **libraryDirs;
 	int numLibraryDirs;
 	int state, opt;
@@ -120,12 +158,17 @@ int main(int argc, char *argv[])
 	int sawException;
 	int registerMode = 0;
 	char *ilprogram;
+	char *newExePath;
 	int ilprogramLen;
+	ILExecValue execValue;
+	ILExecValue retValue;
+	int flags=0;
 #ifndef IL_CONFIG_REDUCE_CODE
 	int dumpInsnProfile = 0;
-	int dumpMethodProfile = 0;
 	int dumpVarProfile = 0;
+	int dumpMethodProfile = 0;
 	int dumpParams = 0;
+	int dumpConfig = 0;
 #endif
 
 	/* Initialize the locale routines */
@@ -166,6 +209,18 @@ int main(int argc, char *argv[])
 			}
 			break;
 
+			case 'C':
+			{
+				methodCachePageSize = 0;
+				while(*param >= '0' && *param <= '9')
+				{
+					methodCachePageSize = methodCachePageSize * 10 + (unsigned long)(*param - '0');
+					++param;
+				}
+				methodCachePageSize *= 1024;
+			}
+			break;
+
 			case 'L':
 			{
 				if(libraryDirs != 0)
@@ -190,6 +245,7 @@ int main(int argc, char *argv[])
 
 			case 'M':
 			{
+				flags |= IL_CODER_FLAG_METHOD_PROFILE;
 				dumpMethodProfile = 1;
 			}
 			break;
@@ -205,6 +261,12 @@ int main(int argc, char *argv[])
 				dumpParams = 1;
 			}
 			break;
+
+			case 'D':
+			{
+				dumpConfig+=1;
+			}
+			break;
 		#endif
 
 			case 'v':
@@ -240,6 +302,15 @@ int main(int argc, char *argv[])
 		return _ILUnregisterFromKernel();
 	}
 
+#ifndef IL_CONFIG_REDUCE_CODE
+
+	if(dumpConfig!=0)
+	{
+		_ILDumpConfig(stdout,dumpConfig);
+		return 0;
+	}
+#endif 
+
 	/* We need at least one input file argument */
 	if(argc <= 1)
 	{
@@ -251,13 +322,17 @@ int main(int argc, char *argv[])
 	ILExecInit(heapSize);
 
 	/* Create a process to load the program into */
-	process = ILExecProcessCreate(stackSize);
+	process = ILExecProcessCreate(stackSize, methodCachePageSize);
 	if(!process)
 	{
+	#ifndef REDUCED_STDIO
 		fprintf(stderr, "%s: could not create process\n", progname);
+	#endif
 		return 1;
 	}
 
+	ILExecProcessSetCoderFlags(process,flags);
+
 	/* Set the list of directories to use for path searching */
 	if(numLibraryDirs > 0)
 	{
@@ -267,15 +342,26 @@ int main(int argc, char *argv[])
 	/* Get the name of the IL program, appending ".exe" if necessary */
 	ilprogram = argv[1];
 	ilprogramLen = strlen(ilprogram);
-	if(!ILFileExists(ilprogram, (char **)0) &&
-	   (ilprogramLen < 4 ||
-	    ILStrICmp(ilprogram + ilprogramLen - 4, ".exe") != 0))
+	if(ILFileExists(ilprogram, &newExePath))
+	{
+		if(newExePath)
+		{
+			ilprogram = newExePath;
+		}
+	}
+	else if(ilprogramLen < 4 ||
+	        ILStrICmp(ilprogram + ilprogramLen - 4, ".exe") != 0)
 	{
 		ilprogram = (char *)ILMalloc(ilprogramLen + 5);
 		if(ilprogram)
 		{
 			strcpy(ilprogram, argv[1]);
-			strcat(ilprogram, ".exe");
+			strcat(ilprogram, ".EXE");
+			if(!ILFileExists(ilprogram, (char **)0))
+			{
+				strcpy(ilprogram, argv[1]);
+				strcat(ilprogram, ".exe");
+			}
 		}
 		else
 		{
@@ -287,9 +373,14 @@ int main(int argc, char *argv[])
 	error = ILExecProcessLoadFile(process, ilprogram);
 	if(error < 0)
 	{
+	#ifndef REDUCED_STDIO
 		perror(ilprogram);
+	#else
+		printf("%s: could not open file", ilprogram);
+	#endif
 		return 1;
 	}
+#if !defined(__palmos__)
 	else if(error == IL_LOADERR_NOT_IL)
 	{
 		/* This is a regular Windows executable */
@@ -309,9 +400,14 @@ int main(int argc, char *argv[])
 		return ILSpawnProcess(argv + 1);
 	#endif
 	}
+#endif
 	else if(error > 0)
 	{
+	#ifndef REDUCED_STDIO
 		fprintf(stderr, "%s: %s\n", ilprogram, ILImageLoadError(error));
+	#else
+		printf("%s: %s\n", ilprogram, ILImageLoadError(error));
+	#endif
 		return 1;
 	}
 
@@ -319,7 +415,11 @@ int main(int argc, char *argv[])
 	method = ILExecProcessGetEntry(process);
 	if(!method)
 	{
+	#ifndef REDUCED_STDIO
 		fprintf(stderr, "%s: no program entry point\n", ilprogram);
+	#else
+		printf("%s: no program entry point\n", ilprogram);
+	#endif
 		ILExecProcessDestroy(process);
 		return 1;
 	}
@@ -327,7 +427,11 @@ int main(int argc, char *argv[])
 	/* Validate the entry point */
 	if(ILExecProcessEntryType(method) == IL_ENTRY_INVALID)
 	{
+	#ifndef REDUCED_STDIO
 		fprintf(stderr, "%s: invalid entry point\n", ilprogram);
+	#else
+		printf("%s: invalid entry point\n", ilprogram);
+	#endif
 		ILExecProcessDestroy(process);
 		return 1;
 	}
@@ -340,13 +444,21 @@ int main(int argc, char *argv[])
 	sawException = 0;
 	if(args != 0 && !ILExecThreadHasException(thread))
 	{
-		retval = 0;
-		if(ILExecThreadCall(thread, method, &retval, args))
+		retval = CallStaticConstructor(thread, method);
+		sawException = retval;
+		execValue.ptrValue = args;
+		ILMemZero(&retValue, sizeof(retValue));
+		if(!sawException && ILExecThreadCallV
+				(thread, method, &retValue, &execValue))
 		{
 			/* An exception was thrown while executing the program */
 			sawException = 1;
 			retval = 1;
 		}
+		else if(!sawException)
+		{
+			retval = retValue.int32Value;
+		}
 	}
 	else
 	{
@@ -360,7 +472,10 @@ int main(int argc, char *argv[])
 		ILExecThreadPrintException(thread);
 	}
 
-#ifndef IL_CONFIG_REDUCE_CODE
+	/* Wait for all foreground threads to finish */
+	ILThreadWaitForForegroundThreads(-1);
+ 
+#if !defined(IL_CONFIG_REDUCE_CODE) && !defined(IL_WITHOUT_TOOLS)
 	/* Print profile information if requested */
 	if(dumpInsnProfile)
 	{
@@ -404,16 +519,18 @@ int main(int argc, char *argv[])
 	/* Clean up the process and exit */
 	error = ILExecProcessGetStatus(process);
 	ILExecProcessDestroy(process);
+	ILExecDeinit();
+
 	return (int)retval;
 }
 
 static void usage(const char *progname)
 {
-	fprintf(stdout, "ILRUN " VERSION " - IL Program Runtime\n");
-	fprintf(stdout, "Copyright (c) 2001 Southern Storm Software, Pty Ltd.\n");
-	fprintf(stdout, "\n");
-	fprintf(stdout, "Usage: %s [options] program [args]\n", progname);
-	fprintf(stdout, "\n");
+	printf("ILRUN " VERSION " - IL Program Runtime\n");
+	printf("Copyright (c) 2001 Southern Storm Software, Pty Ltd.\n");
+	printf("\n");
+	printf("Usage: %s [options] program [args]\n", progname);
+	printf("\n");
 	ILCmdLineHelp(options);
 }
 
