@@ -1,7 +1,7 @@
 /*
  * lib_reflect.c - Internalcall methods for the reflection classes.
  *
- * Copyright (C) 2001, 2002  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2001, 2002, 2003  Southern Storm Software, Pty Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 #include "engine_private.h"
 #include "lib_defs.h"
 #include "il_serialize.h"
+#include "il_crypt.h"
 
 #ifdef	__cplusplus
 extern	"C" {
@@ -115,8 +116,13 @@ static ILObject *DeserializeObject(ILExecThread *thread,
 	const char *strValue;
 	int strLen;
 	int arrayLen;
+	int boxedType;
 	System_Array * arrayVal;
 	ILObject ** buf;
+	ILType *typeAttr;
+	ILType *systemType=ILExecThreadLookupType(thread,"oSystem.Type;");
+	ILClass *classInfo;
+	char *copyStr;
 	
 	if(serialType!=ILSerializeGetType(type))
 	{
@@ -163,6 +169,94 @@ static ILObject *DeserializeObject(ILExecThread *thread,
 					}
 					return (ILObject*)((System_String*)ILStringCreateLen(thread,
 									strValue,strLen));	
+			case IL_META_SERIALTYPE_TYPE:
+					strLen = ILSerializeReaderGetString(reader,&strValue);
+					if(strLen == -1) 
+					{
+						return NULL;
+					}
+					/* NOTE: Make sure the string used for lookup is NULL
+					 * terminated */
+					copyStr=(char*) ILCalloc(strLen+1,sizeof(char));
+					ILMemCpy(copyStr,strValue,strLen);
+					classInfo = ILExecThreadLookupClass(thread,copyStr);
+					ILFree(copyStr);
+					if(!classInfo)
+					{
+						return NULL;
+					}
+					classInfo = ILClassResolve(classInfo);
+					if(ILClass_IsValueType(classInfo))
+					{
+						typeAttr = ILType_FromValueType(classInfo);
+					}
+					else
+					{
+						typeAttr = ILType_FromClass(classInfo);
+					}
+					return _ILGetClrTypeForILType(thread,typeAttr);
+					
+			case IL_META_SERIALTYPE_VARIANT:
+					boxedType=ILSerializeReaderGetBoxedPrefix(reader);
+					if(boxedType==-1)
+					{
+						return NULL;
+					}
+					switch(boxedType)
+					{
+						case IL_META_SERIALTYPE_BOOLEAN :
+						case IL_META_SERIALTYPE_CHAR :
+						case IL_META_SERIALTYPE_I1 :
+						case IL_META_SERIALTYPE_U1 :
+						case IL_META_SERIALTYPE_I2 :
+						case IL_META_SERIALTYPE_U2 :
+						case IL_META_SERIALTYPE_I4 :
+						case IL_META_SERIALTYPE_U4 :
+						case IL_META_SERIALTYPE_I8 :
+						case IL_META_SERIALTYPE_U8 :
+						case IL_META_SERIALTYPE_R4 :
+						case IL_META_SERIALTYPE_R8 :
+						case IL_META_SERIALTYPE_STRING :
+						{
+							/* NOTE: Right now all the serialtypes and
+							 * Elemtypes are the same for primitives */
+							return DeserializeObject(thread,reader,
+									ILType_FromElement(boxedType),boxedType);
+						}
+						/* not reached */
+						case IL_META_SERIALTYPE_TYPE:
+						{
+							return DeserializeObject(thread,reader,
+									systemType, IL_META_SERIALTYPE_TYPE);
+						}
+						case IL_META_SERIALTYPE_ENUM:
+						{
+							strLen = ILSerializeReaderGetString(reader,
+															&strValue);
+							if(strLen == -1) 
+							{
+								return NULL;
+							}
+							/* NOTE: Make sure the string used for lookup 
+							 * is NULL terminated */
+							copyStr=(char*) ILCalloc(strLen+1,sizeof(char));
+							ILMemCpy(copyStr,strValue,strLen);
+							classInfo = ILExecThreadLookupClass(thread,copyStr);
+							ILFree(copyStr);
+							if(!classInfo)
+							{
+								return NULL;
+							}	
+							classInfo=ILClassResolve(classInfo);
+							
+							typeAttr = ILType_FromValueType(classInfo);
+							
+							return DeserializeObject(thread, reader,
+											typeAttr, 
+											ILSerializeGetType(typeAttr));
+						}
+					}
+					
 			default:
 					if((serialType & IL_META_SERIALTYPE_ARRAYOF)!=0)
 					{
@@ -310,13 +404,16 @@ System_Array *_IL_ClrHelpers_GetCustomAttributes
 	ILObject **buffer;
 	ILClass *classInfo;
 	ILAttribute *attr;
+	ILClass *attributeClass=ILExecThreadLookupClass(thread,"System.Attribute");
 
 	/* Check that we have reflection access to the item */
 	if(item && _ILClrCheckItemAccess(thread, item))
 	{
 		num = NumMatchingAttrs(thread, item, type, inherit);
-		array = (System_Array *)ILExecThreadNew(thread, "[oSystem.Object;",
-												"(Ti)V", (ILVaInt)num);
+		array = (System_Array*) _IL_Array_CreateArray_jiiii
+								(thread, (type ? (ILNativeInt)type 
+								: (ILNativeInt)attributeClass), 
+								 1, num, 0, 0);
 		if(!array)
 		{
 			return 0;
@@ -363,8 +460,10 @@ System_Array *_IL_ClrHelpers_GetCustomAttributes
 	}
 
 	/* Invalid item, or insufficient access: return a zero-element array */
-	return (System_Array *)ILExecThreadNew
-				(thread, "[oSystem.Object;", "(Ti)V", (ILVaInt)0);
+	return (System_Array*) _IL_Array_CreateArray_jiiii
+								(thread, (type ? (ILNativeInt)type 
+								: (ILNativeInt)attributeClass), 
+								 1, 0, 0, 0);
 }
 
 /*
@@ -658,6 +757,79 @@ static ILObject *ImageToAssembly(ILExecThread *thread, ILImage *image)
 	return 0;
 }
 
+/*
+ * public Assembly[] GetAssemblies();
+ */
+System_Array *_IL_AppDomain_GetAssemblies(ILExecThread *thread,
+                                          ILObject *_this)
+{
+	ILContext *context;
+	ILImage *image;
+	ILInt32 num;
+	System_Array *array;
+	ILObject **buffer;
+	ILImage **images;
+	ILImage **ptr;
+
+	IL_METADATA_RDLOCK(thread);
+
+	context = thread->process->context;
+
+	/* count out the number of images */
+	image = 0;
+	num = 0;
+	while((image = ILContextNextImage(context, image)) != 0)
+	{
+		++num;
+	}
+
+	/* create the image array */
+	if (!(images = (ILImage **)ILMalloc(sizeof(ILImage *)*num)))
+	{
+		IL_METADATA_UNLOCK(thread);
+		ILExecThreadThrowOutOfMemory(thread);
+		return 0;
+	}
+
+	/* fill the image array */
+	image = 0;
+	ptr = images;
+	while((image = ILContextNextImage(context, image)) != 0)
+	{
+		*ptr = image;
+		++ptr;
+	}
+
+	IL_METADATA_UNLOCK(thread);
+
+	/* create the assembly array */
+	array = (System_Array *)ILExecThreadNew(thread,
+	                                        "[oSystem.Reflection.Assembly;",
+	                                        "(Ti)V", (ILVaInt)num);
+	if(!array)
+	{
+		ILExecThreadThrowOutOfMemory(thread);
+		return 0;
+	}
+
+	/* fill the assembly array */
+	ptr = images;
+	buffer = (ILObject **)(ArrayToBuffer(array));
+	while(num > 0)
+	{
+		*buffer = ImageToAssembly(thread, *ptr);
+		++buffer;
+		++ptr;
+		--num;
+	}
+
+	/* cleanup */
+	ILFree(images);
+
+	/* return the assembly array */
+	return array;
+}
+
 /*
  * public static Assembly GetCallingAssembly();
  */
@@ -852,6 +1024,23 @@ System_Array *_IL_Assembly_GetFiles(ILExecThread *thread,
 	return 0;
 }
 
+/*
+ * private String GetLocation();
+ */
+ILString *_IL_Assembly_GetLocation(ILExecThread *thread, ILObject *_this)
+{
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	if(image)
+	{
+		return ILStringCreate(thread, image->filename);
+	}
+	else
+	{
+		return 0;
+	}
+}
+
 /*
  * public virtual ManifestResourceInfo GetManifestResourceInfo
  *				(String resourceName);
@@ -1070,8 +1259,18 @@ ILObject *_IL_Assembly_LoadFromName(ILExecThread *thread,
 	char *str = ILStringToUTF8(thread, name);
 	if(image && str)
 	{
+		int len;
 		int loadError;
 		ILImage *newImage;
+		len = strlen(str);
+		if(len > 4 && str[len - 4] == '.' &&
+		   (str[len - 3] == 'd' || str[len - 3] == 'D') &&
+		   (str[len - 2] == 'l' || str[len - 2] == 'L') &&
+		   (str[len - 1] == 'l' || str[len - 1] == 'L'))
+		{
+			/* Remove ".dll", to get the assembly name */
+			str[len - 4] = '\0';
+		}
 		loadError = ILImageLoadAssembly(str, thread->process->context,
 										image, &newImage);
 		if(loadError == 0)
@@ -1157,6 +1356,59 @@ ILObject *_IL_Assembly_LoadFromFile(ILExecThread *thread,
 	return 0;
 }
 
+/*
+ * internal static Assembly LoadFromBytes(byte[] bytes, out int error,
+ *										  Assembly parent);
+ */
+ILObject *_IL_Assembly_LoadFromBytes(ILExecThread *thread,
+									 System_Array *bytes,
+									 ILInt32 *error,
+									 ILObject *parent)
+{
+	ILImage *image;
+	int loadError;
+
+	/* Bail out if "bytes" is NULL (should be trapped higher
+	   up the stack, but let's be paranoid anyway) */
+	if(!bytes)
+	{
+		*error = LoadError_FileNotFound;
+		return 0;
+	}
+
+	/* TODO: check security permissions */
+
+	/* Attempt to load the image.  Note: we deliberately don't supply
+	   the "IL_LOADFLAG_IN_PLACE" flag because we don't want the user
+	   to be able to modify the IL binary after we have loaded it.
+	   Or worse, have the garbage collector throw the "bytes" array away */
+	loadError = ILImageLoadFromMemory(ArrayToBuffer(bytes),
+									  (unsigned long)(long)(bytes->length),
+									  thread->process->context,
+									  &image, IL_LOADFLAG_FORCE_32BIT, 0);
+	if(loadError == 0)
+	{
+		*error = LoadError_OK;
+		return ImageToAssembly(thread, image);
+	}
+
+	/* Convert the error code into something the C# library knows about */
+	if(loadError == -1)
+	{
+		*error = LoadError_FileNotFound;
+	}
+	else if(loadError == IL_LOADERR_MEMORY)
+	{
+		*error = LoadError_FileNotFound;
+		ILExecThreadThrowOutOfMemory(thread);
+	}
+	else
+	{
+		*error = LoadError_BadImage;
+	}
+	return 0;
+}
+
 /*
  * private static Type GetType(String typeName, bool throwOnError,
  *							   bool ignoreCase)
@@ -1205,6 +1457,350 @@ void _IL_Assembly_GetEntryPoint(ILExecThread *thread,
 	}
 }
 
+/*
+ * Builtin public key values, to shortcut hashing for common assemblies.
+ */
+static unsigned char const NeutralKey[] =
+	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+	 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
+static unsigned char const NeutralKeyToken[] =
+	{0xb7, 0x7a, 0x5c, 0x56, 0x19, 0x34, 0xe0, 0x89};
+static char const MicrosoftKey[] =
+	{0x00, 0x24, 0x00, 0x00, 0x04, 0x80, 0x00, 0x00,
+	 0x94, 0x00, 0x00, 0x00, 0x06, 0x02, 0x00, 0x00,
+	 0x00, 0x24, 0x00, 0x00, 0x52, 0x53, 0x41, 0x31,
+	 0x00, 0x04, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00,
+	 0x07, 0xd1, 0xfa, 0x57, 0xc4, 0xae, 0xd9, 0xf0,
+	 0xa3, 0x2e, 0x84, 0xaa, 0x0f, 0xae, 0xfd, 0x0d,
+	 0xe9, 0xe8, 0xfd, 0x6a, 0xec, 0x8f, 0x87, 0xfb,
+	 0x03, 0x76, 0x6c, 0x83, 0x4c, 0x99, 0x92, 0x1e,
+	 0xb2, 0x3b, 0xe7, 0x9a, 0xd9, 0xd5, 0xdc, 0xc1,
+	 0xdd, 0x9a, 0xd2, 0x36, 0x13, 0x21, 0x02, 0x90,
+	 0x0b, 0x72, 0x3c, 0xf9, 0x80, 0x95, 0x7f, 0xc4,
+	 0xe1, 0x77, 0x10, 0x8f, 0xc6, 0x07, 0x77, 0x4f,
+	 0x29, 0xe8, 0x32, 0x0e, 0x92, 0xea, 0x05, 0xec,
+	 0xe4, 0xe8, 0x21, 0xc0, 0xa5, 0xef, 0xe8, 0xf1,
+	 0x64, 0x5c, 0x4c, 0x0c, 0x93, 0xc1, 0xab, 0x99,
+	 0x28, 0x5d, 0x62, 0x2c, 0xaa, 0x65, 0x2c, 0x1d,
+	 0xfa, 0xd6, 0x3d, 0x74, 0x5d, 0x6f, 0x2d, 0xe5,
+	 0xf1, 0x7e, 0x5e, 0xaf, 0x0f, 0xc4, 0x96, 0x3d,
+	 0x26, 0x1c, 0x8a, 0x12, 0x43, 0x65, 0x18, 0x20,
+	 0x6d, 0xc0, 0x93, 0x34, 0x4d, 0x5a, 0xd2, 0x93};
+static unsigned char const MicrosoftKeyToken[] =
+	{0xb0, 0x3f, 0x5f, 0x7f, 0x11, 0xd5, 0x0a, 0x3a};
+
+/*
+ * private String GetFullName();
+ */
+ILString *_IL_Assembly_GetFullName(ILExecThread *thread, ILObject *_this)
+{
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	if(image)
+	{
+		ILAssembly *assem;
+		const char *name;
+		const ILUInt16 *version;
+		const char *locale;
+		const void *publicKey;
+		unsigned long publicKeyLen;
+		unsigned long publicKeyLenInChars;
+		char versbuf[64];
+		char *buf;
+		int posn;
+		ILString *str;
+		ILSHAContext sha;
+		unsigned char hash[IL_SHA_HASH_SIZE];
+
+		/* Collect up the assembly name components */
+		name = ILImageGetAssemblyName(image);
+		if(!name)
+		{
+			return 0;
+		}
+		assem = (ILAssembly *)ILImageTokenInfo
+					(image, IL_META_TOKEN_ASSEMBLY | 1);
+		version = ILAssemblyGetVersion(assem);
+		sprintf(versbuf, "%u.%u.%u.%u",
+				(unsigned)(version[0]), (unsigned)(version[1]),
+				(unsigned)(version[2]), (unsigned)(version[3]));
+		locale = ILAssembly_Locale(assem);
+		if(!locale)
+		{
+			locale = "neutral";
+		}
+		publicKeyLen = 0;
+		publicKey = ILAssemblyGetOriginator(assem, &publicKeyLen);
+		if(publicKey != 0 && publicKeyLen != 0)
+		{
+			publicKeyLenInChars = 16;
+		}
+		else
+		{
+			publicKeyLenInChars = 4;	/* "null" */
+		}
+
+		/* Allocate a buffer big enough to hold the entire name */
+		buf = (char *)ILMalloc(strlen(name) + 10    /* ", Version=" */ +
+							   strlen(versbuf) + 10 /* ", Culture=" */ +
+							   strlen(locale) + 17	/* ", PublicKeyToken=" */ +
+							   publicKeyLenInChars + 1);
+		if(!buf)
+		{
+			ILExecThreadThrowOutOfMemory(thread);
+			return 0;
+		}
+
+		/* Format the name */
+		sprintf(buf, "%s, Version=%s, Culture=%s, PublicKeyToken=",
+				name, versbuf, locale);
+		posn = strlen(buf);
+		if(publicKey != 0 && publicKeyLen != 0)
+		{
+			if(publicKeyLen == sizeof(NeutralKey) &&
+			   !ILMemCmp(publicKey, NeutralKey, publicKeyLen))
+			{
+				sprintf(buf + posn, "%02x%02x%02x%02x%02x%02x%02x%02x",
+						NeutralKeyToken[0], NeutralKeyToken[1],
+						NeutralKeyToken[2], NeutralKeyToken[3],
+						NeutralKeyToken[4], NeutralKeyToken[5],
+						NeutralKeyToken[6], NeutralKeyToken[7]);
+			}
+			else if(publicKeyLen == sizeof(MicrosoftKey) &&
+			        !ILMemCmp(publicKey, MicrosoftKey, publicKeyLen))
+			{
+				sprintf(buf + posn, "%02x%02x%02x%02x%02x%02x%02x%02x",
+						MicrosoftKeyToken[0], MicrosoftKeyToken[1],
+						MicrosoftKeyToken[2], MicrosoftKeyToken[3],
+						MicrosoftKeyToken[4], MicrosoftKeyToken[5],
+						MicrosoftKeyToken[6], MicrosoftKeyToken[7]);
+			}
+			else
+			{
+				ILSHAInit(&sha);
+				ILSHAData(&sha, publicKey, publicKeyLen);
+				ILSHAFinalize(&sha, hash);
+				sprintf(buf + posn, "%02x%02x%02x%02x%02x%02x%02x%02x",
+						hash[19], hash[18], hash[17], hash[16],
+						hash[15], hash[14], hash[13], hash[12]);
+			}
+		}
+		else
+		{
+			strcpy(buf + posn, "null");
+		}
+
+		/* Convert the buffer into a string object and return it */
+		str = ILStringCreateUTF8(thread, buf);
+		ILFree(buf);
+		return str;
+	}
+	else
+	{
+		return 0;
+	}
+}
+
+/*
+ * Call a method on the "AssemblyName" class.
+ */
+static int CallAssemblyNameMethod(ILExecThread *thread, ILObject *nameInfo,
+								  const char *methodName,
+								  const char *signature, ILObject *param)
+{
+	ILExecValue value;
+	value.objValue = param;
+	return ILExecThreadCallNamedVirtualV
+				(thread, "System.Reflection.AssemblyName",
+				 methodName, signature, &value, nameInfo, &value);
+}
+static int CallAssemblyNameMethodI4(ILExecThread *thread, ILObject *nameInfo,
+								    const char *methodName,
+								    const char *signature, ILInt32 param)
+{
+	ILExecValue value;
+	value.int32Value = param;
+	return ILExecThreadCallNamedVirtualV
+				(thread, "System.Reflection.AssemblyName",
+				 methodName, signature, &value, nameInfo, &value);
+}
+static int CallAssemblyNameMethodV(ILExecThread *thread, ILObject *nameInfo,
+								   const ILUInt16 *param)
+{
+	ILExecValue value[4];
+	value[0].int32Value = param[0];
+	value[1].int32Value = param[1];
+	value[2].int32Value = param[2];
+	value[3].int32Value = param[3];
+	return ILExecThreadCallNamedVirtualV
+				(thread, "System.Reflection.AssemblyName",
+				 "SetVersion", "(Tiiii)V", value, nameInfo, value);
+}
+
+/*
+ * Fill in an "AssemblyName" object with an image's name information.
+ */
+static void FillAssemblyNameFromImage(ILExecThread *thread,
+									  ILObject *nameInfo, ILImage *image)
+{
+	ILString *name;
+	ILAssembly *assem;
+	const void *publicKey;
+	unsigned long publicKeyLen;
+	System_Array *array;
+
+	/* Set the assembly name */
+	name = ILStringCreateUTF8(thread, ILImageGetAssemblyName(image));
+	if(!name)
+	{
+		return;
+	}
+	if(CallAssemblyNameMethod(thread, nameInfo, "set_Name",
+							  "(ToSystem.String;)V", (ILObject *)name))
+	{
+		return;
+	}
+
+	/* Get the first assembly structure for the image */
+	assem = (ILAssembly *)ILImageTokenInfo(image, IL_META_TOKEN_ASSEMBLY | 1);
+
+	/* Set the locale name */
+	if(ILAssembly_Locale(assem))
+	{
+		name = ILStringCreateUTF8(thread, ILAssembly_Locale(assem));
+		if(!name)
+		{
+			return;
+		}
+		if(CallAssemblyNameMethod(thread, nameInfo, "SetCultureByName",
+								  "(ToSystem.String;)V", (ILObject *)name))
+		{
+			return;
+		}
+	}
+
+	/* Set the assembly name flags */
+	if(ILAssembly_HasPublicKey(assem))
+	{
+		if(CallAssemblyNameMethodI4(thread, nameInfo, "set_Flags",
+									"(TvSystem.Reflection.AssemblyNameFlags;)V",
+									(ILInt32)0x0001))
+		{
+			return;
+		}
+	}
+
+	/* Set the version information */
+	if(CallAssemblyNameMethodV(thread, nameInfo, ILAssemblyGetVersion(assem)))
+	{
+		return;
+	}
+
+	/* Set the assembly's hash algorithm */
+	if(CallAssemblyNameMethodI4(thread, nameInfo, "set_HashAlgorithm",
+			"(TvSystem.Configuration.Assemblies.AssemblyHashAlgorithm;)V",
+			(ILInt32)(ILAssembly_HashAlg(assem))))
+	{
+		return;
+	}
+
+	/* Set the assembly's public key */
+	publicKeyLen = 0;
+	publicKey = ILAssemblyGetOriginator(assem, &publicKeyLen);
+	if(publicKey != 0 && publicKeyLen != 0)
+	{
+		array = (System_Array *)ILExecThreadNew(thread, "[B", "(Ti)V",
+												(ILVaInt)publicKeyLen);
+		if(!array)
+		{
+			return;
+		}
+		ILMemCpy(ArrayToBuffer(array), publicKey, publicKeyLen);
+		if(CallAssemblyNameMethod(thread, nameInfo, "SetPublicKey",
+								  "(T[B)V", (ILObject *)array))
+		{
+			return;
+		}
+	}
+}
+
+/*
+ * private void FillAssemblyName(AssemblyName nameInfo);
+ */
+void _IL_Assembly_FillAssemblyName(ILExecThread *thread, ILObject *_this,
+								   ILObject *nameInfo)
+{
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	if(image)
+	{
+		FillAssemblyNameFromImage(thread, nameInfo, image);
+	}
+}
+
+/*
+ * private static int FillAssemblyNameFromFile
+ *		(AssemblyName nameInfo, String assemblyFile, Assembly caller);
+ */
+ILInt32 _IL_AssemblyName_FillAssemblyNameFromFile(ILExecThread *thread,
+												  ILObject *nameInfo,
+												  ILString *assemblyFile,
+												  ILObject *caller)
+{
+	ILContext *context;
+	ILImage *image;
+	int loadError;
+	char *filename;
+
+	/* Convert the filename into an ANSI string */
+	filename = ILStringToAnsi(thread, assemblyFile);
+	if(!filename)
+	{
+		return 0;
+	}
+
+	/* TODO: check security permissions */
+
+	/* We need to temporarily load the image to get its name information,
+	   and then immediately discard it.  Use a new context for that */
+	context = ILContextCreate();
+	if(!context)
+	{
+		ILExecThreadThrowOutOfMemory(thread);
+		return 0;
+	}
+
+	/* Attempt to load the file */
+	loadError = ILImageLoadFromFile(filename, context, &image,
+									IL_LOADFLAG_FORCE_32BIT |
+									IL_LOADFLAG_NO_RESOLVE, 0);
+	if(loadError != 0)
+	{
+		/* Convert the error code into something the C# library knows about */
+		ILContextDestroy(context);
+		if(loadError == -1)
+		{
+			return LoadError_FileNotFound;
+		}
+		else if(loadError == IL_LOADERR_MEMORY)
+		{
+			ILExecThreadThrowOutOfMemory(thread);
+			return 0;
+		}
+		else
+		{
+			return LoadError_BadImage;
+		}
+	}
+
+	/* Fill the "AssemblyName" object with the image's name details */
+	FillAssemblyNameFromImage(thread, nameInfo, image);
+
+	/* Clean up the temporary context and exit */
+	ILContextDestroy(context);
+	return 0;
+}
+
 ILString *_IL_Assembly_GetSatellitePath(ILExecThread *thread,
 										ILObject *_this,
 										ILString *filename)
@@ -1259,6 +1855,187 @@ ILString *_IL_Assembly_GetSatellitePath(ILExecThread *thread,
 	}
 }
 
+ILString *_IL_Assembly_GetImageRuntimeVersion(ILExecThread *_thread,
+											  ILObject *_this)
+{
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	const char *version;
+	int length;
+	if(image)
+	{
+		version = ILImageMetaRuntimeVersion(image, &length);
+		if(version)
+		{
+			return ILStringCreateLen(_thread, version, length);
+		}
+		else
+		{
+			return 0;
+		}
+	}
+	else
+	{
+		return 0;
+	}
+}
+
+ILObject *_IL_Assembly_GetModuleInternal(ILExecThread *_thread,
+										 ILObject *_this, ILString *name)
+{
+	char *nameutf = ILStringToUTF8(_thread, name);
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	ILModule *module;
+	ILFileDecl *file;
+	if(nameutf && image)
+	{
+		/* Search the module table for the name */
+		module = 0;
+		while((module = (ILModule *)ILImageNextToken
+				(image, IL_META_TOKEN_MODULE, module)) != 0)
+		{
+			if(!ILStrICmp(ILModule_Name(module), nameutf))
+			{
+				return _ILClrToObject
+					(_thread, module, "System.Reflection.Module");
+			}
+		}
+
+		/* Search the file table for the name */
+		file = 0;
+		while((file = (ILFileDecl *)ILImageNextToken
+				(image, IL_META_TOKEN_FILE, file)) != 0)
+		{
+			if(!ILStrICmp(ILFileDecl_Name(file), nameutf))
+			{
+				return _ILClrToObject
+					(_thread, file, "System.Reflection.Module");
+			}
+		}
+	}
+	return 0;
+}
+
+System_Array *_IL_Assembly_GetModules(ILExecThread *_thread,
+									  ILObject *_this,
+									  ILBool getResourceModules)
+{
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	unsigned long size;
+	ILModule *module;
+	ILFileDecl *file;
+	System_Array *array;
+	ILObject **buf;
+	if(image)
+	{
+		/* Count the number of relevant program items */
+		size = ILImageNumTokens(image, IL_META_TOKEN_MODULE);
+		file = 0;
+		while((file = (ILFileDecl *)ILImageNextToken
+				(image, IL_META_TOKEN_FILE, file)) != 0)
+		{
+			if(getResourceModules || ILFileDecl_HasMetaData(file))
+			{
+				++size;
+			}
+		}
+
+		/* Allocate the module array */
+		array = (System_Array *)ILExecThreadNew
+			(_thread, "[oSystem.Reflection.Module;", "(Ti)V", (ILVaInt)size);
+		if(!array)
+		{
+			return 0;
+		}
+		buf = (ILObject **)(ArrayToBuffer(array));
+
+		/* Fill the module array */
+		size = 0;
+		module = 0;
+		while((module = (ILModule *)ILImageNextToken
+				(image, IL_META_TOKEN_MODULE, module)) != 0)
+		{
+			buf[size] = _ILClrToObject
+				(_thread, module, "System.Reflection.Module");
+			if(!(buf[size]))
+			{
+				return 0;
+			}
+			++size;
+		}
+		file = 0;
+		while((file = (ILFileDecl *)ILImageNextToken
+				(image, IL_META_TOKEN_FILE, file)) != 0)
+		{
+			if(getResourceModules || ILFileDecl_HasMetaData(file))
+			{
+				buf[size] = _ILClrToObject
+					(_thread, file, "System.Reflection.Module");
+				if(!(buf[size]))
+				{
+					return 0;
+				}
+				++size;
+			}
+		}
+
+		/* Return the array to the caller */
+		return array;
+	}
+	else
+	{
+		return 0;
+	}
+}
+
+System_Array *_IL_Assembly_GetReferencedAssembliesInternal
+					(ILExecThread *_thread, ILObject *_this)
+{
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	unsigned long size;
+	ILAssembly *assem;
+	System_Array *array;
+	ILObject **buf;
+	if(image)
+	{
+		/* Get the number of assembly references */
+		size = ILImageNumTokens(image, IL_META_TOKEN_ASSEMBLY_REF);
+
+		/* Allocate the assembly array */
+		array = (System_Array *)ILExecThreadNew
+			(_thread, "[oSystem.Reflection.Assembly;", "(Ti)V", (ILVaInt)size);
+		if(!array)
+		{
+			return 0;
+		}
+		buf = (ILObject **)(ArrayToBuffer(array));
+
+		/* Fill the assembly */
+		size = 0;
+		assem = 0;
+		while((assem = (ILAssembly *)ILImageNextToken
+				(image, IL_META_TOKEN_ASSEMBLY_REF, assem)) != 0)
+		{
+			buf[size] = ImageToAssembly(_thread, ILAssemblyToImage(assem));
+			if(!(buf[size]))
+			{
+				return 0;
+			}
+			++size;
+		}
+
+		/* Return the array to the caller */
+		return array;
+	}
+	else
+	{
+		return 0;
+	}
+}
+
 #ifdef IL_CONFIG_REFLECTION
 
 /*
@@ -1667,8 +2444,19 @@ ILObject *_IL_MethodBase_GetCurrentMethod(ILExecThread *thread)
 ILNativeInt _IL_RuntimeMethodHandle_GetFunctionPointer
 				(ILExecThread *thread, void *_this)
 {
-	/* Function pointers don't make any sense for us, so always return NULL */
-	return 0;
+#if defined(HAVE_LIBFFI)
+	ILMethod *method = *((ILMethod **)_this);
+	if(method)
+	{
+		/* Create a closure for the method, without a delegate around it */
+		return (ILNativeInt)(_ILMakeClosureForDelegate(0, method));
+	}
+	else
+#endif
+	{
+		/* Invalid RuntimeMethodHandle value */
+		return 0;
+	}
 }
 
 /*
@@ -1752,8 +2540,8 @@ static ILObject *InvokeMethod(ILExecThread *thread, ILMethod *method,
 			if(ILType_IsPrimitive(paramType))
 			{
 				/* Unbox primitive and enumerated types into the argument */
-				if(!ILExecThreadUnbox(thread, objectType, paramObject,
-									  &(args[argNum])))
+				if(!ILExecThreadUnboxFloat
+							(thread, objectType, paramObject, &(args[argNum])))
 				{
 					ILExecThreadThrowSystem
 						(thread, "System.ArgumentException", 0);
@@ -1850,7 +2638,8 @@ static ILObject *InvokeMethod(ILExecThread *thread, ILMethod *method,
 		{
 			return 0;
 		}
-		return ILExecThreadBox(thread, ILTypeGetReturn(signature), &result);
+		return ILExecThreadBoxFloat
+			(thread, ILTypeGetReturn(signature), &result);
 	}
 	else if(ILType_IsClass(paramType))
 	{
@@ -1972,8 +2761,9 @@ ILObject *_IL_ClrMethod_Invoke(ILExecThread *thread,
 				ThrowTargetException(thread);
 				return 0;
 			}
-			method = _ILLookupInterfaceMethod(targetClass, classInfo,
-											  method->index);
+			method = _ILLookupInterfaceMethod
+					(((ILClassPrivate *)(targetClass->userData)),
+					 classInfo, method->index);
 		}
 		else
 		{
@@ -2043,6 +2833,104 @@ ILObject *_IL_ClrMethod_Invoke(ILExecThread *thread,
  */
 ILObject *_IL_ClrMethod_GetBaseDefinition(ILExecThread *thread,
 										  ILObject *_this)
+{
+	ILMethod *method;
+
+	/* Extract the method item from the "this" object */
+	method = ILProgramItemToMethod(_ILClrFromObject(thread, _this));
+	if(!method)
+	{
+		/* Something is wrong with the object */
+		ILExecThreadThrowSystem(thread, "System.MissingMethodException", 0);
+		return 0;
+	}
+
+	/* If the method is part of an interface, return it as-is */
+	if(ILClass_IsInterface(ILMethod_Owner(method)))
+	{
+		return _this;
+	}
+
+	/* If the method is "new", static, or a constructor, return it as-is */
+	if(ILMethod_IsNewSlot(method))
+	{
+		return _this;
+	}
+	if(ILMethod_IsStatic(method))
+	{
+		return _this;
+	}
+	if(ILMethod_IsConstructor(method))
+	{
+		return _this;
+	}
+
+	/* Find the base method */
+	method = (ILMethod *)ILMemberGetBase((ILMember *)method);
+	if(method)
+	{
+		return _ILClrToObject(thread, method, "System.Reflection.ClrMethod");
+	}
+
+	/* We were unable to find a base definition, so return the method itself */
+	return _this;
+}
+
+/*
+ * protected override bool HasGenericArgumentsImpl();
+ */
+ILBool _IL_ClrMethod_HasGenericArgumentsImpl(ILExecThread *_thread,
+											 ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * protected override bool HasGenericParametersImpl();
+ */
+ILBool _IL_ClrMethod_HasGenericParametersImpl(ILExecThread *_thread,
+											  ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * private Type[] GetGenericArgumentsImpl();
+ */
+System_Array *_IL_ClrMethod_GetGenericArgumentsImpl(ILExecThread *_thread,
+													ILObject * _this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * private ClrMethod GetGenericMethodDefinitionImpl();
+ */
+ILObject *_IL_ClrMethod_GetGenericMethodDefinitionImpl(ILExecThread *_thread,
+													   ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * private int GetArity();
+ */
+ILInt32 _IL_ClrMethod_GetArity(ILExecThread *_thread, ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * private MethodInfo BindGenericParametersImpl(Type[] typeArgs);
+ */
+ILObject *_IL_ClrMethod_BindGenericParametersImpl(ILExecThread *_thread,
+												  ILObject *_this,
+												  System_Array *typeArgs)
 {
 	/* TODO */
 	return 0;
@@ -2056,7 +2944,18 @@ ILObject *_IL_Module_GetType(ILExecThread *_thread, ILObject *_this,
 							 ILString *name, ILBool throwOnError,
 							 ILBool ignoreCase)
 {
-	/* TODO */
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILModule *module;
+	ILFileDecl *file;
+	if((module = ILProgramItemToModule(item)) != 0)
+	{
+		return _ILGetTypeFromImage(_thread, ILProgramItem_Image(module),
+								   name, throwOnError, ignoreCase);
+	}
+	else if((file = ILProgramItemToFileDecl(item)) != 0)
+	{
+		/* TODO: metadata-based file modules */
+	}
 	return 0;
 }
 
@@ -2074,8 +2973,16 @@ System_Array *_IL_Module_GetTypes(ILExecThread *_thread, ILObject *_this)
  */
 ILBool _IL_Module_IsResource(ILExecThread *_thread, ILObject *_this)
 {
-	/* TODO */
-	return 0;
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILFileDecl *file = ILProgramItemToFileDecl(item);
+	if(file != 0)
+	{
+		return !ILFileDecl_HasMetaData(file);
+	}
+	else
+	{
+		return 0;
+	}
 }
 
 /*
@@ -2116,7 +3023,26 @@ ILObject *_IL_Module_GetModuleType(ILExecThread *_thread, ILObject *_this)
  */
 ILObject *_IL_Module_GetAssembly(ILExecThread *_thread, ILObject *_this)
 {
-	/* TODO */
+	ILProgramItem *item;
+	ILImage *image;
+	ILAssembly *assembly;
+	const char *name;
+
+	item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	if (image)
+	{
+		assembly = ILAssembly_FromToken(image, (IL_META_TOKEN_ASSEMBLY | 1));
+		if (ILImageType(image) == IL_IMAGETYPE_BUILDING)
+		{
+			name = "System.Reflection.Emit.AssemblyBuilder";
+		}
+		else
+		{
+			name = "System.Reflection.Assembly";
+		}
+		return _ILClrToObject(_thread, assembly, name);
+	}
 	return 0;
 }
 
