@@ -1,7 +1,7 @@
 /*
  * lib_string.c - Internalcall methods for "System.String".
  *
- * Copyright (C) 2001  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2001, 2002, 2003  Southern Storm Software, Pty Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -351,8 +351,6 @@ System_String *_IL_String_ctor_pb(ILExecThread *thread, ILInt8 *value)
 
 /*
  * Compare two Unicode strings.
- *
- * TODO: replace this with something better and move to "support".
  */
 static int ILStrCmpUnicode(const ILUInt16 *str1, ILInt32 length1,
 						   const ILUInt16 *str2, ILInt32 length2)
@@ -386,53 +384,6 @@ static int ILStrCmpUnicode(const ILUInt16 *str1, ILInt32 length1,
 	}
 }
 
-/*
- * Compare two Unicode strings, while ignoring case.
- *
- * TODO: replace this with something better and move to "support".
- */
-static int ILStrICmpUnicode(const ILUInt16 *str1, ILInt32 length1,
-							const ILUInt16 *str2, ILInt32 length2)
-{
-	ILUInt16 ch1;
-	ILUInt16 ch2;
-	while(length1 > 0 && length2 > 0)
-	{
-		ch1 = *str1++;
-		ch2 = *str2++;
-		if(ch1 >= 'a' && ch1 <= 'z')
-		{
-			ch1 = (ILUInt16)(ch1 - 'a' + 'A');
-		}
-		if(ch2 >= 'a' && ch2 <= 'z')
-		{
-			ch2 = (ILUInt16)(ch2 - 'a' + 'A');
-		}
-		if(ch1 < ch2)
-		{
-			return -1;
-		}
-		else if(ch1 > ch2)
-		{
-			return 1;
-		}
-		--length1;
-		--length2;
-	}
-	if(length1 > 0)
-	{
-		return 1;
-	}
-	else if(length2 > 0)
-	{
-		return -1;
-	}
-	else
-	{
-		return 0;
-	}
-}
-
 /*
  * public static int Compare(String strA, String strB);
  */
@@ -467,18 +418,19 @@ ILInt32 _IL_String_Compare(ILExecThread *thread,
 }
 
 /*
- * public static int InternalCompare(String strA, int indexA, int lengthA,
+ * public static int CompareInternal(String strA, int indexA, int lengthA,
  *									 String strB, int indexB, int lengthB,
- *							 		 bool ignoreCase, CultureInfo culture);
+ *							 		 bool ignoreCase);
  */
-ILInt32 _IL_String_InternalCompare(ILExecThread *thread,
+ILInt32 _IL_String_CompareInternal(ILExecThread *thread,
 						   		   System_String *strA,
 								   ILInt32 indexA, ILInt32 lengthA,
 						   		   System_String *strB,
 								   ILInt32 indexB, ILInt32 lengthB,
-						   		   ILBool ignoreCase,
-						   		   ILObject *culture)
+						   		   ILBool ignoreCase)
 {
+	int cmp;
+
 	/* Handle the easy cases first */
 	if(!strA)
 	{
@@ -496,13 +448,40 @@ ILInt32 _IL_String_InternalCompare(ILExecThread *thread,
 		return 1;
 	}
 
-	/* TODO: handle culture information */
-
 	/* Compare the two strings */
 	if(ignoreCase)
 	{
-		return ILStrICmpUnicode(StringToBuffer(strA) + indexA, lengthA,
-							    StringToBuffer(strB) + indexB, lengthB);
+		if(lengthA >= lengthB)
+		{
+			cmp = ILUnicodeStringCompare
+					(StringToBuffer(strA) + indexA,
+					 StringToBuffer(strB) + indexB,
+					 (unsigned long)(long)lengthB);
+			if(cmp != 0)
+			{
+				return cmp;
+			}
+			if(lengthA > lengthB)
+			{
+				return 1;
+			}
+			else
+			{
+				return 0;
+			}
+		}
+		else
+		{
+			cmp = ILUnicodeStringCompare
+					(StringToBuffer(strA) + indexA,
+					 StringToBuffer(strB) + indexB,
+					 (unsigned long)(long)lengthA);
+			if(cmp != 0)
+			{
+				return cmp;
+			}
+			return -1;
+		}
 	}
 	else
 	{
@@ -788,11 +767,11 @@ ILInt32 _IL_String_GetHashCode(ILExecThread *thread, System_String *_this)
 /*
  * public int IndexOf(char value, int startIndex, int count);
  */
-int _IL_String_IndexOf(ILExecThread *thread,
-			 	       System_String *_this,
-					   ILUInt16 value,
-					   ILInt32 startIndex,
-					   ILInt32 count)
+ILInt32 _IL_String_IndexOf(ILExecThread *thread,
+			 	           System_String *_this,
+					       ILUInt16 value,
+					       ILInt32 startIndex,
+					       ILInt32 count)
 {
 	ILUInt16 *buf;
 
@@ -827,11 +806,11 @@ int _IL_String_IndexOf(ILExecThread *thread,
 /*
  * public int IndexOfAny(char[] anyOf, int startIndex, int count);
  */
-int _IL_String_IndexOfAny(ILExecThread *thread,
-				 	      System_String *_this,
-					      System_Array *anyOf,
-					      ILInt32 startIndex,
-					      ILInt32 count)
+ILInt32 _IL_String_IndexOfAny(ILExecThread *thread,
+				 	          System_String *_this,
+					          System_Array *anyOf,
+					          ILInt32 startIndex,
+					          ILInt32 count)
 {
 	ILUInt16 *buf;
 	ILUInt16 *anyBuf;
@@ -887,11 +866,11 @@ int _IL_String_IndexOfAny(ILExecThread *thread,
 /*
  * public int LastIndexOf(char value, int startIndex, int count);
  */
-int _IL_String_LastIndexOf(ILExecThread *thread,
-				 	 	   System_String *_this,
-					 	   ILUInt16 value,
-					 	   ILInt32 startIndex,
-					 	   ILInt32 count)
+ILInt32 _IL_String_LastIndexOf(ILExecThread *thread,
+				 	 	       System_String *_this,
+					 	       ILUInt16 value,
+					 	       ILInt32 startIndex,
+					 	       ILInt32 count)
 {
 	ILUInt16 *buf;
 
@@ -933,11 +912,11 @@ int _IL_String_LastIndexOf(ILExecThread *thread,
 /*
  * public int LastIndexOfAny(char[] anyOf, int startIndex, int count);
  */
-int _IL_String_LastIndexOfAny(ILExecThread *thread,
-							  System_String *_this,
-							  System_Array *anyOf,
-							  ILInt32 startIndex,
-							  ILInt32 count)
+ILInt32 _IL_String_LastIndexOfAny(ILExecThread *thread,
+							      System_String *_this,
+							      System_Array *anyOf,
+							      ILInt32 startIndex,
+							      ILInt32 count)
 {
 	ILUInt16 *buf;
 	ILUInt16 *anyBuf;
@@ -1041,7 +1020,7 @@ ILInt32 _IL_String_FindInRange(ILExecThread *thread, System_String *_this,
 /*
  * Size of the intern'ed string hash table.
  */
-#define	IL_INTERN_HASH_SIZE		512
+#define	IL_INTERN_HASH_SIZE		509
 
 /*
  * Structure of a intern'ed string hash table entry.
@@ -1564,13 +1543,13 @@ int ILStringCompare(ILExecThread *thread, ILString *strA, ILString *strB)
 int ILStringCompareIgnoreCase(ILExecThread *thread, ILString *strA,
 							  ILString *strB)
 {
-	return (int)(_IL_String_InternalCompare
+	return (int)(_IL_String_CompareInternal
 						(thread,
 						 (System_String *)strA, 0,
 						 ((strA != 0) ? ((System_String *)strA)->length : 0),
 						 (System_String *)strB, 0,
 						 ((strB != 0) ? ((System_String *)strB)->length : 0),
-						 (ILBool)1, (void *)0));
+						 (ILBool)1));
 }
 
 int ILStringCompareOrdinal(ILExecThread *thread, ILString *strA,
@@ -1973,6 +1952,39 @@ char *ILStringToUTF8(ILExecThread *thread, ILString *str)
 	return newStr;
 }
 
+ILUInt16 *ILStringToUTF16(ILExecThread *thread, ILString *str)
+{
+	ILUInt16 *buffer;
+	ILInt32 length;
+	ILUInt16 *newStr;
+
+	/* Bail out immediately if the string is NULL */
+	if(!str)
+	{
+		return 0;
+	}
+
+	/* Determine the length of the string in UTF-16 characters */
+	buffer = StringToBuffer(str);
+	length = ((System_String *)str)->length;
+
+	/* Allocate space within the garbage-collected heap */
+	newStr = (ILUInt16 *)ILGCAllocAtomic((length + 1) * sizeof(ILUInt16));
+	if(!newStr)
+	{
+		ILExecThreadThrowOutOfMemory(thread);
+		return 0;
+	}
+
+	/* Copy the characters into the allocated buffer */
+	if(length > 0)
+	{
+		ILMemCpy(newStr, buffer, length * sizeof(ILUInt16));
+	}
+	newStr[length] = 0;
+	return newStr;
+}
+
 char *ILStringToAnsi(ILExecThread *thread, ILString *str)
 {
 	if(str)
@@ -1997,6 +2009,96 @@ char *ILStringToAnsi(ILExecThread *thread, ILString *str)
 	}
 }
 
+/*
+ * public virtual char ToLower(char c);
+ */
+ILUInt16 _IL_TextInfo_ToLower_c(ILExecThread *_thread, ILObject *_this,
+								ILUInt16 c)
+{
+	if(c >= 'A' && c <= 'Z')
+	{
+		return (ILUInt16)(c - 'A' + 'a');
+	}
+	else if(c < 0x0080)
+	{
+		return c;
+	}
+	else
+	{
+		return ILUnicodeCharToLower(c);
+	}
+}
+
+/*
+ * public virtual char ToUpper(char c);
+ */
+ILUInt16 _IL_TextInfo_ToUpper_c(ILExecThread *_thread, ILObject *_this,
+								ILUInt16 c)
+{
+	if(c >= 'a' && c <= 'z')
+	{
+		return (ILUInt16)(c - 'a' + 'A');
+	}
+	else if(c < 0x0080)
+	{
+		return c;
+	}
+	else
+	{
+		return ILUnicodeCharToUpper(c);
+	}
+}
+
+/*
+ * public virtual String ToLower(String str);
+ */
+ILString *_IL_TextInfo_ToLower_String(ILExecThread *_thread, ILObject *_this,
+									  ILString *str)
+{
+	System_String *newStr;
+	ILInt32 len;
+	if(!str)
+	{
+		ILExecThreadThrowArgNull(_thread, "str");
+		return 0;
+	}
+	len = ((System_String *)str)->length;
+	newStr = AllocString(_thread, len);
+	if(!newStr)
+	{
+		return 0;
+	}
+	ILUnicodeStringToLower(StringToBuffer(newStr),
+						   StringToBuffer(str),
+						   (unsigned long)(long)len);
+	return (ILString *)newStr;
+}
+
+/*
+ * public virtual String ToUpper(String str);
+ */
+ILString * _IL_TextInfo_ToUpper_String(ILExecThread *_thread, ILObject *_this,
+									   ILString *str)
+{
+	System_String *newStr;
+	ILInt32 len;
+	if(!str)
+	{
+		ILExecThreadThrowArgNull(_thread, "str");
+		return 0;
+	}
+	len = ((System_String *)str)->length;
+	newStr = AllocString(_thread, len);
+	if(!newStr)
+	{
+		return 0;
+	}
+	ILUnicodeStringToUpper(StringToBuffer(newStr),
+						   StringToBuffer(str),
+						   (unsigned long)(long)len);
+	return (ILString *)newStr;
+}
+
 #ifdef	__cplusplus
 };
 #endif
