@@ -146,6 +146,7 @@ ILObject *_ILGetClrTypeForILType(ILExecThread *thread, ILType *type)
 	/* Convert the type into an "ILClass" structure */
 	classInfo = ILClassFromType(ILProgramItem_Image(thread->method),
 								0, type, 0);
+	classInfo = ILClassResolve(classInfo);
 
 	/* Get the "ClrType" object for the "ILClass" structure */
 	if(classInfo)
@@ -315,8 +316,10 @@ ILObject *_ILGetTypeFromImage(ILExecThread *thread,
 	ILInt32 nameEnd;
 	ILInt32 nameSpecial;
 	int brackets;
+	int rank;
 	ILProgramItem *scope;
 	ILClass *classInfo;
+	ILType *typeInfo;
 
 	/* Validate the parameters */
 	if(!name)
@@ -545,19 +548,68 @@ ILObject *_ILGetTypeFromImage(ILExecThread *thread,
 		return 0;
 	}
 
+	typeInfo = ILClassToType(classInfo);
+
 	/* Resolve special suffixes for array and pointer designations */
 	if(nameSpecial < nameEnd)
 	{
-		/* TODO */
-		if(throwOnError)
+		posn=nameSpecial;
+		brackets=0;
+		rank=0;
+		while(posn < nameEnd)
 		{
-			ThrowTypeLoad(thread, name);
+			switch(buf[posn])
+			{
+				case '[':
+				{
+					rank=1;
+					brackets++;
+				}
+				break;
+				
+				case ']':
+				{
+					if(brackets==0)
+					{
+						if(throwOnError)
+						{
+							ThrowTypeLoad(thread, name);
+						}
+						return 0;
+					}
+					typeInfo = ILTypeFindOrCreateArray(
+									thread->process->context,
+									(unsigned long) rank,
+									typeInfo);
+					rank=0;
+					brackets--;
+				}
+				break;
+
+				case ',':
+				{
+					rank++;
+				}
+				break;
+
+				case '*': /* TODO */
+				case '&': /* TODO */
+				default:
+				{
+					if(throwOnError)
+					{
+						ThrowTypeLoad(thread, name);
+					}
+					return 0;
+				}
+				break;
+			}
+			posn++;
 		}
-		return 0;
 	}
 
 	/* Convert the class information block into a "ClrType" instance */
-	return _ILGetClrType(thread, classInfo);
+	return _ILGetClrTypeForILType(thread, typeInfo);
 }
 
 void _ILClrNotImplemented(ILExecThread *thread)
@@ -647,11 +699,27 @@ ILObject *_IL_ClrType_GetElementType(ILExecThread *thread, ILObject *_this)
 /*
  * Determine if we have an interface name match for "GetInterface".
  */
-static int InterfaceNameMatch(ILClass *interface, ILString *name,
-							  ILBool ignoreCase)
+static int InterfaceNameMatch(ILExecThread *thread, ILClass *interface,
+                              ILString *name, ILBool ignoreCase)
 {
-	/* TODO */
-	return 0;
+	const char *nameUtf8;
+
+	if (!name) { return 0; }
+
+	if (!(nameUtf8 = ILStringToUTF8(thread, name)))
+	{
+		ILExecThreadThrowOutOfMemory(thread);
+		return 0;
+	}
+
+	if (!ignoreCase)
+	{
+		return (!strcmp(interface->className->name, nameUtf8));
+	}
+	else
+	{
+		return (!ILStrICmp(interface->className->name, nameUtf8));
+	}
 }
 
 /*
@@ -665,7 +733,7 @@ static ILClass *GetInterface(ILExecThread *thread, ILClass *classInfo,
 	while((impl = ILClassNextImplements(classInfo, impl)) != 0)
 	{
 		interface = ILImplementsGetInterface(impl);
-		if(InterfaceNameMatch(interface, name, ignoreCase))
+		if(InterfaceNameMatch(thread, interface, name, ignoreCase))
 		{
 			return interface;
 		}
@@ -1157,7 +1225,7 @@ static ILInt32 NameOutputTypeSuffixes(ILUInt16 *buf, ILType *type)
 			{
 				++len;
 			}
-			while(rank > 0)
+			while(rank > 1)
 			{
 				if(buf != 0)
 				{
@@ -1270,6 +1338,7 @@ static ILString *GetTypeName(ILExecThread *thread, ILObject *_this,
 			ILExecThreadThrowOutOfMemory(thread);
 			return 0;
 		}
+		elemInfo = ILClassResolve(elemInfo);
 	}
 
 	/* Compute the size of the full name */
@@ -1397,6 +1466,7 @@ ILString *_IL_ClrType_GetClrNamespace(ILExecThread *thread, ILObject *_this)
 				ILExecThreadThrowOutOfMemory(thread);
 				return 0;
 			}
+			classInfo = ILClassResolve(classInfo);
 		}
 
 		/* Find the outermost type in the nesting levels, to
@@ -1808,7 +1878,7 @@ static int ParameterTypeMatch(ILExecThread *thread, ILImage *image,
 		}
 		else
 		{
-			if(!ILTypeAssignCompatible(image, typeInfo,
+			if(!ILTypeAssignCompatibleNonBoxing(image, typeInfo,
 									   ILTypeGetParam(signature, paramNum + 1)))
 			{
 				return 0;
@@ -1820,6 +1890,88 @@ static int ParameterTypeMatch(ILExecThread *thread, ILImage *image,
 	return 1;
 }
 
+/*
+ * Get the method underlying a member, for permission and access checks.
+ * Returns NULL if there is no underlying method.
+ */
+static ILMethod *GetUnderlyingMethod(ILMember *member)
+{
+	ILMethod *method;
+
+	switch(ILMemberGetKind(member))
+	{
+		case IL_META_MEMBERKIND_METHOD:
+		{
+			return (ILMethod *)member;
+		}
+		/* Not reached */
+
+		case IL_META_MEMBERKIND_PROPERTY:
+		{
+			method = ILProperty_Getter((ILProperty *)member);
+			if(method)
+			{
+				return method;
+			}
+			method = ILProperty_Setter((ILProperty *)member);
+			if(method)
+			{
+				return method;
+			}
+		}
+		break;
+
+		case IL_META_MEMBERKIND_EVENT:
+		{
+			method = ILEvent_AddOn((ILEvent *)member);
+			if(method)
+			{
+				return method;
+			}
+			method = ILEvent_RemoveOn((ILEvent *)member);
+			if(method)
+			{
+				return method;
+			}
+		}
+		break;
+	}
+
+	return 0;
+}
+
+/* 
+ * Check if member1 override member2 or vice versa , return overriding 
+ * member , else return NULL.
+ *
+ * TODO: handle "new" instance methods, fields and methods
+ */
+static ILMember * CheckMemberOverride(ILMember *member1, ILMember *member2)
+{
+	ILMethod *method1, *method2;
+	ILClass *class1, *class2;
+
+	method1=GetUnderlyingMethod(member1);
+	method2=GetUnderlyingMethod(member2);
+	if(!method1 || !method2) return NULL;
+	if(!ILMethod_IsVirtual(method1) || !ILMethod_IsVirtual(method2)) 
+	{
+		return NULL;
+	}
+	class1=ILMember_Owner(member1);
+	class2=ILMember_Owner(member2);
+	/* Note: I'm assuming here that the object heirarchy says it all */
+	if(ILClassInheritsFrom(class1,class2))
+	{
+		return member1;
+	}
+	if(ILClassInheritsFrom(class2, class1))
+	{
+		return member2;
+	}
+	return NULL;
+}
+
 #endif /* IL_CONFIG_REFLECTION */
 
 /*
@@ -1844,9 +1996,12 @@ ILObject *_IL_ClrType_GetMemberImpl(ILExecThread *thread,
 	ILClass *classInfo;
 	char *nameUtf8;
 	ILMember *member;
+	ILMember *other;
 	ILProgramItem *foundItem;
+	ILProgramItem *foundExact;
 	ILNestedInfo *nested;
 	ILClass *child;
+	int inexactIsAmbig;
 
 	/* Validate the parameters */
 	if(!name)
@@ -1884,6 +2039,8 @@ ILObject *_IL_ClrType_GetMemberImpl(ILExecThread *thread,
 
 	/* Scan the class hierarchy looking for a member match */
 	foundItem = 0;
+	foundExact = 0;
+	inexactIsAmbig = 0;
 	do
 	{
 		member = classInfo->firstMember;
@@ -1894,71 +2051,110 @@ ILObject *_IL_ClrType_GetMemberImpl(ILExecThread *thread,
 			{
 				if(strcmp(member->name, nameUtf8) != 0)
 				{
-					member = member->nextMember;
-					continue;
+					goto advance;
 				}
 			}
 			else
 			{
 				if(ILStrICmp(member->name, nameUtf8) != 0)
 				{
-					member = member->nextMember;
-					continue;
+					goto advance;
 				}
 			}
 	
 			/* Filter based on the member type */
 			if(!MemberTypeMatch(member, memberTypes))
 			{
-				member = member->nextMember;
-				continue;
+				goto advance;
 			}
 	
 			/* Filter based on the binding attributes and calling conventions */
 			if(!BindingAttrMatch(member, bindingAttrs, callConv))
 			{
-				member = member->nextMember;
-				continue;
+				goto advance;
 			}
-	
-			/* Filter based on the parameter types */
-			if(types != 0)
-			{
-				if(member->kind == IL_META_MEMBERKIND_METHOD ||
-				   member->kind == IL_META_MEMBERKIND_PROPERTY)
-				{
-					if(!ParameterTypeMatch
-							(thread, member->programItem.image,
-						     member->signature, (System_Array *)types,
-						     (bindingAttrs & BF_ExactBinding) != 0))
-					{
-						member = member->nextMember;
-						continue;
-					}
-				}
-			}
-	
+
 			/* Check that we have reflection access for this member */
 			if(!_ILClrCheckAccess(thread, (ILClass *)0, member))
 			{
-				member = member->nextMember;
-				continue;
+				goto advance;
 			}
-	
-			/* This is a candidate member */
-			if(foundItem)
+
+			/* Filter based on the parameter types */
+			if(types &&
+			   (member->kind == IL_META_MEMBERKIND_METHOD ||
+			    member->kind == IL_META_MEMBERKIND_PROPERTY))
 			{
+				/* Check for an exact match */
+				if(!ParameterTypeMatch(thread, member->programItem.image,
+				                       member->signature, types, 1))
+				{
+					/* Check for an inexact match if neccessary */
+					if(!foundExact && (bindingAttrs & BF_ExactBinding) == 0)
+					{
+						if(!ParameterTypeMatch(thread,
+						                       member->programItem.image,
+						                       member->signature, types, 0))
+						{
+							goto advance;
+						}
+					}
+					else
+					{
+						goto advance;
+					}
+				}
+				else
+				{
+					/* This is a candidate member */
+					if(foundExact)
+					{
+						other = ILProgramItemToMember(foundExact);
+						if(member->kind == other->kind &&
+						   (other = CheckMemberOverride(other, member)))
+						{
+							foundExact = &(other->programItem);
+							goto advance;
+						}
+						/* The member match is ambiguous */
+						goto ambiguous;
+					}
+					foundExact = &(member->programItem);
+				}
+			}
+			/* This is a candidate member */
+			if(!foundExact && foundItem)
+			{
+				other = ILProgramItemToMember(foundItem);
+
+				if(member->kind == other->kind &&
+				   (other = CheckMemberOverride(other, member)))
+				{
+					foundItem = &(other->programItem);
+					goto advance;
+				}
 				/* The member match is ambiguous */
-			ambiguous:
-				ILExecThreadThrowSystem
-						(thread, "System.Reflection.AmbiguousMatchException",
-						 (const char *)0);
-				return 0;
+				inexactIsAmbig = 1;
 			}
 			foundItem = &(member->programItem);
-	
+
+		advance:
 			/* Advance to the next member */
 			member = member->nextMember;
+
+			/* Final ambiguouity check */
+			if(!member && !foundExact && inexactIsAmbig)
+			{
+				goto ambiguous;
+			}
+			continue;
+
+		ambiguous:
+			/* The member match is ambiguous */
+			ILExecThreadThrowSystem
+					(thread, "System.Reflection.AmbiguousMatchException",
+					 (const char *)0);
+			return 0;
 		}
 
 		/* Scan the nested types also */
@@ -1973,14 +2169,14 @@ ILObject *_IL_ClrType_GetMemberImpl(ILExecThread *thread,
 					/* Filter the child based on its name */
 					if((bindingAttrs & (ILInt32)BF_IgnoreCase) == 0)
 					{
-						if(strcmp(child->name, nameUtf8) != 0)
+						if(strcmp(ILClass_Name(child), nameUtf8) != 0)
 						{
 							continue;
 						}
 					}
 					else
 					{
-						if(ILStrICmp(child->name, nameUtf8) != 0)
+						if(ILStrICmp(ILClass_Name(child), nameUtf8) != 0)
 						{
 							continue;
 						}
@@ -2016,7 +2212,11 @@ ILObject *_IL_ClrType_GetMemberImpl(ILExecThread *thread,
 	      (bindingAttrs & (ILInt32)BF_DeclaredOnly) == 0);
 
 	/* Return the item that was found to the caller */
-	if(foundItem)
+	if(foundExact)
+	{
+		return ItemToClrObject(thread, foundExact);
+	}
+	else if(foundItem)
 	{
 		return ItemToClrObject(thread, foundItem);
 	}
@@ -2195,14 +2395,14 @@ ILObject *_IL_ClrType_GetMembersImpl(ILExecThread *thread,
 					{
 						if((bindingAttrs & (ILInt32)BF_IgnoreCase) == 0)
 						{
-							if(strcmp(child->name, nameUtf8) != 0)
+							if(strcmp(ILClass_Name(child), nameUtf8) != 0)
 							{
 								continue;
 							}
 						}
 						else
 						{
-							if(ILStrICmp(child->name, nameUtf8) != 0)
+							if(ILStrICmp(ILClass_Name(child), nameUtf8) != 0)
 							{
 								continue;
 							}
@@ -2308,6 +2508,126 @@ ILObject *_IL_Type_GetTypeFromHandle(ILExecThread *thread, void *handle)
 	}
 }
 
+#ifdef IL_CONFIG_REFLECTION
+
+/*
+ * private static MemberInfo[] InternalGetSerializableMembers(type);
+ */
+System_Array *_IL_FormatterServices_InternalGetSerializableMembers
+			(ILExecThread *_thread, ILObject *type)
+{
+	ILClass *classInfo;
+	ILClass *info;
+	ILField *field;
+	ILInt32 size;
+	System_Array *array;
+	ILObject **buf;
+
+	/* Convert the type into an ILClass structure */
+	classInfo = _ILGetClrClass(_thread, type);
+
+	/* Count the number of serializable fields in the type */
+	info = classInfo;
+	size = 0;
+	while(info != 0)
+	{
+		field = 0;
+		while((field = (ILField *)ILClassNextMemberByKind
+					(info, (ILMember *)field, IL_META_MEMBERKIND_FIELD)) != 0)
+		{
+			if(!ILField_IsStatic(field) && !ILField_IsNotSerialized(field))
+			{
+				++size;
+			}
+		}
+		info = ILClass_Parent(info);
+	}
+
+	/* Allocate an array to hold the serializable fields */
+	array = (System_Array *)ILExecThreadNew
+				(_thread, "[oSystem.Reflection.MemberInfo;",
+				 "(Ti)V", (ILVaInt)size);
+	if(!array)
+	{
+		return 0;
+	}
+	buf = (ILObject **)(ArrayToBuffer(array));
+
+	/* Populate the array */
+	info = classInfo;
+	size = 0;
+	while(info != 0)
+	{
+		field = 0;
+		while((field = (ILField *)ILClassNextMemberByKind
+					(info, (ILMember *)field, IL_META_MEMBERKIND_FIELD)) != 0)
+		{
+			if(!ILField_IsStatic(field) && !ILField_IsNotSerialized(field))
+			{
+				buf[size++] = ItemToClrObject
+					(_thread, &(field->member.programItem));
+			}
+		}
+		info = ILClass_Parent(info);
+	}
+
+	/* Return the final array to the caller */
+	return array;
+}
+
+#endif /* IL_CONFIG_REFLECTION */
+
+/*
+ * protected override bool HasGenericArgumentsImpl();
+ */
+ILBool _IL_ClrType_HasGenericArgumentsImpl(ILExecThread *thread,
+										   ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * protected override bool HasGenericParametersImpl();
+ */
+ILBool _IL_ClrType_HasGenericParametersImpl(ILExecThread *thread,
+										    ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * public override Type[] GetGenericArguments();
+ */
+System_Array *_IL_ClrType_GetGenericArguments(ILExecThread *thread,
+										      ILObject * _this)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * public override Type BindGenericParameters(Type[] inst);
+ */
+ILObject *_IL_ClrType_BindGenericParameters(ILExecThread *thread,
+										    ILObject *_this,
+											System_Array *inst)
+{
+	/* TODO */
+	return 0;
+}
+
+/*
+ * public override Type GetGenericTypeDefinition();
+ */
+ILObject *_IL_ClrType_GetGenericTypeDefinition(ILExecThread *thread,
+											   ILObject *_this)
+{
+	/* TODO */
+	return 0;
+}
+
 #ifdef	__cplusplus
 };
 #endif
