@@ -440,13 +440,25 @@ void *_ILMakeCifForConstructor(ILMethod *method, int isInternal)
 				} \
 			} while (0)
 
+/*
+ * User data for "PackDelegateParams".
+ */
+typedef struct
+{
+	void     **args;
+	ILMethod  *pinvokeInfo;
+	int		   needThis;
+
+} PackDelegateUserData;
+
 /*
  * Pack the parameters for a delegate closure call onto the CVM stack.
  */
 static int PackDelegateParams(ILExecThread *thread, ILMethod *method,
 					          int isCtor, void *_this, void *userData)
 {
-	void **args = (void **)userData;
+	void **args = ((PackDelegateUserData *)userData)->args;
+	ILMethod *pinvokeInfo = ((PackDelegateUserData *)userData)->pinvokeInfo;
 	ILType *signature = ILMethod_Signature(method);
 	CVMWord *stacktop, *stacklimit;
 	ILUInt32 param, numParams;
@@ -455,6 +467,10 @@ static int PackDelegateParams(ILExecThread *thread, ILMethod *method,
 	ILUInt32 size, sizeInWords;
 	ILNativeFloat tempFloat;
 	ILUInt32 marshalType;
+	char *customName;
+	int customNameLen;
+	char *customCookie;
+	int customCookieLen;
 	char *strValue;
 
 	/* Get the top and extent of the stack */
@@ -466,14 +482,27 @@ static int PackDelegateParams(ILExecThread *thread, ILMethod *method,
 	{
 		/* Push the "this" argument */
 		CHECK_SPACE(1);
-		stacktop->ptrValue = _this;
+		if(((PackDelegateUserData *)userData)->needThis)
+		{
+			/* We get the "this" value from the incoming arguments */
+			stacktop->ptrValue = *((void **)(*args));
+			++args;
+		}
+		else
+		{
+			/* We get the "this" value from the delegate object */
+			stacktop->ptrValue = _this;
+		}
 		++stacktop;
 	}
 	numParams = ILTypeNumParams(signature);
 	for(param = 1; param <= numParams; ++param)
 	{
 		/* Marshal parameters that need special handling */
-		marshalType = ILPInvokeGetMarshalType(0, method, param);
+		marshalType = ILPInvokeGetMarshalType(0, pinvokeInfo, param,
+											  &customName, &customNameLen,
+											  &customCookie, &customCookieLen,
+											  ILTypeGetParam(signature, param));
 		if(marshalType != IL_META_MARSHAL_DIRECT)
 		{
 			switch(marshalType)
@@ -522,6 +551,46 @@ static int PackDelegateParams(ILExecThread *thread, ILMethod *method,
 					++stacktop;
 				}
 				continue;
+
+				case IL_META_MARSHAL_UTF16_STRING:
+				{
+					/* Marshal a UTF-16 string from the native world */
+					CHECK_SPACE(1);
+					strValue = *((char **)(*args));
+					if(strValue)
+					{
+						stacktop->ptrValue =
+							ILStringWCreate(thread, (ILUInt16 *)strValue);
+						if(!(stacktop->ptrValue))
+						{
+							return 1;
+						}
+					}
+					else
+					{
+						stacktop->ptrValue = 0;
+					}
+					++args;
+					++stacktop;
+				}
+				continue;
+
+				case IL_META_MARSHAL_CUSTOM:
+				{
+					/* Marshal a custom value from the native world */
+					CHECK_SPACE(1);
+					stacktop->ptrValue = _ILCustomToObject
+						(thread, *((void **)(*args)),
+						 customName, customNameLen,
+						 customCookie, customCookieLen);
+					if(ILExecThreadHasException(thread))
+					{
+						return 1;
+					}
+					++args;
+					++stacktop;
+				}
+				continue;
 			}
 		}
 
@@ -654,16 +723,23 @@ static int PackDelegateParams(ILExecThread *thread, ILMethod *method,
  * Unpack the result of a delegate closure call.
  */
 static void UnpackDelegateResult(ILExecThread *thread, ILMethod *method,
-					             int isCtor, void *result)
+					             int isCtor, void *result, void *userData)
 {
+	ILMethod *pinvokeInfo = ((PackDelegateUserData *)userData)->pinvokeInfo;
 	ILType *signature = ILMethod_Signature(method);
 	ILType *paramType;
 	ILUInt32 size, sizeInWords;
 	ILNativeFloat tempFloat;
 	ILUInt32 marshalType;
+	char *customName;
+	int customNameLen;
+	char *customCookie;
+	int customCookieLen;
 
 	/* Marshal return types that need special handling */
-	marshalType = ILPInvokeGetMarshalType(0, method, 0);
+	marshalType = ILPInvokeGetMarshalType
+			(0, pinvokeInfo, 0, &customName, &customNameLen,
+			 &customCookie, &customCookieLen, ILTypeGetReturn(signature));
 	if(marshalType != IL_META_MARSHAL_DIRECT)
 	{
 		switch(marshalType)
@@ -686,11 +762,20 @@ static void UnpackDelegateResult(ILExecThread *thread, ILMethod *method,
 			}
 			return;
 
+			case IL_META_MARSHAL_UTF16_STRING:
+			{
+				/* Marshal a UTF-16 string back to the native world */
+				*((ILUInt16 **)result) = ILStringToUTF16
+					(thread, (ILString *)(thread->stackTop[-1].ptrValue));
+				--(thread->stackTop);
+			}
+			return;
+
 			case IL_META_MARSHAL_FNPTR:
 			{
 				/* Convert a delegate into a function closure pointer */
 				*((void **)result) = _ILDelegateGetClosure
-					((ILObject *)(thread->stackTop[-1].ptrValue));
+					(thread, (ILObject *)(thread->stackTop[-1].ptrValue));
 				--(thread->stackTop);
 			}
 			return;
@@ -710,6 +795,16 @@ static void UnpackDelegateResult(ILExecThread *thread, ILMethod *method,
 				}
 			}
 			return;
+
+			case IL_META_MARSHAL_CUSTOM:
+			{
+				/* Marshal a custom value to the native world */
+				*((void **)result) = _ILObjectToCustom
+					(thread, (ILObject *)(thread->stackTop[-1].ptrValue),
+					 customName, customNameLen, customCookie, customCookieLen);
+				--(thread->stackTop);
+			}
+			return;
 		}
 	}
 
@@ -825,6 +920,7 @@ static void DelegateInvoke(ffi_cif *cif, void *result,
 	ILMethod *method;
 	ILType *type;
 	ILUInt32 size;
+	PackDelegateUserData userData;
 
 	/* If this is a multicast delegate, then execute "prev" first */
 	if(((System_Delegate *)delegate)->prev)
@@ -847,10 +943,47 @@ static void DelegateInvoke(ffi_cif *cif, void *result,
 	}
 
 	/* Call the method */
+	userData.args = args;
+	userData.pinvokeInfo = (ILMethod *)ILTypeGetDelegateMethod
+		(ILType_FromClass(GetObjectClass(delegate)));
+	userData.needThis = 0;
 	if(_ILCallMethod(thread, method,
 				     UnpackDelegateResult, result,
 				     0, ((System_Delegate *)delegate)->target,
-				     PackDelegateParams, args))
+				     PackDelegateParams, &userData))
+	{
+		/* An exception occurred, which is already stored in the thread */
+		type = ILMethod_Signature(method);
+		type = ILTypeGetEnumType(ILTypeGetReturn(type));
+		if(type != ILType_Void)
+		{
+			/* Clear the native return value, because we cannot assume
+			   that the native caller knows how to handle exceptions */
+			size = ILSizeOfType(thread, type);
+			ILMemZero(result, size);
+		}
+	}
+}
+
+/*
+ * Invoke a method from a closure.
+ */
+static void MethodInvoke(ffi_cif *cif, void *result,
+						 void **args, void *_method)
+{
+	ILExecThread *thread = ILExecThreadCurrent();
+	ILMethod *method = (ILMethod *)_method;
+	ILType *type;
+	ILUInt32 size;
+	PackDelegateUserData userData;
+
+	/* Call the method */
+	userData.args = args;
+	userData.pinvokeInfo = method;
+	userData.needThis = 0;
+	if(_ILCallMethod(thread, method,
+				     UnpackDelegateResult, result, 0, 0,
+				     PackDelegateParams, &userData))
 	{
 		/* An exception occurred, which is already stored in the thread */
 		type = ILMethod_Signature(method);
@@ -885,7 +1018,7 @@ void *_ILMakeClosureForDelegate(ILObject *delegate, ILMethod *method)
 
 	/* Allocate space for the cif */
 	cif = (ffi_cif *)ILMalloc(sizeof(ffi_cif) +
-							  sizeof(ffi_type *) * numArgs);
+							  sizeof(ffi_type *) * (numArgs + 1));
 	if(!cif)
 	{
 		return 0;
@@ -897,6 +1030,11 @@ void *_ILMakeClosureForDelegate(ILObject *delegate, ILMethod *method)
 
 	/* Convert the argument types */
 	arg = 0;
+	if(!delegate && ILType_HasThis(signature))
+	{
+		/* We need an extra argument for the method's "this" value */
+		args[arg++] = &ffi_type_pointer;
+	}
 	for(param = 1; param <= numArgs; ++param)
 	{
 		args[arg++] = TypeToFFI(ILTypeGetEnumType
@@ -904,7 +1042,7 @@ void *_ILMakeClosureForDelegate(ILObject *delegate, ILMethod *method)
 	}
 
 	/* Prepare the "ffi_cif" structure for the call */
-	if(ffi_prep_cif(cif, FFI_DEFAULT_ABI, numArgs, rtype, args) != FFI_OK)
+	if(ffi_prep_cif(cif, FFI_DEFAULT_ABI, arg, rtype, args) != FFI_OK)
 	{
 		fprintf(stderr, "Cannot marshal a type in the definition of %s::%s\n",
 				ILClass_Name(ILMethod_Owner(method)), ILMethod_Name(method));
@@ -921,12 +1059,25 @@ void *_ILMakeClosureForDelegate(ILObject *delegate, ILMethod *method)
 	}
 
 	/* Prepare the closure using the call parameters */
-	if(ffi_prep_closure(closure, cif, DelegateInvoke, (void *)delegate)
-			!= FFI_OK)
+	if(delegate)
 	{
-		fprintf(stderr, "Cannot create a closure for %s::%s\n",
+		if(ffi_prep_closure(closure, cif, DelegateInvoke, (void *)delegate)
+				!= FFI_OK)
+		{
+			fprintf(stderr, "Cannot create a closure for %s::%s\n",
 				ILClass_Name(ILMethod_Owner(method)), ILMethod_Name(method));
-		return 0;
+			return 0;
+		}
+	}
+	else
+	{
+		if(ffi_prep_closure(closure, cif, MethodInvoke, (void *)method)
+				!= FFI_OK)
+		{
+			fprintf(stderr, "Cannot create a closure for %s::%s\n",
+				ILClass_Name(ILMethod_Owner(method)), ILMethod_Name(method));
+			return 0;
+		}
 	}
 
 	/* The closure is ready to go */
