@@ -172,7 +172,8 @@ static ILType *GetCallSiteSig(ILMethod *method, unsigned char *pc)
 static ILInt32 MatchSignature(ILCoder *coder, ILEngineStackItem *stack,
 						      ILUInt32 stackSize, ILType *signature,
 						      ILMethod *method, int unsafeAllowed,
-							  int suppressThis, int indirectCall)
+							  int suppressThis, int indirectCall,
+							  ILCoderMethodInfo *callInfo, int tailCall)
 {
 	ILClass *owner = (method ? ILMethod_Owner(method) : 0);
 	ILUInt32 numParams;
@@ -568,6 +569,11 @@ static ILInt32 MatchSignature(ILCoder *coder, ILEngineStackItem *stack,
 				{
 					return -1;
 				}
+				if(tailCall && !unsafeAllowed)
+				{
+					/* Cannot use managed pointers in tail calls */
+					return -1;
+				}
 			}
 			break;
 
@@ -612,12 +618,21 @@ static ILInt32 MatchSignature(ILCoder *coder, ILEngineStackItem *stack,
 		}
 	}
 
+	/* Initialize the call argument information */
+	callInfo->args = stack;
+	callInfo->numBaseArgs = numParams;
+	callInfo->numVarArgs = totalParams - numParams;
+	callInfo->hasParamArray = 0;
+	callInfo->tailCall = tailCall;
+
 #ifdef IL_CONFIG_VARARGS
 	/* Convert the vararg parameters into an "Object[]" array */
 	if(isVarArg)
 	{
-		ILCoderPackVarArgs(coder, signature, numParams + 2,
+		ILCoderPackVarArgs(coder, signature,
+						   (hasThis ? numParams + 1 : numParams + 2),
 						   stack + numParams, totalParams - numParams);
+		callInfo->hasParamArray = 1;
 	}
 #endif /* IL_CONFIG_VARARGS */
 
@@ -791,7 +806,15 @@ static void InsertCtorArgs(ILEngineStackItem *stack, ILUInt32 stackSize,
 		--stackSize;
 		stack[stackSize + 2] = stack[stackSize];
 	}
-	stack[insertPosn].engineType = engineType1;
+	if(typeInfo1 == ILType_Int || typeInfo1 == ILType_UInt)
+	{
+		/* Convert MV into I for the IntPtr and UIntPtr types */
+		stack[insertPosn].engineType = ILEngineType_I;
+	}
+	else
+	{
+		stack[insertPosn].engineType = engineType1;
+	}
 	stack[insertPosn].typeInfo = typeInfo1;
 	stack[insertPosn + 1].engineType = engineType2;
 	stack[insertPosn + 1].typeInfo = typeInfo2;
@@ -815,6 +838,55 @@ static int Is2DArrayClass(ILClass *classInfo)
 	return 0;
 }
 
+/*
+ * Table of inline methods.
+ */
+typedef struct
+{
+	const char *name;
+	const char *namespace;
+	const char *methodName;
+	const char *signature;
+	int         inlineType;
+
+} InlineMethodInfo;
+static InlineMethodInfo const InlineMethods[] = {
+	{"String", "System", "get_Length", "(T)i", IL_INLINEMETHOD_STRING_LENGTH},
+	{"String", "System", "Concat",
+	 "(oSystem.String;oSystem.String;)oSystem.String;",
+	 IL_INLINEMETHOD_STRING_CONCAT_2},
+	{"String", "System", "Concat",
+	 "(oSystem.String;oSystem.String;oSystem.String;)oSystem.String;",
+	 IL_INLINEMETHOD_STRING_CONCAT_3},
+	{"String", "System", "Concat",
+	 "(oSystem.String;oSystem.String;oSystem.String;oSystem.String;)"
+	 		"oSystem.String;",
+	 IL_INLINEMETHOD_STRING_CONCAT_4},
+	{"String", "System", "op_Equality", "(oSystem.String;oSystem.String;)Z",
+	 IL_INLINEMETHOD_STRING_EQUALS},
+	{"String", "System", "Equals", "(oSystem.String;oSystem.String;)Z",
+	 IL_INLINEMETHOD_STRING_EQUALS},
+	{"String", "System", "op_Inequality", "(oSystem.String;oSystem.String;)Z",
+	 IL_INLINEMETHOD_STRING_NOT_EQUALS},
+	{"String", "System", "get_Chars", "(Ti)c",
+	 IL_INLINEMETHOD_STRING_GET_CHAR},
+
+	{"Monitor", "System.Threading", "Enter", "(oSystem.Object;)V",
+	 IL_INLINEMETHOD_MONITOR_ENTER},
+	{"Monitor", "System.Threading", "Exit", "(oSystem.Object;)V",
+	 IL_INLINEMETHOD_MONITOR_EXIT},
+
+	{"Type", "System", "GetTypeFromHandle",
+	 "(vSystem.RuntimeTypeHandle;)oSystem.Type;",
+	 IL_INLINEMETHOD_TYPE_FROM_HANDLE},
+
+	{"StringBuilder", "System.Text", "Append",
+	 "(Tc)oSystem.Text.StringBuilder;", IL_INLINEMETHOD_BUILDER_APPEND_CHAR},
+
+	{"Char", "System", "IsWhiteSpace", "(c)Z", IL_INLINEMETHOD_IS_WHITE_SPACE},
+};
+#define	NumInlineMethods	(sizeof(InlineMethods) / sizeof(InlineMethodInfo))
+
 /*
  * Determine if a method is inlineable, and return its inline type.
  * Returns -1 if the method is not inlineable.
@@ -823,136 +895,45 @@ static int GetInlineMethodType(ILMethod *method)
 {
 	ILClass *owner;
 	const char *name;
+	const char *namespace;
+	const char *methodName;
+	int posn;
 	ILImage *image;
 	ILImage *systemImage;
 	ILType *signature;
 
+	/* Extract the information that we require to find the inline */
 	owner = ILMethod_Owner(method);
 	name = ILClass_Name(owner);
-	if(!strcmp(name, "String"))
-	{
-		/* Make sure that this is really "System.String" */
-		name = ILClass_Namespace(owner);
-		if(!name || strcmp(name, "System") != 0)
-		{
-			return -1;
-		}
-		image = ILClassToImage(owner);
-		systemImage = ILContextGetSystem(ILImageToContext(image));
-		if(systemImage && systemImage != image)
-		{
-			return -1;
-		}
+	namespace = ILClass_Namespace(owner);
+	methodName = ILMethod_Name(method);
+	signature = ILMethod_Signature(method);
 
-		/* Check for known inlineable methods */
-		name = ILMethod_Name(method);
-		signature = ILMethod_Signature(method);
-		if(!strcmp(name, "get_Length") &&
-		   _ILLookupTypeMatch(signature, "(T)i"))
+	/* Check for types in the runtime library */
+	if(namespace)
+	{
+		for(posn = 0; posn < NumInlineMethods; ++posn)
 		{
-			return IL_INLINEMETHOD_STRING_LENGTH;
-		}
-		else if(!strcmp(name, "Concat"))
-		{
-			if(_ILLookupTypeMatch(signature,
-				"(oSystem.String;oSystem.String;)oSystem.String;"))
+			if(!strcmp(name, InlineMethods[posn].name) &&
+			   !strcmp(namespace, InlineMethods[posn].namespace) &&
+			   !strcmp(methodName, InlineMethods[posn].methodName) &&
+			   _ILLookupTypeMatch(signature, InlineMethods[posn].signature))
 			{
-				return IL_INLINEMETHOD_STRING_CONCAT_2;
-			}
-			else if(_ILLookupTypeMatch(signature,
-						"(oSystem.String;oSystem.String;"
-						"oSystem.String;)oSystem.String;"))
-			{
-				return IL_INLINEMETHOD_STRING_CONCAT_3;
-			}
-			else if(_ILLookupTypeMatch(signature,
-						"(oSystem.String;oSystem.String;oSystem.String;"
-						"oSystem.String;)oSystem.String;"))
-			{
-				return IL_INLINEMETHOD_STRING_CONCAT_4;
+				image = ILClassToImage(owner);
+				systemImage = ILContextGetSystem(ILImageToContext(image));
+				if(!systemImage || systemImage == image)
+				{
+					return InlineMethods[posn].inlineType;
+				}
 			}
 		}
-		else if((!strcmp(name, "Equals") || !strcmp(name, "op_Equality")) &&
-				_ILLookupTypeMatch(signature,
-					"(oSystem.String;oSystem.String;)Z"))
-		{
-			return IL_INLINEMETHOD_STRING_EQUALS;
-		}
-		else if(!strcmp(name, "op_Inequality") &&
-				_ILLookupTypeMatch(signature,
-					"(oSystem.String;oSystem.String;)Z"))
-		{
-			return IL_INLINEMETHOD_STRING_NOT_EQUALS;
-		}
-		else if(!strcmp(name, "get_Chars") &&
-				_ILLookupTypeMatch(signature, "(Ti)c"))
-		{
-			return IL_INLINEMETHOD_STRING_GET_CHAR;
-		}
-		return -1;
 	}
-	else if(!strcmp(name, "Monitor"))
-	{
-		/* Make sure that this is really "System.Threading.Monitor" */
-		name = ILClass_Namespace(owner);
-		if(!name || strcmp(name, "System.Threading") != 0)
-		{
-			return -1;
-		}
-		image = ILClassToImage(owner);
-		systemImage = ILContextGetSystem(ILImageToContext(image));
-		if(systemImage && systemImage != image)
-		{
-			return -1;
-		}
 
-		/* Check for known inlineable methods */
-		name = ILMethod_Name(method);
-		signature = ILMethod_Signature(method);
-		if(!strcmp(name, "Enter") &&
-		   _ILLookupTypeMatch(signature, "(oSystem.Object;)V"))
-		{
-			return IL_INLINEMETHOD_MONITOR_ENTER;
-		}
-		else if(!strcmp(name, "Exit") &&
-		        _ILLookupTypeMatch(signature, "(oSystem.Object;)V"))
-		{
-			return IL_INLINEMETHOD_MONITOR_EXIT;
-		}
-		return -1;
-	}
-	else if(!strcmp(name, "Type"))
-	{
-		/* Make sure that this is really "System.Type" */
-		name = ILClass_Namespace(owner);
-		if(!name || strcmp(name, "System") != 0)
-		{
-			return -1;
-		}
-		image = ILClassToImage(owner);
-		systemImage = ILContextGetSystem(ILImageToContext(image));
-		if(systemImage && systemImage != image)
-		{
-			return -1;
-		}
-
-		/* Check for known inlineable methods */
-		name = ILMethod_Name(method);
-		signature = ILMethod_Signature(method);
-		if(!strcmp(name, "GetTypeFromHandle") &&
-		   _ILLookupTypeMatch(signature,
-		   					  "(vSystem.RuntimeTypeHandle;)oSystem.Type;"))
-		{
-			return IL_INLINEMETHOD_TYPE_FROM_HANDLE;
-		}
-		return -1;
-	}
-	else if(Is2DArrayClass(owner))
+	/* Check for 2D array types */
+	if(Is2DArrayClass(owner))
 	{
 		/* Two-dimensional array operation */
-		name = ILMethod_Name(method);
-		signature = ILMethod_Signature(method);
-		if(!strcmp(name, "Get"))
+		if(!strcmp(methodName, "Get"))
 		{
 			signature = ILTypeGetReturn(signature);
 			if(signature == ILType_Int32)
@@ -968,7 +949,7 @@ static int GetInlineMethodType(ILMethod *method)
 				return IL_INLINEMETHOD_GET2D_OBJECT;
 			}
 		}
-		else if(!strcmp(name, "Set") && ILTypeNumParams(signature) > 0)
+		else if(!strcmp(methodName, "Set") && ILTypeNumParams(signature) > 0)
 		{
 			signature = ILTypeGetParam(signature, ILTypeNumParams(signature));
 			if(signature == ILType_Int32)
@@ -986,10 +967,24 @@ static int GetInlineMethodType(ILMethod *method)
 		}
 		return -1;
 	}
+
+	/* This is not an inlineable method */
+	return -1;
+}
+
+/*
+ * Set return type information within a stack item.
+ */
+static void SetReturnType(ILEngineStackItem *item, ILType *returnType)
+{
+	item->engineType = TypeToEngineType(returnType);
+	if(item->engineType != ILEngineType_M)
+	{
+		item->typeInfo = returnType;
+	}
 	else
 	{
-		/* This class does not have inlineable methods */
-		return -1;
+		item->typeInfo = ILType_Ref(ILTypeStripPrefixes(returnType));
 	}
 }
 
@@ -997,6 +992,7 @@ static int GetInlineMethodType(ILMethod *method)
 
 ILType *methodSignature;
 ILType *returnType;
+ILCoderMethodInfo callInfo;
 ILInt32 numParams;
 int inlineType;
 
@@ -1051,14 +1047,15 @@ case IL_OP_CALL:
 		{
 			numParams = MatchSignature(coder, stack, stackSize,
 									   methodSignature, methodInfo,
-									   unsafeAllowed, 0, 0);
+									   unsafeAllowed, 0, 0,
+									   &callInfo, tailCall);
+			tailCall = 0;
 			if(numParams >= 0)
 			{
 				returnType = ILTypeGetReturn(methodSignature);
 				if(returnType != ILType_Void)
 				{
-					stack[stackSize].engineType = TypeToEngineType(returnType);
-					stack[stackSize].typeInfo = returnType;
+					SetReturnType(&(stack[stackSize]), returnType);
 				}
 				else
 				{
@@ -1066,10 +1063,9 @@ case IL_OP_CALL:
 				}
 				inlineType = GetInlineMethodType(methodInfo);
 				if(inlineType == -1 ||
-				   !ILCoderCallInlineable(coder, inlineType))
+				   !ILCoderCallInlineable(coder, inlineType, methodInfo))
 				{
-					ILCoderCallMethod(coder, stack + stackSize - numParams,
-									  (ILUInt32)numParams, &(stack[stackSize]),
+					ILCoderCallMethod(coder, &callInfo, &(stack[stackSize]),
 									  methodInfo);
 				}
 				stackSize -= (ILUInt32)numParams;
@@ -1146,21 +1142,21 @@ case IL_OP_CALLI:
 		/* Match the signature against the current stack contents */
 		numParams = MatchSignature(coder, stack, stackSize,
 								   methodSignature, 0,
-								   unsafeAllowed, 0, 1);
+								   unsafeAllowed, 0, 1,
+								   &callInfo, tailCall);
+		tailCall = 0;
 		if(numParams >= 0)
 		{
 			returnType = ILTypeGetReturn(methodSignature);
 			if(returnType != ILType_Void)
 			{
-				stack[stackSize].engineType = TypeToEngineType(returnType);
-				stack[stackSize].typeInfo = returnType;
+				SetReturnType(&(stack[stackSize]), returnType);
 			}
 			else
 			{
 				stack[stackSize].engineType = ILEngineType_Invalid;
 			}
-			ILCoderCallIndirect(coder, stack + stackSize - numParams,
-							    (ILUInt32)numParams, &(stack[stackSize]));
+			ILCoderCallIndirect(coder, &callInfo, &(stack[stackSize]));
 			stackSize -= (ILUInt32)numParams;
 			if(returnType != ILType_Void)
 			{
@@ -1229,14 +1225,15 @@ case IL_OP_CALLVIRT:
 		{
 			numParams = MatchSignature(coder, stack, stackSize,
 									   methodSignature, methodInfo,
-									   unsafeAllowed, 0, 0);
+									   unsafeAllowed, 0, 0,
+									   &callInfo, tailCall);
+			tailCall = 0;
 			if(numParams >= 0)
 			{
 				returnType = ILTypeGetReturn(methodSignature);
 				if(returnType != ILType_Void)
 				{
-					stack[stackSize].engineType = TypeToEngineType(returnType);
-					stack[stackSize].typeInfo = returnType;
+					SetReturnType(&(stack[stackSize]), returnType);
 				}
 				else
 				{
@@ -1249,32 +1246,27 @@ case IL_OP_CALLVIRT:
 					   "call" is probably a better way to do it */
 					inlineType = GetInlineMethodType(methodInfo);
 					if(inlineType == -1 ||
-					   !ILCoderCallInlineable(coder, inlineType))
+					   !ILCoderCallInlineable(coder, inlineType, methodInfo))
 					{
 						if(numParams > 0 && !ILMethod_IsStatic(methodInfo) &&
 						   (stack + stackSize - numParams)->engineType
 						   		== ILEngineType_O)
 						{
 							/* Check the first parameter against "null" */
-							ILCoderCheckCallNull
-								(coder, stack + stackSize - numParams,
-								 (ILUInt32)numParams);
+							ILCoderCheckCallNull(coder, &callInfo);
 						}
-						ILCoderCallMethod(coder, stack + stackSize - numParams,
-										  (ILUInt32)numParams,
+						ILCoderCallMethod(coder, &callInfo,
 										  &(stack[stackSize]), methodInfo);
 					}
 				}
 				else if(ILClass_IsInterface(ILMethod_Owner(methodInfo)))
 				{
-					ILCoderCallInterface(coder, stack + stackSize - numParams,
-									     (ILUInt32)numParams,
+					ILCoderCallInterface(coder, &callInfo,
 										 &(stack[stackSize]), methodInfo);
 				}
 				else
 				{
-					ILCoderCallVirtual(coder, stack + stackSize - numParams,
-									   (ILUInt32)numParams,
+					ILCoderCallVirtual(coder, &callInfo,
 									   &(stack[stackSize]), methodInfo);
 				}
 				stackSize -= (ILUInt32)numParams;
@@ -1328,11 +1320,14 @@ case IL_OP_NEWOBJ:
 			}
 
 			/* Call the allocation constructor for the delegate */
-			ILCoderCallCtor(coder, stack + stackSize - numParams,
-						    (ILUInt32)numParams, methodInfo);
+			callInfo.args = stack + stackSize - numParams;
+			callInfo.numBaseArgs = (ILUInt32)numParams;
+			callInfo.numVarArgs = 0;
+			callInfo.hasParamArray = 0;
+			ILCoderCallCtor(coder, &callInfo, methodInfo);
 			stackSize -= (ILUInt32)numParams;
 			stack[stackSize].engineType = ILEngineType_O;
-			stack[stackSize].typeInfo = ILType_FromClass(classInfo);
+			stack[stackSize].typeInfo = ILClassToType(classInfo);
 			++stackSize;
 		}
 		else if(!ILClassIsValueType(classInfo))
@@ -1340,20 +1335,21 @@ case IL_OP_NEWOBJ:
 			/* Match the signature for the allocation constructor */
 			numParams = MatchSignature(coder, stack, stackSize,
 									   methodSignature, methodInfo,
-									   unsafeAllowed, 1, 0);
+									   unsafeAllowed, 1, 0,
+									   &callInfo, tailCall);
+			tailCall = 0;
 			if(numParams < 0)
 			{
 				VERIFY_TYPE_ERROR();
 			}
 
 			/* Call the allocation constructor for the class */
-			ILCoderCallCtor(coder, stack + stackSize - numParams,
-						    (ILUInt32)numParams, methodInfo);
+			ILCoderCallCtor(coder, &callInfo, methodInfo);
 			stackSize -= (ILUInt32)numParams;
 			if(stackSize < code->maxStack)
 			{
 				stack[stackSize].engineType = ILEngineType_O;
-				stack[stackSize].typeInfo = ILType_FromClass(classInfo);
+				stack[stackSize].typeInfo = ILClassToType(classInfo);
 				++stackSize;
 			}
 			else
@@ -1380,15 +1376,16 @@ case IL_OP_NEWOBJ:
 			/* Match the constructor signature */
 			numParams = MatchSignature(coder, stack, stackSize,
 									   methodSignature, methodInfo,
-									   unsafeAllowed, 0, 0);
+									   unsafeAllowed, 0, 0,
+									   &callInfo, tailCall);
+			tailCall = 0;
 			if(numParams < 0)
 			{
 				VERIFY_TYPE_ERROR();
 			}
 
 			/* Call the constructor and pop all of its arguments */
-			ILCoderCallMethod(coder, stack + stackSize - numParams,
-							  (ILUInt32)numParams, 0, methodInfo);
+			ILCoderCallMethod(coder, &callInfo, 0, methodInfo);
 			stackSize -= (ILUInt32)numParams;
 
 			/* Make sure that we had at least 1 real stack slot
@@ -1443,7 +1440,7 @@ case IL_OP_PREFIX + IL_PREFIX_OP_LDVIRTFTN:
 	methodInfo = GetMethodToken(method, pc, (ILType **)0);
 	if(methodInfo && STK_UNARY == ILEngineType_O)
 	{
-		classInfo = ILMethod_Owner(method);
+		classInfo = ILMethod_Owner(methodInfo);
 		if(AssignCompatible(method, &(stack[stackSize - 1]),
 							ILType_FromClass(classInfo),
 							unsafeAllowed))
@@ -1502,86 +1499,27 @@ break;
 
 case IL_OP_PREFIX + IL_PREFIX_OP_TAIL:
 {
-	/* Missing: (?)
-	 *
-	 * Correct CIL must not branch to the call instruction, but it is
-	 * permitted to branch to ret.
-	 *
-	 * The tail.call (or calli or callvirt) instruction cannot be used
-	 * to transfer control out of a try, filter, catch, or finally block.
-	 */
-
-	/* Confirm that the .tail precedes a call */
-	if ((pc[2] == IL_OP_CALL || pc[2] == IL_OP_CALLI || 
-		 pc[2] == IL_OP_CALLVIRT) && pc[7] == IL_OP_RET)
+	/* Need at least 8 bytes to make up the full tail call sequence */
+	if(len < 8)
 	{
-		/*  Initialization */
-		methodInfo = 0;
-		methodSignature = 0;
-		numParams = 0;
+		VERIFY_TRUNCATED();
+	}
 
-		/*  Get called method information as appropriate. */
-		switch(pc[2]) 
-		{
-		case IL_OP_CALL:
-			/* Test that only the target method's arguments are on the stack */
-			methodInfo = GetMethodToken(method, &pc[2], &methodSignature);
-			numParams = MatchSignature(coder, stack, stackSize,
-								   methodSignature, methodInfo,
-								   unsafeAllowed, 0, 0);
+	/* Check that the instruction after the tail call is not a branch target */
+	if(IsJumpTarget(jumpMask, offset + 2))
+	{
+		VERIFY_BRANCH_ERROR();
+	}
 
-		case IL_OP_CALLI:
-		case IL_OP_CALLVIRT:
-			/*  TODO - Implement CALLI and CALLVIRT tails */
-			break;
-		}
-
-		/*  A method exists and it's not synchronized? */
-		if (methodInfo && !ILMethod_IsSynchronized(methodInfo)) 
-		{
-			/*  Test that the stack contents are sane */
-			if (stackSize == numParams) 
-			{
-				if (unsafeAllowed) 
-				{
-					/*  Don't bother checking for pointers */
-					ILCoderTailCall(coder, method);
-				} 
-				else 
-				{
-					/*  Test for managed pointers on the stack - a no-no */
-					int i, mPtrFlag;
-					mPtrFlag = 0;
-					for (i=0; i<numParams; i++)
-					{
-						if (stack[i].engineType == ILEngineType_M ||
-								stack[i].engineType == ILEngineType_T)
-						{
-							mPtrFlag = 1;
-							break;
-						}
-					}
-
-					if (mPtrFlag) 
-					{
-						VERIFY_TYPE_ERROR();
-					} 
-					else 
-					{
-						ILCoderTailCall(coder, method);
-					}
-				}
-			}
-			else
-			{
-				VERIFY_STACK_ERROR();
-			}
-		}
-	} 
-	else 
+	/* Check that the instruction sequence is as expected */
+	if((pc[2] != IL_OP_CALL && pc[2] == IL_OP_CALLI &&
+	    pc[2] != IL_OP_CALLVIRT) || pc[7] != IL_OP_RET)
 	{
 		VERIFY_INSN_ERROR();
 	}
+
+	/* Set the tail call flag and then advance to the next instruction */
+	tailCall = 1;
 }
 break;
 
