@@ -556,7 +556,7 @@ case IL_OP_NEWARR:
 	if(classType != 0 &&
 	   (STK_UNARY == ILEngineType_I || STK_UNARY == ILEngineType_I4))
 	{
-		classType = ILTypeCreateArray
+		classType = ILTypeFindOrCreateArray
 				(ILImageToContext(ILProgramItem_Image(method)), 1, classType);
 		if(!classType)
 		{
@@ -568,6 +568,7 @@ case IL_OP_NEWARR:
 		{
 			VERIFY_MEMORY_ERROR();
 		}
+		classInfo = ILClassResolve(classInfo);
 		ILCoderNewArray(coder, classType, classInfo, STK_UNARY);
 		stack[stackSize - 1].engineType = ILEngineType_O;
 		stack[stackSize - 1].typeInfo = classType;
