@@ -1,7 +1,7 @@
 /*
  * link_library.c - Process libraries within a linker context.
  *
- * Copyright (C) 2001  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2001, 2003  Southern Storm Software, Pty Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
  */
 
 #include "linker.h"
+#include "il_crypt.h"
 #ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
@@ -30,12 +31,6 @@
 extern	"C" {
 #endif
 
-/*
- * Size of the name hash table associated with a library.
- * Must be a power of 2.
- */
-#define	IL_LINK_HASH_SIZE		512
-
 /*
  * Determine if a directory exists.
  */
@@ -179,6 +174,10 @@ static void LibraryDestroy(ILLibrary *library)
 		next = library->altNames;
 		ILFree(library->name);
 		ILFree(library->filename);
+		if(library->publicKey)
+		{
+			ILFree(library->publicKey);
+		}
 		ILHashDestroy(library->classHash);
 		ILHashDestroy(library->symbolHash);
 		ILMemPoolDestroy(&(library->pool));
@@ -233,7 +232,10 @@ static int ClassHash_Match(const ILLibraryClass *libClass,
 	{
 		if(key->namespace)
 		{
-			return !strcmp(libClass->namespace, key->namespace);
+			if(strcmp(libClass->namespace, key->namespace) != 0)
+			{
+				return 0;
+			}
 		}
 		else
 		{
@@ -242,7 +244,10 @@ static int ClassHash_Match(const ILLibraryClass *libClass,
 	}
 	else
 	{
-		return (key->namespace == 0);
+		if(key->namespace != 0)
+		{
+			return 0;
+		}
 	}
 	return (key->parent == libClass->parent);
 }
@@ -284,6 +289,8 @@ static ILLibrary *ScanAssemblies(ILLinker *linker, ILContext *context,
 	ILLibrary *lastLibrary;
 	ILLibrary *nextLibrary;
 	ILAssembly *assem;
+	const void *originator;
+	unsigned long originatorLen;
 
 	/* Scan the assembly definitions */
 	library = 0;
@@ -299,6 +306,8 @@ static ILLibrary *ScanAssemblies(ILLinker *linker, ILContext *context,
 			LibraryDestroy(library);
 			return 0;
 		}
+		nextLibrary->publicKey = 0;
+		nextLibrary->publicKeyLen = 0;
 		nextLibrary->name = ILDupString(ILAssembly_Name(assem));
 		if(!(nextLibrary->name))
 		{
@@ -323,9 +332,24 @@ static ILLibrary *ScanAssemblies(ILLinker *linker, ILContext *context,
 		{
 			nextLibrary->classHash = 0;
 			nextLibrary->symbolHash = 0;
+			nextLibrary->memoryModel = library->memoryModel;
+			nextLibrary->modelFlags = library->modelFlags;
 		}
 		else
 		{
+			nextLibrary->memoryModel =
+				ILLinkerCMemoryModel(image, &(nextLibrary->modelFlags));
+			originator = ILAssemblyGetOriginator(assem, &originatorLen);
+			if(originator && originatorLen)
+			{
+				nextLibrary->publicKey =
+					(unsigned char *)ILMalloc(originatorLen + 1);
+				if(nextLibrary->publicKey)
+				{
+					ILMemCpy(nextLibrary->publicKey, originator, originatorLen);
+					nextLibrary->publicKeyLen = (ILUInt32)originatorLen;
+				}
+			}
 			nextLibrary->classHash =
 				ILHashCreate(0, (ILHashComputeFunc)ClassHash_Compute,
 								(ILHashKeyComputeFunc)ClassHash_KeyCompute,
@@ -333,6 +357,10 @@ static ILLibrary *ScanAssemblies(ILLinker *linker, ILContext *context,
 								(ILHashFreeFunc)0);
 			if(!(nextLibrary->classHash))
 			{
+				if(nextLibrary->publicKey)
+				{
+					ILFree(nextLibrary->publicKey);
+				}
 				ILFree(nextLibrary->filename);
 				ILFree(nextLibrary->name);
 				ILFree(nextLibrary);
@@ -347,6 +375,10 @@ static ILLibrary *ScanAssemblies(ILLinker *linker, ILContext *context,
 			if(!(nextLibrary->symbolHash))
 			{
 				ILHashDestroy(nextLibrary->classHash);
+				if(nextLibrary->publicKey)
+				{
+					ILFree(nextLibrary->publicKey);
+				}
 				ILFree(nextLibrary->filename);
 				ILFree(nextLibrary->name);
 				ILFree(nextLibrary);
@@ -471,7 +503,29 @@ static int AddGlobalSymbol(ILLinker *linker, ILLibrary *library,
 		libSymbol = ILHashFindType(linker->symbolHash, name, ILLibrarySymbol);
 		if(libSymbol)
 		{
-			fprintf(stderr, "%s : multiply defined\n", name);
+			/* Don't report an error if it looks like a PInvoke or
+			   field RVA definition.  This can happen with code
+			   generated by certain C compilers */
+			if(ILMember_IsMethod(member))
+			{
+				if(ILMethod_HasPInvokeImpl((ILMethod *)member))
+				{
+					return 1;
+				}
+			}
+			if(ILMember_IsField(member))
+			{
+				if((ILField_Attrs((ILField *)member) &
+						IL_META_FIELDDEF_HAS_FIELD_RVA) != 0)
+				{
+					return 1;
+				}
+			}
+			if((libSymbol->flags & IL_LINKSYM_SAW_UNDEF) == 0)
+			{
+				fprintf(stderr, "%s : multiply defined\n", name);
+				libSymbol->flags |= IL_LINKSYM_SAW_UNDEF;
+			}
 			linker->error = 1;
 			return 1;
 		}
@@ -880,6 +934,32 @@ void _ILLinkerPrintClass(ILLibraryFind *find, const char *name,
 	fputs(name, stderr);
 }
 
+/*
+ * Set the public key token for an assembly reference.
+ */
+static int SetAssemblyRefToken(ILAssembly *assem, unsigned char *key,
+							   ILUInt32 keyLen)
+{
+	ILSHAContext sha;
+	unsigned char hash[IL_SHA_HASH_SIZE];
+	unsigned char token[8];
+	int posn;
+
+	/* Compute the SHA1 hash of the key value */
+	ILSHAInit(&sha);
+	ILSHAData(&sha, key, (unsigned long)keyLen);
+	ILSHAFinalize(&sha, hash);
+
+	/* The token is the last 8 bytes of the hash, reversed */
+	for(posn = 0; posn < 8; ++posn)
+	{
+		token[posn] = hash[IL_SHA_HASH_SIZE - posn - 1];
+	}
+
+	/* Set the token on the assembly reference record */
+	return ILAssemblySetOriginator(assem, token, 8);
+}
+
 /*
  * Make a type reference within an image.
  */
@@ -937,6 +1017,15 @@ static ILClass *MakeTypeRef(ILLibraryFind *find, ILLibraryClass *libClass,
 				return 0;
 			}
 			ILAssemblySetVersion(assem, find->library->version);
+			if(find->library->publicKey)
+			{
+				if(!SetAssemblyRefToken(assem, find->library->publicKey,
+								        find->library->publicKeyLen))
+				{
+					_ILLinkerOutOfMemory(find->linker);
+					return 0;
+				}
+			}
 		}
 
 		/* See if we already have a reference */
