@@ -1,7 +1,7 @@
 /*
  * link_main.c - Link IL images together to form an assembly.
  *
- * Copyright (C) 2001, 2002  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2001, 2002, 2003  Southern Storm Software, Pty Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -88,6 +88,9 @@ static ILCmdLineOption const options[] = {
 	{"-M", 'M', 1,
 		"-fmodule-name=name          or -M name",
 		"Specify the name of the module to embed in the output."},
+	{"-K", 'K', 1,
+		"-fassembly-key=key          or -K key",
+		"Specify the assembly's public key value."},
 	{"-E", 'E', 1, 0, 0},
 	{"--entry-point", 'E', 1,
 		"--entry-point name          or -E name",
@@ -102,6 +105,15 @@ static ILCmdLineOption const options[] = {
 	{"-p", 'p', 0,
 		"-fprivate-resources         or -p",
 		"Mark the resources as private to the output assembly."},
+	{"-z", 'z', 0,
+		"-fminimize-parameters       or -z",
+		"Minimize the size of the parameter table in the output."},
+	{"-W", 'W', 1,
+		"-fwin32res=file             or -W file",
+		"Specify a Win32 resource file to add to the image."},
+	{"-i", 'i', 1,
+		"-fwin32icon=file            or -i file",
+		"Specify a Win32 icon file to add to the image."},
 	{"-H", 'H', 1,
 		"-fhash-algorithm=name       or -H name",
 		"Specify the algorithm to use to hash files (SHA1 or MD5)"},
@@ -155,8 +167,8 @@ static ILCmdLineOption const options[] = {
 	{"/trade*", '?', 1, 0, 0},		/* "/trademark:text" */
 	{"/t*", 'F', 1, 0, 0},			/* "/target:lib|exe|win" */
 	{"/v*", 'A', 1, 0, 0},			/* "/version:version" */
-	{"/win32icon", '?', 1, 0, 0},	/* "/win32icon:filename" */
-	{"/win32res", '?', 1, 0, 0},	/* "/win32res:filename" */
+	{"/win32icon", 'i', 1, 0, 0},	/* "/win32icon:filename" */
+	{"/win32res", 'W', 1, 0, 0},	/* "/win32res:filename" */
 	{"/?", 'h', 0, 0, 0},			/* "/?" */
 
 	{0, 0, 0, 0, 0}
@@ -190,6 +202,7 @@ int ILLinkerMain(int argc, char *argv[])
 	char *assemblyName = NULL;
 	ILUInt16 assemblyVersion[4] = {0, 0, 0, 0};
 	char *moduleName = NULL;
+	const char *publicKey = NULL;
 	char *entryPoint = NULL;
 	int hashAlgorithm = IL_META_HASHALG_SHA1;
 	int len;
@@ -209,6 +222,9 @@ int ILLinkerMain(int argc, char *argv[])
 	int isStatic = 0;
 	int privateResources = 0;
 	int defaultIsLib = 0;
+	int linkerFlags = 0;
+	const char *win32Res = 0;
+	const char *win32Icon = 0;
 	int firstFile;
 	int temp, temp2;
 	ILLinker *linker;
@@ -346,12 +362,30 @@ int ILLinkerMain(int argc, char *argv[])
 			}
 			break;
 
+			case 'K':
+			{
+				publicKey = param;
+			}
+			break;
+
 			case 'E':
 			{
 				entryPoint = param;
 			}
 			break;
 
+			case 'W':
+			{
+				win32Res = param;
+			}
+			break;
+
+			case 'i':
+			{
+				win32Icon = param;
+			}
+			break;
+
 			case 'r':
 			{
 				resources[numResources] = param;
@@ -415,6 +449,12 @@ int ILLinkerMain(int argc, char *argv[])
 			}
 			break;
 
+			case 'z':
+			{
+				linkerFlags |= IL_LINKFLAG_MINIMIZE_PARAMS;
+			}
+			break;
+
 			case 'S':
 			{
 				stdLibrary = param;
@@ -486,6 +526,10 @@ int ILLinkerMain(int argc, char *argv[])
 						return 1;
 					}
 				}
+				else if(!strncmp(param, "assembly-key=", 13))
+				{
+					publicKey = param + 13;
+				}
 				else if(!strncmp(param, "module-name=", 12))
 				{
 					moduleName = param + 12;
@@ -502,10 +546,22 @@ int ILLinkerMain(int argc, char *argv[])
 					param += 15;
 					goto parseHashAlg;
 				}
+				else if(!strncmp(param, "win32res=", 9))
+				{
+					win32Res = param + 9;
+				}
+				else if(!strncmp(param, "win32icon=", 10))
+				{
+					win32Icon = param + 10;
+				}
 				else if(!strcmp(param, "private-resources"))
 				{
 					privateResources = 1;
 				}
+				else if(!strcmp(param, "minimize-parameters"))
+				{
+					linkerFlags |= IL_LINKFLAG_MINIMIZE_PARAMS;
+				}
 				else
 				{
 					/* All other flags are ignored, because they may
@@ -642,15 +698,19 @@ int ILLinkerMain(int argc, char *argv[])
 	{
 		if(format == IL_IMAGETYPE_EXE)
 		{
-			outputFile = "a.out.exe";
+		#ifdef IL_WIN32_PLATFORM
+			outputFile = "a.exe";
+		#else
+			outputFile = "a.out";
+		#endif
 		}
 		else if(format == IL_IMAGETYPE_DLL)
 		{
-			outputFile = "a.out.dll";
+			outputFile = "a.dll";
 		}
 		else
 		{
-			outputFile = "a.out.obj";
+			outputFile = "a.obj";
 		}
 	}
 
@@ -680,10 +740,10 @@ int ILLinkerMain(int argc, char *argv[])
 	}
 
 	/* Open the output file */
-	if((outfile = fopen(outputFile, "wb")) == NULL)
+	if((outfile = fopen(outputFile, "w+b")) == NULL)
 	{
 		/* Try again in case libc does not understand "wb" */
-		if((outfile = fopen(outputFile, "w")) == NULL)
+		if((outfile = fopen(outputFile, "w+")) == NULL)
 		{
 			perror(outputFile);
 			return 1;
@@ -698,11 +758,12 @@ int ILLinkerMain(int argc, char *argv[])
 		ILDeleteFile(outputFile);
 		outOfMemory();
 	}
+	ILLinkerSetFlags(linker, linkerFlags);
 
 	/* Create the global module and assembly */
 	if(!ILLinkerCreateModuleAndAssembly(linker, moduleName,
 										assemblyName, assemblyVersion,
-										hashAlgorithm))
+										publicKey, hashAlgorithm))
 	{
 		ILLinkerDestroy(linker);
 		fclose(outfile);
@@ -827,6 +888,20 @@ int ILLinkerMain(int argc, char *argv[])
 		}
 	}
 
+	/* Add Win32 resource, icon, or version information */
+	if(win32Res)
+	{
+		ILLinkerAddWin32Resource(linker, win32Res);
+	}
+	else if(win32Icon)
+	{
+		ILLinkerAddWin32Icon(linker, win32Icon);
+	}
+	else
+	{
+		ILLinkerAddWin32Version(linker);
+	}
+
 	/* Set the default entry point, or report an error about
 	   incorrect entry point specifications */
 	if(!resourcesOnly)
@@ -848,15 +923,6 @@ int ILLinkerMain(int argc, char *argv[])
 				errors = 1;
 			}
 		}
-		else if(format == IL_IMAGETYPE_DLL)
-		{
-			if(ILLinkerHasEntryPoint(linker))
-			{
-				fprintf(stderr, "%s: DLL's cannot have entry points\n",
-						outputFile);
-				errors = 1;
-			}
-		}
 	}
 
 	/* Destroy the linker context */
@@ -888,7 +954,7 @@ int ILLinkerMain(int argc, char *argv[])
 static void usage(const char *progname)
 {
 	fprintf(stdout, "ILALINK " VERSION " - IL Assembly Link Utility\n");
-	fprintf(stdout, "Copyright (c) 2001 Southern Storm Software, Pty Ltd.\n");
+	fprintf(stdout, "Copyright (c) 2001, 2002, 2003 Southern Storm Software, Pty Ltd.\n");
 	fprintf(stdout, "\n");
 	fprintf(stdout, "Usage: %s [options] input ...\n", progname);
 	fprintf(stdout, "\n");
@@ -899,7 +965,7 @@ static void version(void)
 {
 
 	printf("ILALINK " VERSION " - IL Assembly Link Utility\n");
-	printf("Copyright (c) 2001 Southern Storm Software, Pty Ltd.\n");
+	printf("Copyright (c) 2001, 2002, 2003 Southern Storm Software, Pty Ltd.\n");
 	printf("\n");
 	printf("ILALINK comes with ABSOLUTELY NO WARRANTY.  This is free software,\n");
 	printf("and you are welcome to redistribute it under the terms of the\n");
@@ -1127,7 +1193,7 @@ static int processFile(ILLinker *linker, const char *filename,
 				{
 					--temp;
 				}
-				if(temp > 0 && arhdr.ar_name[temp - 1] == '/')
+				while(temp > 0 && arhdr.ar_name[temp - 1] == '/')
 				{
 					--temp;
 				}
