@@ -65,6 +65,7 @@ static int ConvertToken(ILLinker *linker, ILMethod *method,
 	ILType *type;
 	ILStandAloneSig *sig;
 	ILTypeSpec *spec;
+	ILMethodSpec *mspec;
 
 	if(!(*token))
 	{
@@ -106,7 +107,7 @@ static int ConvertToken(ILLinker *linker, ILMethod *method,
 			*token = ILMember_Token(member);
 			return 1;
 		}
-		break;
+		/* Not reached */
 
 		case IL_META_TOKEN_STAND_ALONE_SIG:
 		{
@@ -126,7 +127,7 @@ static int ConvertToken(ILLinker *linker, ILMethod *method,
 			*token = ILStandAloneSig_Token(sig);
 			return 1;
 		}
-		break;
+		/* Not reached */
 
 		case IL_META_TOKEN_TYPE_SPEC:
 		{
@@ -140,7 +141,33 @@ static int ConvertToken(ILLinker *linker, ILMethod *method,
 			*token = ILTypeSpec_Token(spec);
 			return 1;
 		}
-		break;
+		/* Not reached */
+
+		case IL_META_TOKEN_METHOD_SPEC:
+		{
+			/* Convert a generic method specification */
+			member = ILMethodSpec_Method((ILMethodSpec *)item);
+			member = _ILLinkerConvertMemberRef(linker, member);
+			if(!member)
+			{
+				return -1;
+			}
+			type = ILMethodSpec_Type((ILMethodSpec *)item);
+			type = _ILLinkerConvertType(linker, type);
+			if(!type)
+			{
+				return -1;
+			}
+			mspec = ILMethodSpecCreate(linker->image, 0, member, type);
+			if(!mspec)
+			{
+				_ILLinkerOutOfMemory(linker);
+				return -1;
+			}
+			*token = ILMethodSpec_Token(mspec);
+			return 1;
+		}
+		/* Not reached */
 
 		default:
 		{
@@ -573,6 +600,11 @@ int _ILLinkerConvertMethod(ILLinker *linker, ILMethod *method,
 	ILException *exceptions;
 	ILUInt32 numParams;
 	char *newName = 0;
+	ILUInt32 genericNum;
+	ILGenericPar *genPar;
+	ILGenericPar *newGenPar;
+	ILProgramItem *constraint;
+	ILTypeSpec *spec;
 
 	/* Rename the method if it is within the "<Module>" class and private */
 	if(ILMethod_IsPrivate(method) && ILMethod_IsStatic(method) &&
@@ -596,38 +628,33 @@ int _ILLinkerConvertMethod(ILLinker *linker, ILMethod *method,
 		}
 		return 0;
 	}
-	while((newMethod = (ILMethod *)ILClassNextMemberByKind
-				(newClass, (ILMember *)newMethod, IL_META_MEMBERKIND_METHOD))
-						!= 0)
+	if((newMethod = (ILMethod *)ILClassNextMemberMatch
+			(newClass, (ILMember *)0,
+			 IL_META_MEMBERKIND_METHOD, name, signature)) != 0)
 	{
-		if(!strcmp(ILMethod_Name(newMethod), name) &&
-		   ILTypeIdentical(ILMethod_Signature(newMethod), signature))
+		/* Bail out if the method is already defined.  This shouldn't
+		   happen very often because duplicate classes are trapped long
+		   before control gets to here.  Global methods may result in
+		   this code being used, however */
+		if((ILMethod_Token(newMethod) & IL_META_TOKEN_MASK)
+				!= IL_META_TOKEN_MEMBER_REF)
 		{
-			/* Bail out if the method is already defined.  This shouldn't
-			   happen very often because duplicate classes are trapped long
-			   before control gets to here.  Global methods may result in
-			   this code being used, however */
-			if((ILMethod_Token(newMethod) & IL_META_TOKEN_MASK)
-					!= IL_META_TOKEN_MEMBER_REF)
+			if(newName)
 			{
-				if(newName)
-				{
-					ILFree(newName);
-				}
-				return 1;
+				ILFree(newName);
 			}
+			return 1;
+		}
 
-			/* Allocate a new token for the method */
-			if(!ILMethodNewToken(newMethod))
+		/* Allocate a new token for the method */
+		if(!ILMethodNewToken(newMethod))
+		{
+			_ILLinkerOutOfMemory(linker);
+			if(newName)
 			{
-				_ILLinkerOutOfMemory(linker);
-				if(newName)
-				{
-					ILFree(newName);
-				}
-				return 0;
+				ILFree(newName);
 			}
-			break;
+			return 0;
 		}
 	}
 	if(!newMethod)
@@ -674,6 +701,14 @@ int _ILLinkerConvertMethod(ILLinker *linker, ILMethod *method,
 	param = 0;
 	while((param = ILMethodNextParam(method, param)) != 0)
 	{
+		if((linker->linkerFlags & IL_LINKFLAG_MINIMIZE_PARAMS) != 0 &&
+		   !ILParameter_Attrs(param) &&
+		   !ILProgramItem_HasAttrs(param) &&
+		   !ILFieldMarshalGetFromOwner(ILToProgramItem(param)))
+		{
+			/* This parameter block doesn't add anything useful */
+			continue;
+		}
 		newParam = ILParameterCreate(newMethod, 0, ILParameter_Name(param),
 									 ILParameter_Attrs(param),
 									 ILParameter_Num(param));
@@ -785,6 +820,46 @@ int _ILLinkerConvertMethod(ILLinker *linker, ILMethod *method,
 		}
 	}
 
+	/* Convert the generic parameters, if any */
+	genericNum = 0;
+	while((genPar = ILGenericParGetFromOwner
+				(ILToProgramItem(method), genericNum)) != 0)
+	{
+		newGenPar = ILGenericParCreate(linker->image, 0,
+									   ILToProgramItem(newMethod), genericNum);
+		if(!newGenPar)
+		{
+			_ILLinkerOutOfMemory(linker);
+			return 0;
+		}
+		if(!ILGenericParSetName(newGenPar, ILGenericPar_Name(genPar)))
+		{
+			_ILLinkerOutOfMemory(linker);
+			return 0;
+		}
+		constraint = ILGenericPar_Constraint(genPar);
+		if(constraint)
+		{
+			spec = ILProgramItemToTypeSpec(constraint);
+			if(spec)
+			{
+				constraint = ILToProgramItem
+					(_ILLinkerConvertTypeSpec(linker, ILTypeSpec_Type(spec)));
+			}
+			else
+			{
+				constraint = ILToProgramItem
+					(_ILLinkerConvertClassRef(linker, (ILClass *)constraint));
+			}
+			if(!constraint)
+			{
+				return 0;
+			}
+			ILGenericParSetConstraint(newGenPar, constraint);
+		}
+		++genericNum;
+	}
+
 	/* Get the method's code.  If there is no code, then we are done */
 	if(!ILMethodGetCode(method, &code))
 	{
@@ -930,19 +1005,15 @@ ILMember *_ILLinkerConvertMemberRef(ILLinker *linker, ILMember *member)
 	if(ILMember_IsMethod(member))
 	{
 		/* Searching for a method */
-		method = 0;
-		while((method = (ILMethod *)ILClassNextMemberByKind
-				(owner, (ILMember *)method, IL_META_MEMBERKIND_METHOD)) != 0)
+		if((method = (ILMethod *)ILClassNextMemberMatch
+				(owner, (ILMember *)0,
+				 IL_META_MEMBERKIND_METHOD, name, signature)) != 0)
 		{
-			if(!strcmp(ILMethod_Name(method), name) &&
-			   ILTypeIdentical(ILMethod_Signature(method), signature))
+			if(newName)
 			{
-				if(newName)
-				{
-					ILFree(newName);
-				}
-				return (ILMember *)method;
+				ILFree(newName);
 			}
+			return (ILMember *)method;
 		}
 
 		/* Create a MemberRef for the method */
@@ -967,19 +1038,15 @@ ILMember *_ILLinkerConvertMemberRef(ILLinker *linker, ILMember *member)
 	else
 	{
 		/* Searching for a field */
-		field = 0;
-		while((field = (ILField *)ILClassNextMemberByKind
-				(owner, (ILMember *)field, IL_META_MEMBERKIND_FIELD)) != 0)
+		if((field = (ILField *)ILClassNextMemberMatch
+				(owner, (ILMember *)0,
+				 IL_META_MEMBERKIND_FIELD, name, signature)) != 0)
 		{
-			if(!strcmp(ILField_Name(field), name) &&
-			   ILTypeIdentical(ILFieldGetTypeWithPrefixes(field), signature))
+			if(newName)
 			{
-				if(newName)
-				{
-					ILFree(newName);
-				}
-				return (ILMember *)field;
+				ILFree(newName);
 			}
+			return (ILMember *)field;
 		}
 
 		/* Create a MemberRef for the field */
