@@ -54,17 +54,29 @@ static ILUInt32		  length = 0;
  * Other information for the current method.
  */
 static ILType        *localVars = 0;
-static ILUInt32		  maxStack = 0;
+static ILUInt32		  maxStack = 8;
 static int			  initLocals = 0;
 static ILUInt32		  localIndex = 0;
+typedef struct _tagLocalBlock
+{
+	char *blockStart;
+	char *blockEnd;
+	struct _tagLocalBlock *outer;
+	struct _tagLocalBlock *next;
+
+} LocalBlock;
 typedef struct _tagLocalInfo
 {
+	LocalBlock *block;
 	char *name;
 	ILUInt32 index;
+	int isParam;
 	struct _tagLocalInfo *next;
 
 } LocalInfo;
 static LocalInfo	 *localNames = 0;
+static LocalBlock	 *localBlocks = 0;
+static LocalBlock	 *localCurrentBlock = 0;
 static ILAsmOutException *exceptionList = 0;
 static ILAsmOutException *lastException = 0;
 static unsigned long  numExceptions = 0;
@@ -129,12 +141,14 @@ typedef struct _tagLabelInfo
 	struct _tagLabelInfo *next;
 	char		 *debugFilename;
 	ILUInt32	  debugLine;
+	ILUInt32	  debugColumn;
 
 } LabelInfo;
 static ILMemPool labelPool;
 static ILMemPool labelRefPool;
 static int       initLabelPool = 0;
 static LabelInfo *labels = 0;
+static int		 haveColumnInfo = 0;
 
 /*
  * Emit a warning when an attempt was made to encode an
@@ -251,6 +265,7 @@ static void OutShortVar(ILInt32 lopcode, ILInt32 sopcode,
 void ILAsmOutReset(void)
 {
 	LocalInfo *local, *nextLocal;
+	LocalBlock *block, *nextBlock;
 	ILAsmOutException *exception, *nextException;
 
 	/* Destroy allocated data structures */
@@ -269,6 +284,13 @@ void ILAsmOutReset(void)
 		ILFree(local);
 		local = nextLocal;
 	}
+	block = localBlocks;
+	while(block != 0)
+	{
+		nextBlock = block->next;
+		ILFree(block);
+		block = nextBlock;
+	}
 	exception = exceptionList;
 	while(exception != 0)
 	{
@@ -287,10 +309,11 @@ void ILAsmOutReset(void)
 	buffer = 0;
 	length = 0;
 	localVars = 0;
-	maxStack = 0;
+	maxStack = 8;
 	initLocals = 0;
 	localIndex = 0;
 	localNames = 0;
+	localBlocks = 0;
 	exceptionList = 0;
 	lastException = 0;
 	numExceptions = 0;
@@ -304,6 +327,7 @@ void ILAsmOutReset(void)
 	defaultLookupSwitchLabel = 0;
 	initLabelPool = 0;
 	labels = 0;
+	haveColumnInfo = 0;
 }
 
 void ILAsmOutCreate(FILE *stream, int seekable, int type, int flags)
@@ -594,6 +618,7 @@ static LabelInfo *GetLabel(char *name)
 	label->next = labels;
 	label->debugFilename = 0;
 	label->debugLine = 0;
+	label->debugColumn = 0;
 	labels = label;
 	return label;
 }
@@ -1283,7 +1308,7 @@ char *ILAsmOutUniqueLabel(void)
 	return name;
 }
 
-void ILAsmOutDebugLine(char *filename, ILUInt32 line)
+void ILAsmOutDebugLine(char *filename, ILUInt32 line, ILUInt32 column)
 {
 	/* Output a unique label at this position */
 	char *label = ILAsmOutUniqueLabel();
@@ -1292,7 +1317,9 @@ void ILAsmOutDebugLine(char *filename, ILUInt32 line)
 	/* Attach the debug information to the label */
 	info->debugFilename = filename;
 	info->debugLine = line;
+	info->debugColumn = column;
 	haveDebug = 1;
+	haveColumnInfo |= (column != 0);
 }
 
 void ILAsmOutSSAStart(ILInt32 opcode)
@@ -1334,16 +1361,18 @@ void ILAsmOutZeroInit(void)
 /*
  * Add a local variable name and index to the current method.
  */
-static void AddLocalName(char *name, ILUInt32 index)
+static void AddLocalName(char *name, ILUInt32 index, int isParam)
 {
 	LocalInfo *local = (LocalInfo *)ILMalloc(sizeof(LocalInfo));
 	if(!local)
 	{
 		ILAsmOutOfMemory();
 	}
+	local->block = localCurrentBlock;
 	local->name = name;
 	local->index = index;
 	local->next = localNames;
+	local->isParam = isParam;
 	localNames = local;
 }
 
@@ -1366,7 +1395,7 @@ void ILAsmOutAddLocals(ILAsmParamInfo *vars)
 		}
 		if(vars->name)
 		{
-			AddLocalName(vars->name, localIndex);
+			AddLocalName(vars->name, localIndex, 0);
 		}
 		++localIndex;
 		nextVar = vars->next;
@@ -1391,7 +1420,7 @@ void ILAsmOutAddParams(ILAsmParamInfo *vars, ILUInt32 callConv)
 	{
 		if(vars->name)
 		{
-			AddLocalName(vars->name, paramIndex);
+			AddLocalName(vars->name, paramIndex, 1);
 		}
 		++paramIndex;
 		vars = vars->next;
@@ -1668,6 +1697,16 @@ static void OutputDebugInfo(ILMethod *method)
 	char *prevFilename = 0;
 	LabelInfo *label = labels;
 	unsigned long len = 0;
+	int type;
+	LocalInfo *local;
+	LocalBlock *lastBlock;
+
+	/* Output the line number information for the method */
+	type = IL_DEBUGTYPE_LINE_OFFSETS;
+	if(haveColumnInfo)
+	{
+		type = IL_DEBUGTYPE_LINE_COL_OFFSETS;
+	}
 	while(label != 0)
 	{
 		if(label->debugFilename)
@@ -1678,7 +1717,7 @@ static void OutputDebugInfo(ILMethod *method)
 				if(len > 0)
 				{
 					ILWriterDebugAdd(ILAsmWriter, (ILProgramItem *)method,
-									 IL_DEBUGTYPE_LINE_OFFSETS, buf, len);
+									 type, buf, len);
 					len = 0;
 				}
 				len += ILMetaCompressData
@@ -1689,13 +1728,18 @@ static void OutputDebugInfo(ILMethod *method)
 
 			/* Add the line and offset information */
 			len += ILMetaCompressData(buf + len, label->debugLine);
+			if(haveColumnInfo)
+			{
+				len += ILMetaCompressData(buf + len, label->debugColumn);
+			}
 			len += ILMetaCompressData(buf + len, label->address);
 
 			/* Flush the buffer if it is nearly full */
-			if(len >= (sizeof(buf) - IL_META_COMPRESS_MAX_SIZE * 3))
+			if(len >= (sizeof(buf) - IL_META_COMPRESS_MAX_SIZE * 4))
 			{
 				ILWriterDebugAdd(ILAsmWriter, (ILProgramItem *)method,
-								 IL_DEBUGTYPE_LINE_OFFSETS, buf, len);
+								 type, buf, len);
+				prevFilename = 0;
 				len = 0;
 			}
 		}
@@ -1705,7 +1749,58 @@ static void OutputDebugInfo(ILMethod *method)
 	{
 		/* Flush the remainder of the debug information */
 		ILWriterDebugAdd(ILAsmWriter, (ILProgramItem *)method,
-						 IL_DEBUGTYPE_LINE_OFFSETS, buf, len);
+						 type, buf, len);
+		len = 0;
+	}
+
+	/* Output the local variable name information for the method */
+	local = localNames;
+	lastBlock = (LocalBlock *)(ILNativeInt)(-1);
+	type = IL_DEBUGTYPE_VARS;
+	while(local != 0)
+	{
+		if(local->isParam)
+		{
+			local = local->next;
+			continue;
+		}
+		if(!len || len >= (sizeof(buf) - IL_META_COMPRESS_MAX_SIZE * 4) ||
+		   local->block != lastBlock)
+		{
+			/* We've encountered a change in local variable scope,
+			   or the buffer is almost full */
+			if(len > 0)
+			{
+				ILWriterDebugAdd(ILAsmWriter, (ILProgramItem *)method,
+								 type, buf, len);
+				len = 0;
+			}
+			lastBlock = local->block;
+			if(!lastBlock)
+			{
+				/* Outermost local variable scope for the method */
+				type = IL_DEBUGTYPE_VARS;
+			}
+			else
+			{
+				/* Inner local variable scope for the method */
+				type = IL_DEBUGTYPE_VARS_OFFSETS;
+				label = GetLabel(lastBlock->blockStart);
+				len += ILMetaCompressData(buf + len, label->address);
+				label = GetLabel(lastBlock->blockEnd);
+				len += ILMetaCompressData(buf + len, label->address);
+			}
+		}
+		len += ILMetaCompressData
+			(buf + len, ILWriterDebugString(ILAsmWriter, local->name));
+		len += ILMetaCompressData(buf + len, local->index);
+		local = local->next;
+	}
+	if(len > 0)
+	{
+		/* Flush the remainder of the debug information */
+		ILWriterDebugAdd(ILAsmWriter, (ILProgramItem *)method,
+						 type, buf, len);
 	}
 }
 
@@ -1716,6 +1811,8 @@ void ILAsmOutFinalizeMethod(ILMethod *method)
 	ILToken token;
 	LocalInfo *local;
 	LocalInfo *nextLocal;
+	LocalBlock *block;
+	LocalBlock *nextBlock;
 	ILAsmOutException *exception;
 	ILAsmOutException *nextException;
 	int fatExceptions;
@@ -1723,6 +1820,7 @@ void ILAsmOutFinalizeMethod(ILMethod *method)
 	unsigned long startRVA;
 	LabelInfo *label;
 
+#ifdef IL_CONFIG_JAVA
 	/* check if parsing java method */
 	if(ILAsmParseJava) 
 	{
@@ -1811,6 +1909,7 @@ void ILAsmOutFinalizeMethod(ILMethod *method)
 		ILJavaAppendCode(ILAsmWriter, ILMethod_Owner(method), method, header, 2);
 	} 
 	else
+#endif
 	{
 
 	/* Finish the label processing for the method */
@@ -2035,7 +2134,7 @@ void ILAsmOutFinalizeMethod(ILMethod *method)
 cleanup:
 	offset = 0;
 	localVars = 0;
-	maxStack = 0;
+	maxStack = 8;
 	initLocals = 0;
 	localIndex = 0;
 	local = localNames;
@@ -2046,6 +2145,14 @@ cleanup:
 		local = nextLocal;
 	}
 	localNames = 0;
+	block = localBlocks;
+	while(block != 0)
+	{
+		nextBlock = block->next;
+		ILFree(block);
+		block = nextBlock;
+	}
+	localBlocks = 0;
 	exception = exceptionList;
 	while(exception != 0)
 	{
@@ -2063,6 +2170,7 @@ cleanup:
 	}
 	labels = 0;
 	haveDebug = 0;
+	haveColumnInfo = 0;
 }
 
 void ILAsmOutAddResource(const char *name, FILE *stream)
@@ -2116,9 +2224,40 @@ void ILAsmOutAddResource(const char *name, FILE *stream)
 	}
 }
 
+void ILAsmOutDeclareVarName(char *name, ILUInt32 index)
+{
+	AddLocalName(name, index, 0);
+}
+
+void ILAsmOutPushVarScope(char *name)
+{
+	LocalBlock *block = (LocalBlock *)ILMalloc(sizeof(LocalBlock));
+	if(!block)
+	{
+		ILAsmOutOfMemory();
+	}
+	block->blockStart = name;
+	block->blockEnd = 0;
+	block->outer = localCurrentBlock;
+	block->next = localBlocks;
+	localBlocks = block;
+	localCurrentBlock = block;
+}
+
+void ILAsmOutPopVarScope(char *name)
+{
+	if(localCurrentBlock)
+	{
+		localCurrentBlock->blockEnd = name;
+		localCurrentBlock = localCurrentBlock->outer;
+	}
+}
+
 void ILJavaAsmInitPool()
 {
+#ifdef IL_CONFIG_JAVA
 	ILJavaInitPool(ILAsmWriter, ILAsmClass);
+#endif
 }
 
 void ILJavaAsmOutSimple(ILInt32 opcode)
@@ -2158,6 +2297,8 @@ void ILJavaAsmOutInt(ILInt32 opcode, ILInt64 value)
 	}
 }
 
+#ifdef IL_CONFIG_JAVA
+
 static void OutConst32(int index)
 {
 	if(index < 256)
@@ -2180,45 +2321,58 @@ static void OutConst64(int index)
 	OUT_BYTE(index);
 }
 
+#endif
+
 void ILJavaAsmOutConstInt32(ILInt32 opcode, ILInt64 value)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetInteger(ILAsmWriter, ILAsmClass, (ILInt32)value);
 
 	OutConst32(index);
+#endif
 }
 
 void ILJavaAsmOutConstInt64(ILInt32 opcode, ILInt64 value)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetLong(ILAsmWriter, ILAsmClass, value);
 
 	OutConst64(index);
+#endif
 }
 
 void ILJavaAsmOutConstFloat32(ILInt32 opcode, ILUInt8 *value)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetFloat(ILAsmWriter, ILAsmClass, IL_READ_FLOAT(value));
 	
 	OutConst32(index);
+#endif
 }
 
 void ILJavaAsmOutConstFloat64(ILInt32 opcode, ILUInt8 *value)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetDouble(ILAsmWriter, ILAsmClass, IL_READ_DOUBLE(value));
 
 	OutConst64(index);
+#endif
 }
 
 void ILJavaAsmOutString(ILIntString interned) 
 {
+#ifdef IL_CONFIG_JAVA
 	int strIndex = ILJavaSetUTF8String(ILAsmWriter, ILAsmClass, interned.string, 
 									   interned.len);
 	int index = ILJavaSetString(ILAsmWriter, ILAsmClass, strIndex);
 
 	OutConst32(index);
+#endif
 }
 
 void ILJavaAsmOutToken(ILInt32 opcode, ILUInt32 token)
 {
+#ifdef IL_CONFIG_JAVA
 	ILMember *member;
 	ILClass *class;
 	ILUInt32 index = 0;
@@ -2262,11 +2416,13 @@ void ILJavaAsmOutToken(ILInt32 opcode, ILUInt32 token)
 	OUT_BYTE(opcode);
 	OUT_BYTE(index >> 8);
 	OUT_BYTE(index);
+#endif
 }
 
 void ILJavaAsmOutRef(ILInt32 opcode, int isMethod, char *className, 
 					 char *refName, char *sigName)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = 
 		ILJavaSetrefFromName(ILAsmWriter, ILAsmClass, 
 							 isMethod ? JAVA_CONST_METHODREF : JAVA_CONST_FIELDREF,
@@ -2275,15 +2431,18 @@ void ILJavaAsmOutRef(ILInt32 opcode, int isMethod, char *className,
 	OUT_BYTE(opcode);
 	OUT_BYTE(index >> 8);
 	OUT_BYTE(index);
+#endif
 }
 
 void ILJavaAsmOutType(ILInt32 opcode, char *className)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetClassFromName(ILAsmWriter, ILAsmClass, className);
 
 	OUT_BYTE(opcode);
 	OUT_BYTE(index >> 8);
 	OUT_BYTE(index);
+#endif
 }
 
 void ILJavaAsmOutNewarray(ILInt32 opcode, ILInt64 type)
@@ -2294,22 +2453,26 @@ void ILJavaAsmOutNewarray(ILInt32 opcode, ILInt64 type)
 
 void ILJavaAsmOutMultinewarray(ILInt32 opcode, ILType *type, ILInt64 dim)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetClassFromType(ILAsmWriter, ILAsmClass, type);
 
 	OUT_BYTE(opcode);
 	OUT_BYTE(index >> 8);
 	OUT_BYTE(index);
 	OUT_BYTE(dim);
+#endif
 }
 
 void ILJavaAsmOutMultinewarrayFromName(ILInt32 opcode, char *typeName, ILInt64 dim)
 {
+#ifdef IL_CONFIG_JAVA
 	int index = ILJavaSetClassFromName(ILAsmWriter, ILAsmClass, typeName);
 
 	OUT_BYTE(opcode);
 	OUT_BYTE(index >> 8);
 	OUT_BYTE(index);
 	OUT_BYTE(dim);
+#endif
 }
 
 /*
