@@ -25,7 +25,12 @@
 #include "il_system.h"
 #include "il_utils.h"
 #include "il_program.h"
+#if !defined(__palmos__)
+#define	IL_USE_WRITER
+#endif
+#ifdef	IL_USE_WRITER
 #include "il_writer.h"
+#endif
 
 #ifdef	__cplusplus
 extern	"C" {
@@ -57,7 +62,8 @@ extern	"C" {
  * Structure of a context, which holds multiple loaded images,
  * and the information about the classes, methods, etc, in them.
  */
-#define	IL_CONTEXT_HASH_SIZE	512
+#define	IL_CONTEXT_HASH_SIZE	509
+#define	IL_CONTEXT_NS_HASH_SIZE	67
 struct _tagILContext
 {
 	/* Hash table that maps class names to class information blocks */
@@ -80,18 +86,25 @@ struct _tagILContext
 	char		  **libraryDirs;
 	int				numLibraryDirs;
 
+	/* Hash table to keeps track of valid namespaces */
+	ILHashTable	   *namespaceHash;
+
+	/* User data for the runtime engine */
+	void *userData;
 };
 
 /*
  * Key information to use when looking up an entry in the class hash.
  */
+typedef struct _tagILClassName ILClassName;
 typedef struct
 {
 	const char	   *name;
 	int				nameLen;
 	const char	   *namespace;
 	int				namespaceLen;
-	ILProgramItem  *scope;
+	ILProgramItem  *scopeItem;
+	ILClassName    *scopeName;
 	ILImage		   *image;
 	int				wantGlobal;
 	int				ignoreCase;
@@ -143,12 +156,16 @@ struct _tagILImage
 	/* File that the image was loaded file */
 	char		   *filename;		/* NULL if filename is unknown */
 
+	/* Flags that were used during loading */
+	int				loadFlags;
+
 	/* Housekeeping for managing the raw PE/COFF for of an image */
 	int				type : 8;		/* Type of image */
 	int				secure : 1;		/* Non-zero if image source is secure */
 	int				hadNative : 1;	/* Non-zero if image had native code */
 	int				only32Bit : 1;	/* Non-zero if image is 32-bit only */
 	int				mapped : 1;		/* Non-zero if mmap used to load image */
+	int				inPlace : 1;	/* Non-zero if in-place execution */
 	int				strRefBig : 1;	/* Non-zero if STRREF's are 32-bit */
 	int				blobRefBig : 1;	/* Non-zero if BLOBREF's are 32-bit */
 	int				guidRefBig : 1;	/* Non-zero if GUIDREF's are 32-bit */
@@ -166,6 +183,8 @@ struct _tagILImage
 	unsigned long	dataSize;		/* Size of ".sdata" */
 	unsigned long	tlsRVA;			/* RVA for the start of ".tls" */
 	unsigned long	tlsSize;		/* Size of ".tls" */
+	unsigned long	rsrcRVA;		/* RVA for the start of ".rsrc" */
+	unsigned long	rsrcSize;		/* Size of ".rsrc" */
 
 	/* Memory stack that is used to allocate program objects */
 	ILMemStack		memStack;
@@ -216,6 +235,36 @@ struct _tagILStringHash
 #define	IL_STRING_HASH_UNICODE	((ILUInt32)0x40000000)
 #define	IL_STRING_HASH_BLOB		((ILUInt32)0x80000000)
 
+/*
+ * Information about a resource entry.
+ */
+typedef struct _tagILResourceEntry ILResourceEntry;
+struct _tagILResourceEntry
+{
+	int					isDirectory : 1;
+	int					isMallocData : 1;
+	int					isNumeric : 1;
+	char			   *name;
+	int					nameLen;
+	ILResourceEntry	   *children;
+	ILResourceEntry	   *next;
+	unsigned char	   *data;
+	unsigned long		length;
+
+};
+
+/*
+ * Information about the resource section.
+ */
+struct _tagILResourceSection
+{
+	ILImage			   *image;
+	unsigned char	   *data;
+	unsigned long		length;
+	ILResourceEntry	   *rootDirectory;
+
+};
+
 /*
  * Subclass of ILImage which is used when building an image in memory.
  */
@@ -275,6 +324,7 @@ typedef struct _tagILWBufferList
 typedef struct _tagILDebugToken
 {
 	ILProgramItem  *item;
+	unsigned long	pseudo;
 	unsigned long	offset;
 
 } ILDebugToken;
@@ -298,6 +348,8 @@ struct _tagILFixup
 #define	IL_FIXUP_TOKEN		0
 #define	IL_FIXUP_FIELD_RVA	1
 
+#ifdef IL_USE_WRITER
+
 /*
  * Internal structure used when writing IL binaries.
  */
@@ -350,8 +402,16 @@ struct _tagILWriter
 	int				outOfMemory;	/* We ran out of memory during the write */
 	int				writeFailed;	/* The write failed at some point */
 
+	/* Back-patch cache */
+	int				backpatching;	/* Non-zero if last write was back-patch */
+	unsigned long	backpatchSeek;	/* Seek position of back-patch cache */
+	unsigned long	backpatchLen;	/* Length of back-patch cache */
+	unsigned char  *backpatchBuf;	/* Buffer containing back-patch data */
+
 };
 
+#endif /* IL_USE_WRITER */
+
 /*
  * Create a persistent version of a string if necessary.
  * This is typically used on class names and the like which
@@ -372,12 +432,16 @@ const char *_ILContextPersistMalloc(ILImage *image, char *str);
  */
 ILImage *_ILImageCreate(ILContext *context, unsigned size);
 
+#ifdef IL_CONFIG_JAVA
+
 /*
  * Declare the Java image loading function for use by "ILImageLoad".
  */
 int _ILImageJavaLoad(FILE *file, const char *filename, ILContext *context,
 					 ILImage **image, int flags, char *buffer);
 
+#endif /* IL_CONFIG_JAVA */
+
 /*
  * Free a section map.
  */
@@ -408,6 +472,16 @@ int _ILImageParseMeta(ILImage *image, const char *filename, int flags);
 int _ILImageBuildMetaStructures(ILImage *image, const char *filename,
 								int loadFlags);
 
+/*
+ * Load a token information block on demand.
+ */
+void *_ILImageLoadOnDemand(ILImage *image, ILToken token);
+
+/*
+ * Determine if a token has already been loaded.
+ */
+int _ILImageTokenAlreadyLoaded(ILImage *image, ILToken token);
+
 /*
  * Perform dynamic linking on an image to resolve all
  * assemblies that it refers to.  Returns a load error.
@@ -426,6 +500,20 @@ void _ILImageFreeTokens(ILImage *image);
 int _ILImageSetToken(ILImage *image, ILProgramItem *item,
 					 unsigned long token, unsigned long tokenKind);
 
+/*
+ * Search the raw values in an owned item table for a token match.
+ * Returns the first token that matches in the "ownedType" table
+ * in "firstMatch".  The number of tokens that match is returned
+ * from the function, or zero if there are no matches.
+ */
+typedef int (*ILSearchRawFunc)(ILUInt32 *values, ILToken searchFor,
+							   int valueField);
+ILUInt32 _ILSearchForRawToken(ILImage *image, ILSearchRawFunc func,
+							  ILToken ownedType, ILToken *firstMatch,
+							  ILToken searchFor, int valueField);
+
+#ifdef IL_USE_WRITER
+
 /*
  * Compute the size of all token types prior to writing
  * the metadata to an image.
@@ -468,6 +556,8 @@ void _ILWBufferListDestroy(ILWBufferList *list);
  */
 int _ILWBufferListAdd(ILWBufferList *list, const void *buffer, unsigned size);
 
+#endif /* IL_USE_WRITER */
+
 /*
  * Remove all classes for an image from the class lookup hash.
  */
@@ -485,6 +575,15 @@ ILClass *_ILTypeToSyntheticArray(ILImage *image, ILType *type, int singleDim);
  */
 ILClass *_ILTypeToSyntheticOther(ILImage *image, ILType *type);
 
+/*
+ * Look up a synthetic class for an instantiated type.
+ * Returns NULL if not possible, or out of memory.
+ */
+ILClass *_ILTypeToSyntheticInstantiation
+		(ILImage *image, ILType *type, ILType *classParams);
+
+#ifdef IL_USE_WRITER
+
 /*
  * Compact all type and member references in an image to
  * remove tokens that have been replaced with definitions.
@@ -512,6 +611,8 @@ void _ILWriteJavaHeaders(ILWriter *writer);
  */
 void WriteJavaClass(ILWriter *writer, ILClass *class);
 
+#endif /* IL_USE_WRITER */
+
 /*
  * Recommended number of token columns for use with "ILImageRawTokenData".
  */
@@ -717,6 +818,18 @@ void WriteJavaClass(ILWriter *writer, ILClass *class);
 #define	IL_OFFSET_NESTEDCLASS_CHILD			0
 #define	IL_OFFSET_NESTEDCLASS_PARENT		1
 
+#define	IL_OFFSET_GENERICPAR_NUMBER			0
+#define	IL_OFFSET_GENERICPAR_FLAGS			1
+#define	IL_OFFSET_GENERICPAR_OWNER			2
+#define	IL_OFFSET_GENERICPAR_NAME			3
+#define	IL_OFFSET_GENERICPAR_KIND			4
+#define	IL_OFFSET_GENERICPAR_CONSTRAINT		5
+
+#define	IL_OFFSET_METHODSPEC_METHOD			0
+#define	IL_OFFSET_METHODSPEC_INST_RAW		1
+#define	IL_OFFSET_METHODSPEC_INST			2
+#define	IL_OFFSET_METHODSPEC_INST_LEN		3
+
 /*
  * Get the raw data associated with a particular token.
  * Returns zero if the token does not exist.  The "values"
@@ -735,7 +848,11 @@ void _ILImageRawTokenEncode(ILImage *image, unsigned char *ptr,
 /*
  * Define this to 1 to enable debugging.
  */
+#ifndef	REDUCED_STDIO
 #define	IL_DEBUG_META		1
+#else
+#define	IL_DEBUG_META		0
+#endif
 
 /*
  * Debugging support.
