@@ -31,7 +31,7 @@ extern	"C" {
 #define	CSCC_LIB_PATH_DEFAULT	\
 			"/usr/local/lib/cscc/lib:/usr/lib/cscc/lib"
 #define	LD_LIBRARY_PATH_DEFAULT	\
-			"/usr/local/lib:/usr/lib"
+			"/usr/local/lib:/usr/X11R6/lib:/usr/lib"
 
 /*
  * The cached system path, loaded from the environment.
@@ -174,6 +174,7 @@ static void SplitPathString(char *list, char *stdpath, char *defaultPath)
  */
 static void LoadSystemPath(void)
 {
+#if !defined(__palmos__)
 	if(!systemPath)
 	{
 		SplitPathString(getenv("CSCC_LIB_PATH"),
@@ -186,6 +187,7 @@ static void LoadSystemPath(void)
 		SplitPathString(getenv("PATH"), 0, 0);
 	#endif
 	}
+#endif
 }
 
 /*
@@ -216,7 +218,7 @@ static char *TestPathForFile(const char *pathname, int pathlen,
 	}
 
 	/* Build the full pathname */
-	path = (char *)ILMalloc(fullLen);
+	path = (char *)ILMalloc(fullLen + 3);
 	if(!path)
 	{
 		return 0;
@@ -282,6 +284,25 @@ static char *TestPathForFile(const char *pathname, int pathlen,
 		return path;
 	}
 
+	/* If the suffix is ".dll", then try again with ".DLL" */
+	if(suffix && !strcmp(suffix, ".dll"))
+	{
+		strcpy(path + posn - 3, "DLL");
+		if(ILFileExists(path, 0))
+		{
+			return path;
+		}
+
+#ifdef IL_CONFIG_GZIP
+		/* Try ".dll.gz" as well, in case the library was compressed */
+		strcpy(path + posn - 3, "dll.gz");
+		if(ILFileExists(path, 0))
+		{
+			return path;
+		}
+#endif
+	}
+
 	/* Not found */
 	ILFree(path);
 	return 0;
@@ -500,10 +521,13 @@ int _ILImageDynamicLink(ILImage *image, const char *filename, int flags)
 {
 	ILContext *context = ILImageToContext(image);
 	ILAssembly *assem;
+	ILFileDecl *file;
 	char *pathname;
 	int error;
 	ILImage *newImage;
 	int sameDir;
+	int loadFlags;
+	int len, retryLower;
 
 	/* Scan the AssemblyRef table for the assemblies that we require */
 	assem = 0;
@@ -537,7 +561,90 @@ int _ILImageDynamicLink(ILImage *image, const char *filename, int flags)
 		   can assume that it is being loaded from a secure location */
 		if(!sameDir)
 		{
-			flags &= ~IL_LOADFLAG_INSECURE;
+			loadFlags = flags & ~IL_LOADFLAG_INSECURE;
+		}
+		else
+		{
+			loadFlags = flags;
+		}
+
+		/* Load the image */
+		error = ILImageLoadFromFile(pathname, image->context,
+							        &newImage, flags, 0);
+		ILFree(pathname);
+		if(error != 0)
+		{
+			if(error == -1)
+			{
+				return IL_LOADERR_UNRESOLVED;
+			}
+			else
+			{
+				return error;
+			}
+		}
+	}
+
+	/* If we loaded the parent from an insecure source, then bail out
+	   without attempting to load the module files */
+	if((flags & IL_LOADFLAG_INSECURE) != 0 || !filename)
+	{
+		return 0;
+	}
+
+	/* Strip the final component from the filename */
+	len = strlen(filename);
+	while(len > 0 && filename[len - 1] != '/' && filename[len - 1] != '\\')
+	{
+		--len;
+	}
+	if(len > 0)
+	{
+		--len;
+	}
+
+	/* Scan the File table for the external module files that we require */
+	file = 0;
+	while((file = (ILFileDecl *)ILImageNextToken
+				(image, IL_META_TOKEN_FILE, file)) != 0)
+	{
+		/* Ignore this file if it does not contain metadata */
+		if(!ILFileDecl_HasMetaData(file))
+		{
+			continue;
+		}
+
+		/* Ignore this file if we already have it */
+		if(ILContextGetFile(image->context, file->name) != 0)
+		{
+			continue;
+		}
+
+		/* Ignore this file if its name contains a '/' or '\', because
+		   files in other directories may be a security risk */
+		if(ILMemChr(file->name, '/', strlen(file->name)) != 0 ||
+		   ILMemChr(file->name, '\\', strlen(file->name)) != 0)
+		{
+			continue;
+		}
+
+		/* Get the full pathname of the referenced file */
+		retryLower = 0;
+		pathname = TestPathForFile(filename, len,
+								   file->name, strlen(file->name),
+								   0, 0, &retryLower, 0);
+		if(!pathname && retryLower)
+		{
+			pathname = TestPathForFile(filename, len,
+									   file->name, strlen(file->name),
+									   0, 0, &retryLower, 1);
+		}
+		if(!pathname)
+		{
+		#if IL_DEBUG_META
+			fprintf(stderr, "could not locate the file %s\n", file->name);
+		#endif
+			return IL_LOADERR_UNRESOLVED;
 		}
 
 		/* Load the image */
@@ -737,11 +844,42 @@ static int SearchForDllMap(ILProgramItem *item, const char *name,
 	const char *newName;
 	int newNameLen;
 
+#ifndef IL_WIN32_PLATFORM
+	/* Hack around a hard-wired library names in Gtk#.  This is temporary
+	   until we can come up with a better solution that does not require
+	   the user to manually edit configuration files */
+	if(!strcmp(name, "libgtk-win32-2.0-0.dll"))
+	{
+		*remapName = "libgtk-x11-2.0.so";
+		return strlen(*remapName);
+	}
+	else if(!strcmp(name, "libgobject-2.0-0.dll"))
+	{
+		*remapName = "libgobject-2.0.so";
+		return strlen(*remapName);
+	}
+	else if(!strcmp(name, "libglib-2.0-0.dll"))
+	{
+		*remapName = "libglib-2.0.so";
+		return strlen(*remapName);
+	}
+	else if(!strcmp(name, "libpango-1.0-0.dll"))
+	{
+		*remapName = "libpango-1.0.so";
+		return strlen(*remapName);
+	}
+	else if(!strcmp(name, "libatk-1.0-0.dll"))
+	{
+		*remapName = "libatk-1.0.so";
+		return strlen(*remapName);
+	}
+#endif
+
 	attr = 0;
 	while((attr = ILProgramItemNextAttribute(item, attr)) != 0)
 	{
 		method = ILProgramItemToMethod(ILAttributeTypeAsItem(attr));
-		if(method != 0 && !strcmp(method->member.owner->name,
+		if(method != 0 && !strcmp(method->member.owner->className->name,
 								  "DllImportMapAttribute"))
 		{
 			blob = ILAttributeGetValue(attr, &blobLen);
@@ -814,11 +952,11 @@ char *ILPInvokeResolveModule(ILPInvoke *pinvoke)
 	}
 
 	/* Does the name need to be remapped for this platform? */
-	namelen = SearchForDllMap(&(pinvoke->method->member.programItem),
+	namelen = SearchForDllMap(&(pinvoke->memberInfo->programItem),
 							  name, &remapName);
 	if(namelen == -1)
 	{
-		namelen = SearchForDllMap(&(pinvoke->method->member.owner->programItem),
+		namelen = SearchForDllMap(&(pinvoke->memberInfo->owner->programItem),
 							  	  name, &remapName);
 		if(namelen == -1)
 		{
@@ -861,7 +999,7 @@ char *ILPInvokeResolveModule(ILPInvoke *pinvoke)
 			}
 			++posn;
 		}
-		if(!needSuffix)
+		if(needSuffix)
 		{
 			/* Strip ".dll" from the end of the name if present */
 			if(namelen >= 4 && name[namelen - 4] == '.' &&
@@ -875,7 +1013,11 @@ char *ILPInvokeResolveModule(ILPInvoke *pinvoke)
 		}
 		if(needSuffix)
 		{
+		#if defined(__APPLE__) && defined(__MACH__)
+			suffix = ".dylib";
+		#else
 			suffix = ".so";
+		#endif
 		}
 	}
 #else
@@ -982,7 +1124,10 @@ char *ILPInvokeResolveModule(ILPInvoke *pinvoke)
 		{
 			strcpy(fullName, baseName);
 		}
-		strcat(fullName, suffix);
+		if(suffix)
+		{
+			strcat(fullName, suffix);
+		}
 		ILFree(baseName);
 		return fullName;
 	}
