@@ -351,6 +351,78 @@ void _ILMemberSetSignatureIndex(ILMember *member, ILUInt32 index)
 	member->signatureBlob = index;
 }
 
+static ILMethod *MemberToMethod(ILMember *member)
+{
+	ILMethod *accessor;
+	member = ILMemberResolve(member);
+	if(ILMember_IsMethod(member))
+	{
+		return (ILMethod *)member;
+	}
+	else if(ILMember_IsProperty(member))
+	{
+		accessor = ILProperty_Getter((ILProperty *)member);
+		if(!accessor)
+		{
+			accessor = ILProperty_Setter((ILProperty *)member);
+		}
+		return accessor;
+	}
+	else if(ILMember_IsEvent(member))
+	{
+		accessor = ILEvent_AddOn((ILEvent *)member);
+		if(!accessor)
+		{
+			accessor = ILEvent_RemoveOn((ILEvent *)member);
+		}
+		return accessor;
+	}
+	else
+	{
+		return 0;
+	}
+}
+
+ILMember *ILMemberGetBase(ILMember *member)
+{
+	ILClass *classInfo = ILClassResolve(ILMember_Owner(member));
+	ILClass *origClass = classInfo;
+	ILMember *testMember;
+	ILMethod *underlying;
+	if(classInfo)
+	{
+		classInfo = ILClass_Parent(classInfo);
+	}
+	while(classInfo != 0)
+	{
+		testMember = 0;
+		while((testMember = ILClassNextMemberByKind
+					(classInfo, testMember, ILMemberGetKind(member))) != 0)
+		{
+			if(!strcmp(ILMember_Name(testMember), ILMember_Name(member)) &&
+			   ILTypeIdentical(ILMember_Signature(testMember),
+			   				   ILMember_Signature(member)))
+			{
+				/* The member must be accessible from the original
+				   class to be considered a candidate.  This allows
+				   us to skip "private" members in parent classes
+				   that might otherwise indicate a match */
+				if(ILMemberAccessible(testMember, origClass))
+				{
+					underlying = MemberToMethod(testMember);
+					if(underlying && ILMethod_IsVirtual(underlying) &&
+					   ILMethod_IsNewSlot(underlying))
+					{
+						return testMember;
+					}
+				}
+			}
+		}
+		classInfo = ILClass_Parent(classInfo);
+	}
+	return 0;
+}
+
 ILMethod *ILMethodCreate(ILClass *info, ILToken token,
 						 const char *name, ILUInt32 attributes)
 {
@@ -420,8 +492,13 @@ ILParameter *ILMethodNextParam(ILMethod *method, ILParameter *last)
 	{
 		return last->next;
 	}
+	else if(method->parameters)
+	{
+		return method->parameters;
+	}
 	else
 	{
+		_ILMethodLoadParams(method);
 		return method->parameters;
 	}
 }
@@ -978,7 +1055,10 @@ ILParameter *ILParameterCreate(ILMethod *method, ILToken token,
 
 	/* Convert the parameter name into a persistent string */
 	param->name = _ILContextPersistString(image, name);
-	if(!(param->name))
+
+	/* Check persistent string only if name is not NULL and paramNum is
+	 * not Zero (retval) */
+	if(name && paramNum && !(param->name))
 	{
 		return 0;
 	}
@@ -1090,6 +1170,34 @@ ILType *ILFieldGetTypeWithPrefixes(ILField *field)
 	return ILMember_Signature(field);
 }
 
+int ILFieldIsThreadStatic(ILField *field)
+{
+	ILAttribute *attr;
+	ILMethod *ctor;
+	ILClass *classInfo;
+	if(!field || !ILField_IsStatic(field))
+	{
+		return 0;
+	}
+	attr = 0;
+	while((attr = ILProgramItemNextAttribute
+				(ILToProgramItem(field), attr)) != 0)
+	{
+		ctor = ILProgramItemToMethod(ILAttributeTypeAsItem(attr));
+		if(ctor)
+		{
+			classInfo = ILMethod_Owner(ctor);
+			if(!strcmp(ILClass_Name(classInfo), "ThreadStaticAttribute") &&
+			   ILClass_Namespace(classInfo) != 0 &&
+			   !strcmp(ILClass_Namespace(classInfo), "System"))
+			{
+				return 1;
+			}
+		}
+	}
+	return 0;
+}
+
 ILEvent *ILEventCreate(ILClass *info, ILToken token,
 					   const char *name, ILUInt32 attributes,
 					   ILClass *type)
@@ -1165,12 +1273,20 @@ ILProperty *ILPropertyCreate(ILClass *info, ILToken token,
 
 ILMethod *ILPropertyGetGetter(ILProperty *property)
 {
+	if(property->getter)
+	{
+		return property->getter;
+	}
 	return ILMethodSemGetByType(&(property->member.programItem),
 								IL_META_METHODSEM_GETTER);
 }
 
 ILMethod *ILPropertyGetSetter(ILProperty *property)
 {
+	if(property->setter)
+	{
+		return property->setter;
+	}
 	return ILMethodSemGetByType(&(property->member.programItem),
 								IL_META_METHODSEM_SETTER);
 }
@@ -1201,7 +1317,7 @@ ILPInvoke *ILPInvokeCreate(ILMethod *method, ILToken token,
 	}
 
 	/* Initialize the other PInvoke fields */
-	pinvoke->method = method;
+	pinvoke->memberInfo = (ILMember *)method;
 	pinvoke->module = module;
 	if(aliasName && aliasName[0] != '\0')
 	{
@@ -1220,9 +1336,22 @@ ILPInvoke *ILPInvokeCreate(ILMethod *method, ILToken token,
 	return pinvoke;
 }
 
+ILPInvoke *ILPInvokeFieldCreate(ILField *field, ILToken token,
+						        ILUInt32 attributes, ILModule *module,
+						        const char *aliasName)
+{
+	return ILPInvokeCreate((ILMethod *)field, token, attributes,
+						   module, aliasName);
+}
+
 ILMethod *ILPInvokeGetMethod(ILPInvoke *pinvoke)
 {
-	return pinvoke->method;
+	return ILProgramItemToMethod(ILToProgramItem(pinvoke->memberInfo));
+}
+
+ILField *ILPInvokeGetField(ILPInvoke *pinvoke)
+{
+	return ILProgramItemToField(ILToProgramItem(pinvoke->memberInfo));
 }
 
 ILModule *ILPInvokeGetModule(ILPInvoke *pinvoke)
@@ -1241,7 +1370,7 @@ ILPInvoke *ILPInvokeFind(ILMethod *method)
 	while(member != 0)
 	{
 		if(member->kind == IL_META_MEMBERKIND_PINVOKE &&
-		   ((ILPInvoke *)member)->method == method)
+		   ((ILPInvoke *)member)->memberInfo == (ILMember *)method)
 		{
 			return ((ILPInvoke *)member);
 		}
@@ -1250,6 +1379,11 @@ ILPInvoke *ILPInvokeFind(ILMethod *method)
 	return 0;
 }
 
+ILPInvoke *ILPInvokeFindField(ILField *field)
+{
+	return ILPInvokeFind((ILMethod *)field);
+}
+
 ILOverride *ILOverrideCreate(ILClass *info, ILToken token,
 					   		 ILMethod *decl, ILMethod *body)
 {
@@ -1463,6 +1597,19 @@ ILMethod *ILMethodSemGetByType(ILProgramItem *item, ILUInt32 type)
 		}
 		sem = sem->next;
 	}
+	if(property != 0)
+	{
+		/* Check the cached getter and setter values, in case
+		   we didn't actually create an ILMethodSem block */
+		if(type == IL_META_METHODSEM_GETTER)
+		{
+			return property->getter;
+		}
+		if(type == IL_META_METHODSEM_SETTER)
+		{
+			return property->setter;
+		}
+	}
 	return 0;
 }
 
