@@ -158,6 +158,12 @@ static ILUInt32 const ExportedType_Desc[] =
 	 1,
 	 END_DESC};
 
+static ILUInt32 const GenericParOwner_Desc[] =
+	{1,
+	 IL_META_TOKEN_TYPE_DEF,
+	 IL_META_TOKEN_METHOD_DEF,
+	 END_DESC};
+
 /*
  * Special field types.
  */
@@ -302,6 +308,14 @@ static const ILUInt32 * const Fields_ExeLocation[] =
 static const ILUInt32 * const Fields_NestedClass[] =
 	{TKREF_FIELD(TYPE_DEF), TKREF_FIELD(TYPE_DEF), END_FIELD};
 
+static const ILUInt32 * const Fields_GenericPar[] =
+	{UINT16_FIELD, UINT16_FIELD, GenericParOwner_Desc, 
+	 STRREF_FIELD, TypeDefRefOrSpec_Desc, TypeDefRefOrSpec_Desc,
+	 END_FIELD};
+
+static const ILUInt32 * const Fields_MethodSpec[] =
+	{MethodDefOrRef_Desc, BLOBREF_FIELD, END_FIELD};
+
 /*
  * Table of all field description types.
  */
@@ -348,8 +362,8 @@ static const ILUInt32 * const * const FieldDescriptions[] = {
 	Fields_ExportedType,
 	Fields_ManifestResource,	/* 28 */
 	Fields_NestedClass,
-	0,
-	0,
+	Fields_GenericPar,
+	Fields_MethodSpec,
 	0,
 	0,
 	0,
@@ -1224,6 +1238,7 @@ unsigned long ILImageNumTokens(ILImage *image, ILToken tokenType)
 void *ILImageTokenInfo(ILImage *image, ILToken token)
 {
 	void **data;
+	void *item;
 	if(token < (unsigned long)0x40000000)
 	{
 		ILToken tokenId = (token & (unsigned long)0x00FFFFFF);
@@ -1233,7 +1248,38 @@ void *ILImageTokenInfo(ILImage *image, ILToken token)
 			data = image->tokenData[tokenType];
 			if(data)
 			{
-				return data[tokenId - 1];
+				item = data[tokenId - 1];
+				if(item)
+				{
+					return item;
+				}
+			}
+			if(image->type != IL_IMAGETYPE_BUILDING)
+			{
+				/* Perform on-demand loading of the token */
+				return _ILImageLoadOnDemand(image, token);
+			}
+		}
+	}
+	return 0;
+}
+
+int _ILImageTokenAlreadyLoaded(ILImage *image, ILToken token)
+{
+	void **data;
+	if(token < (unsigned long)0x40000000)
+	{
+		ILToken tokenId = (token & (unsigned long)0x00FFFFFF);
+		ILToken tokenType = (token >> 24);
+		if(tokenId >= 1 && tokenId <= image->tokenCount[tokenType])
+		{
+			data = image->tokenData[tokenType];
+			if(data)
+			{
+				if(data[tokenId - 1])
+				{
+					return 1;
+				}
 			}
 		}
 	}
@@ -1254,8 +1300,8 @@ void *ILImageSearchForToken(ILImage *image, ILToken tokenType,
 	ILToken maxToken;
 	ILToken token;
 	ILToken left, right;
-	void **data;
 	void *item;
+	void **data;
 	int cmp;
 
 	/* Find the table in question */
@@ -1271,7 +1317,7 @@ void *ILImageSearchForToken(ILImage *image, ILToken tokenType,
 		while(left <= right)
 		{
 			token = ((left + right) / 2);
-			item = data[(token & ~IL_META_TOKEN_MASK) - 1];
+			item = ILImageTokenInfo(image, token);
 			if(!item)
 			{
 				/* There is a gap in the table: revert to linear search */
@@ -1298,10 +1344,20 @@ void *ILImageSearchForToken(ILImage *image, ILToken tokenType,
 	linearSearch:
 		for(token = (tokenType | 1); token <= maxToken; ++token)
 		{
-			item = data[(token & ~IL_META_TOKEN_MASK) - 1];
-			if(item && (*compareFunc)(item, userData) == 0)
+			if(data && (item = data[(token & ~IL_META_TOKEN_MASK) - 1]) != 0)
 			{
-				return item;
+				if((*compareFunc)(item, userData) == 0)
+				{
+					return item;
+				}
+			}
+			else
+			{
+				item = ILImageTokenInfo(image, token);
+				if(item && (*compareFunc)(item, userData) == 0)
+				{
+					return item;
+				}
 			}
 		}
 	}
@@ -1340,7 +1396,10 @@ int _ILImageSetToken(ILImage *image, ILProgramItem *item,
 	if(image->type != IL_IMAGETYPE_BUILDING)
 	{
 		/* We are loading an image, so use the supplied token */
-		item->token = (ILUInt32)token;
+		if(!(item->token))
+		{
+			item->token = (ILUInt32)token;
+		}
 		if(token < (unsigned long)0x40000000)
 		{
 			tokenId = (token & (unsigned long)0x00FFFFFF);
