@@ -49,6 +49,10 @@ ILType *ILTypeCreateArray(ILContext *context, unsigned long rank,
 		{
 			elem = type;
 			type = ILMemPoolCalloc(&(context->typePool), ILType);
+			if(!type)
+			{
+				return 0;
+			}
 			type->kind__ = IL_TYPE_COMPLEX_ARRAY_CONTINUE;
 			type->un.array__.elemType__ = elem;
 			type->un.array__.size__ = 0;
@@ -59,6 +63,46 @@ ILType *ILTypeCreateArray(ILContext *context, unsigned long rank,
 	return type;
 }
 
+ILType *ILTypeFindOrCreateArray(ILContext *context, unsigned long rank,
+						        ILType *elem)
+{
+	ILType *type;
+	ILType dims[8];
+	unsigned long index;
+	ILClass *info;
+
+	/* Bail out if too many dimensions to be worth checking */
+	if(rank > 8)
+	{
+		return ILTypeCreateArray(context, rank, elem);
+	}
+
+	/* Create a pseudo type on the stack, for performing hash lookups */
+	ILMemZero(dims, sizeof(ILType) * rank);
+	for(index = 0; index < (rank - 1); ++index)
+	{
+		dims[index].kind__ = IL_TYPE_COMPLEX_ARRAY_CONTINUE;
+		dims[index].un.array__.elemType__ = &(dims[index + 1]);
+		dims[index].un.array__.size__ = 0;
+		dims[index].un.array__.lowBound__ = 0;
+	}
+	dims[index].kind__ = IL_TYPE_COMPLEX_ARRAY;
+	dims[index].un.array__.elemType__ = elem;
+	dims[index].un.array__.size__ = 0;
+	dims[index].un.array__.lowBound__ = 0;
+	type = &(dims[0]);
+
+	/* Look for the type in the synthetic hash */
+	info = ILHashFindType(context->syntheticHash, type, ILClass);
+	if(info)
+	{
+		return info->synthetic;
+	}
+
+	/* Create the type for the first time */
+	return ILTypeCreateArray(context, rank, elem);
+}
+
 void ILTypeSetSize(ILType *array, unsigned long dimension, long value)
 {
 	while(dimension > 0)
@@ -292,7 +336,8 @@ int ILTypeAddSentinel(ILContext *context, ILType *method)
 
 unsigned long ILTypeNumParams(ILType *method)
 {
-	if(ILType_IsMethod(method) || ILType_IsProperty(method))
+	if(ILType_IsMethod(method) || ILType_IsProperty(method) ||
+	   ILType_IsWith(method))
 	{
 		return method->num__;
 	}
@@ -309,7 +354,8 @@ ILType *ILTypeGetParam(ILType *method, unsigned long index)
 
 ILType *ILTypeGetParamWithPrefixes(ILType *method, unsigned long index)
 {
-	if(!ILType_IsMethod(method) && !ILType_IsProperty(method))
+	if(!ILType_IsMethod(method) && !ILType_IsProperty(method) &&
+	   !ILType_IsWith(method))
 	{
 		return 0;
 	}
@@ -356,7 +402,7 @@ void ILTypeSetReturn(ILType *type, ILType *retType)
 
 ILType *ILTypeGetReturn(ILType *type)
 {
-	if(ILType_IsMethod(type) || ILType_IsProperty(type))
+	if(ILType_IsMethod(type) || ILType_IsProperty(type) || ILType_IsWith(type))
 	{
 		return ILTypeStripPrefixes(type->un.method__.retType__);
 	}
@@ -368,7 +414,7 @@ ILType *ILTypeGetReturn(ILType *type)
 
 ILType *ILTypeGetReturnWithPrefixes(ILType *type)
 {
-	if(ILType_IsMethod(type) || ILType_IsProperty(type))
+	if(ILType_IsMethod(type) || ILType_IsProperty(type) || ILType_IsWith(type))
 	{
 		return type->un.method__.retType__;
 	}
@@ -489,11 +535,19 @@ int ILTypeIdentical(ILType *type1, ILType *type2)
 		}
 		/* Not reached */
 
+		case IL_TYPE_COMPLEX_MVAR:
+		case IL_TYPE_COMPLEX_VAR:
+		{
+			return (ILType_VarNum(type1) == ILType_VarNum(type2));
+		}
+		/* Not reached */
+
 		default:
 		{
-			/* Probably a method or property */
+			/* Probably a method, property, or generic type specification */
 			if((type1->kind__ & 0xFF) == IL_TYPE_COMPLEX_PROPERTY ||
-			   (type1->kind__ & IL_TYPE_COMPLEX_METHOD) != 0)
+			   (type1->kind__ & IL_TYPE_COMPLEX_METHOD) != 0 ||
+			   (type1->kind__ & 0xFF) == IL_TYPE_COMPLEX_WITH)
 			{
 				/* Check the property or method signature */
 				if(type1->num__ != type2->num__)
@@ -545,11 +599,14 @@ static char *AppendString(char *str1, const char *str2)
 char *ILTypeToName(ILType *type)
 {
 	char *name;
+	char *elemName;
 	ILClass *info;
 	int len;
 	const char *assemName;
 	char numbuf[80];
 	ILType *elemType;
+	unsigned long numParams;
+	unsigned long param;
 
 	/* Strip unnecessary prefixes from the type */
 	type = ILTypeStripPrefixes(type);
@@ -584,11 +641,11 @@ char *ILTypeToName(ILType *type)
 		{
 			len = 0;
 		}
-		if(info->namespace)
+		if(info->className->namespace)
 		{
-			len += strlen(info->namespace) + 1;
+			len += strlen(info->className->namespace) + 1;
 		}
-		len += strlen(info->name) + 1;
+		len += strlen(info->className->name) + 1;
 
 		/* Build the fully-qualified class name */
 		name = (char *)ILMalloc(len);
@@ -605,13 +662,13 @@ char *ILTypeToName(ILType *type)
 			{
 				len = 0;
 			}
-			if(info->namespace)
+			if(info->className->namespace)
 			{
-				strcpy(name + len, info->namespace);
-				len += strlen(info->namespace);
+				strcpy(name + len, info->className->namespace);
+				len += strlen(info->className->namespace);
 				name[len++] = '.';
 			}
-			strcpy(name + len, info->name);
+			strcpy(name + len, info->className->name);
 		}
 		return name;
 	}
@@ -744,9 +801,80 @@ char *ILTypeToName(ILType *type)
 
 			case IL_TYPE_COMPLEX_METHOD:
 			{
-				/* TODO */
+				name = ILTypeToName(ILTypeGetReturn(type));
+				if(ILType_HasThis(type))
+				{
+					name = AppendString(name, " * instance (");
+				}
+				else
+				{
+					name = AppendString(name, " * (");
+				}
+				numParams = ILTypeNumWithParams(type);
+				for(param = 1; param <= numParams; ++param)
+				{
+					if(param != 1)
+					{
+						name = AppendString(name, ", ");
+					}
+					elemType = ILTypeGetParam(type, param);
+					elemName = ILTypeToName(elemType);
+					if(!elemName)
+					{
+						if(name)
+						{
+							ILFree(name);
+						}
+						return 0;
+					}
+					name = AppendString(name, elemName);
+					ILFree(elemName);
+				}
+				return AppendString(name, ")");
 			}
-			break;
+			/* Not reached */
+
+			case IL_TYPE_COMPLEX_WITH:
+			{
+				name = ILTypeToName(ILTypeGetWithMain(type));
+				name = AppendString(name, "<");
+				numParams = ILTypeNumWithParams(type);
+				for(param = 1; param <= numParams; ++param)
+				{
+					if(param != 1)
+					{
+						name = AppendString(name, ", ");
+					}
+					elemType = ILTypeGetWithParam(type, param);
+					elemName = ILTypeToName(elemType);
+					if(!elemName)
+					{
+						if(name)
+						{
+							ILFree(name);
+						}
+						return 0;
+					}
+					name = AppendString(name, elemName);
+					ILFree(elemName);
+				}
+				return AppendString(name, ">");
+			}
+			/* Not reached */
+
+			case IL_TYPE_COMPLEX_MVAR:
+			{
+				sprintf(numbuf, "!!%d", ILType_VarNum(type));
+				return ILDupString(numbuf);
+			}
+			/* Not reached */
+
+			case IL_TYPE_COMPLEX_VAR:
+			{
+				sprintf(numbuf, "!%d", ILType_VarNum(type));
+				return ILDupString(numbuf);
+			}
+			/* Not reached */
 
 			default: break;
 		}
@@ -827,8 +955,9 @@ int ILTypeIsStringClass(ILType *type)
 	{
 		/* Check the name against "System.String" */
 		info = ILClassResolve(ILType_ToClass(type));
-		if(!strcmp(info->name, "String") &&
-		   info->namespace && !strcmp(info->namespace, "System"))
+		if(!strcmp(info->className->name, "String") &&
+		   info->className->namespace &&
+		   !strcmp(info->className->namespace, "System"))
 		{
 			/* Check that it is within the system image, to prevent
 			   applications from fooling us into believing that their
@@ -852,8 +981,9 @@ int ILTypeIsObjectClass(ILType *type)
 	{
 		/* Check the name against "System.String" */
 		info = ILClassResolve(ILType_ToClass(type));
-		if(!strcmp(info->name, "Object") &&
-		   info->namespace && !strcmp(info->namespace, "System"))
+		if(!strcmp(info->className->name, "Object") &&
+		   info->className->namespace &&
+		   !strcmp(info->className->namespace, "System"))
 		{
 			/* Check that it is within the system image, to prevent
 			   applications from fooling us into believing that their
@@ -869,7 +999,8 @@ int ILTypeIsObjectClass(ILType *type)
 	return 0;
 }
 
-int ILTypeAssignCompatible(ILImage *image, ILType *src, ILType *dest)
+static int TypeAssignCompatible(ILImage *image, ILType *src,
+								ILType *dest, int allowBoxing)
 {
 	ILClass *classInfo;
 	ILClass *classInfo2;
@@ -880,7 +1011,7 @@ int ILTypeAssignCompatible(ILImage *image, ILType *src, ILType *dest)
 	dest = ILTypeGetEnumType(ILTypeStripPrefixes(dest));
 
 	/* Determine how to compare the types based on their kind */
-	if(ILType_IsPrimitive(src))
+	if(ILType_IsPrimitive(src) && !allowBoxing)
 	{
 		/* Primitive type assignments must be identical */
 		return (src == dest);
@@ -903,6 +1034,7 @@ int ILTypeAssignCompatible(ILImage *image, ILType *src, ILType *dest)
 			/* Both types must be object references */
 			return 0;
 		}
+	boxCheck:
 		classInfo = ILClassFromType(image, 0, dest, 0);
 		classInfo2 = ILClassFromType(image, 0, src, 0);
 		if(classInfo && classInfo2)
@@ -940,6 +1072,11 @@ int ILTypeAssignCompatible(ILImage *image, ILType *src, ILType *dest)
 			return 0;
 		}
 	}
+	else if(allowBoxing && ILTypeIsReference(dest))
+	{
+		/* Check for boxing conversions */
+		goto boxCheck;
+	}
 	else
 	{
 		/* Everything else must have type identity to be assignable */
@@ -947,6 +1084,16 @@ int ILTypeAssignCompatible(ILImage *image, ILType *src, ILType *dest)
 	}
 }
 
+int ILTypeAssignCompatible(ILImage *image, ILType *src, ILType *dest)
+{
+	return TypeAssignCompatible(image, src, dest, 1);
+}
+
+int ILTypeAssignCompatibleNonBoxing(ILImage *image, ILType *src, ILType *dest)
+{
+	return TypeAssignCompatible(image, src, dest, 0);
+}
+
 int ILTypeHasModifier(ILType *type, ILClass *classInfo)
 {
 	classInfo = ILClassResolve(classInfo);
@@ -994,7 +1141,21 @@ int ILTypeIsReference(ILType *type)
 
 int ILTypeIsEnum(ILType *type)
 {
-	return (ILTypeGetEnumType(type) != type);
+	if(ILType_IsValueType(type))
+	{
+		ILClass *classInfo = ILClassResolve(ILType_ToValueType(type));
+		ILClass *parent = ILClass_Parent(classInfo);
+		if(parent)
+		{
+			const char *namespace = ILClass_Namespace(parent);
+			if(namespace && !strcmp(namespace, "System") &&
+			   !strcmp(ILClass_Name(parent), "Enum"))
+			{
+				return 1;
+			}
+		}
+	}
+	return 0;
 }
 
 int ILTypeIsValue(ILType *type)
@@ -1140,6 +1301,59 @@ int ILTypeIsDelegateSubClass(ILType *type)
 	return 0;
 }
 
+ILType *ILTypeCreateVarNum(ILContext *context, int kind, int num)
+{
+	ILType *type = ILMemPoolCalloc(&(context->typePool), ILType);
+	if(type)
+	{
+		type->kind__ = kind;
+		type->un.num__ = num;
+	}
+	return type;
+}
+
+ILType *ILTypeCreateWith(ILContext *context, ILType *mainType)
+{
+	ILType *type = ILMemPoolCalloc(&(context->typePool), ILType);
+	if(type)
+	{
+		type->kind__ = IL_TYPE_COMPLEX_WITH;
+		type->num__ = 0;
+		type->un.method__.retType__ = mainType;
+	}
+	return type;
+}
+
+int ILTypeAddWithParam(ILContext *context, ILType *type, ILType *paramType)
+{
+	return ILTypeAddParam(context, type, paramType);
+}
+
+unsigned long ILTypeNumWithParams(ILType *type)
+{
+	return ILTypeNumParams(type);
+}
+
+ILType *ILTypeGetWithParam(ILType *type, unsigned long num)
+{
+	return ILTypeGetParam(type, num);
+}
+
+ILType *ILTypeGetWithParamWithPrefixes(ILType *type, unsigned long num)
+{
+	return ILTypeGetParamWithPrefixes(type, num);
+}
+
+ILType *ILTypeGetWithMain(ILType *type)
+{
+	return ILTypeGetReturn(type);
+}
+
+ILType *ILTypeGetWithMainWithPrefixes(ILType *type)
+{
+	return ILTypeGetReturnWithPrefixes(type);
+}
+
 #ifdef	__cplusplus
 };
 #endif
