@@ -236,6 +236,34 @@ const void *ILConstantGetValue(ILConstant *constant, unsigned long *len)
  * uses the encoded token value rather than the raw token
  * values that we are using here.
  */
+static int OwnedItemCompareRaw(ILUInt32 *values, ILToken searchFor,
+							   int valueField)
+{
+	ILToken token1 = searchFor;
+	ILToken token1Stripped = (token1 & ~IL_META_TOKEN_MASK);
+	ILToken token2 = (ILToken)(values[valueField]);
+	ILToken token2Stripped = (token2 & ~IL_META_TOKEN_MASK);
+	if(token1Stripped < token2Stripped)
+	{
+		return -1;
+	}
+	else if(token1Stripped > token2Stripped)
+	{
+		return 1;
+	}
+	else if(token1 < token2)
+	{
+		return -1;
+	}
+	else if(token1 > token2)
+	{
+		return 1;
+	}
+	else
+	{
+		return 0;
+	}
+}
 static int OwnedItemCompare(void *item, void *userData)
 {
 	ILOwnedItem *owned = (ILOwnedItem *)item;
@@ -265,11 +293,65 @@ static int OwnedItemCompare(void *item, void *userData)
 	}
 }
 
+/*
+ * Search for an owned item.
+ */
+static void *SearchForOwnedItem(ILImage *image, ILToken tokenType,
+								void *owner, int valueField)
+{
+	ILUInt32 token, num;
+	if(!owner)
+	{
+		return 0;
+	}
+	if(image->type == IL_IMAGETYPE_BUILDING)
+	{
+		return ILImageSearchForToken(image, tokenType, OwnedItemCompare, owner);
+	}
+	num = _ILSearchForRawToken(image, OwnedItemCompareRaw, tokenType, &token,
+							   ((ILProgramItem *)owner)->token,
+							   valueField);
+	if(num != 1)
+	{
+		return 0;
+	}
+	return ILImageTokenInfo(image, token);
+}
+
 ILConstant *ILConstantGetFromOwner(ILProgramItem *owner)
 {
-	return (ILConstant *)ILImageSearchForToken
+	ILField *field;
+	ILParameter *param;
+	ILProperty *property;
+
+	/* Filter out program items that obviously don't have a value,
+	   to avoid searching the constant table unnecessarily */
+	if((field = ILProgramItemToField(owner)) != 0)
+	{
+		if((field->member.attributes & IL_META_FIELDDEF_HAS_DEFAULT) == 0)
+		{
+			return 0;
+		}
+	}
+	else if((param = ILProgramItemToParameter(owner)) != 0)
+	{
+		if((param->attributes & IL_META_PARAMDEF_HAS_DEFAULT) == 0)
+		{
+			return 0;
+		}
+	}
+	else if((property = ILProgramItemToProperty(owner)) != 0)
+	{
+		if((property->member.attributes & IL_META_PROPDEF_HAS_DEFAULT) == 0)
+		{
+			return 0;
+		}
+	}
+
+	/* Search for the constant token */
+	return (ILConstant *)SearchForOwnedItem
 				(owner->image, IL_META_TOKEN_CONSTANT,
-				 OwnedItemCompare, (void *)owner);
+				 (void *)owner, IL_OFFSET_CONSTANT_REFERENCE);
 }
 
 ILFieldRVA *ILFieldRVACreate(ILImage *image, ILToken token,
@@ -315,12 +397,21 @@ ILUInt32 ILFieldRVAGetRVA(ILFieldRVA *rva)
 	return rva->rva;
 }
 
+void ILFieldRVASetRVA(ILFieldRVA *rva, ILUInt32 value)
+{
+	rva->rva = value;
+}
+
 ILFieldRVA *ILFieldRVAGetFromOwner(ILField *owner)
 {
-	return (ILFieldRVA *)ILImageSearchForToken
+	if((owner->member.attributes & IL_META_FIELDDEF_HAS_FIELD_RVA) == 0)
+	{
+		return 0;
+	}
+	return (ILFieldRVA *)SearchForOwnedItem
 				(owner->member.programItem.image,
 				 IL_META_TOKEN_FIELD_RVA,
-				 OwnedItemCompare, (void *)owner);
+				 (void *)owner, IL_OFFSET_FIELDRVA_FIELD);
 }
 
 ILFieldLayout *ILFieldLayoutCreate(ILImage *image, ILToken token,
@@ -361,12 +452,25 @@ ILUInt32 ILFieldLayoutGetOffset(ILFieldLayout *layout)
 	return layout->offset;
 }
 
+void ILFieldLayoutSetOffset(ILFieldLayout *layout, ILUInt32 offset)
+{
+	layout->offset = offset;
+}
+
 ILFieldLayout *ILFieldLayoutGetFromOwner(ILField *owner)
 {
-	return (ILFieldLayout *)ILImageSearchForToken
+	/* Field layout information can only exist on explicit-layout classes,
+	   so bail out early if we have some other kind of class */
+	if(!ILClass_IsExplicitLayout(owner->member.owner))
+	{
+		return 0;
+	}
+
+	/* Search for the layout token */
+	return (ILFieldLayout *)SearchForOwnedItem
 				(owner->member.programItem.image,
 				 IL_META_TOKEN_FIELD_LAYOUT,
-				 OwnedItemCompare, (void *)owner);
+				 (void *)owner, IL_OFFSET_FIELDLAYOUT_FIELD);
 }
 
 ILFieldMarshal *ILFieldMarshalCreate(ILImage *image, ILToken token,
@@ -446,9 +550,29 @@ const void *ILFieldMarshalGetType(ILFieldMarshal *marshal, unsigned long *len)
 
 ILFieldMarshal *ILFieldMarshalGetFromOwner(ILProgramItem *owner)
 {
-	return (ILFieldMarshal *)ILImageSearchForToken
+	ILField *field;
+	ILParameter *param;
+
+	/* Filter out members that obviously don't have marshal information */
+	if((field = ILProgramItemToField(owner)) != 0)
+	{
+		if((field->member.attributes & IL_META_FIELDDEF_HAS_FIELD_MARSHAL) == 0)
+		{
+			return 0;
+		}
+	}
+	else if((param = ILProgramItemToParameter(owner)) != 0)
+	{
+		if((param->attributes & IL_META_PARAMDEF_HAS_FIELD_MARSHAL) == 0)
+		{
+			return 0;
+		}
+	}
+
+	/* Search for the token */
+	return (ILFieldMarshal *)SearchForOwnedItem
 				(owner->image, IL_META_TOKEN_FIELD_MARSHAL,
-				 OwnedItemCompare, (void *)owner);
+				 (void *)owner, IL_OFFSET_FIELDMARSHAL_TOKEN);
 }
 
 ILClassLayout *ILClassLayoutCreate(ILImage *image, ILToken token,
@@ -509,9 +633,16 @@ void ILClassLayoutSetClassSize(ILClassLayout *layout, ILUInt32 size)
 
 ILClassLayout *ILClassLayoutGetFromOwner(ILClass *owner)
 {
-	return (ILClassLayout *)ILImageSearchForToken
+	/* Class layout information cannot exist on auto-layout classes */
+	if(ILClass_IsAutoLayout(owner))
+	{
+		return 0;
+	}
+
+	/* Search for the layout token */
+	return (ILClassLayout *)SearchForOwnedItem
 				(owner->programItem.image, IL_META_TOKEN_CLASS_LAYOUT,
-				 OwnedItemCompare, (void *)owner);
+				 (void *)owner, IL_OFFSET_CLASSLAYOUT_TYPE);
 }
 
 ILDeclSecurity *ILDeclSecurityCreate(ILImage *image, ILToken token,
@@ -597,9 +728,9 @@ const void *ILDeclSecurityGetBlob(ILDeclSecurity *security, unsigned long *len)
 
 ILDeclSecurity *ILDeclSecurityGetFromOwner(ILProgramItem *owner)
 {
-	return (ILDeclSecurity *)ILImageSearchForToken
+	return (ILDeclSecurity *)SearchForOwnedItem
 				(owner->image, IL_META_TOKEN_DECL_SECURITY,
-				 OwnedItemCompare, (void *)owner);
+				 (void *)owner, IL_OFFSET_DECLSECURITY_TOKEN);
 }
 
 ILFileDecl *ILFileDeclCreate(ILImage *image, ILToken token,
@@ -765,29 +896,33 @@ ILExportedType *ILExportedTypeCreate(ILImage *image, ILToken token,
 	}
 
 	/* Set the exported type fields */
-	type->programItem.image = image;
-	type->attributes = attributes;
+	type->classItem.programItem.image = image;
+	type->classItem.attributes = attributes | IL_META_TYPEDEF_REFERENCE;
+	type->classItem.className =
+		_ILClassNameLookup(image, 0, 0, name, namespace);
+	if(!(type->classItem.className))
+	{
+		type->classItem.className =
+			_ILClassNameCreate(image, token, name, namespace, 0, 0);
+	}
+	if(!(type->classItem.className))
+	{
+		return 0;
+	}
 	type->identifier = 0;
-	type->name = _ILContextPersistString(image, name);
-	type->namespace = _ILContextPersistString(image, namespace);
-	type->scope = 0;
 
 	/* Assign a token code to the exported type */
-	if(!_ILImageSetToken(image, &(type->programItem), token,
+	if(!_ILImageSetToken(image, &(type->classItem.programItem), token,
 						 IL_META_TOKEN_EXPORTED_TYPE))
 	{
 		return 0;
 	}
+	_ILClassNameUpdateToken(&(type->classItem));
 
 	/* Return the exported type to the caller */
 	return type;
 }
 
-ILUInt32 ILExportedTypeGetAttrs(ILExportedType *type)
-{
-	return type->attributes;
-}
-
 void ILExportedTypeSetId(ILExportedType *type, ILUInt32 identifier)
 {
 	type->identifier = identifier;
@@ -798,34 +933,19 @@ ILUInt32 ILExportedTypeGetId(ILExportedType *type)
 	return type->identifier;
 }
 
-const char *ILExportedTypeGetName(ILExportedType *type)
-{
-	return type->name;
-}
-
-const char *ILExportedTypeGetNamespace(ILExportedType *type)
-{
-	return type->namespace;
-}
-
 void ILExportedTypeSetScopeFile(ILExportedType *type, ILFileDecl *decl)
 {
-	type->scope = (ILProgramItem *)decl;
+	type->classItem.className->scope = (ILProgramItem *)decl;
 }
 
 void ILExportedTypeSetScopeAssembly(ILExportedType *type, ILAssembly *assem)
 {
-	type->scope = (ILProgramItem *)assem;
+	type->classItem.className->scope = (ILProgramItem *)assem;
 }
 
 void ILExportedTypeSetScopeType(ILExportedType *type, ILExportedType *scope)
 {
-	type->scope = (ILProgramItem *)scope;
-}
-
-ILProgramItem *ILExportedTypeGetScope(ILExportedType *type)
-{
-	return type->scope;
+	type->classItem.className->scope = (ILProgramItem *)scope;
 }
 
 ILExportedType *ILExportedTypeFind(ILImage *image,
@@ -841,18 +961,18 @@ ILExportedType *ILExportedTypeFind(ILImage *image,
 	while((type = (ILExportedType *)ILImageNextToken
 				(image, IL_META_TOKEN_EXPORTED_TYPE, type)) != 0)
 	{
-		if(!strcmp(type->name, name))
+		if(!strcmp(type->classItem.className->name, name))
 		{
 			if(!namespace)
 			{
-				if(!(type->namespace))
+				if(!(type->classItem.className->namespace))
 				{
 					return type;
 				}
 			}
-			else if(type->namespace)
+			else if(type->classItem.className->namespace)
 			{
-				if(!strcmp(type->namespace, namespace))
+				if(!strcmp(type->classItem.className->namespace, namespace))
 				{
 					return type;
 				}
