@@ -1,7 +1,7 @@
 /*
  * pecoff_loader.c - Deal with the ugly parts of loading PE/COFF images.
  *
- * Copyright (C) 2001  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2001, 2003  Southern Storm Software, Pty Ltd.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,18 +19,302 @@
  */
 
 #include "image.h"
+#if defined(IL_CONFIG_GZIP) && defined(HAVE_ZLIB_H) && defined(HAVE_LIBZ)
+	#define	IL_USE_GZIP	1
+#endif
+#ifdef IL_USE_GZIP
+	#include <zlib.h>
+#endif
 
 #ifdef	__cplusplus
 extern	"C" {
 #endif
 
+/*
+ * Input context.
+ */
+typedef struct _tagILInputContext ILInputContext;
+struct _tagILInputContext
+{
+#ifndef REDUCED_STDIO
+	FILE		   *stream;
+#endif
+	const char	   *buffer;
+	unsigned long	bufLen;
+	int (*readFunc)(ILInputContext *ctx, void *buf, unsigned len);
+#ifdef IL_USE_GZIP
+	int				sawEOF;
+	z_streamp		gzStream;
+	ILInputContext *linked;
+#endif
+};
+
+#ifndef REDUCED_STDIO
+
+/*
+ * Stdio-based read operation.
+ */
+static int StdioRead(ILInputContext *ctx, void *buf, unsigned len)
+{
+	return (int)fread(buf, 1, len, ctx->stream);
+}
+
+#else
+
+#define	fileno(f)	0
+
+#endif
+
+/*
+ * Memory-based read operation.
+ */
+static int MemoryRead(ILInputContext *ctx, void *buf, unsigned len)
+{
+	if(ctx->bufLen < len)
+	{
+		len = (unsigned)(ctx->bufLen);
+	}
+	if(len > 0)
+	{
+		ILMemCpy(buf, ctx->buffer, len);
+		ctx->buffer += len;
+		ctx->bufLen -= len;
+	}
+	return (int)len;
+}
+
+#ifdef IL_USE_GZIP
+
+/*
+ * Gzip flag bits in the header.
+ */
+#define GZ_ASCII_FLAG   0x01
+#define GZ_HEAD_CRC     0x02
+#define GZ_EXTRA_FIELD  0x04
+#define GZ_ORIG_NAME    0x08
+#define GZ_COMMENT      0x10
+#define GZ_RESERVED     0xE0
+
+/*
+ * Fill the gzip read buffer if necessary.
+ */
+static void GzipFillRead(ILInputContext *ctx)
+{
+	if(ctx->gzStream->avail_in == 0 && !(ctx->sawEOF))
+	{
+		int temp = (*(ctx->linked->readFunc))
+			(ctx->linked, (void *)(ctx->buffer), BUFSIZ);
+		if(temp <= 0)
+		{
+			ctx->gzStream->avail_in = 0;
+			ctx->sawEOF = 1;
+		}
+		else
+		{
+			ctx->gzStream->avail_in = (unsigned)temp;
+		}
+		ctx->gzStream->next_in = (Bytef *)(ctx->buffer);
+	}
+}
+
+/*
+ * Read a single byte from the read buffer.
+ */
+static int GzipReadByte(ILInputContext *ctx)
+{
+	GzipFillRead(ctx);
+	if(ctx->gzStream->avail_in > 0)
+	{
+		--(ctx->gzStream->avail_in);
+		return *(ctx->gzStream->next_in)++;
+	}
+	else
+	{
+		return -1;
+	}
+}
+
+/*
+ * Read a short value from the read buffer.
+ */
+static int GzipReadShort(ILInputContext *ctx)
+{
+	int b1 = GzipReadByte(ctx);
+	int b2 = GzipReadByte(ctx);
+	return (b1 | (b2 << 8));
+}
+
+/*
+ * Skip a number of bytes in the read buffer.
+ */
+static void GzipSkip(ILInputContext *ctx, int len)
+{
+	while(len > 0 && GzipReadByte(ctx) != -1)
+	{
+		--len;
+	}
+}
+
+/*
+ * Skip a NUL-terminated string in the read buffer.
+ */
+static void GzipSkipString(ILInputContext *ctx)
+{
+	int b;
+	while((b = GzipReadByte(ctx)) != -1 && b != 0)
+	{
+		/* Nothing to do here */
+	}
+}
+
+/*
+ * Gzip-based read operation.  This version doesn't check CRC's
+ * or handle concatenated gzip files.  Fix later if required.
+ */
+static int GzipRead(ILInputContext *ctx, void *buf, unsigned len)
+{
+	int error;
+
+	/* Initialize the zlib output buffer */
+	ctx->gzStream->next_out = (Bytef *)buf;
+	ctx->gzStream->avail_out = len;
+
+	/* Read data and write it to the output buffer */
+	while(ctx->gzStream->avail_out != 0 && !(ctx->sawEOF))
+	{
+		/* Read more data from the input stream if necessary */
+		GzipFillRead(ctx);
+
+		/* Inflate the input data */
+		error = inflate(ctx->gzStream, Z_NO_FLUSH);
+		if(error == Z_STREAM_END)
+		{
+			/* We've reached the end of the gzip data: skip the CRC value */
+			GzipSkip(ctx, 4);
+			ctx->sawEOF = 1;
+		}
+		else if(error != Z_OK)
+		{
+			/* Some kind of error occurred in the input data */
+			ctx->sawEOF = 1;
+		}
+	}
+
+	/* Read the length of the uncompressed data to the caller */
+	return (int)(len - ctx->gzStream->avail_out);
+}
+
+
+/*
+ * Finalize gzip control structures in an input context.
+ */
+static void GzipFinalize(ILInputContext *ctx)
+{
+	ILInputContext *linked;
+	if(ctx->gzStream)
+	{
+		inflateEnd(ctx->gzStream);
+		ILFree(ctx->gzStream);
+		ILFree((char *)(ctx->buffer));
+		linked = ctx->linked;
+		*ctx = *linked;
+		ILFree(linked);
+	}
+}
+
+/*
+ * Gzip initialize operation.  It is assumed that the first two
+ * bytes of the gzip stream (containing the magic number) have
+ * already been read.
+ */
+static int GzipInitialize(ILInputContext *ctx)
+{
+	ILInputContext *linked;
+	char *buffer;
+	z_streamp gzStream;
+	int method, flags, len;
+
+	/* Make a copy of the context for performing linked reads */
+	linked = (ILInputContext *)ILMalloc(sizeof(ILInputContext));
+	if(!linked)
+	{
+		return 0;
+	}
+	*linked = *ctx;
+
+	/* Allocate a buffer for holding the gzip input data */
+	buffer = (char *)ILMalloc(BUFSIZ);
+	if(!buffer)
+	{
+		ILFree(linked);
+		return 0;
+	}
+
+	/* Initialize the gzip stream details */
+	gzStream = (z_streamp)ILCalloc(1, sizeof(z_stream));
+	if(!gzStream)
+	{
+		ILFree(buffer);
+		ILFree(linked);
+		return 0;
+	}
+	if(inflateInit2(gzStream, -MAX_WBITS) != Z_OK)
+	{
+		ILFree(gzStream);
+		ILFree(buffer);
+		ILFree(linked);
+		return 0;
+	}
+
+	/* Set up the input context for gzip-based reads */
+	ctx->stream = 0;
+	ctx->buffer = (const char *)buffer;
+	ctx->bufLen = 0;
+	ctx->readFunc = GzipRead;
+	ctx->gzStream = gzStream;
+	ctx->linked = linked;
+	ctx->sawEOF = 0;
+
+	/* Skip past the gzip header */
+	method = GzipReadByte(ctx);
+	flags = GzipReadByte(ctx);
+	if(method != Z_DEFLATED || (flags & GZ_RESERVED) != 0)
+	{
+		GzipFinalize(ctx);
+		return 0;
+	}
+	GzipSkip(ctx, 6);		/* Time, extra flags, and OS code */
+	if((flags & GZ_EXTRA_FIELD) != 0)
+	{
+		len = GzipReadShort(ctx);
+		GzipSkip(ctx, len);
+	}
+	if((flags & GZ_ORIG_NAME) != 0)
+	{
+		GzipSkipString(ctx);
+	}
+	if((flags & GZ_COMMENT) != 0)
+	{
+		GzipSkipString(ctx);
+	}
+	if((flags & GZ_HEAD_CRC) != 0)
+	{
+		GzipSkip(ctx, 2);
+	}
+
+	/* The stream is ready to go */
+	return 1;
+}
+
+#endif /* IL_USE_GZIP */
+
 /*
  * Seek to a particular offset within a stream by reading
  * and discarding data.  This is designed to work on streams
  * that may not necessarily be file-based.  Returns zero if
  * the seek failed.
  */
-static int SeekWithinStream(FILE *file, char *buffer,
+static int SeekWithinStream(ILInputContext *ctx, char *buffer,
 							unsigned long currentOffset,
 							unsigned long destOffset)
 {
@@ -45,7 +329,7 @@ static int SeekWithinStream(FILE *file, char *buffer,
 		{
 			size = (unsigned)(destOffset - currentOffset);
 		}
-		if(fread(buffer, 1, size, file) != size)
+		if((*(ctx->readFunc))(ctx, buffer, size) != (int)size)
 		{
 			return 0;
 		}
@@ -112,14 +396,14 @@ void _ILImageFreeMemory(ILImage *image)
 	{
 		ILUnmapFileFromMemory(image->mapAddress, image->mapLength);
 	}
-	else if(image->data)
+	else if(image->data && !(image->inPlace))
 	{
 		ILFree(image->data);
 	}
 }
 
-int ILImageLoad(FILE *file, const char *filename,
-				ILContext *context, ILImage **image, int flags)
+static int ImageLoad(ILInputContext *ctx, const char *filename,
+					 ILContext *context, ILImage **image, int flags)
 {
 	char buffer[1024];
 	int isDLL;
@@ -142,11 +426,14 @@ int ILImageLoad(FILE *file, const char *filename,
 	unsigned long dataSize;
 	unsigned long tlsRVA;
 	unsigned long tlsSize;
+	unsigned long rsrcRVA;
+	unsigned long rsrcSize;
 	ILSectionMap *map;
 	ILSectionMap *newMap;
 	char *data;
 	unsigned char *runtimeHdr;
 	int isMapped;
+	int isInPlace;
 	void *mapAddress;
 	unsigned long mapLength;
 	int error;
@@ -154,14 +441,37 @@ int ILImageLoad(FILE *file, const char *filename,
 	/* Read the first 2 bytes and look for either an MS-DOS
 	   stub (executables and DLL's), or the beginning of a
 	   PE/COFF header (object files) */
-	if(fread(buffer, 1, 2, file) != 2)
+	if((*(ctx->readFunc))(ctx, buffer, 2) != 2)
 	{
 		return IL_LOADERR_TRUNCATED;
 	}
+#ifdef IL_USE_GZIP
+	if(buffer[0] == (char)0x1F && buffer[1] == (char)0x8B)
+	{
+		/* The image has been compressed with gzip, so layer a decompression
+		   object on top of the underlying input context */
+		if(!GzipInitialize(ctx))
+		{
+			return IL_LOADERR_NOT_PE;
+		}
+
+		/* Suppress the use of mmap to load the stream's contents */
+		flags |= IL_LOADFLAG_NO_MAP;
+
+		/* Disable in-place execution */
+		flags &= ~IL_LOADFLAG_IN_PLACE;
+
+		/* Re-read the magic number bytes from the uncompressed data */
+		if((*(ctx->readFunc))(ctx, buffer, 2) != 2)
+		{
+			return IL_LOADERR_TRUNCATED;
+		}
+	}
+#endif
 	if(buffer[0] == 'M' && buffer[1] == 'Z')
 	{
 		/* Read the MS-DOS stub and find the start of the PE header */
-		if(fread(buffer + 2, 1, 62, file) != 62)
+		if((*(ctx->readFunc))(ctx, buffer + 2, 62) != 62)
 		{
 			return IL_LOADERR_TRUNCATED;
 		}
@@ -171,12 +481,12 @@ int ILImageLoad(FILE *file, const char *filename,
 		{
 			return IL_LOADERR_BACKWARDS;
 		}
-		if(!SeekWithinStream(file, buffer, offset, base))
+		if(!SeekWithinStream(ctx, buffer, offset, base))
 		{
 			return IL_LOADERR_TRUNCATED;
 		}
 		offset = base;
-		if(fread(buffer, 1, 4, file) != 4)
+		if((*(ctx->readFunc))(ctx, buffer, 4) != 4)
 		{
 			return IL_LOADERR_TRUNCATED;
 		}
@@ -186,7 +496,7 @@ int ILImageLoad(FILE *file, const char *filename,
 		{
 			return IL_LOADERR_NOT_PE;
 		}
-		if(fread(buffer, 1, 20, file) != 20)
+		if((*(ctx->readFunc))(ctx, buffer, 20) != 20)
 		{
 			return IL_LOADERR_TRUNCATED;
 		}
@@ -195,7 +505,7 @@ int ILImageLoad(FILE *file, const char *filename,
 	else if(buffer[0] == (char)0x4C && buffer[1] == (char)0x01)
 	{
 		/* This is an i386 PE/COFF object file: read the rest of the header */
-		if(fread(buffer + 2, 1, 18, file) != 18)
+		if((*(ctx->readFunc))(ctx, buffer + 2, 18) != 18)
 		{
 			return IL_LOADERR_TRUNCATED;
 		}
@@ -208,13 +518,21 @@ int ILImageLoad(FILE *file, const char *filename,
 	{
 		/* This looks like a Java ".class" or ".jar" file, which
 		   we need to pass off to "_ILImageJavaLoad" to handle */
-		return _ILImageJavaLoad(file, filename, context, image, flags, buffer);
+		if(ctx->stream)
+		{
+			return _ILImageJavaLoad(ctx->stream, filename, context,
+									image, flags, buffer);
+		}
+		else
+		{
+			return IL_LOADERR_NOT_PE;
+		}
 	}
 #endif
 	else if(buffer[0] == '!' && buffer[1] == '<')
 	{
 		/* This may be an "ar" archive file: read the rest of the header */
-		if(fread(buffer, 1, 6, file) != 6)
+		if((*(ctx->readFunc))(ctx, buffer, 6) != 6)
 		{
 			return IL_LOADERR_NOT_PE;
 		}
@@ -248,7 +566,7 @@ int ILImageLoad(FILE *file, const char *filename,
 	   the data directory information for the IL runtime header */
 	if(headerSize != 0)
 	{
-		if(fread(buffer, 1, headerSize, file) != headerSize)
+		if((*(ctx->readFunc))(ctx, buffer, headerSize) != headerSize)
 		{
 			return IL_LOADERR_TRUNCATED;
 		}
@@ -288,9 +606,11 @@ int ILImageLoad(FILE *file, const char *filename,
 	dataSize = 0;
 	tlsRVA = 0;
 	tlsSize = 0;
+	rsrcRVA = 0;
+	rsrcSize = 0;
 	while(numSections > 0)
 	{
-		if(fread(buffer, 1, 40, file) != 40)
+		if((*(ctx->readFunc))(ctx, buffer, 40) != 40)
 		{
 			_ILFreeSectionMap(map);
 			return IL_LOADERR_TRUNCATED;
@@ -342,7 +662,12 @@ int ILImageLoad(FILE *file, const char *filename,
 			debugRVA = newMap->virtAddr;
 			debugSize = newMap->virtSize;
 		}
-		else if(!ILMemCmp(buffer, ".sdata\0\0", 8))
+		else if(!ILMemCmp(buffer, ".sdata\0\0", 8) && !dataRVA)
+		{
+			dataRVA = newMap->virtAddr;
+			dataSize = newMap->virtSize;
+		}
+		else if(!ILMemCmp(buffer, ".data\0\0\0", 8) && !dataRVA)
 		{
 			dataRVA = newMap->virtAddr;
 			dataSize = newMap->virtSize;
@@ -352,6 +677,11 @@ int ILImageLoad(FILE *file, const char *filename,
 			tlsRVA = newMap->virtAddr;
 			tlsSize = newMap->virtSize;
 		}
+		else if(!ILMemCmp(buffer, ".rsrc\0\0\0", 8))
+		{
+			rsrcRVA = newMap->virtAddr;
+			rsrcSize = newMap->virtSize;
+		}
 		if(newMap->realSize > 0)
 		{
 			if(newMap->realAddr < minAddress)
@@ -384,7 +714,7 @@ int ILImageLoad(FILE *file, const char *filename,
 	}
 
 	/* Seek to the beginning of the first section */
-	if(!SeekWithinStream(file, buffer, offset, minAddress))
+	if(!SeekWithinStream(ctx, buffer, offset, minAddress))
 	{
 		_ILFreeSectionMap(map);
 		return IL_LOADERR_TRUNCATED;
@@ -395,10 +725,24 @@ int ILImageLoad(FILE *file, const char *filename,
 	   there is too much variation in how compilers lay out binaries.
 	   In particular, the IL bytecode can be pretty much anywhere */
 	if((flags & IL_LOADFLAG_NO_MAP) == 0 &&
-	   ILMapFileToMemory(fileno(file), minAddress, maxAddress,
+	   ILMapFileToMemory(fileno(ctx->stream), minAddress, maxAddress,
 					     &mapAddress, &mapLength, &data))
 	{
 		isMapped = 1;
+		isInPlace = 0;
+	}
+#ifndef REDUCED_STDIO
+	else if(!(ctx->stream) && (flags & IL_LOADFLAG_IN_PLACE) != 0)
+#else
+	else if((flags & IL_LOADFLAG_IN_PLACE) != 0)
+#endif
+	{
+		/* Execute directly from the supplied buffer */
+		data = (char *)(ctx->buffer);
+		isMapped = 0;
+		mapAddress = 0;
+		mapLength = 0;
+		isInPlace = 1;
 	}
 	else
 	{
@@ -408,7 +752,7 @@ int ILImageLoad(FILE *file, const char *filename,
 			_ILFreeSectionMap(map);
 			return IL_LOADERR_MEMORY;
 		}
-		if(fread(data, 1, maxAddress - minAddress, file) !=
+		if((*(ctx->readFunc))(ctx, data, maxAddress - minAddress) !=
 					(maxAddress - minAddress))
 		{
 			ILFree(data);
@@ -418,6 +762,7 @@ int ILImageLoad(FILE *file, const char *filename,
 		isMapped = 0;
 		mapAddress = 0;
 		mapLength = 0;
+		isInPlace = 0;
 	}
 
 	/* Adjust the section map to account for the new location of the program */
@@ -505,7 +850,7 @@ int ILImageLoad(FILE *file, const char *filename,
 		{
 			ILUnmapFileFromMemory(mapAddress, mapLength);
 		}
-		else
+		else if(!isInPlace)
 		{
 			ILFree(data);
 		}
@@ -522,7 +867,7 @@ int ILImageLoad(FILE *file, const char *filename,
 			{
 				ILUnmapFileFromMemory(mapAddress, mapLength);
 			}
-			else
+			else if(!isInPlace)
 			{
 				ILFree(data);
 			}
@@ -530,12 +875,14 @@ int ILImageLoad(FILE *file, const char *filename,
 			return IL_LOADERR_MEMORY;
 		}
 	}
+	(*image)->loadFlags = flags;
 	(*image)->type = (isOBJ ? IL_IMAGETYPE_OBJ :
 						(isDLL ? IL_IMAGETYPE_DLL : IL_IMAGETYPE_EXE));
 	(*image)->secure = ((flags & IL_LOADFLAG_INSECURE) == 0);
 	(*image)->hadNative = hadNative;
 	(*image)->only32Bit = only32Bit;
 	(*image)->mapped = isMapped;
+	(*image)->inPlace = isInPlace;
 	(*image)->map = map;
 	(*image)->data = data;
 	(*image)->len = (maxAddress - minAddress);
@@ -550,6 +897,8 @@ int ILImageLoad(FILE *file, const char *filename,
 	(*image)->dataSize = dataSize;
 	(*image)->tlsRVA = tlsRVA;
 	(*image)->tlsSize = tlsSize;
+	(*image)->rsrcRVA = rsrcRVA;
+	(*image)->rsrcSize = rsrcSize;
 
 	/* Load the meta information from the image */
 	if((flags & IL_LOADFLAG_NO_METADATA) == 0)
@@ -565,6 +914,29 @@ int ILImageLoad(FILE *file, const char *filename,
 	return 0;
 }
 
+#ifndef REDUCED_STDIO
+
+int ILImageLoad(FILE *file, const char *filename,
+				ILContext *context, ILImage **image, int flags)
+{
+	ILInputContext ctx;
+	int error;
+	ctx.stream = file;
+	ctx.buffer = 0;
+	ctx.bufLen = 0;
+	ctx.readFunc = StdioRead;
+#ifdef IL_USE_GZIP
+	ctx.sawEOF = 0;
+	ctx.gzStream = 0;
+	ctx.linked = 0;
+#endif
+	error = ImageLoad(&ctx, filename, context, image, flags);
+#ifdef IL_USE_GZIP
+	GzipFinalize(&ctx);
+#endif
+	return error;
+}
+
 int ILImageLoadFromFile(const char *filename, ILContext *context,
 						ILImage **image, int flags, int printErrors)
 {
@@ -614,6 +986,33 @@ int ILImageLoadFromFile(const char *filename, ILContext *context,
 	return loadError;
 }
 
+#endif /* !REDUCED_STDIO */
+
+int ILImageLoadFromMemory(const void *buffer, unsigned long bufLen,
+						  ILContext *context, ILImage **image,
+						  int flags, const char *filename)
+{
+	ILInputContext ctx;
+	int error;
+#ifndef REDUCED_STDIO
+	ctx.stream = 0;
+#endif
+	ctx.buffer = (const char *)buffer;
+	ctx.bufLen = bufLen;
+	ctx.readFunc = MemoryRead;
+#ifdef IL_USE_GZIP
+	ctx.sawEOF = 0;
+	ctx.gzStream = 0;
+	ctx.linked = 0;
+#endif
+	error = ImageLoad(&ctx, filename, context,
+					  image, flags | IL_LOADFLAG_NO_MAP);
+#ifdef IL_USE_GZIP
+	GzipFinalize(&ctx);
+#endif
+	return error;
+}
+
 void *ILImageMapAddress(ILImage *image, unsigned long address)
 {
 	unsigned long realAddr = ConvertVirtAddrToReal(image->map, address);
@@ -964,6 +1363,14 @@ int _ILImageGetSection(ILImage *image, int section,
 		}
 		break;
 
+		case IL_SECTION_WINRES:
+		{
+			/* Windows resource section */
+			*address = image->rsrcRVA;
+			*size = image->rsrcSize;
+		}
+		break;
+
 		default: return 0;
 	}
 	return (*address != 0 && *size != 0);
