@@ -19,12 +19,43 @@
  */
 
 #include "program.h"
+
+#ifdef IL_USE_WRITER
+
 #include <time.h>
 
 #ifdef	__cplusplus
 extern	"C" {
 #endif
 
+/*
+ * Flush the back-patch cache if necessary.
+ */
+static void FlushBackPatchCache(ILWriter *writer)
+{
+	if(writer->backpatching)
+	{
+		/* Seek to the beginning of the back-patch area */
+		if(fseek(writer->stream, writer->backpatchSeek, 0) < 0)
+		{
+			writer->writeFailed = 1;
+			return;
+		}
+
+		/* Write the modified contents of the back-patch buffer */
+		if(fwrite(writer->backpatchBuf, 1, (unsigned)(writer->backpatchLen),
+				  writer->stream) != (unsigned)(writer->backpatchLen))
+		{
+			writer->writeFailed = 1;
+			return;
+		}
+
+		/* Record the new seek position and exit the back-patching mode */
+		writer->currSeek = writer->backpatchSeek + writer->backpatchLen;
+		writer->backpatching = 0;
+	}
+}
+
 /*
  * Write a block of data to an output image.
  */
@@ -37,6 +68,9 @@ static void WriteToImage(ILWriter *writer, const void *buffer, unsigned size)
 	}
 	if(writer->seekable)
 	{
+		/* Flush the back-patching cache, if present */
+		FlushBackPatchCache(writer);
+
 		/* This is a seekable stream, so write directly to it */
 		if(writer->offset != writer->currSeek)
 		{
@@ -109,6 +143,11 @@ static void WriteFlush(ILWriter *writer)
 			wbuffer = wbuffer->next;
 		}
 	}
+	else
+	{
+		/* Flush the back-patching cache */
+		FlushBackPatchCache(writer);
+	}
 }
 
 /*
@@ -140,7 +179,56 @@ static void WriteBackPatchBytes(ILWriter *writer, unsigned long posn,
 	}
 	if(writer->seekable)
 	{
+		/* Can we add the bytes to the current back-patch cache? */
+		if(writer->backpatching && posn >= writer->backpatchSeek &&
+		   (posn + size) <= (writer->backpatchSeek + writer->backpatchLen))
+		{
+			ILMemCpy(writer->backpatchBuf +
+					 (unsigned)(posn - writer->backpatchSeek), buffer, size);
+			return;
+		}
+
+		/* Flush the current back-patch cache contents */
+		FlushBackPatchCache(writer);
+
+		/* Only do caching for small values (usually 32-bit quantities) */
+		if(size <= 4 && writer->backpatchBuf)
+		{
+			/* Align the back-patch cache for better filesystem performance */
+			writer->backpatchSeek = posn & ~511;
+			writer->backpatchLen = writer->offset - writer->backpatchSeek;
+			if(writer->backpatchLen > IL_WRITE_BUFFER_SIZE)
+			{
+				writer->backpatchLen = IL_WRITE_BUFFER_SIZE;
+			}
+
+			/* Read the block to be patched into memory */
+			if(fseek(writer->stream, writer->backpatchSeek, 0) < 0)
+			{
+				writer->writeFailed = 1;
+				return;
+			}
+			if(fread(writer->backpatchBuf, 1, (unsigned)(writer->backpatchLen),
+					 writer->stream) != (unsigned)(writer->backpatchLen))
+			{
+				/* The stream may not have been opened in a read-write mode,
+				   so fall back to the simple "seek and write" method */
+				writer->currSeek = ~((unsigned long)0);
+				ILFree(writer->backpatchBuf);
+				writer->backpatchBuf = 0;
+				goto fallback;
+			}
+			writer->currSeek = writer->backpatchSeek + writer->backpatchLen;
+
+			/* Patch the required bytes and return */
+			ILMemCpy(writer->backpatchBuf +
+					 (unsigned)(posn - writer->backpatchSeek), buffer, size);
+			writer->backpatching = 1;
+			return;
+		}
+
 		/* Seek back in the file and write the change */
+	fallback:
 		if(posn != writer->currSeek)
 		{
 			if(fseek(writer->stream, posn, 0) < 0)
@@ -540,6 +628,8 @@ void _ILWriteFinal(ILWriter *writer)
 	unsigned long imageSize;
 	unsigned long offset;
 	ILWSection *section;
+	ILWSection *prevSection;
+	ILWSection *debugSection;
 	int numSections;
 	unsigned long table;
 	unsigned long textLength;
@@ -692,6 +782,44 @@ void _ILWriteFinal(ILWriter *writer)
 		realTextSize = ((textLength + 511) & ~511);
 	}
 
+	/* If we have ".ildebug" data, then move it to the end of the list */
+	section = writer->sections;
+	prevSection = 0;
+	debugSection = 0;
+	while(section != 0)
+	{
+		if(!strcmp(section->name, ".ildebug"))
+		{
+			debugSection = section;
+			if(prevSection)
+			{
+				prevSection->next = section->next;
+			}
+			else
+			{
+				writer->sections = section->next;
+			}
+			section = section->next;
+		}
+		else
+		{
+			prevSection = section;
+			section = section->next;
+		}
+	}
+	if(debugSection)
+	{
+		if(prevSection)
+		{
+			prevSection->next = debugSection;
+		}
+		else
+		{
+			writer->sections = debugSection;
+		}
+		debugSection->next = 0;
+	}
+
 	/* Dump the contents of other sections */
 	section = writer->sections;
 	while(section != 0)
@@ -943,6 +1071,75 @@ void _ILWriteFieldRVAFixups(ILWriter *writer, unsigned long dataSection,
 	}
 }
 
+void ILWriterSetStream(ILWriter *writer, FILE *stream, int seekable)
+{
+	if (!(writer->stream) && stream)
+	{
+		writer->stream = stream;
+		if (seekable)
+		{
+			WriteFlush(writer);
+		}
+		writer->seekable = seekable;
+	}
+}
+
+int ILWriterResetTypeAndFlags(ILWriter *writer, int type, int flags)
+{
+	if (type != IL_IMAGETYPE_DLL &&
+	    type != IL_IMAGETYPE_EXE)
+	{
+		return 0;
+	}
+	if (writer->type != IL_IMAGETYPE_DLL &&
+	    writer->type != IL_IMAGETYPE_EXE)
+	{
+		return 0;
+	}
+	if ((flags & IL_WRITEFLAG_JVM_MODE) != 0 ||
+	    (writer->flags & IL_WRITEFLAG_JVM_MODE) != 0)
+	{
+		return 0;
+	}
+	if (writer->type != type)
+	{
+		writer->type = type;
+		if(type == IL_IMAGETYPE_DLL)
+		{
+			WriteBackPatch16(writer, (writer->peOffset)+18, 0x210E);
+			WriteBackPatch32(writer, (writer->optOffset)+28, 0x10000000);
+		}
+		else
+		{
+			WriteBackPatch16(writer, (writer->peOffset)+18, 0x010E);
+			WriteBackPatch32(writer, (writer->optOffset)+28, 0x00400000);
+		}
+	}
+	if (writer->flags != flags)
+	{
+		writer->flags = flags;
+		if((flags & IL_WRITEFLAG_SUBSYS_GUI) != 0)
+		{
+			WriteBackPatch16(writer, (writer->optOffset)+68, 2);
+		}
+		else
+		{
+			WriteBackPatch16(writer, (writer->optOffset)+68, 3);
+		}
+		if((flags & IL_WRITEFLAG_32BIT_ONLY) != 0)
+		{
+			WriteBackPatch32(writer, (writer->runtimeOffset)+16, 3);
+		}
+		else
+		{
+			WriteBackPatch32(writer, (writer->runtimeOffset)+16, 1);
+		}
+	}
+	return 1;
+}
+
 #ifdef	__cplusplus
 };
 #endif
+
+#endif /* IL_USE_WRITER */
