@@ -81,6 +81,19 @@ struct _tagILCoder
 
 };
 
+/*
+ * Argument information for a method call.
+ */
+typedef struct
+{
+	ILEngineStackItem *args;
+	ILUInt32 numBaseArgs;
+	ILUInt32 numVarArgs;
+	int hasParamArray;
+	int tailCall;
+
+} ILCoderMethodInfo;
+
 /*
  * Inlineable method calls.
  */
@@ -100,6 +113,8 @@ struct _tagILCoder
 #define	IL_INLINEMETHOD_SET2D_INT			13
 #define	IL_INLINEMETHOD_SET2D_OBJECT		14
 #define	IL_INLINEMETHOD_SET2D_DOUBLE		15
+#define	IL_INLINEMETHOD_BUILDER_APPEND_CHAR	16
+#define	IL_INLINEMETHOD_IS_WHITE_SPACE		17
 
 /*
  * Return values for "ILCoderFinish".
@@ -108,6 +123,12 @@ struct _tagILCoder
 #define	IL_CODER_END_RESTART	1		/* Restart is required */
 #define	IL_CODER_END_TOO_BIG	2		/* Method is too big for the cache */
 
+/* 
+ * Profiling flags
+ */
+
+#define IL_CODER_FLAG_IR_DUMP 1
+#define IL_CODER_FLAG_METHOD_PROFILE 2
 /*
  * Coder class definition.
  */
@@ -116,7 +137,7 @@ struct _tagILCoderClass
 	/*
 	 * Create a coder instance.  Returns NULL if not possible.
 	 */
-	ILCoder *(*create)(ILUInt32 size);
+	ILCoder *(*create)(ILUInt32 size,unsigned long cachePageSize);
 	
 	/*
 	 * Enable debug mode.  The coder will output extra information
@@ -521,47 +542,43 @@ struct _tagILCoderClass
 	/*
 	 * Check a method call's first parameter for null.
 	 */
-	void (*checkCallNull)(ILCoder *coder, ILEngineStackItem *args,
-					      ILUInt32 numArgs);
+	void (*checkCallNull)(ILCoder *coder, ILCoderMethodInfo *info);
 
 	/*
 	 * Call a method directly.
 	 */
-	void (*callMethod)(ILCoder *coder, ILEngineStackItem *args,
-					   ILUInt32 numArgs, ILEngineStackItem *returnItem,
-					   ILMethod *methodInfo);
+	void (*callMethod)(ILCoder *coder, ILCoderMethodInfo *info,
+					   ILEngineStackItem *returnItem, ILMethod *methodInfo);
 
 	/*
 	 * Call a method using an indirect pointer.
 	 */
-	void (*callIndirect)(ILCoder *coder, ILEngineStackItem *args,
-					     ILUInt32 numArgs, ILEngineStackItem *returnItem);
+	void (*callIndirect)(ILCoder *coder, ILCoderMethodInfo *info,
+					     ILEngineStackItem *returnItem);
 
 	/*
 	 * Call a constructor method directly.
 	 */
-	void (*callCtor)(ILCoder *coder, ILEngineStackItem *args,
-					 ILUInt32 numArgs, ILMethod *methodInfo);
+	void (*callCtor)(ILCoder *coder, ILCoderMethodInfo *info,
+					 ILMethod *methodInfo);
 
 	/*
 	 * Call a virtual method.
 	 */
-	void (*callVirtual)(ILCoder *coder, ILEngineStackItem *args,
-					    ILUInt32 numArgs, ILEngineStackItem *returnItem,
-						ILMethod *methodInfo);
+	void (*callVirtual)(ILCoder *coder, ILCoderMethodInfo *info,
+					    ILEngineStackItem *returnItem, ILMethod *methodInfo);
 
 	/*
 	 * Call an interface method.
 	 */
-	void (*callInterface)(ILCoder *coder, ILEngineStackItem *args,
-					      ILUInt32 numArgs, ILEngineStackItem *returnItem,
-						  ILMethod *methodInfo);
+	void (*callInterface)(ILCoder *coder, ILCoderMethodInfo *info,
+					      ILEngineStackItem *returnItem, ILMethod *methodInfo);
 
 	/*
 	 * Call an inlineable method.  Returns zero if the coder
 	 * cannot inline the method.
 	 */
-	int (*callInlineable)(ILCoder *coder, int inlineType);
+	int (*callInlineable)(ILCoder *coder, int inlineType, ILMethod *methodInfo);
 
 	/*
 	 * Jump to a method with the same signature as the current method.
@@ -590,12 +607,6 @@ struct _tagILCoderClass
 	 */
 	void (*loadInterfaceAddr)(ILCoder *coder, ILMethod *methodInfo);
 
-	/*
-	 * Mark the following call, calli, or callvirt as being a
-	 * tail call -- pop current frame before calling.
-	 */
-	void (*tailCall)(ILCoder *coder, ILMethod *methodInfo);
-
 	/*
 	 * Set up exception handling for the current method.
 	 */
@@ -610,6 +621,12 @@ struct _tagILCoderClass
 	 */
 	void (*throwException)(ILCoder *coder, int inCurrentMethod);
 
+
+	/*
+	 * Set a stack trace. This is often used in conjunction with 
+	 * throwException to mark exceptions with stack traces */
+	void (*setStackTrace)(ILCoder *coder);
+
 	/*
 	 * Re-throw the current exception for a particular exception region.
 	 */
@@ -686,6 +703,68 @@ struct _tagILCoderClass
 	 */
 	void (*markEnd)(ILCoder *coder);
 
+	/*
+	 * Set the different flags which enable or disable
+	 * profiling, debugging etc.
+	 */
+	void (*setFlags)(ILCoder *coder, int flags);
+
+	/*
+	 * Allocate an extra local variable in the current method frame.
+	 * Returns the local variable index.
+	 */
+	ILUInt32 (*allocExtraLocal)(ILCoder *coder, ILType *type);
+
+	/*
+	 * Push a thread value onto the stack for an internalcall.
+	 */
+	void (*pushThread)(ILCoder *coder, int useRawCalls);
+
+	/*
+	 * Load the address of an argument onto the native argument stack.
+	 */
+	void (*loadNativeArgAddr)(ILCoder *coder, ILUInt32 num);
+
+	/*
+	 * Load the address of a local onto the native argument stack.
+	 */
+	void (*loadNativeLocalAddr)(ILCoder *coder, ILUInt32 num);
+
+	/*
+	 * Start pushing arguments for a "libffi" call onto the stack.
+	 */
+	void (*startFfiArgs)(ILCoder *coder);
+
+	/*
+	 * Push the address of the raw argument block onto the stack.
+	 */
+	void (*pushRawArgPointer)(ILCoder *coder);
+
+	/*
+	 * Perform a function call using "libffi".
+	 */
+	void (*callFfi)(ILCoder *coder, void *fn, void *cif,
+					int useRawCalls, int hasReturn);
+
+	/*
+	 * Check the top of stack value for NULL.
+	 */
+	void (*checkNull)(ILCoder *coder);
+
+	/*
+	 * Output an instruction to convert the top of stack according
+	 * to a PInvoke marshalling rule.
+	 */
+	void (*convert)(ILCoder *coder, int opcode);
+
+	/*
+	 * Output an instruction to convert the top of stack according
+	 * to a custom marshalling rule.
+	 */
+	void (*convertCustom)(ILCoder *coder, int opcode,
+						  ILUInt32 customNameLen, ILUInt32 customCookieLen,
+						  void *customName, void *customCookie);
+
 	/*
 	 * Sentinel string to catch missing methods in class tables.
 	 */
@@ -696,8 +775,8 @@ struct _tagILCoderClass
 /*
  * Helper macros for calling coder methods.
  */
-#define	ILCoderCreate(classInfo,size)	\
-			((*((classInfo)->create))((size)))
+#define	ILCoderCreate(classInfo,size,cachePageSize)	\
+			((*((classInfo)->create))((size),(cachePageSize)))
 #define	ILCoderEnableDebug(coder)	\
 			((*((coder)->classInfo->enableDebug))((coder)))
 #define	ILCoderAlloc(coder,size)	\
@@ -863,29 +942,29 @@ struct _tagILCoderClass
 #define	ILCoderValueCtorArgs(coder,_classInfo,args,numArgs) \
 			((*((coder)->classInfo->valueCtorArgs))((coder), (_classInfo), \
 												    (args), (numArgs)))
-#define	ILCoderCheckCallNull(coder,args,numArgs) \
-			((*((coder)->classInfo->checkCallNull))((coder), (args), \
-												    (numArgs)))
-#define	ILCoderCallMethod(coder,args,numArgs,returnItem,methodInfo) \
-			((*((coder)->classInfo->callMethod))((coder), (args), \
-												 (numArgs), (returnItem), \
+#define	ILCoderCheckCallNull(coder,info) \
+			((*((coder)->classInfo->checkCallNull))((coder), (info)))
+#define	ILCoderCallMethod(coder,info,returnItem,methodInfo) \
+			((*((coder)->classInfo->callMethod))((coder), (info), \
+												 (returnItem), \
 												 (methodInfo)))
-#define	ILCoderCallIndirect(coder,args,numArgs,returnItem) \
-			((*((coder)->classInfo->callIndirect))((coder), (args), \
-												   (numArgs), (returnItem)))
-#define	ILCoderCallCtor(coder,args,numArgs,methodInfo) \
-			((*((coder)->classInfo->callCtor))((coder), (args), \
-											   (numArgs), (methodInfo)))
-#define	ILCoderCallVirtual(coder,args,numArgs,returnItem,methodInfo) \
-			((*((coder)->classInfo->callVirtual))((coder), (args), \
-												  (numArgs), (returnItem), \
+#define	ILCoderCallIndirect(coder,info,returnItem) \
+			((*((coder)->classInfo->callIndirect))((coder), (info), \
+												   (returnItem)))
+#define	ILCoderCallCtor(coder,info,methodInfo) \
+			((*((coder)->classInfo->callCtor))((coder), (info), \
+											   (methodInfo)))
+#define	ILCoderCallVirtual(coder,info,returnItem,methodInfo) \
+			((*((coder)->classInfo->callVirtual))((coder), (info), \
+												  (returnItem), \
 												  (methodInfo)))
-#define	ILCoderCallInterface(coder,args,numArgs,returnItem,methodInfo) \
-			((*((coder)->classInfo->callInterface))((coder), (args), \
-												    (numArgs), (returnItem), \
+#define	ILCoderCallInterface(coder,info,returnItem,methodInfo) \
+			((*((coder)->classInfo->callInterface))((coder), (info), \
+												    (returnItem), \
 													(methodInfo)))
-#define	ILCoderCallInlineable(coder,inlineType) \
-			((*((coder)->classInfo->callInlineable))((coder), (inlineType)))
+#define	ILCoderCallInlineable(coder,inlineType,methodInfo) \
+			((*((coder)->classInfo->callInlineable))((coder), (inlineType), \
+													 (methodInfo)))
 #define	ILCoderJumpMethod(coder,methodInfo) \
 			((*((coder)->classInfo->jumpMethod))((coder), (methodInfo)))
 #define	ILCoderReturnInsn(coder,engineType,returnType) \
@@ -897,13 +976,13 @@ struct _tagILCoderClass
 			((*((coder)->classInfo->loadVirtualAddr))((coder), (methodInfo)))
 #define	ILCoderLoadInterfaceAddr(coder,methodInfo) \
 			((*((coder)->classInfo->loadInterfaceAddr))((coder), (methodInfo)))
-#define ILCoderTailCall(coder,methodInfo) \
-			((*((coder)->classInfo->tailCall))((coder), (methodInfo)))
 #define	ILCoderSetupExceptions(coder,exceptions,hasRethrow) \
 			((*((coder)->classInfo->setupExceptions))((coder), (exceptions), \
 													  (hasRethrow)))
 #define	ILCoderThrow(coder,inCurrent) \
 			((*((coder)->classInfo->throwException))((coder), (inCurrent)))
+#define	ILCoderSetStackTrace(coder) \
+			((*((coder)->classInfo->setStackTrace))((coder)))
 #define	ILCoderRethrow(coder,exception) \
 			((*((coder)->classInfo->rethrow))((coder), (exception)))
 #define	ILCoderJsr(coder,dest) \
@@ -929,6 +1008,33 @@ struct _tagILCoderClass
 												      (offset), (exact)))
 #define	ILCoderMarkBytecode(coder,offset) \
 			((*((coder)->classInfo->markBytecode))((coder), (offset)))
+#define ILCoderSetFlags(coder,flags) \
+			((*((coder)->classInfo->setFlags))((coder), (flags)))
+#define	ILCoderAllocExtraLocal(coder,type) \
+			((*((coder)->classInfo->allocExtraLocal))((coder), (type)))
+#define	ILCoderPushThread(coder,useRawCalls) \
+			((*((coder)->classInfo->pushThread))((coder), (useRawCalls)))
+#define	ILCoderLoadNativeArgAddr(coder,num) \
+			((*((coder)->classInfo->loadNativeArgAddr))((coder), (num)))
+#define	ILCoderLoadNativeLocalAddr(coder,num) \
+			((*((coder)->classInfo->loadNativeLocalAddr))((coder), (num)))
+#define	ILCoderStartFfiArgs(coder) \
+			((*((coder)->classInfo->startFfiArgs))((coder)))
+#define	ILCoderPushRawArgPointer(coder) \
+			((*((coder)->classInfo->pushRawArgPointer))((coder)))
+#define	ILCoderCallFfi(coder,fn,cif,useRawCalls,hasReturn) \
+			((*((coder)->classInfo->callFfi))((coder), (fn), (cif), \
+											  (useRawCalls), (hasReturn)))
+#define	ILCoderCheckNull(coder) \
+			((*((coder)->classInfo->checkNull))((coder)))
+#define	ILCoderConvert(coder,opcode) \
+			((*((coder)->classInfo->convert))((coder), (opcode)))
+#define	ILCoderConvertCustom(coder,opcode,customNameLen,customCookieLen,customName,customCookie) \
+			((*((coder)->classInfo->convertCustom))((coder), (opcode), \
+													(customNameLen), \
+													(customCookieLen), \
+													(customName), \
+													(customCookie)))
 #define	ILCoderMarkEnd(coder) \
 			((*((coder)->classInfo->markEnd))((coder)))
 
