@@ -61,6 +61,8 @@ typedef struct _tagILDeclSecurity	ILDeclSecurity;
 typedef struct _tagILFileDecl		ILFileDecl;
 typedef struct _tagILManifestRes	ILManifestRes;
 typedef struct _tagILExportedType	ILExportedType;
+typedef struct _tagILGenericPar		ILGenericPar;
+typedef struct _tagILMethodSpec		ILMethodSpec;
 
 /*
  * Iterate over the list of attributes that are associated
@@ -144,6 +146,8 @@ ILDeclSecurity *ILProgramItemToDeclSecurity(ILProgramItem *item);
 ILFileDecl *ILProgramItemToFileDecl(ILProgramItem *item);
 ILManifestRes *ILProgramItemToManifestRes(ILProgramItem *item);
 ILExportedType *ILProgramItemToExportedType(ILProgramItem *item);
+ILGenericPar *ILProgramItemToGenericPar(ILProgramItem *item);
+ILMethodSpec *ILProgramItemToMethodSpec(ILProgramItem *item);
 
 /*
  * Helper macros for querying information about a program item.
@@ -513,6 +517,12 @@ int ILClassIsRef(ILClass *info);
  */
 ILClass *ILClassGetParent(ILClass *info);
 
+/*
+ * Set the parent of a class, if the class hasn't been
+ * marked completed.
+ */
+void ILClassSetParent(ILClass *info, ILClass *parent);
+
 /*
  * Get the parent of a particular class, but don't cross
  * image boundaries.  A reference will be returned if
@@ -696,6 +706,15 @@ ILMember *ILClassNextMember(ILClass *info, ILMember *last);
  */
 ILMember *ILClassNextMemberByKind(ILClass *info, ILMember *last, int kind);
 
+/*
+ * Iterate over the members of a class that match the given criteria.
+ * If "kind" is zero, then match all kinds.  If "name" is NULL, then
+ * match all names.  If "signature" is NULL, then match all signatures.
+ * The conditions are AND'ed together.
+ */
+ILMember *ILClassNextMemberMatch(ILClass *info, ILMember *last, int kind,
+								 const char *name, ILType *signature);
+
 /*
  * Mark a class definition as "complete".  This is typically
  * used by compilers and assemblers to prevent further members
@@ -766,6 +785,15 @@ void ILClassSetUserData(ILClass *info, void *data);
  */
 ILMethod *ILClassGetMethodImpl(ILClass *info, ILMethod *method);
 
+/*
+ * Get the implementation of a particular interface method
+ * where a regular implementing method is not available, but
+ * we can define a proxy to call through to a non-virtual method
+ * with the same name and signature.  Returns NULL if no such
+ * proxy implementation was found.
+ */
+ILMethod *ILClassGetMethodImplForProxy(ILClass *info, ILMethod *method);
+
 /*
  * Detach a member from its owning class.
  */
@@ -776,6 +804,25 @@ void ILClassDetachMember(ILMember *member);
  */
 void ILClassAttachMember(ILClass *info, ILMember *member);
 
+/*
+ * Instantiate a class type that involves generic parameters.
+ * Returns a synthetic class for the instantiation, or NULL
+ * if insufficient memory to perform the instantiation.
+ */
+ILClass *ILClassInstantiate(ILImage *image, ILType *classType,
+							ILType *classParams);
+
+/*
+ * Determine if a namespace is valid for a context.  A namespace
+ * is valid if there is at least one class with that namespace.
+ */
+int ILClassNamespaceIsValid(ILContext *context, const char *nspace);
+
+/*
+ * Get the underlying class from a generic class reference.
+ */
+ILClass *ILClassGetUnderlying(ILClass *info);
+
 /*
  * Helper macros for querying information about a class.
  */
@@ -935,6 +982,11 @@ int ILMemberAccessible(ILMember *member, ILClass *scope);
  */
 ILMember *ILMemberImport(ILImage *image, ILMember *member);
 
+/*
+ * Get the base "virtual" definition for an "override" member.
+ */
+ILMember *ILMemberGetBase(ILMember *member);
+
 /*
  * Helper macros for querying information about members.
  */
@@ -1019,9 +1071,9 @@ typedef struct
 	void		   *code;				/* Beginning of the code */
 	ILUInt32		codeLen;			/* Length of the code */
 	ILStandAloneSig	*localVarSig;		/* Local variable signature */
-	int				initLocals : 1;		/* Non-zero to initialize locals */
-	int				moreSections : 1;	/* Non-zero if more sections */
-	int				javaLocals : 30;	/* Maximum number of Java locals */
+	ILInt32			initLocals : 1;		/* Non-zero to initialize locals */
+	ILInt32			moreSections : 1;	/* Non-zero if more sections */
+	ILInt32			javaLocals : 30;	/* Maximum number of Java locals */
 	ILUInt32		remaining;			/* Remaining space in code section */
 
 } ILMethodCode;
@@ -1276,6 +1328,11 @@ ILType *ILFieldGetType(ILField *field);
  */
 ILType *ILFieldGetTypeWithPrefixes(ILField *field);
 
+/*
+ * Determine if a field is marked as thread-static.
+ */
+int ILFieldIsThreadStatic(ILField *field);
+
 /*
  * Helper macros for querying information about a field.
  */
@@ -1435,11 +1492,23 @@ ILPInvoke *ILPInvokeCreate(ILMethod *method, ILToken token,
 						   ILUInt32 attributes, ILModule *module,
 						   const char *aliasName);
 
+/*
+ * Create a new PInvoke member and attach it to a field.
+ */
+ILPInvoke *ILPInvokeFieldCreate(ILField *field, ILToken token,
+						        ILUInt32 attributes, ILModule *module,
+						        const char *aliasName);
+
 /*
  * Get the method associated with a PInvoke member.
  */
 ILMethod *ILPInvokeGetMethod(ILPInvoke *pinvoke);
 
+/*
+ * Get the field associated with a PInvoke member.
+ */
+ILField *ILPInvokeGetField(ILPInvoke *pinvoke);
+
 /*
  * Get the module associated with a PInvoke member.
  */
@@ -1456,12 +1525,20 @@ const char *ILPInvokeGetAlias(ILPInvoke *pinvoke);
  */
 ILPInvoke *ILPInvokeFind(ILMethod *method);
 
+/*
+ * Find the PInvoke member for a particular field.
+ * Returns NULL if no PInvoke member.
+ */
+ILPInvoke *ILPInvokeFindField(ILField *field);
+
 /*
  * Get the marshal conversion type for a method parameter.
  * If "param" is 0, then report about the return type.
  */
 ILUInt32 ILPInvokeGetMarshalType(ILPInvoke *pinvoke, ILMethod *method,
-								 unsigned long param);
+								 unsigned long param, char **customName,
+								 int *customNameLen, char **customCookie,
+								 int *customCookieLen, ILType *type);
 
 /*
  * Resolve the module information for a PInvoke declaration
@@ -1479,6 +1556,7 @@ char *ILPInvokeResolveModule(ILPInvoke *pinvoke);
 #define	ILPInvoke_Token(pinvoke)		(ILProgramItem_Token((pinvoke)))
 #define	ILPInvoke_Owner(pinvoke)		(ILMember_Owner((pinvoke)))
 #define	ILPInvoke_Method(pinvoke)		(ILPInvokeGetMethod((pinvoke)))
+#define	ILPInvoke_Field(pinvoke)		(ILPInvokeGetField((pinvoke)))
 #define	ILPInvoke_Module(pinvoke)		(ILPInvokeGetModule((pinvoke)))
 #define	ILPInvoke_Alias(pinvoke)		(ILPInvokeGetAlias((pinvoke)))
 #define	ILPInvoke_Attrs(pinvoke)		(ILMember_Attrs((pinvoke)))
@@ -1842,6 +1920,11 @@ ILField *ILFieldRVAGetOwner(ILFieldRVA *rva);
  */
 ILUInt32 ILFieldRVAGetRVA(ILFieldRVA *rva);
 
+/*
+ * Set the RVA associated with a field RVA record.
+ */
+void ILFieldRVASetRVA(ILFieldRVA *rva, ILUInt32 value);
+
 /*
  * Get the field RVA that is associated with a field.
  * Returns NULL if no such field RVA.
@@ -1873,6 +1956,11 @@ ILField *ILFieldLayoutGetOwner(ILFieldLayout *layout);
  */
 ILUInt32 ILFieldLayoutGetOffset(ILFieldLayout *layout);
 
+/*
+ * Set the offset associated with a field layout record.
+ */
+void ILFieldLayoutSetOffset(ILFieldLayout *layout, ILUInt32 offset);
+
 /*
  * Get the field layout that is associated with a field.
  * Returns NULL if no such field layout.
@@ -2137,11 +2225,6 @@ ILExportedType *ILExportedTypeCreate(ILImage *image, ILToken token,
 									 const char *name,
 									 const char *nspace);
 
-/*
- * Get the attributes associated with an exported type declaration.
- */
-ILUInt32 ILExportedTypeGetAttrs(ILExportedType *type);
-
 /*
  * Set the foreign type ID associated with an exported type declaration.
  */
@@ -2152,12 +2235,6 @@ void ILExportedTypeSetId(ILExportedType *type, ILUInt32 identifier);
  */
 ILUInt32 ILExportedTypeGetId(ILExportedType *type);
 
-/*
- * Get the name and namespace associated with an exported type declaration.
- */
-const char *ILExportedTypeGetName(ILExportedType *type);
-const char *ILExportedTypeGetNamespace(ILExportedType *type);
-
 /*
  * Set the scope in which an exported type can be found to a file.
  */
@@ -2174,12 +2251,6 @@ void ILExportedTypeSetScopeAssembly(ILExportedType *type, ILAssembly *assem);
  */
 void ILExportedTypeSetScopeType(ILExportedType *type, ILExportedType *scope);
 
-/*
- * Get the scope in which an exported type can be found.
- * Normally a File, ExportedType, or AssemblyRef.
- */
-ILProgramItem *ILExportedTypeGetScope(ILExportedType *type);
-
 /*
  * Find an exported type from its name and namespace.
  * Returns NULL if not found.
@@ -2194,11 +2265,117 @@ ILExportedType *ILExportedTypeFind(ILImage *image,
 #define	ILExportedType_FromToken(image,token)	\
 				((ILExportedType *)ILImageTokenInfo((image), (token)))
 #define	ILExportedType_Token(type)		(ILProgramItem_Token((type)))
-#define	ILExportedType_Attrs(type)		(ILExportedTypeGetAttrs((type)))
+#define	ILExportedType_Attrs(type)		(ILClassGetAttrs((ILClass *)(type)))
 #define	ILExportedType_Id(type)			(ILExportedTypeGetId((type)))
-#define	ILExportedType_Name(type)		(ILExportedTypeGetName((type)))
-#define	ILExportedType_Namespace(type)	(ILExportedTypeGetNamespace((type)))
-#define	ILExportedType_Scope(type)		(ILExportedTypeGetScope((type)))
+#define	ILExportedType_Name(type)		(ILClassGetName((ILClass *)(type)))
+#define	ILExportedType_Namespace(type)	(ILClassGetNamespace((ILClass *)(type)))
+#define	ILExportedType_Scope(type)		(ILClassGetScope((ILClass *)(type)))
+
+/*
+ * Create a generic parameter record.
+ */
+ILGenericPar *ILGenericParCreate(ILImage *image, ILToken token,
+								 ILProgramItem *owner, ILUInt32 number);
+
+/*
+ * Get the number associated with a generic parameter.
+ */
+ILUInt32 ILGenericParGetNumber(ILGenericPar *genPar);
+
+/*
+ * Get the flags associated with a generic parameter.
+ */
+ILUInt32 ILGenericParGetFlags(ILGenericPar *genPar);
+
+/*
+ * Set the flags associated with a generic parameter.
+ */
+void ILGenericParSetFlags(ILGenericPar *genPar, ILUInt32 mask, ILUInt32 value);
+
+/*
+ * Get the owner associated with a generic parameter.
+ */
+ILProgramItem *ILGenericParGetOwner(ILGenericPar *genPar);
+
+/*
+ * Get the name associated with a generic parameter.
+ */
+const char *ILGenericParGetName(ILGenericPar *genPar);
+
+/*
+ * Set the name associated with a generic parameter.  Returns
+ * zero if out of memory.
+ */
+int ILGenericParSetName(ILGenericPar *genPar, const char *name);
+
+/*
+ * Get the kind associated with a generic parameter.
+ */
+ILProgramItem *ILGenericParGetKind(ILGenericPar *genPar);
+
+/*
+ * Set the kind associated with a generic parameter.
+ */
+void ILGenericParSetKind(ILGenericPar *genPar, ILProgramItem *type);
+
+/*
+ * Get the constraint associated with a generic parameter.
+ */
+ILProgramItem *ILGenericParGetConstraint(ILGenericPar *genPar);
+
+/*
+ * Set the constraint associated with a generic parameter.
+ */
+void ILGenericParSetConstraint(ILGenericPar *genPar, ILProgramItem *type);
+
+/*
+ * Get a generic parameter record for a particular owner and number.
+ * Returns NULL if the record could not be found.
+ */
+ILGenericPar *ILGenericParGetFromOwner(ILProgramItem *owner, ILUInt32 number);
+
+/*
+ * Get the number of generic parameters that are owned by an item.
+ */
+ILUInt32 ILGenericParGetNumParams(ILProgramItem *owner);
+
+/*
+ * Helper macros for querying information about generic parameters.
+ */
+#define	ILGenericPar_FromToken(image,token)	\
+				((ILGenericPar *)ILImageTokenInfo((image), (token)))
+#define	ILGenericPar_Token(genPar)		(ILProgramItem_Token((genPar)))
+#define	ILGenericPar_Number(genPar)		(ILGenericParGetNumber((genPar)))
+#define	ILGenericPar_Flags(genPar)		(ILGenericParGetFlags((genPar)))
+#define	ILGenericPar_Owner(genPar)		(ILGenericParGetOwner((genPar)))
+#define	ILGenericPar_Name(genPar)		(ILGenericParGetName((genPar)))
+#define	ILGenericPar_Kind(genPar)		(ILGenericParGetKind((genPar)))
+#define	ILGenericPar_Constraint(genPar)	(ILGenericParGetConstraint((genPar)))
+
+/*
+ * Create a method specification record.
+ */
+ILMethodSpec *ILMethodSpecCreate(ILImage *image, ILToken token,
+								 ILMember *method, ILType *type);
+
+/*
+ * Get the method associated with a MethodSpec token.
+ */
+ILMember *ILMethodSpecGetMethod(ILMethodSpec *spec);
+
+/*
+ * Get the type associated with a MethodSpec token.
+ */
+ILType *ILMethodSpecGetType(ILMethodSpec *spec);
+
+/*
+ * Helper macros for querying information about method specifications.
+ */
+#define	ILMethodSpec_FromToken(image,token)	\
+				((ILMethodSpec *)ILImageTokenInfo((image), (token)))
+#define	ILMethodSpec_Token(spec)		(ILProgramItem_Token((spec)))
+#define	ILMethodSpec_Method(spec)		(ILMethodSpecGetMethod((spec)))
+#define	ILMethodSpec_Type(spec)			(ILMethodSpecGetType((spec)))
 
 /*
  * Get the type of a Java constant pool entry for a class.
