@@ -42,11 +42,6 @@ extern	"C" {
 #define	IL_TS_ABORT_REQUESTED	0x0080
 #define	IL_TS_ABORTED			0x0100
 
-/*
- * Type for a thread start function.
- */
-typedef void (*ILThreadStartFunc)(void *objectArg);
-
 /*
  * Opaque types for thread-related objects.
  */
@@ -55,6 +50,16 @@ typedef struct _tagILMutex      ILMutex;
 typedef struct _tagILRWLock     ILRWLock;
 typedef struct _tagILWaitHandle ILWaitHandle;
 
+/*
+ * Type for a thread start function.
+ */
+typedef void (*ILThreadStartFunc)(void *objectArg);
+
+/*
+ *	Cleanup function for ILThread.
+ */
+typedef void (*ILThreadCleanupFunc)(ILThread *thread);
+
 /*
  * Determine if the system has thread support.  This can
  * be called either before or after "ILThreadInit".
@@ -67,6 +72,12 @@ int ILHasThreads(void);
  */
 void ILThreadInit(void);
 
+/*
+ * Cleans up the threading subsystem.  Safe to call multiple times.  The thread
+ * subsystem can't be reinitialized after it has been deinitialized.
+ */
+void ILThreadDeinit(void);
+
 /*
  * Create a new thread, initially in the "unstarted" state.
  * When the thread is started, it will call "startFunc" with
@@ -99,7 +110,7 @@ void *ILThreadGetObject(ILThread *thread);
 
 /*
  * Set the object reference that is associated with a thread.
- * This is typically only required for the "main" thread.
+ * This is used by the engine to store a pointer to an ILExecThread.
  */
 void ILThreadSetObject(ILThread *thread, void *objectArg);
 
@@ -159,6 +170,25 @@ int ILThreadIsAborting(void);
  */
 int ILThreadAbortReset(void);
 
+/*
+ *	Thread priorities.
+ */
+#define IL_TP_LOWEST					0x0
+#define IL_TP_BELOW_NORMAL		0x1
+#define IL_TP_NORMAL					0x2
+#define IL_TP_ABOVE_NORMAL		0x3
+#define IL_TP_HIGHEST					0x4
+
+/*
+ *	Sets the thread priority.
+ */
+void ILThreadSetPriority(ILThread *thread, int priority);
+
+/*
+ *	Gets the thread priority.
+ */
+int ILThreadGetPriority(ILThread *thread);
+
 /*
  * Join result codes.
  */
@@ -222,6 +252,26 @@ void ILThreadGetCounts(unsigned long *numForeground,
  */
 int ILThreadSleep(ILUInt32 ms);
 
+/*
+ *	Wait for a specified amount of time (in ms) for a all foreground threads
+ * (except the main thread) to finish.  A timeout of -1 means infinite.
+ */
+void ILThreadWaitForForegroundThreads(int timeout);
+
+/*
+ *	Registers a function that will be called by the thread when it has thread finished.
+ *
+ * Returns 0 on success.
+ */
+int ILThreadRegisterCleanup(ILThread *thread, ILThreadCleanupFunc func);
+
+/*
+ *	Unregisters a thread cleanup function.
+ *
+ * Returns 0 on success.
+ */
+int ILThreadUnregisterCleanup(ILThread *thread, ILThreadCleanupFunc func);
+
 /*
  * Create a mutex.  Note: this type of mutex will not
  * necessarily update the thread's "wait/sleep/join"
@@ -300,6 +350,11 @@ int ILWaitHandleClose(ILWaitHandle *handle);
  */
 int ILWaitOne(ILWaitHandle *handle, ILUInt32 timeout);
 
+/*
+ *	Signals and waits for a handle atomically.
+ */
+int ILSignalAndWait(ILWaitHandle *signalHandle, ILWaitHandle *waitHandle, ILUInt32 timeout);
+
 /*
  * Wait for any wait handle in a set to become available.
  * Returns the index of the handle that was acquired.
@@ -332,6 +387,55 @@ ILWaitHandle *ILWaitMutexNamedCreate(const char *name, int initiallyOwned,
  */
 int ILWaitMutexRelease(ILWaitHandle *handle);
 
+/*
+ * Create a wait handle that corresponds to a monitor.
+ */
+ILWaitHandle *ILWaitMonitorCreate(void);
+
+/*
+ * Wait on a monitor for a "pulse" operation.  Returns zero if we
+ * don't own the monitor, 1 if waiting, or an "IL_WAIT_*" error
+ * code otherwise.
+ */
+int ILWaitMonitorWait(ILWaitHandle *handle, ILUInt32 timeout);
+
+/*
+ * Pulse a single waiting thread on a monitor.
+ */
+int ILWaitMonitorPulse(ILWaitHandle *handle);
+
+/*
+ * Pulse all waiting threads on a monitor.
+ */
+int ILWaitMonitorPulseAll(ILWaitHandle *handle);
+
+/*
+ *	Checks if the monitor has no owner and no waiters.
+ */
+int ILWaitMonitorCanClose(ILWaitHandle *handle);
+
+#define IL_WAIT_LEAVE_STILL_OWNS		(2)
+
+/*
+ *	Similar to ILWaitMonitorLeave except that waiting moniters aren't signalled.
+ */
+int ILWaitMonitorSpeculativeLeave(ILWaitHandle *handle);
+
+/*
+ *	Signals waiting moniters when call after ILWaitMonitorSpeculativeLeave.
+ */
+int ILWaitMonitorCompleteLeave(ILWaitHandle *handle);
+
+/*
+ * Syntactic sugar.
+ */
+#define	ILWaitMonitorEnter(monitor)		\
+			(ILWaitOne((monitor), IL_WAIT_INFINITE))
+#define	ILWaitMonitorTryEnter(monitor,timeout)		\
+			(ILWaitOne((monitor), (timeout)))
+#define	ILWaitMonitorLeave(monitor)		\
+			(ILWaitMutexRelease((monitor)))
+
 /*
  * Wait Event definitions
  *
@@ -339,20 +443,29 @@ int ILWaitMutexRelease(ILWaitHandle *handle);
  */
 
 /*
- * Create a wait event.
+ * Creates an event.
  */
 ILWaitHandle *ILWaitEventCreate(int manualReset, int initialState);
 
 /*
- * Set the event.
+ * Sets an event.
  */
 int ILWaitEventSet(ILWaitHandle *event);
 
 /*
- * Reset the event.
+ * Resets an event.
  */
 int ILWaitEventReset(ILWaitHandle *event);
 
+/*
+ *	Pulses an event.
+ *
+ * If the wait event is a manual reset event, all waiting threads are signalled.
+ * If the wait event is an auto reset event, a single waiting thread is signalled.
+ * On return, the wait event is unsignalled.
+ */
+int ILWaitEventPulse(ILWaitHandle *event);
+
 #ifdef	__cplusplus 
 };
 #endif
