@@ -3,6 +3,8 @@
  *
  * Copyright (C) 2002  Southern Storm Software, Pty Ltd.
  *
+ * Contributions from Thong Nguyen (tum) [tum@veridicus.com]
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -38,7 +40,7 @@ extern	"C" {
 static _ILMutex threadLockAll;
 static long volatile numThreads;
 static long volatile numBackgroundThreads;
-static _ILSemaphore startupAck;
+static ILWaitHandle *foregroundThreadsFinished;
 
 /*
  * Global mutex for atomic operations.
@@ -64,8 +66,7 @@ static void ThreadInit(void)
 	_ILThreadInitSystem(&mainThread);
 
 	/* Initialize synchronization objects that we need */
-	_ILMutexCreate(&threadLockAll);
-	_ILSemaphoreCreate(&startupAck);
+	_ILMutexCreate(&threadLockAll);	
 	_ILMutexCreate(&atomicLock);
 
 	/* Set up the "main" thread.  "_ILThreadInitSystem" has already
@@ -82,6 +83,8 @@ static void ThreadInit(void)
 	_ILWakeupCreate(&(mainThread.wakeup));
 	_ILWakeupQueueCreate(&(mainThread.joinQueue));
 
+	foregroundThreadsFinished = ILWaitEventCreate(1, 1);
+
 	/* Set the thread object for the "main" thread */
 	_ILThreadSetSelf(&mainThread);
 
@@ -90,20 +93,28 @@ static void ThreadInit(void)
 	numBackgroundThreads = 0;
 }
 
+void ThreadDeinit(void)
+{
+	if(foregroundThreadsFinished != 0)
+	{
+		ILWaitHandleClose(foregroundThreadsFinished);
+	}
+}
+
 void ILThreadInit(void)
 {
 	_ILCallOnce(ThreadInit);
 }
 
+void ILThreadDeinit(void)
+{
+	_ILCallOnce(ThreadDeinit);
+}
+
 void _ILThreadRun(ILThread *thread)
 {
 	int isbg;
-
-	/* Signal "ILThreadCreate" that the thread is ready to go */
-	_ILSemaphorePost(&startupAck);
-
-	/* Wait until the thread is really started by a call to "ILThreadStart" */
-	_ILThreadSuspendSelf(thread);
+	ILThreadCleanupEntry *entry, *next;
 
 	/* If we still have a startup function, then execute it.
 	   The field may have been replaced with NULL if the thread
@@ -113,10 +124,30 @@ void _ILThreadRun(ILThread *thread)
 		(*(thread->startFunc))(thread->objectArg);
 	}
 
-	/* Mark the thread as stopped, and wakeup everyone waiting to join */
+	/* Mark the thread as stopped */
 	_ILMutexLock(&(thread->lock));
 	thread->state |= IL_TS_STOPPED;
-	isbg = ((thread->state & IL_TS_BACKGROUND) != 0);
+	isbg = ((thread->state & IL_TS_BACKGROUND) != 0);	
+	_ILMutexUnlock(&(thread->lock));
+
+	/* Notify and remove the cleanup handlers */
+
+	entry = thread->firstCleanupEntry;
+
+	while (entry)
+	{
+		/* Call the cleanup function */
+		entry->cleanup(thread);
+		next = entry->next;
+
+		/* Free the entry */
+		ILFree(entry);
+
+		entry = next;
+	}
+
+	/* Wakeup everyone waiting to join */
+	_ILMutexLock(&(thread->lock));
 	_ILWakeupQueueWakeAll(&(thread->joinQueue));
 	_ILMutexUnlock(&(thread->lock));
 
@@ -127,6 +158,13 @@ void _ILThreadRun(ILThread *thread)
 	{
 		numBackgroundThreads -= 1;
 	}
+
+	/* If there are no more foreground threads (except the main one) then set
+		the event that signals that state */
+	if(numThreads - numBackgroundThreads == 1)
+	{
+		ILWaitEventSet(foregroundThreadsFinished);
+	}
 	_ILMutexUnlock(&threadLockAll);
 }
 
@@ -146,38 +184,25 @@ ILThread *ILThreadCreate(ILThreadStartFunc startFunc, void *objectArg)
 	{
 		return 0;
 	}
+
 	_ILMutexCreate(&(thread->lock));
 	thread->state = IL_TS_UNSTARTED;
 	thread->resumeRequested = 0;
 	thread->suspendRequested = 0;
 	thread->numLocksHeld = 0;
+	thread->firstCleanupEntry = 0;
+	thread->lastCleanupEntry = 0;
 	_ILSemaphoreCreate(&(thread->resumeAck));
 	_ILSemaphoreCreate(&(thread->suspendAck));
 	thread->startFunc = startFunc;
 	thread->objectArg = objectArg;
 	_ILWakeupCreate(&(thread->wakeup));
 	_ILWakeupQueueCreate(&(thread->joinQueue));
+	thread->handle = 0;
 
 	/* Lock out the thread system */
 	_ILMutexLock(&threadLockAll);
-
-	/* Create the new thread */
-	if(!_ILThreadCreateSystem(thread))
-	{
-		_ILMutexUnlock(&threadLockAll);
-		_ILWakeupQueueDestroy(&(thread->joinQueue));
-		_ILWakeupDestroy(&(thread->wakeup));
-		_ILSemaphoreDestroy(&(thread->suspendAck));
-		_ILSemaphoreDestroy(&(thread->resumeAck));
-		_ILMutexDestroy(&(thread->lock));
-		ILFree(thread);
-		return 0;
-	}
-
-	/* Wait until the thread startup ack is posted, so that
-	   we know that the thread is ready to go */
-	_ILSemaphoreWait(&startupAck);
-
+	
 	/* We have one extra thread in the system at present */
 	++numThreads;
 
@@ -196,10 +221,29 @@ int ILThreadStart(ILThread *thread)
 	/* Are we in the correct state to start? */
 	if((thread->state & IL_TS_UNSTARTED) != 0)
 	{
-		/* Resume the thread, which is suspended in "ThreadStart" */
-		thread->state &= ~IL_TS_UNSTARTED;
-		_ILThreadResumeSelf(thread);
-		result = 1;
+		/* Create the new thread */
+		if(!_ILThreadCreateSystem(thread))
+		{
+			result = 0;
+		}
+		else
+		{
+			/* Set the thread state to running (0) */
+			thread->state &= ~IL_TS_UNSTARTED;
+			
+			/* If this thread isn't a background thread then unset the
+				foregroundThreadsFinished event.
+				This occurs here rather than in ILThreadCreate so that unstarted
+				threads won't unset the foregroundThreadsFinished event without ever 
+				setting it again (which normally happens at the end of _ILThreadRun) */
+
+			if(((thread->state & IL_TS_BACKGROUND) == 0))
+			{
+				ILWaitEventReset(foregroundThreadsFinished);
+			}
+
+			result = 1;
+		}
 	}
 	else
 	{
@@ -231,14 +275,25 @@ void ILThreadDestroy(ILThread *thread)
 	if((thread->state & IL_TS_STOPPED) == 0)
 	{
 		thread->state |= IL_TS_STOPPED;
-		_ILThreadTerminate(thread);
+
+		/* Only terminate the system thread if one was created */
+		if((thread->state & IL_TS_UNSTARTED) == 0)
+		{
+			_ILThreadTerminate(thread);
+		}
+
 		iscounted = 1;
 		isbg = ((thread->state & IL_TS_BACKGROUND) != 0);
 	}
 
 	/* Unlock the thread object and free it */
 	_ILMutexUnlock(&(thread->lock));
-	_ILThreadDestroy(thread);
+
+	/* Only destroy the system thread if one was created */
+	if((thread->state & IL_TS_UNSTARTED) == 0)
+	{
+		_ILThreadDestroy(thread);
+	}
 	_ILMutexDestroy(&(thread->lock));
 	_ILSemaphoreDestroy(&(thread->suspendAck));
 	_ILSemaphoreDestroy(&(thread->resumeAck));
@@ -256,6 +311,14 @@ void ILThreadDestroy(ILThread *thread)
 	{
 		--numBackgroundThreads;
 	}
+
+	/* If there are no more foreground threads (except the main one) then set
+		the event that signals that state */
+	if(numThreads - numBackgroundThreads == 1)
+	{
+		ILWaitEventSet(foregroundThreadsFinished);
+	}
+
 	_ILMutexUnlock(&threadLockAll);
 }
 
@@ -461,6 +524,7 @@ int ILThreadAbort(ILThread *thread)
 		{
 			_ILMutexUnlock(&(thread->lock));
 			_ILWakeupInterrupt(&(thread->wakeup));
+
 			return 0;
 		}
 
@@ -501,6 +565,10 @@ int ILThreadAbortReset(void)
 	if((thread->state & (IL_TS_ABORTED | IL_TS_ABORT_REQUESTED)) != 0)
 	{
 		thread->state &= ~(IL_TS_ABORTED | IL_TS_ABORT_REQUESTED);
+		thread->state &= ~(IL_TS_INTERRUPTED);
+
+		_ILWakeupCancelInterrupt(&thread->wakeup);
+
 		result = 1;
 	}
 	else
@@ -528,77 +596,85 @@ int ILThreadJoin(ILThread *thread, ILUInt32 ms)
 	_ILMutexLock(&(thread->lock));
 
 	/* Determine what to do based on the thread's state */
-	if((thread->state & IL_TS_STOPPED) == 0)
+	if((thread->state & IL_TS_STOPPED) != 0)
 	{
 		/* The thread is already stopped, so return immediately */
 		result = IL_JOIN_OK;
 	}
 	else
 	{
-		/* Add ourselves to the foreign thread's join queue */
-		if(!_ILWakeupQueueAdd(&(thread->joinQueue), &(self->wakeup), self))
+		/* Note: We must set our wait limit before adding ourselves to any wait queues.
+		    Failure to do so may mean we may miss some signals because they will be 
+			aborted by the signal invoker (which reads us as having a null wait limit).
+			In this specific case, the order doesn't actually matter because we have locked 
+			the queue owner (the thread) but both operations must be performed before we
+			unlock the thread - Tum */
+
+		/* Set our wait limit to 1 */
+		if(!_ILWakeupSetLimit(&(self->wakeup), 1))
 		{
-			result = IL_JOIN_MEMORY;
+			result = -1;
 		}
 		else
-		{
-			/* Unlock the foreign thread */
-			_ILMutexUnlock(&(thread->lock));
-
-			/* Put ourselves into the "wait/sleep/join" state */
-			_ILMutexLock(&(self->lock));
-			if((self->state & (IL_TS_ABORTED | IL_TS_ABORT_REQUESTED)) != 0)
+		{			
+			/* Add ourselves to the foreign thread's join queue */
+			if(!_ILWakeupQueueAdd(&(thread->joinQueue), &(self->wakeup), self))
 			{
-				/* The current thread is aborted */
-				_ILMutexUnlock(&(self->lock));
-				_ILMutexLock(&(thread->lock));
-				_ILWakeupQueueRemove(&(thread->joinQueue), &(self->wakeup));
+				result = IL_JOIN_MEMORY;
+			}
+			else
+			{
+				/* Unlock the foreign thread */
 				_ILMutexUnlock(&(thread->lock));
-				return IL_JOIN_ABORTED;
-			}
-			self->state |= IL_TS_WAIT_SLEEP_JOIN;
-			_ILMutexUnlock(&(self->lock));
 
-			/* Wait until we are woken or a timeout occurs */
-			if(_ILWakeupSetLimit(&(self->wakeup), 1))
-			{
+				/* Put ourselves into the "wait/sleep/join" state */
+				_ILMutexLock(&(self->lock));
+				if((self->state & (IL_TS_ABORTED | IL_TS_ABORT_REQUESTED)) != 0)
+				{
+					/* The current thread is aborted */
+					_ILMutexUnlock(&(self->lock));
+					_ILMutexLock(&(thread->lock));
+					_ILWakeupQueueRemove(&(thread->joinQueue), &(self->wakeup));
+					_ILMutexUnlock(&(thread->lock));
+					return IL_JOIN_ABORTED;
+				}
+				self->state |= IL_TS_WAIT_SLEEP_JOIN;
+				_ILMutexUnlock(&(self->lock));
+
 				result = _ILWakeupWait(&(self->wakeup), ms, (void **)0);
-			}
-			else
-			{
-				result = -1;
-			}
-			if(result < 0)
-			{
-				/* The wakeup was interrupted.  It may be either an
-				   "interrupt" or an "abort request".  We assume abort
-				   for now until we can inspect "self->state" below */
-				result = IL_JOIN_ABORTED;
-			}
-			else if(result > 0)
-			{
-				result = IL_JOIN_OK;
-			}
-			else
-			{
-				result = IL_JOIN_TIMEOUT;
-			}
+				
+				if(result < 0)
+				{
+					/* The wakeup was interrupted.  It may be either an
+					"interrupt" or an "abort request".  We assume abort
+					for now until we can inspect "self->state" below */
+					result = IL_JOIN_ABORTED;
+				}
+				else if(result > 0)
+				{
+					result = IL_JOIN_OK;
+				}
+				else
+				{
+					result = IL_JOIN_TIMEOUT;
+				}
 
-			/* Remove ourselves from the "wait/sleep/join" state,
-			   and check for a pending interrupt */
-			_ILMutexLock(&(self->lock));
-			if((self->state & IL_TS_INTERRUPTED) != 0)
-			{
-				result = IL_JOIN_INTERRUPTED;
+				/* Remove ourselves from the "wait/sleep/join" state,
+				and check for a pending interrupt */
+				_ILMutexLock(&(self->lock));
+				if((self->state & IL_TS_INTERRUPTED) != 0)
+				{
+					result = IL_JOIN_INTERRUPTED;
+				}
+				self->state &= ~(IL_TS_WAIT_SLEEP_JOIN | IL_TS_INTERRUPTED);
+				_ILMutexUnlock(&(self->lock));
+
+				/* Lock down the foreign thread again */
+				_ILMutexLock(&(thread->lock));
+
+				/* Remove ourselves from the foreign thread's join queue */
+				_ILWakeupQueueRemove(&(thread->joinQueue), &(self->wakeup));
 			}
-			self->state &= ~(IL_TS_WAIT_SLEEP_JOIN | IL_TS_INTERRUPTED);
-			_ILMutexUnlock(&(self->lock));
-
-			/* Lock down the foreign thread again */
-			_ILMutexLock(&(thread->lock));
-
-			/* Remove ourselves from the foreign thread's join queue */
-			_ILWakeupQueueRemove(&(thread->joinQueue), &(self->wakeup));
 		}
 	}
 
@@ -653,6 +729,14 @@ void ILThreadSetBackground(ILThread *thread, int flag)
 	/* Adjust "numBackgroundThreads" */
 	_ILMutexLock(&threadLockAll);
 	numBackgroundThreads += change;
+
+	/* If there are no more foreground threads (except the main one) then set
+		the event that signals that state */
+	if(numThreads - numBackgroundThreads == 1)
+	{
+		ILWaitEventSet(foregroundThreadsFinished);
+	}
+
 	_ILMutexUnlock(&threadLockAll);
 }
 
@@ -770,6 +854,149 @@ int ILThreadSleep(ILUInt32 ms)
 	return result;
 }
 
+void ILThreadWaitForForegroundThreads(int timeout)
+{
+#ifdef IL_NO_THREADS
+	/* Nothing to do */
+#else	
+	ILWaitOne(foregroundThreadsFinished, timeout);
+#endif
+}
+
+int ILThreadRegisterCleanup(ILThread *thread, ILThreadCleanupFunc func)
+{
+	ILThreadCleanupEntry *entry;
+
+	_ILMutexLock(&(thread->lock));
+
+	if((thread->state & IL_TS_STOPPED))
+	{
+		/* Thread has stopped */
+
+		_ILMutexUnlock(&(thread->lock));
+
+		return -1;
+	}
+
+	entry = thread->firstCleanupEntry;
+
+	while(entry)
+	{
+		if(entry->cleanup == func)
+		{
+			/* Function already registered */
+
+			_ILMutexUnlock(&(thread->lock));
+
+			return -1;
+		}
+	}
+
+	if((entry = (ILThreadCleanupEntry *)ILMalloc(sizeof(ILThreadCleanupEntry))) == 0)
+	{
+		/* Out of memory */
+
+		_ILMutexUnlock(&(thread->lock));
+
+		return -1;
+	}
+
+	entry->cleanup = func;
+	entry->next = 0;
+
+	/* Add the entry to the end up the cleanup list */
+
+	if(thread->lastCleanupEntry)
+	{
+		thread->lastCleanupEntry->next = entry;
+		thread->lastCleanupEntry = entry;
+	}
+	else
+	{
+		thread->firstCleanupEntry = thread->lastCleanupEntry = entry;
+	}
+
+	_ILMutexUnlock(&(thread->lock));
+	
+	return 0;
+}
+
+int ILThreadUnregisterCleanup(ILThread *thread, ILThreadCleanupFunc func)
+{
+	ILThreadCleanupEntry *entry, *prev;
+
+	_ILMutexLock(&(thread->lock));
+
+	if((thread->state & IL_TS_STOPPED))
+	{
+		/* Thread has stopped */
+
+		_ILMutexUnlock(&(thread->lock));
+
+		return -1;
+	}
+
+	/* Walk the list and remove the cleanup function */
+
+	prev = 0;
+	entry = thread->firstCleanupEntry;
+
+	while(entry)
+	{
+		if(entry->cleanup == func)
+		{
+			/* Remove the entry from the list */
+
+			if(prev)
+			{
+				/* Entry is in the tail of the list */
+
+				prev->next = entry->next;
+
+				if(prev->next == 0)
+				{
+					thread->lastCleanupEntry = prev;
+				}
+
+				ILFree(entry);
+			}
+			else
+			{
+				/* Entry is in the head of the list */
+
+				thread->firstCleanupEntry = entry->next;
+				
+				if(thread->firstCleanupEntry == 0)
+				{
+					thread->lastCleanupEntry = 0;
+				}
+			}
+
+			_ILMutexUnlock(&(thread->lock));
+
+			/* Found and removed */
+
+			return 0;
+		}
+	}
+
+	_ILMutexUnlock(&(thread->lock));
+
+	/* Not found */
+
+	return -1;
+}
+
+void ILThreadSetPriority(ILThread *thread, int priority)
+{
+	_ILThreadSetPriority(thread, priority);
+}
+
+int ILThreadGetPriority(ILThread *thread)
+{
+	_ILThreadGetPriority(thread);
+}
+
 void _ILThreadSuspendRequest(ILThread *thread)
 {
 	/* Clear the "suspendRequested" and "resumeRequested" flags */
